;==============================================================================

; file: "_kernel.scm"

; Copyright (C) 1994-2002 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##include "header.scm")

;==============================================================================

(c-declare "#include \"os.h\"")
(c-declare "#include \"setup.h\"")
(c-declare "#include \"mem.h\"")
(c-declare "#include \"c_intf.h\"")

;;;;;;;;;;;;;;;;********************
(c-declare "
#ifndef ___CC_CMD
#define ___CC_CMD 0
#endif

#ifndef ___LD_CMD
#define ___LD_CMD 0
#endif
")

;------------------------------------------------------------------------------

; Kernel handlers.

; The procedure "##kernel-handlers" is only a container for the kernel
; handlers.  It must never be called.  The function "___setup" in the
; file "setup.c" is responsible for setting up the kernel handlers.

(define-prim (##kernel-handlers)

  (##declare (not interrupts-enabled))

  (##c-code
   "
   /*
    * ___LBL(1)
    *
    * This is the internal return handler.  It is invoked when an
    * internal return point is returned to.
    *
    * Internal return points are used by the compiler when some routine
    * (e.g.  garbage collector) has to be called but it would be space
    * inefficient to generate the code to construct a normal continuation
    * frame (all the live GVM registers would have to be pushed on the
    * stack).  With internal return points it is not necessary for the
    * caller to save the live GVM registers before the routine is called,
    * and restore the GVM registers when the routine returns.  Instead it
    * is the called routine which saves and restores all the registers.
    *
    * The continuation frame for an internal return point has this layout:
    *
    *              STACK
    *          |            |
    *          +------------+
    *  ___fp ->|  RESERVED  | ^ <-- added to make stack and heap length equal
    *          |<ALIGN PAD> | | <-- words added for alignment of frame
    *          |  ret_adr   | | <-- internal return address (back to caller)
    *          | GVM reg N  | | ADDED BY ROUTINE
    *          |    ...     | |
    *          | GVM reg 0  | |
    *          |<ALIGN PAD> | v <-- to force known offset from ___fp to ret_adr
    *          |  slot fs   | ^
    *          |    ...     | | PUT ON STACK BY CALLER OF ROUTINE
    *          |  slot 1    | v
    *          +------------+
    *          |    ...     |
    */

   int fs;
   ___WORD ira;

   /* save result in case we are returning from ##force-undetermined */

   ___ps->temp2 = ___R1;

   /* make ___fp point to internal return address */

   ___ADJFP(___RETI_RA)

   /* pop return address and all GVM registers */

   ira = ___POP;
   ___POP_REGS

   /* get number of slots put on stack by caller */

   ___RETI_GET_CFS(ira,fs)

   /* adjust ___fp so that slot fs is on top of stack */

   ___ADJFP(fs-___ROUND_TO_MULT(fs,___FRAME_ALIGN))

   /* jump to internal return point */

   ___COVER_INTERNAL_RETURN_HANDLER_END;

   ___JUMPEXTPRM(___NOTHING,ira)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(2)
    *
    * This is the break handler.  It is invoked when a procedure
    * attempts to return to its caller and the caller's stack frame
    * is not on top of the stack because it has been captured.
    *
    * At this point the callee will have cleaned up the stack so that the
    * frame pointer (___fp) points to the break frame.  The break frame
    * contains a pointer to the caller's continuation frame, which
    * contains the address in the caller where control will return (the
    * return address ret_adr1).  The caller's continuation frame can either
    * be in the stack or in the heap.  The two situations are depicted below:
    *
    *              STACK                      STACK              HEAP
    *          |            |             |            |               caller's
    *          +------------+             +------------+    +------------+frame
    *  ___fp ->| call frame ---+  ___fp ->| call frame ---->|    HEAD    |
    *          |<ALIGN PAD> |  |          |<ALIGN PAD> |    |  ret_adr1  |
    *          +------------+  |          +------------+    |  slot fs   |
    *          |     .      |  |          |     .      |    |    ...     |
    *          |     .      |  |          |     .      |    | slot link ----+
    *          |     .      |  |          |     .      |    |    ...     |  |
    *          |     .      |  |          |     .      |    |  slot 1    |  |
    *          +------------+  |          |            |    +------------+  |
    *          |  ret_adr1  |<-+          |            |                    |
    *          |  RESERVED  |             |            |                    |
    * caller's |  slot fs   |             |            |                    |
    * frame    |    ...     |             |            |                    |
    *          |  ret_adr2  |             |            |                    |
    *          |    ...     |             |            |                    |
    *          |  slot 1    |             |     .      |    +------------+  |
    *          +------------+             |     .      |    |    HEAD    |<-+
    *          |  RESERVED  |             |     .      |    |  ret_adr2  |
    *          |    ...     |             |     .      |    |    ...    ----+
    *          +------------+             +------------+    +------------+  V
    *                                                                      ...
    *
    * These cases are distinguished by the tag on the pointer to the
    * caller's frame (i.e. 'call frame').
    *
    * The break handler puts a copy of the caller's frame on the top of
    * the stack except that the slot 'link' is set to the address of the
    * break handler.  The frame pointer in the break frame is modified
    * so that it points to the frame of the caller's caller.  Finally a
    * jump to the return address in the caller's frame (ret_adr1) is
    * performed.  At that point the stack will be in the following state
    * respectively:
    *
    *              STACK                      STACK              HEAP
    *          |            |             |            |
    *          +------------+             +------------+
    *  ___fp ->|  RESERVED  |     ___fp ->|  RESERVED  | ^
    *          |  slot fs   |             |  slot fs   | |
    *          |    ...     |             |    ...     | | ADDED BY BREAK
    *          | break hdlr |             | break hdlr | | HANDLER
    *          |    ...     |             |    ...     | |
    *          |  slot 1    |             |  slot 1    | v
    *          +------------+             +------------+
    *          |new call fr.---+          |new call fr.--+  +------------+
    *          |<ALIGN PAD> |  |          |<ALIGN PAD> | |  |    HEAD    |
    *          +------------+  |          +------------+ |  |  ret_adr1  |
    *          |     .      |  |          |     .      | |  |  slot fs   |
    *          |     .      |  |          |     .      | |  |    ...     |
    *          |     .      |  |          |     .      | |  | slot link ----+
    *          |     .      |  |          |     .      | |  |    ...     |  |
    *          +------------+  |          |            | |  |  slot 1    |  |
    *          |  ret_adr1  |  |          |            | |  +------------+  |
    *          |  RESERVED  |  |          |            | |                  |
    *          |  slot fs   |  |          |            | |                  |
    *          |    ...     |  |          |            | |                  |
    *          |  ret_adr2  |  |          |            | |                  |
    *          |    ...     |  |          |            | |                  |
    *          |  slot 1    |  |          |     .      | |  +------------+  |
    *          +------------+  |          |     .      | +->|    HEAD    |<-+
    *          |  ret_adr2  |<-+          |     .      |    |  ret_adr2  |
    *          |    ...     | (see note)  |     .      |    |    ...    ----+
    *          +------------+             +------------+    +------------+  V
    *                                                                      ...
    *
    * Note that in the first case, the pointer to the caller's frame is
    * normally advanced to the frame following the caller's frame and the
    * return address ret_adr2 is saved in it.  However, if the frame
    * following the caller's frame is a break frame, then the content of
    * that break frame is copied to the topmost break frame.  This
    * ensures that break frames never contain pointers to other break
    * frames which is needed to properly implement tail-calls.
    */

   int fs, link, i;
   ___WORD *fp;
   ___WORD ra1;
   ___WORD ra2;
   ___WORD cf;

   cf = ___STK(-___BREAK_FRAME_NEXT); /* pointer to caller's frame */

   if (___TYP(cf) == ___tFIXNUM)
     {
       /* caller's frame is in the stack */

       /* cf can't be equal to ___FIX(0) */

       fp = ___CAST(___WORD*,cf);

       ra1 = ___FP_STK(fp,-___FRAME_STACK_RA);

       if (ra1 == ___internal_return)
         {
           ___WORD actual_ra = ___FP_STK(fp,___RETI_RA);
           ___RETI_GET_FS_LINK(actual_ra,fs,link)
           ___COVER_BREAK_HANDLER_STACK_RETI;
         }
       else
         {
           ___RETN_GET_FS_LINK(ra1,fs,link)
           ___COVER_BREAK_HANDLER_STACK_RETN;
         }

       ___FP_ADJFP(fp,-___FRAME_SPACE(fs)); /* get base of frame */

       for (i=fs; i>0; i--)
         ___SET_STK(i,___FP_STK(fp,i))

       ra2 = ___STK(link+1);

       if (ra2 == ___ps->handler_break)
         {
           /* first frame of that section */

           ___COVER_BREAK_HANDLER_STACK_FIRST_FRAME;

           ___SET_STK(-___BREAK_FRAME_NEXT,
                      ___FP_STK(fp,-___BREAK_FRAME_NEXT))
         }
       else
         {
           /* not the first frame of that section */

           ___COVER_BREAK_HANDLER_STACK_NOT_FIRST_FRAME;

           ___FP_SET_STK(fp,-___FRAME_STACK_RA,ra2)
           ___SET_STK(-___BREAK_FRAME_NEXT,___CAST(___WORD,fp))
           ___SET_STK(link+1,___ps->handler_break)
         }
     }
   else
     {
       /* caller's frame is in the heap */

       fp = ___BODY_AS(cf,___tSUBTYPED); /* get pointer to frame's body */

       ra1 = fp[___FRAME_RA];

       if (ra1 == ___internal_return)
         {
           ___WORD actual_ra = fp[___FRAME_RETI_RA];
           ___RETI_GET_FS_LINK(actual_ra,fs,link)
           ___COVER_BREAK_HANDLER_HEAP_RETI;
         }
       else
         {
           ___RETN_GET_FS_LINK(ra1,fs,link)
           ___COVER_BREAK_HANDLER_HEAP_RETN;
         }

       fp += fs+1; /* get base of frame */

       for (i=fs; i>0; i--)
         ___SET_STK(i,___FP_STK(fp,i))

       ___SET_STK(-___BREAK_FRAME_NEXT,___STK(link+1))
       ___SET_STK(link+1,___ps->handler_break)
     }

   ___ADJFP(___FRAME_SPACE(fs))

   ___JUMPEXTPRM(___NOTHING,ra1)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(3)
    *
    * This is the stack-limit handler.  It is invoked by the ___POLL(n)
    * macro when the stack limit is reached and when an interrupt is received.
    *
    * This handler checks for which reason it was invoked and dispatches
    * to one of the following procedures:
    *
    *  ##interrupt-handler
    *  ##raise-stack-overflow-exception
    */

   int fs;
   ___WORD ra;

   /* setup internal return continuation frame */

   ra = ___ps->temp1;

   ___RETI_GET_CFS(ra,fs)

   ___ADJFP(___ROUND_TO_MULT(fs,___FRAME_ALIGN)-fs)

   ___PUSH_REGS /* push all GVM registers (live or not) */
   ___PUSH(ra)  /* push return address */

   ___ADJFP(-___RETI_RA)

   ___SET_R0(___internal_return)

   /* check why the handler was called */

   if (___FP_AFTER(___fp,___ps->stack_limit)
#ifdef CALL_GC_FREQUENTLY
       || --___gc_calls_to_punt < 0
#endif
      )
     {
       ___BOOL overflow;

       ___COVER_STACK_LIMIT_HANDLER_STACK_LIMIT;

       ___FRAME_STORE_RA(___R0)
       ___W_ALL
       overflow = ___stack_limit () && ___garbage_collect ();
       ___R_ALL
       ___SET_R0(___FRAME_FETCH_RA)

       if (overflow)
         {
           ___COVER_STACK_LIMIT_HANDLER_HEAP_OVERFLOW;
           ___JUMPPRM(___SET_NARGS(0),
                      ___G__23__23_raise_2d_stack_2d_overflow_2d_exception.prm)
         }
     }

   /* prepare for next interrupt */

   ___begin_interrupt_service ();

   if (___ps->intr_enabled)
     {
       int i;

       ___COVER_STACK_LIMIT_HANDLER_INTR_ENABLED;

       for (i=0; i<___NB_INTRS; i++)
         if (___check_interrupt (i))
           break;

       ___end_interrupt_service (i+1);

       if (i < ___NB_INTRS)
         {
           ___COVER_STACK_LIMIT_HANDLER_INTERRUPT;

           ___SET_R1(___FIX(i))
           ___JUMPPRM(___SET_NARGS(1),
                      ___G__23__23_interrupt_2d_handler.prm)
         }
     }
   else
     ___end_interrupt_service (0);

   ___COVER_STACK_LIMIT_HANDLER_END;

   ___JUMPEXTPRM(___NOTHING,___R0)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(4)
    *
    * This is the heap-limit handler.  It is invoked by the ___CHECK_HEAP(n,m)
    * macro when the heap pointer reaches the end of the current msection.
    *
    * This handler simply calls ##check-heap.
    */

   int fs;
   ___WORD ra;

   /* setup internal return continuation frame */

   ra = ___ps->temp1;

   ___RETI_GET_CFS(ra,fs)

   ___ADJFP(___ROUND_TO_MULT(fs,___FRAME_ALIGN)-fs)

   ___PUSH_REGS /* push all GVM registers (live or not) */
   ___PUSH(ra)  /* push return address */

   ___ADJFP(-___RETI_RA)

   ___SET_R0(___internal_return)

   /* tail call to ##check-heap */

   ___COVER_HEAP_LIMIT_HANDLER_END;

   ___JUMPPRM(___SET_NARGS(0),
              ___G__23__23_check_2d_heap.prm)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(5)
    *
    * This is the call to nonprocedure handler.  It is invoked when
    * there is an attempt to call an object that is not a procedure.
    *
    * This handler simply tail calls ##apply-with-procedure-check-nary
    * (i.e. the continuation will be the same as that of the faulty call).
    * The arguments are the object that was in operator position followed
    * by the arguments of the faulty call.
    */

   int na;
   int i;

   na = ___ps->na;

   ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */
   ___PUSH(___FIX(0))       /* make space for operator */

   for (i=0; i<na; i++)
     ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

   ___SET_STK(-na,___ps->temp1) /* set operator argument */

   ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

   ___COVER_NONPROC_HANDLER_END;

   ___JUMPPRM(___SET_NARGS(na+1),
              ___G__23__23_apply_2d_with_2d_procedure_2d_check_2d_nary.prm)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(6)
    *
    * This is the call to global nonprocedure handler.  It is invoked
    * when there is an attempt to call an object that is not a procedure
    * that is bound to a global variable.
    *
    * This handler simply tail calls ##apply-global-with-procedure-check-nary
    * or ##apply-with-procedure-check-nary (i.e. the continuation will
    * be the same as that of the faulty call).
    * ##apply-global-with-procedure-check-nary is called when the global
    * variable's name is known (in this case the arguments are the variable's
    * name followed by the arguments of the faulty call).  Otherwise,
    * ##apply-with-procedure-check-nary is called and the arguments are the
    * object that was in operator position followed by the arguments of the
    * faulty call.
    */

   int na;
   int i;
   ___WORD result;
   ___WORD handler;

   na = ___ps->na;

   ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */
   ___PUSH(___FIX(0))       /* make space for operator */

   for (i=0; i<na; i++)
     ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

   result = ___CAST(___glo_struct*,___ps->temp4)->val;
   handler = ___G__23__23_apply_2d_with_2d_procedure_2d_check_2d_nary.prm;

#if 0

   i = ___INT(___VECTORLENGTH(___symbol_table)) - 1;

   while (i >= 0)
     {
       ___WORD probe = ___FIELD(___symbol_table,i);

       while (probe != ___NUL)
         {
           if (___ps->temp4 == ___FIELD(probe,___SYMBOL_GLOBAL))
             {
               ___COVER_GLOBAL_NONPROC_HANDLER_FOUND;
               result = probe;
               handler = ___G__23__23_apply_2d_global_2d_with_2d_procedure_2d_check_2d_nary.prm;
               break;
             }
           ___COVER_GLOBAL_NONPROC_HANDLER_NEXT;
           probe = ___FIELD(probe,___SYMKEY_NEXT);
         }
       i--;
     }

#endif

   ___SET_STK(-na,result) /* set operator argument */

   ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

   ___COVER_GLOBAL_NONPROC_HANDLER_END;

   ___JUMPPRM(___SET_NARGS(na+1),handler)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(7)
    *
    * This is the wrong number of arguments handler.  It is invoked when
    * a procedure is called with a number of arguments that it does not
    * accept.
    *
    * This handler simply tail calls
    * ##raise-wrong-number-of-arguments-exception (i.e. the continuation
    * will be the same as that of the faulty call).  The arguments are the
    * object that was in operator position followed by the arguments of
    * the faulty call.
    */

   int na;
   int i;

   na = ___ps->na;

   ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */
   ___PUSH(___FIX(0))       /* make space for operator */

   for (i=0; i<na; i++)
     ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

   /* ___ps->temp1 points to the entry point of the procedure */

   if (___HD_TYP(___HEADER(___ps->temp1)) == ___PERM)
     {
       ___COVER_WRONG_NARGS_HANDLER_NONCLOSURE;
       ___SET_STK(-na,___ps->temp1) /*set operator argument when nonclosure*/
     }
   else
     {
       ___COVER_WRONG_NARGS_HANDLER_CLOSURE;
       ___SET_STK(-na,___SELF) /* set operator argument when closure */
     }

   ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

   ___JUMPPRM(___SET_NARGS(na+1),
              ___G__23__23_raise_2d_wrong_2d_number_2d_of_2d_arguments_2d_exception_2d_nary.prm)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(8)
    *
    * This is the rest parameter handler.  It is invoked when a nonnull
    * rest parameter must be constructed.
    */

   int np;
   int na;
   int i;
   ___WORD rest_param_list;

   np = ___PRD_NBPARMS(___HEADER(___ps->temp1));
   na = ___ps->na;

   ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */

   if (na < np)
     {
       ___COVER_REST_PARAM_HANDLER_WRONG_NARGS;

       ___PUSH(___FIX(0)) /* make space for operator */

       for (i=0; i<na; i++)
         ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

       ___SET_STK(-na,___ps->temp1) /* set operator argument */

       ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

       ___JUMPPRM(___SET_NARGS(na+1),
                  ___G__23__23_raise_2d_wrong_2d_number_2d_of_2d_arguments_2d_exception_2d_nary.prm)
     }

   rest_param_list = ___NUL;

   i = na - np + 1;

   while (i > 0)
     {
       rest_param_list = ___CONS(___POP,rest_param_list);
       i--;

       if (___hp > ___ps->heap_limit)
         {
            ___BOOL need_to_gc;

            ___COVER_REST_PARAM_HANDLER_HEAP_LIMIT;

            ___W_HEAP
            need_to_gc = ___heap_limit ();
            ___R_HEAP

            if (need_to_gc)
              {
                /*
                 * We need to garbage collect, but before we do we have
                 * to remove all arguments from the stack so that the GC
                 * only sees complete continuation frames on the stack.
                 * The arguments are stored in a heap allocated vector.
                 */

                ___COVER_REST_PARAM_HANDLER_NEED_TO_GC;

                while (i > 0)
                  {
                    rest_param_list = ___CONS(___POP,rest_param_list);
                    i--;
                  }

                ___BEGIN_ALLOC_VECTOR(np)
                i = np - 1;
                ___ADD_VECTOR_ELEM(i,rest_param_list)
                while (i > 0)
                  {
                    i--;
                    ___ADD_VECTOR_ELEM(i,___POP)
                  }
                ___END_ALLOC_VECTOR(np)

                ___PUSH_ARGS2(___ps->temp1,___GET_VECTOR(np))

                ___JUMPPRM(___SET_NARGS(2),
                           ___G__23__23_rest_2d_param_2d_check_2d_heap.prm)
              }
          }
     }

   ___COVER_REST_PARAM_HANDLER_NO_NEED_TO_GC;

   ___PUSH(rest_param_list) /* rest parameter is last */

   ___POP_ARGS_IN_REGS(np) /* load register arguments */

   ___JUMPEXTPRM(___SET_NARGS(-1),___ps->temp1)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(9)
    *
    * This is the keyword parameter handler.  It is invoked when keyword
    * parameters must be processed.
    */

   int np;
   int na;
   int nb_req_opt;
   int nb_key;
   int i;
   int j;
   int k;
   ___WORD key_descr;
   ___WORD key_vals[___MAX_NB_PARMS];

   np = ___PRD_NBPARMS(___HEADER(___ps->temp1));
   na = ___ps->na;
   key_descr = ___ps->temp3;
   nb_req_opt = ___ps->temp2;
   nb_key = np - nb_req_opt;

   ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */

   k = na - nb_req_opt;

   if (k < 0)
     {
       ___COVER_KEYWORD_PARAM_HANDLER_WRONG_NARGS1;
       goto wrong_nb_args1;
     }

   for (j=nb_key-1; j>=0; j--)
     key_vals[j] = ___FIELD(key_descr,j*2+1);

   for (i=1; i<=k; i+=2)
     {
       ___WORD key = ___STK(-i);

       for (j=nb_key-1; j>=0; j--)
         if (key == ___FIELD(key_descr,j*2))
           {
             ___COVER_KEYWORD_PARAM_HANDLER_FOUND;
             key_vals[j] = ___STK(-i+1);
             goto continue1;
           }

       /*
        * The argument did not match any of the valid keywords for this
        * procedure, so raise an exception.
        */

       ___COVER_KEYWORD_PARAM_HANDLER_NOT_FOUND;

       ___PUSH(___FIX(0)) /* make space for operator */

       for (i=0; i<na; i++)
         ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

       ___SET_STK(-na,___ps->temp1) /* set operator argument */

       ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

       if (!___KEYWORDP(key))
         ___JUMPPRM(___SET_NARGS(na+1),
                    ___G__23__23_raise_2d_keyword_2d_expected_2d_exception_2d_nary.prm)
       else
         ___JUMPPRM(___SET_NARGS(na+1),
                    ___G__23__23_raise_2d_unknown_2d_keyword_2d_argument_2d_exception_2d_nary.prm)

       continue1:;
     }

   if (k & 1) /* keyword arguments must come in pairs */
     {
       ___COVER_KEYWORD_PARAM_HANDLER_WRONG_NARGS2;

       wrong_nb_args1:

       ___PUSH(___FIX(0)) /* make space for operator */

       for (i=0; i<na; i++)
         ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

       ___SET_STK(-na,___ps->temp1) /* set operator argument */

       ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

       ___JUMPPRM(___SET_NARGS(na+1),
                  ___G__23__23_raise_2d_wrong_2d_number_2d_of_2d_arguments_2d_exception_2d_nary.prm)
     }

   ___ADJFP(-k) /* remove keyword arguments */

   for (i=0; i<nb_key; i++) /* push value of all keyword params */
     ___PUSH(key_vals[i])

   ___POP_ARGS_IN_REGS(np) /* load register arguments */

   ___COVER_KEYWORD_PARAM_HANDLER_END;

   ___JUMPEXTPRM(___SET_NARGS(-1),___ps->temp1)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(10)
    *
    * This is the keyword and rest parameter handler.  It is invoked when
    * keyword parameters must be processed and a rest parameter must be
    * constructed.
    */

   int np;
   int na;
   int nb_req_opt;
   int nb_key;
   int i;
   int j;
   int k;
   ___WORD key_descr;
   ___WORD key_vals[___MAX_NB_PARMS];
   ___WORD rest_param_list;

   np = ___PRD_NBPARMS(___HEADER(___ps->temp1));
   na = ___ps->na;
   key_descr = ___ps->temp3;
   nb_req_opt = ___ps->temp2;
   nb_key = (np-1) - nb_req_opt;

   ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */

   k = na - nb_req_opt;

   if (k < 0)
     {
       ___COVER_KEYWORD_REST_PARAM_HANDLER_WRONG_NARGS1;
       goto wrong_nb_args2;
     }

   for (j=nb_key-1; j>=0; j--)
     key_vals[j] = ___FIELD(key_descr,j*2+1);

   for (i=1; i<=k; i+=2)
     {
       ___WORD key = ___STK(-i);

       for (j=nb_key-1; j>=0; j--)
         if (key == ___FIELD(key_descr,j*2))
           {
             ___COVER_KEYWORD_REST_PARAM_HANDLER_FOUND;
             key_vals[j] = ___STK(-i+1);
             goto continue2;
           }

       /*
        * The argument did not match any of the valid keywords for this
        * procedure.  An exception is raised only if the argument is not
        * a keyword.
        */

       if (!___KEYWORDP(key))
         {
           ___COVER_KEYWORD_REST_PARAM_HANDLER_NOT_FOUND;

           ___PUSH(___FIX(0)) /* make space for operator */

           for (i=0; i<na; i++)
             ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

           ___SET_STK(-na,___ps->temp1) /* set operator argument */

           ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

           ___JUMPPRM(___SET_NARGS(na+1),
                      ___G__23__23_raise_2d_keyword_2d_expected_2d_exception_2d_nary.prm)
         }

       continue2:;
     }

   if (k & 1) /* keyword arguments must come in pairs */
     {
       ___COVER_KEYWORD_REST_PARAM_HANDLER_WRONG_NARGS2;

       wrong_nb_args2:

       ___PUSH(___FIX(0)) /* make space for operator */

       for (i=0; i<na; i++)
         ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

       ___SET_STK(-na,___ps->temp1) /* set operator argument */

       ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

       ___JUMPPRM(___SET_NARGS(na+1),
                  ___G__23__23_raise_2d_wrong_2d_number_2d_of_2d_arguments_2d_exception_2d_nary.prm)
     }

   rest_param_list = ___NUL;

   i = k;

   while (i > 0)
     {
       rest_param_list = ___CONS(___POP,rest_param_list);
       i--;

       if (___hp > ___ps->heap_limit)
         {
            ___BOOL need_to_gc;

            ___COVER_KEYWORD_REST_PARAM_HANDLER_HEAP_LIMIT;

            ___W_HEAP
            need_to_gc = ___heap_limit ();
            ___R_HEAP

            if (need_to_gc)
              {
                /*
                 * We need to garbage collect, but before we do we have
                 * to remove all arguments from the stack so that the GC
                 * only sees complete continuation frames on the stack.
                 * The arguments are stored in a heap allocated vector.
                 */

                ___COVER_KEYWORD_REST_PARAM_HANDLER_NEED_TO_GC;

                while (i > 0)
                  {
                    rest_param_list = ___CONS(___POP,rest_param_list);
                    i--;
                  }

                for (i=0; i<nb_key; i++) /*push value of all keyword params*/
                  ___PUSH(key_vals[i])

                ___BEGIN_ALLOC_VECTOR(np)
                i = np - 1;
                ___ADD_VECTOR_ELEM(i,rest_param_list)
                while (i > 0)
                  {
                    i--;
                    ___ADD_VECTOR_ELEM(i,___POP)
                  }
                ___END_ALLOC_VECTOR(np)

                ___PUSH_ARGS2(___ps->temp1,___GET_VECTOR(np))

                ___JUMPPRM(___SET_NARGS(2),
                           ___G__23__23_rest_2d_param_2d_check_2d_heap.prm)
              }
          }
     }

   ___COVER_KEYWORD_REST_PARAM_HANDLER_NO_NEED_TO_GC;

   for (i=0; i<nb_key; i++) /* push value of all keyword params */
     ___PUSH(key_vals[i])

   ___PUSH(rest_param_list) /* rest parameter is last */

   ___POP_ARGS_IN_REGS(np) /* load register arguments */

   ___JUMPEXTPRM(___SET_NARGS(-1),___ps->temp1)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(11)
    *
    * This is the force handler.  It is invoked when a promise is forced.
    */

   ___WORD ra;
   ___WORD promise;
   ___WORD result;

   ra = ___ps->temp1;
   promise = ___ps->temp2;
   result = ___FIELD(promise,___PROMISE_RESULT);

   if (promise != result)
     {
       /* promise is determined, return cached result */

       ___COVER_FORCE_HANDLER_DETERMINED;

       ___ps->temp2 = result;
       ___JUMPEXTPRM(___NOTHING,ra)
     }
   else
     {
       /* promise is not determined */

       /* setup internal return continuation frame */

       int fs;

       ___RETI_GET_CFS(ra,fs)

       ___ADJFP(___ROUND_TO_MULT(fs,___FRAME_ALIGN)-fs)

       ___PUSH_REGS /* push all GVM registers (live or not) */
       ___PUSH(ra)  /* push return address */

       ___ADJFP(-___RETI_RA)

       ___SET_R0(___internal_return)

       /* tail call to ##force-undetermined */

       ___PUSH_ARGS2(promise,___FIELD(promise,___PROMISE_THUNK))

       ___COVER_FORCE_HANDLER_NOT_DETERMINED;

       ___JUMPPRM(___SET_NARGS(2),
                  ___G__23__23_force_2d_undetermined.prm)
     }
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(12)
    *
    * This is the 'c-lambda' conversion error handler.  It is invoked
    * when an error is detected by the C-interface in a 'c-lambda' during
    * a data representation conversion between Scheme and C.
    */

   int na;
   int i;

   na = ___ps->na;

   /* make space for 3 new arguments (one will replace return address) */

   ___ADJFP((na+1)+2-___FRAME_SPACE(na+1))

   ___SET_R0(___STK(-2)) /* get return address */

   for (i=0; i<na; i++)
     ___SET_STK(-i,___STK(-(i+3))) /* shift arguments up by three */

   ___SET_STK(-(na+2),___FIX(___ps->temp1)) /* arg 1 = error code, integer */
   ___SET_STK(-(na+1),___ps->temp2) /* arg 2 = error message, string or #f */
   ___SET_STK(-na,___ps->temp3) /* arg 3 = procedure */

   ___POP_ARGS_IN_REGS(na+3) /* load register arguments */

   ___COVER_C_LAMBDA_CONVERSION_ERROR_HANDLER;

   ___JUMPPRM(___SET_NARGS(na+3),
              ___G__23__23_raise_2d_c_2d_lambda_2d_conversion_2d_exception_2d_nary.prm)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (##c-code
   "
   /*
    * ___LBL(13)
    *
    * This is the 'c-define' conversion error handler.  It is invoked
    * when an error is detected by the C-interface in a 'c-define' during
    * a data representation conversion between Scheme and C.
    */

   ___PUSH_ARGS3(___FIX(___ps->temp1),___ps->temp2,___ps->temp3)

   ___COVER_C_DEFINE_CONVERSION_ERROR_HANDLER;

   ___JUMPPRM(___SET_NARGS(3),
              ___G__23__23_raise_2d_c_2d_define_2d_conversion_2d_exception.prm)
   "
   (let () (##declare (not warnings)) (0))) ; create a return point

  (let ((stack-marker (##first-argument 0)))
    (##c-code
     "
     /*
      * ___LBL(14)
      *
      * This is the return to C handler.  It is invoked when control
      * must return to C (i.e. back from the call to ___call).
      *
      * Note that the continuation frame is not removed from the stack (it
      * contains the stack marker which is needed by ___call).
      */

     ___WORD unwind_destination = ___STK(2-___FRAME_SPACE(2));

     if (___FIELD(unwind_destination,0) == ___TRU) /* first return? */
       {
         ___COVER_RETURN_TO_C_HANDLER_FIRST_RETURN;
         ___FRAME_STORE_RA(___ps->handler_return_to_c)
         ___W_ALL
         ___THROW(___UNWIND_C_STACK);  /* jump back inside ___call */
       }
     else
       {
         ___COVER_RETURN_TO_C_HANDLER_MULTIPLE_RETURN;
         ___SET_R0(___ps->handler_return_to_c)
         ___JUMPPRM(___SET_NARGS(0),
                    ___G__23__23_raise_2d_multiple_2d_c_2d_return_2d_exception.prm)
       }
     "
     (let () (##declare (not warnings)) (0))) ; create a return point with a
                                              ; frame having the same format as
                                              ; the one created by ___call
    stack-marker))

;------------------------------------------------------------------------------

; Miscellaneous definitions.

(define ##err-code-EAGAIN
  (##c-code "___RESULT = ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));"))

(define ##err-code-EINTR
  (##c-code "___RESULT = ___FIX(___ERR_CODE_FROM_ERRNO(EINTR));"))

(define ##max-char
  (##c-code "___RESULT = ___FIX(___MAX_CHR);"))

(define ##min-fixnum
  (##c-code "___RESULT = ___FIX(___MIN_FIX);"))

(define ##max-fixnum
  (##c-code "___RESULT = ___FIX(___MAX_FIX);"))

(define ##fixnum-width
  (##c-code "___RESULT = ___FIX((8*___WS)-___TB);"))

(define ##bignum.adigit-width
  (##c-code "___RESULT = ___FIX(___BIG_ABASE_WIDTH);"))

(define ##bignum.mdigit-width
  (##c-code "___RESULT = ___FIX(___BIG_MBASE_WIDTH);"))

(define ##bignum.fdigit-width
  (##c-code "___RESULT = ___FIX(___BIG_FBASE_WIDTH);"))

(define-prim (##symbol-table)
  (##declare (not interrupts-enabled))
  (##c-code "___RESULT = ___symbol_table;"))

(define-prim (##keyword-table)
  (##declare (not interrupts-enabled))
  (##c-code "___RESULT = ___keyword_table;"))

(define-prim (##command-line)
  (##declare (not interrupts-enabled))
  (##c-code "___RESULT = ___command_line;"))

(define ##processed-command-line '())
(set! ##processed-command-line (##command-line))

(define-prim (##process-statistics)
  (##declare (not interrupts-enabled))
  (let ((v
         (##f64vector (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0))))
    (##c-code
     "
     int n;
     ___F64 user, sys, real;
     long minflt, majflt;
     ___W_ALL
     ___process_times (&user, &sys, &real);
     ___vm_stats (&minflt, &majflt);
     ___F64VECTORSET(___ARG1,___FIX(0),user)
     ___F64VECTORSET(___ARG1,___FIX(1),sys)
     ___F64VECTORSET(___ARG1,___FIX(2),real)
     ___F64VECTORSET(___ARG1,___FIX(3),___GSTATE->gc_user_time)
     ___F64VECTORSET(___ARG1,___FIX(4),___GSTATE->gc_sys_time)
     ___F64VECTORSET(___ARG1,___FIX(5),___GSTATE->gc_real_time)
     ___F64VECTORSET(___ARG1,___FIX(6),___GSTATE->nb_gcs)
     ___F64VECTORSET(___ARG1,___FIX(7),___bytes_allocated ())
     ___F64VECTORSET(___ARG1,___FIX(8),(2*(1+2)<<___LWS))
     n = (1<<___LWS) + 17 * sizeof (___F64) + (5<<___LWS);
     ___F64VECTORSET(___ARG1,___FIX(9),(2*n))
     ___F64VECTORSET(___ARG1,___FIX(10),minflt)
     ___F64VECTORSET(___ARG1,___FIX(11),majflt)
     ___F64VECTORSET(___ARG1,___FIX(12),___GSTATE->last_gc_real_time)
     ___F64VECTORSET(___ARG1,___FIX(13),___GSTATE->last_gc_heap_size)
     ___F64VECTORSET(___ARG1,___FIX(14),___GSTATE->last_gc_live)
     ___F64VECTORSET(___ARG1,___FIX(15),___GSTATE->last_gc_movable)
     ___F64VECTORSET(___ARG1,___FIX(16),___GSTATE->last_gc_nonmovable)
     ___R_ALL
     ___RESULT = ___ARG1;
     "
     v)))

(define-prim (##process-times)
  (##declare (not interrupts-enabled))
  (let ((v
         (##f64vector (macro-inexact-+0)
                      (macro-inexact-+0)
                      (macro-inexact-+0))))
    (##c-code
     "
     ___WORD result = ___ARG1;
     ___F64 user, sys, real;
     ___process_times (&user, &sys, &real);
     ___F64VECTORSET(result,___FIX(0),user)
     ___F64VECTORSET(result,___FIX(1),sys)
     ___F64VECTORSET(result,___FIX(2),real)
     ___RESULT = result;
     "
     v)))

(define-prim (##get-current-time! floats)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   ___time t;
   ___F64 ft;

   ___time_get_current_time (&t);
   ___time_to_seconds (t, &ft);
   ___F64VECTORSET(___ARG1,___FIX(0),ft)

   ___RESULT = ___VOID;
   "
   floats))

(define-prim (##get-min-heap)
  (##declare (not interrupts-enabled))
  (##c-code "___RESULT = ___FIX(___get_min_heap ());"))

(define-prim (##set-min-heap! bytes)
  (##declare (not interrupts-enabled))
  (##c-code
   "___set_min_heap (___INT(___ARG1)); ___RESULT = ___VOID;"
   bytes))

(define-prim (##get-max-heap)
  (##declare (not interrupts-enabled))
  (##c-code "___RESULT = ___FIX(___get_max_heap ());"))

(define-prim (##set-max-heap! bytes)
  (##declare (not interrupts-enabled))
  (##c-code
   "___set_max_heap (___INT(___ARG1)); ___RESULT = ___VOID;"
   bytes))

(define-prim (##get-live-percent)
  (##declare (not interrupts-enabled))
  (##c-code "___RESULT = ___FIX(___get_live_percent ());"))

(define-prim (##set-live-percent! percent)
  (##declare (not interrupts-enabled))
  (##c-code
   "___set_live_percent (___INT(___ARG1)); ___RESULT = ___VOID;"
   percent))

(define-prim (##get-standard-level)
  (##declare (not interrupts-enabled))
  (##c-code "___RESULT = ___FIX(___get_standard_level ());"))

(define-prim (##set-standard-level! level)
  (##declare (not interrupts-enabled))
  (##c-code
   "___set_standard_level (___INT(___ARG1)); ___RESULT = ___VOID;"
   level))

(define-prim (##set-debug-settings! mask new-settings)
  (##declare (not interrupts-enabled))
  (##c-code
   "___RESULT =
      ___FIX(___set_debug_settings (___INT(___ARG1), ___INT(___ARG2)));"
   mask
   new-settings))

(define-prim (##set-terminal-settings! mask new-settings)
  (##declare (not interrupts-enabled))
  (##c-code
   "___RESULT =
      ___FIX(___set_terminal_settings (___INT(___ARG1), ___INT(___ARG2)));"
   mask
   new-settings))

(define-prim ##err-code->string
  (c-lambda (int)
            nonnull-char-string
   "
   char str[___ERR_MAX_LENGTH+1];
   ___err_code_to_string (___arg1, str);
   ___result = str;
   "))

(define-prim ##format-filepos
  (c-lambda (char-string
             long
             bool)
            char-string
   "___format_filepos"))


(define-prim (##system-version)

  (##define-macro (result)
    (c#system-version))

  (result))

(define-prim (##system-version-string)

  (##define-macro (result)
    (c#system-version-string))

  (result))

;------------------------------------------------------------------------------

; Jobs.

(define-prim (##make-jobs)
  (macro-make-fifo))

(define-prim (##add-job! jobs job)
  (macro-fifo-insert-at-tail! jobs job))

(define-prim (##execute-jobs! jobs)
  (let loop ((lst (macro-fifo->list jobs)))
    (if (##pair? lst)
      (begin
        ((##car lst))
        (loop (##cdr lst))))))

(define-prim (##execute-and-clear-jobs! jobs)
  (let loop ((lst (macro-fifo-remove-all! jobs)))
    (if (##pair? lst)
      (begin
        ((##car lst))
        (loop (##cdr lst))))))

(define-prim (##clear-jobs! jobs)
  (macro-fifo-remove-all! jobs)
  (##void))

;------------------------------------------------------------------------------

; Interrupt system.

(define-prim (##disable-interrupts!)
  (##declare (not interrupts-enabled))
  (##c-code "___disable_interrupts (); ___RESULT = ___VOID;"))

(define-prim (##enable-interrupts!)
  (##declare (not interrupts-enabled))
  (##c-code "___enable_interrupts (); ___RESULT = ___VOID;"))

(define ##interrupt-vector
  (##vector #f #f #f #f #f #f #f #f))

(define-prim (##interrupt-handler code)
  (##declare (not interrupts-enabled))
  (let ((proc (##vector-ref ##interrupt-vector code)))
    (if (##procedure? proc)
      (proc))))

(define-prim (##interrupt-vector-set! code handler)
  (##declare (not interrupts-enabled))
  (##vector-set! ##interrupt-vector code handler))

; (##add-gc-interrupt-job! thunk) can be called to add another job to
; do after a GC.  (##clear-gc-interrupt-jobs!) clears the jobs.

(define ##gc-interrupt-jobs (##make-jobs))

(define-prim (##add-gc-interrupt-job! thunk)
  (##add-job! ##gc-interrupt-jobs thunk))

(define-prim (##clear-gc-interrupt-jobs!)
  (##clear-jobs! ##gc-interrupt-jobs))

(define-prim (##gc-finalize!)
  (##declare (not interrupts-enabled))
  (let ((will
         (##c-code
          "
          ___WORD will = ___ps->executable_wills;
          if (___UNTAG(will) == 0) /* end of list? */
            ___RESULT = ___FAL;
          else
            {
              ___ps->executable_wills = ___BODY(will)[0];
              ___RESULT = will;
            }
          ")))
    (if will
      (begin
        (macro-will-execute! will)
        (##gc-finalize!))
      (##gc-final-will-registry!))))

(define ##final-will-registry (macro-make-fifo))

(define-prim (##execute-final-wills!)
  (##declare (not interrupts-enabled))
  (let ((registry ##final-will-registry))
    (let ((lst (macro-fifo-remove-all! registry)))
      (let loop ((x lst))
        (if (##not (##null? x))
          (begin
            (macro-will-execute! (##car x))
            (loop (##cdr x))))))))

(define-prim (##gc-final-will-registry!)
  (##declare (not interrupts-enabled))
  (let ((registry ##final-will-registry))
    (let loop ((curr registry)
               (next (macro-fifo-next registry)))
      (if (##null? next)
        (macro-fifo-tail-set! registry curr)
        (let* ((will (macro-fifo-elem next))
               (action (macro-will-action will)))
          (if action
            (begin
              (macro-fifo-next-set! curr next)
              (loop next
                    (macro-fifo-next next)))
            (loop curr
                  (macro-fifo-next next))))))))

(define-prim (##make-final-will testator action)
  (let* ((registry ##final-will-registry)
         (will (##make-will testator action)))
    (macro-fifo-insert-at-head! registry will)
    will))

(define-prim (##gc)
  (let ((result
         (##c-code
          "
          ___BOOL overflow;
          ___FRAME_STORE_RA(___R0)
          ___W_ALL
          overflow = ___garbage_collect ();
          ___R_ALL
          ___SET_R0(___FRAME_FETCH_RA)
          ___RESULT = ___BOOLEAN(overflow);
          ")))
    (if result
      (begin
        (##raise-heap-overflow-exception)
        (##gc))
      (##void))))

(##interrupt-vector-set! 2
  (lambda ()
    (##declare (not interrupts-enabled))
    (##gc-finalize!)
    (##execute-jobs! ##gc-interrupt-jobs)))

(define-prim (##make-table)

  (##declare (not interrupts-enabled))

  (let ((table (##vector #f)))
    (##add-gc-interrupt-job!
     (lambda ()
       (##table-gc! table)))
    table))

(define-prim (##table-set! table key obj)

  (##declare (not interrupts-enabled))

  (let loop ((prev table)
             (curr (##vector-ref table 0)))
    (if curr
      (let* ((w (##vector-ref curr 2))
             (k (##will-testator w)))
        (if (##eq? k key)
          (begin
            (##vector-set! curr 1 obj)
            (##void))
          (let ((next (##vector-ref curr 0)))
            (if k
              (loop curr next)
              (begin
                (##vector-set! prev 0 next)
                (loop prev next))))))
      (let ((v (##vector #f obj (##make-will key ##values))))
        (##vector-set! v 0 (##vector-ref table 0))
        (##vector-set! table 0 v)
        (##void)))))

(define-prim (##table-ref table key not-found)

  (##declare (not interrupts-enabled))

  (let loop ((curr (##vector-ref table 0)))
    (if curr
      (let* ((w (##vector-ref curr 2))
             (k (##will-testator w)))
        (if (##eq? k key)
          (##vector-ref curr 1)
          (loop (##vector-ref curr 0))))
      not-found)))

(define-prim (##table-reverse-ref table obj not-found)

  (##declare (not interrupts-enabled))

  (let loop ((curr (##vector-ref table 0)))
    (if curr
      (if (##eq? obj (##vector-ref curr 1))
        (or (##will-testator (##vector-ref curr 2))
            not-found)
        (loop (##vector-ref curr 0)))
      not-found)))

(define-prim (##table-gc! table)

  (##declare (not interrupts-enabled))

  (let loop ((prev table)
             (curr (##vector-ref table 0)))
    (if curr
      (let ((next (##vector-ref curr 0)))
        (if (##will-testator (##vector-ref curr 2))
          (loop curr next)
          (begin
            (##vector-set! prev 0 next)
            (loop prev next)))))))

;------------------------------------------------------------------------------

;**********************

(define-prim (##force-undetermined promise thunk)
  (let ((result (##force (thunk))))
    (##c-code
     "
     if (___FIELD(___ARG1,___PROMISE_RESULT) == ___ARG1)
       {
         ___FIELD(___ARG1,___PROMISE_RESULT) = ___ARG2;
         ___FIELD(___ARG1,___PROMISE_THUNK) = ___FAL;
       }
     ___RESULT = ___FIELD(___ARG1,___PROMISE_RESULT);
     "
     promise
     result)))

;------------------------------------------------------------------------------

; Implementation of exceptions.

(define-prim (##with-no-result-expected thunk)
  (##declare (not interrupts-enabled))
  (##first-argument (thunk)))

(define-procedures-for-heap-overflow-exception)

(define-prim (##raise-heap-overflow-exception)
  (##declare (not interrupts-enabled))
  (##with-no-result-expected
   (lambda ()
     (macro-raise
      (macro-make-heap-overflow-exception)))))

(define-procedures-for-stack-overflow-exception)

(define-prim (##raise-stack-overflow-exception)
  (##declare (not interrupts-enabled))
  (##with-no-result-expected
   (lambda ()
     (macro-raise
      (macro-make-stack-overflow-exception)))))

(define-procedures-for-nonprocedure-operator-exception)

(define-prim (##apply-global-with-procedure-check-nary gv . args)
  (##declare (not interrupts-enabled))
  (##apply-with-procedure-check (##global-var-ref gv) args))

(define-prim (##apply-with-procedure-check-nary oper . args)
  (##declare (not interrupts-enabled))
  (##apply-with-procedure-check oper args))

(define-prim (##apply-with-procedure-check oper args)
  (##declare (not interrupts-enabled))
  (macro-force-vars (oper)
    (if (##procedure? oper)
      (##apply oper args)
      (##raise-nonprocedure-operator-exception oper args #f #f))))

(define-prim (##raise-nonprocedure-operator-exception oper args code rte)
  (##declare (not interrupts-enabled))
  (macro-raise
   (macro-make-nonprocedure-operator-exception oper args code rte)))

(define-procedures-for-wrong-number-of-arguments-exception)

(define-prim (##raise-wrong-number-of-arguments-exception-nary proc . args)
  (##declare (not interrupts-enabled))
  (##raise-wrong-number-of-arguments-exception proc args))

(define-prim (##raise-wrong-number-of-arguments-exception proc args)
  (##declare (not interrupts-enabled))
  (macro-raise
   (macro-make-wrong-number-of-arguments-exception proc args)))

(define-procedures-for-keyword-expected-exception)

(define-prim (##raise-keyword-expected-exception-nary proc . args)
  (##declare (not interrupts-enabled))
  (##raise-keyword-expected-exception proc args))

(define-prim (##raise-keyword-expected-exception proc args)
  (##declare (not interrupts-enabled))
  (macro-raise
   (macro-make-keyword-expected-exception proc args)))

(define-procedures-for-unknown-keyword-argument-exception)

(define-prim (##raise-unknown-keyword-argument-exception-nary proc . args)
  (##declare (not interrupts-enabled))
  (##raise-unknown-keyword-argument-exception proc args))

(define-prim (##raise-unknown-keyword-argument-exception proc args)
  (##declare (not interrupts-enabled))
  (macro-raise
   (macro-make-unknown-keyword-argument-exception proc args)))

(define-procedures-for-c-lambda-conversion-exception)

(define-prim (##raise-c-lambda-conversion-exception-nary code message proc . args)
  (##declare (not interrupts-enabled))
  (macro-raise
   (macro-make-c-lambda-conversion-exception proc args code message)))

(define-procedures-for-c-define-conversion-exception)

(define-prim (##raise-c-define-conversion-exception code message proc)
  (##declare (not interrupts-enabled))
  (macro-raise
   (macro-make-c-define-conversion-exception proc '() code message)))

(define-procedures-for-multiple-c-return-exception)

(define-prim (##raise-multiple-c-return-exception)
  (##declare (not interrupts-enabled))
  (macro-raise
   (macro-make-multiple-c-return-exception)))

(define-procedures-for-number-of-arguments-limit-exception)

(define-prim (##raise-number-of-arguments-limit-exception proc args)
  (##declare (not interrupts-enabled))
  (macro-raise
   (macro-make-number-of-arguments-limit-exception proc args)))

(define-procedures-for-type-exception)

(define-prim (##raise-type-exception arg-num type-id proc args)
  (##extract-procedure-and-arguments
   proc
   args
   arg-num
   type-id
   #f
   (lambda (procedure arguments arg-num type-id dummy)
     (macro-raise
      (macro-make-type-exception procedure arguments arg-num type-id)))))

(define-procedures-for-os-exception)

(define-prim (##raise-os-exception message code proc . args)
  (##extract-procedure-and-arguments
   proc
   args
   message
   code
   #f
   (lambda (procedure arguments message code dummy)
     (macro-raise
      (macro-make-os-exception procedure arguments message code)))))

(define-prim (##extract-procedure-and-arguments proc args val1 val2 val3 cont)
  (cont
    (if (##eq? proc #t) (##car args) proc)
    (##argument-list-remove-absent!
     (if (##eq? proc #t) (##argument-list-fix-rest-param! (##cdr args)) args))
    val1
    val2
    val3))

(define-prim (##argument-list-remove-absent! lst)
  (let loop1 ((lst lst))
    (if (##pair? lst)
      (if (##eq? (##car lst) (macro-absent-obj))
        (loop1 (##cdr lst))
        (begin
          (let loop2 ((curr lst)
                      (next (##cdr lst)))
            (if (##pair? next)
              (if (##eq? (##car next) (macro-absent-obj))
                (loop2 curr
                       (##cdr next))
                (begin
                  (##set-cdr! curr next)
                  (loop2 next
                         (##cdr next))))
              (##set-cdr! curr next)))
          lst))
      lst)))

(define-prim (##argument-list-fix-rest-param! lst)
  (let loop ((curr #f) (next lst))
    (let ((tail (##cdr next)))
      (if (##pair? tail)
        (loop next tail)
        (if curr
          (begin
            (##set-cdr! curr (##car next))
            lst)
          (##car next))))))

;------------------------------------------------------------------------------

; (##heartbeat-interval-set! seconds) sets the heartbeat interrupt
; interval to the time closest to "seconds" seconds (a flonum value).
; If "seconds" is negative, the heartbeat interrupt is turned off.  If
; "seconds" is zero, the smallest possible interval is used.  The
; actual interval in seconds is returned.

(define-prim (##heartbeat-interval-set! seconds)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   ___FLONUM_VAL(___ARG2) = ___set_heartbeat_interval (___FLONUM_VAL(___ARG1));
   ___RESULT = ___ARG2;
   "
   seconds
   (##flonum.<-fixnum 0)))

;------------------------------------------------------------------------------

; Memory allocation.

(define-prim (##check-heap)
  (##declare (not interrupts-enabled))
  (let ((result
         (##c-code
          "
          if (___hp > ___ps->heap_limit)
            {
              ___BOOL overflow;
              ___FRAME_STORE_RA(___R0)
              ___W_ALL
              overflow = ___heap_limit () && ___garbage_collect ();
              ___R_ALL
              ___SET_R0(___FRAME_FETCH_RA)
              if (overflow)
                ___RESULT = ___TRU;
              else
                ___RESULT = ___FAL;
            }
          else
            ___RESULT = ___FAL;
          ")))
    (if result
      (begin
        (##raise-heap-overflow-exception)
        (##check-heap)))))

(define-prim (##rest-param-check-heap proc args)
  (##declare (not interrupts-enabled))
  (let ((overflow (##gc-without-exceptions)))
    (if overflow
      (begin
        (##raise-heap-overflow-exception)
        (##rest-param-check-heap proc args))
      (##rest-param-resume-procedure proc args))))

(define-prim (##rest-param-heap-overflow proc args)
  (##raise-heap-overflow-exception)
  (##apply proc args))

(define-prim (##rest-param-resume-procedure proc args)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   ___WORD proc;
   ___WORD args;
   int np;
   int i;

   ___POP_ARGS2(proc,args)

   np = ___INT(___VECTORLENGTH(args));

   for (i=0; i<np; i++)
     ___PUSH(___FIELD(args,i))

   ___POP_ARGS_IN_REGS(np) /* load register arguments */

   ___COVER_REST_PARAM_RESUME_PROCEDURE;

   ___JUMPEXTPRM(___SET_NARGS(-1),proc)
   "))

(define-prim (##gc-without-exceptions)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   ___BOOL overflow;

   ___FRAME_STORE_RA(___R0)
   ___W_ALL
   overflow = ___garbage_collect ();
   ___R_ALL
   ___SET_R0(___FRAME_FETCH_RA)

   ___RESULT = ___BOOLEAN(overflow);

   ___COVER_GC_WITHOUT_EXCEPTIONS;
   "))

(define-prim (##make-vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
long words = n + 1;
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>(___LF+___LWS)))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sVECTOR, n<<___LWS, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_WORDS(n, ___sVECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result))
  {
    ___WORD fill = ___ARG2;
    if (fill == ___ABSENT)
      fill = ___FIX(0);
    for (i=0; i<n; i++)
      ___VECTORSET(result,___FIX(i),fill)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-vector k fill))
      v)))

(define-prim (##make-string k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((s (##c-code "
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<___LCS)) + 1;
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>(___LF+___LCS)))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sSTRING, n<<___LCS, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<___LCS), ___sSTRING);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___STRINGSET(result,___FIX(i),___ARG2);
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? s)
      (begin
        (##raise-heap-overflow-exception)
        (##make-string k fill))
      s)))

(define-prim (##make-s8vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
long words = ___WORDS(n) + 1;
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>___LF))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sS8VECTOR, n, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES(n, ___sS8VECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___S8VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-s8vector k fill))
      v)))

(define-prim (##make-u8vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
long words = ___WORDS(n) + 1;
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>___LF))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sU8VECTOR, n, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES(n, ___sU8VECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___U8VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-u8vector k fill))
      v)))

(define-prim (##make-s16vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<1)) + 1;
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>(___LF+1)))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sS16VECTOR, n<<1, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<1), ___sS16VECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___S16VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-s16vector k fill))
      v)))

(define-prim (##make-u16vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<1)) + 1;
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>(___LF+1)))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sU16VECTOR, n<<1, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<1), ___sU16VECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___U16VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-u16vector k fill))
      v)))

(define-prim (##make-s32vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<2)) + 1;
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>(___LF+2)))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sS32VECTOR, n<<2, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<2), ___sS32VECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___S32VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-s32vector k fill))
      v)))

(define-prim (##make-u32vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<2)) + 1;
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>(___LF+2)))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sU32VECTOR, n<<2, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<2), ___sU32VECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___U32VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-u32vector k fill))
      v)))

(define-prim (##make-s64vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
#if ___WS == 4
long words = ___WORDS((n<<3)) + 2;
#else
long words = ___WORDS((n<<3)) + 1;
#endif
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>(___LF+3)))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sS64VECTOR, n<<3, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
#if ___WS == 4
        result = ___TAG(___CAST(___WORD*,___CAST(___WORD,___hp+2)&~7)-1,
                        ___tSUBTYPED);
#else
        result = ___TAG(___hp, ___tSUBTYPED);
#endif
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<3), ___sS64VECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___S64VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-s64vector k fill))
      v)))

(define-prim (##make-u64vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
#if ___WS == 4
long words = ___WORDS((n<<3)) + 2;
#else
long words = ___WORDS((n<<3)) + 1;
#endif
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>(___LF+3)))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sU64VECTOR, n<<3, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
#if ___WS == 4
        result = ___TAG(___CAST(___WORD*,___CAST(___WORD,___hp+2)&~7)-1,
                        ___tSUBTYPED);
#else
        result = ___TAG(___hp, ___tSUBTYPED);
#endif
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<3), ___sU64VECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___U64VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-u64vector k fill))
      v)))

(define-prim (##make-f32vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<2)) + 1;
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>(___LF+2)))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sF32VECTOR, n<<2, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<2), ___sF32VECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    ___F64 fill = ___F64UNBOX(___ARG2);
    for (i=0; i<n; i++)
      ___F32VECTORSET(result,___FIX(i),fill)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-f32vector k fill))
      v)))

(define-prim (##make-f64vector k #!optional (fill (macro-absent-obj)))
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
#if ___WS == 4
long words = ___WORDS((n<<3)) + 2;
#else
long words = ___WORDS((n<<3)) + 1;
#endif
___WORD result;
if (n > (long)(unsigned long)(___LMASK>>(___LF+3)))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sF64VECTOR, n<<3, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
#if ___WS == 4
        result = ___TAG(___CAST(___WORD*,___CAST(___WORD,___hp+2)&~7)-1,
                        ___tSUBTYPED);
#else
        result = ___TAG(___hp, ___tSUBTYPED);
#endif
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<3), ___sF64VECTOR);
        ___hp += words;
      }
  }
if (!___FIXNUMP(result) && ___ARG2 != ___ABSENT)
  {
    ___F64 fill = ___F64UNBOX(___ARG2);
    for (i=0; i<n; i++)
      ___F64VECTORSET(result,___FIX(i),fill)
  }
___RESULT = result;
" k fill)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##make-f64vector k fill))
      v)))

(define-prim (##bignum.make k x complement?)
  (##declare (not interrupts-enabled))
  (let ((v (##c-code "
long i;
long n = ___INT(___ARG1);
#if ___BIG_ABASE_WIDTH == 32
long words = ___WORDS((n*(___BIG_ABASE_WIDTH/8))) + 1;
#else
#if ___WS == 4
long words = ___WORDS((n*(___BIG_ABASE_WIDTH/8))) + 2;
#else
long words = ___WORDS((n*(___BIG_ABASE_WIDTH/8))) + 1;
#endif
#endif
___WORD result;

if (n > (long)(unsigned long)(___LMASK>>___LF)/(___BIG_ABASE_WIDTH/8))
  result = ___FIX(___HEAP_OVERFLOW_ERR); /* requested object is too big! */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
#if ___BIG_ABASE_WIDTH == 32
    result = ___alloc_scmobj (___sBIGNUM, n<<2, ___STILL);
#else
    result = ___alloc_scmobj (___sBIGNUM, n<<3, ___STILL);
#endif
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (!___FIXNUMP(result))
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FIX(___HEAP_OVERFLOW_ERR);
    else
      {
#if ___BIG_ABASE_WIDTH == 32
        result = ___TAG(___hp, ___tSUBTYPED);
#else
#if ___WS == 4
        result = ___TAG(___CAST(___WORD*,___CAST(___WORD,___hp+2)&~7)-1,
                        ___tSUBTYPED);
#else
        result = ___TAG(___hp, ___tSUBTYPED);
#endif
#endif
#if ___BIG_ABASE_WIDTH == 32
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<2), ___sBIGNUM);
#else
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<3), ___sBIGNUM);
#endif
        ___hp += words;
      }
  }
if (!___FIXNUMP(result))
  {
    ___WORD x = ___ARG2;
    ___WORD len;
    if (x == ___FAL)
      len = 0;
    else
      {
        len = ___INT(___BIGALENGTH(x));
        if (len > n)
          len = n;
      }
#if ___BIG_ABASE_WIDTH == 32
    if (___ARG3 == ___FAL)
      {
        for (i=0; i<len; i++)
          ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,
                       ___FETCH_U32(___BODY_AS(x,___tSUBTYPED),i));
        if (x == ___FAL ||
            ___FETCH_U32(___BODY_AS(x,___tSUBTYPED),(i-1))<___BIG_ABASE_DIV_2)
          for (; i<n; i++)
            ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,0);
        else
          for (; i<n; i++)
            ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,___BIG_ABASE_MIN_1);
      }
    else
      {
        for (i=0; i<len; i++)
          ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,
                       ~___FETCH_U32(___BODY_AS(x,___tSUBTYPED),i));
        if (x == ___FAL ||
            ___FETCH_U32(___BODY_AS(x,___tSUBTYPED),(i-1))<___BIG_ABASE_DIV_2)
          for (; i<n; i++)
            ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,___BIG_ABASE_MIN_1);
        else
          for (; i<n; i++)
            ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,0);
      }
#else
    if (___ARG3 == ___FAL)
      {
        for (i=0; i<len; i++)
          ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,
                       ___FETCH_U64(___BODY_AS(x,___tSUBTYPED),i));
        if (x == ___FAL ||
            ___FETCH_U64(___BODY_AS(x,___tSUBTYPED),(i-1))<___BIG_ABASE_DIV_2)
          for (; i<n; i++)
            ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,0);
        else
          for (; i<n; i++)
            ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,___BIG_ABASE_MIN_1);
      }
    else
      {
        for (i=0; i<len; i++)
          ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,
                       ~___FETCH_U64(___BODY_AS(x,___tSUBTYPED),i));
        if (x == ___FAL ||
            ___FETCH_U64(___BODY_AS(x,___tSUBTYPED),(i-1))<___BIG_ABASE_DIV_2)
          for (; i<n; i++)
            ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,___BIG_ABASE_MIN_1);
        else
          for (; i<n; i++)
            ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,0);
      }
#endif
  }
___RESULT = result;
" k x complement?)))
    (if (##fixnum? v)
      (begin
        (##raise-heap-overflow-exception)
        (##bignum.make k x complement?))
      v)))

;------------------------------------------------------------------------------

; Apply.

(define-prim (##apply proc args)

  (##declare (not inline))

  (define (app proc args)
    (##declare (not interrupts-enabled))
    (##c-code
     "
     ___WORD proc;
     ___WORD args;
     ___WORD lst;
     int na;

     ___POP_ARGS2(proc,args)

     lst = args;
     na = 0;

     while (___PAIRP(lst))
       {
         ___PUSH(___CAR(lst))
         lst = ___CDR(lst);
         na++;

         if (na > ___MAX_NB_ARGS)
           {
             ___ADJFP(-na); /* remove pushed arguments */

             ___PUSH_ARGS2(proc,args)

             ___COVER_APPLY_ARGUMENT_LIMIT;

             ___JUMPPRM(___SET_NARGS(2),
                        ___G__23__23_raise_2d_number_2d_of_2d_arguments_2d_limit_2d_exception.prm)
           }
       }

     ___POP_ARGS_IN_REGS(na) /* load register arguments */

     ___COVER_APPLY_ARGUMENT_LIMIT_END;

     ___JUMPEXTNOTSAFE(___SET_NARGS(na),proc)
     "))

  (app proc args))

;------------------------------------------------------------------------------

; Continuation objects.

(define-prim (##continuation-ret cont)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   ___WORD frame = ___FIELD(___ARG1,___CONTINUATION_FRAME);
   ___WORD ra;

   if (___TYP(frame) == ___tSUBTYPED)
     {
       /* continuation frame is in the heap */

       ra = ___FIELD(frame,0);

       if (ra == ___internal_return)
         ra = ___FIELD(frame,___FRAME_RETI_RA);
     }
   else
     {
       /* continuation frame is in the stack */

       ___WORD *fp = ___CAST(___WORD*,frame);

       ra = fp[___FRAME_STACK_RA];

       if (ra == ___internal_return)
         ra = fp[-___RETI_RA];
     }

   ___RESULT = ra;
   "
   cont))

(define-prim (##continuation-denv cont)
  (##declare (not interrupts-enabled))
  (macro-continuation-denv cont))

(define-prim (##continuation-fs cont)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   ___WORD frame = ___FIELD(___ARG1,___CONTINUATION_FRAME);
   ___WORD ra;
   int fs;

   if (___TYP(frame) == ___tSUBTYPED)
     {
       /* continuation frame is in the heap */

       ra = ___FIELD(frame,0);

       if (ra == ___internal_return)
         ___RETI_GET_FS(___FIELD(frame,___FRAME_RETI_RA),fs)
       else
         ___RETN_GET_FS(ra,fs)
     }
   else
     {
       /* continuation frame is in the stack */

       ___WORD *fp = ___CAST(___WORD*,frame);

       ra = fp[___FRAME_STACK_RA];

       if (ra == ___internal_return)
         ___RETI_GET_FS(fp[-___RETI_RA],fs)
       else
         ___RETN_GET_FS(ra,fs)
     }

   ___RESULT = ___FIX(fs);
   "
   cont))

(define-prim (##continuation-ref cont i)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   ___WORD frame = ___FIELD(___ARG1,___CONTINUATION_FRAME);
   int i = ___INT(___ARG2);
   ___WORD ra;
   int fs, link;

   if (___TYP(frame) == ___tSUBTYPED)
     {
       /* continuation frame is in the heap */

       ra = ___FIELD(frame,0);

       if (ra == ___internal_return)
         ___RETI_GET_FS_LINK(___BODY_AS(frame,___tSUBTYPED)[___FRAME_RETI_RA],fs,link)
       else
         ___RETN_GET_FS_LINK(ra,fs,link)

       ___RESULT = ___BODY_AS(frame,___tSUBTYPED)[fs-i+1];  /* what if i==link and frame is first in section???? */
 /*      if (i == link) ___RESULT = ___FIX(999999);/***********/
     }
   else
     {
       /* continuation frame is in the stack */

       ra = ___CAST(___WORD*,frame)[___FRAME_STACK_RA];

       if (ra == ___internal_return)
         ___RETI_GET_FS_LINK(___CAST(___WORD*,frame)[-___RETI_RA],fs,link)
       else
         ___RETN_GET_FS_LINK(ra,fs,link)

       ___RESULT = ___CAST(___WORD*,frame)[___FRAME_SPACE(fs)-i];  /* what if i==link and frame is first in section???? */
 /*      if (i == link) ___RESULT = ___FIX(999999);/***********/
     }
   "
   cont
   i))

(define-prim (##continuation-next cont)
  (##declare (not interrupts-enabled))
  (let ((next
         (##c-code
          "
          ___WORD f = ___ARG1;
          ___WORD frame = ___FIELD(f,___CONTINUATION_FRAME);
          ___WORD denv  = ___FIELD(f,___CONTINUATION_DENV);
          ___WORD ra;
          ___WORD *fp, frame_ra, next_frame;
          int fs, link;

          if (___TYP(frame)==___tSUBTYPED)
            {
              /* continuation frame is in the heap */

              ra = ___FIELD(frame,0);

              fp = ___BODY_AS(frame,___tSUBTYPED);

              if (ra == ___internal_return)
                ___RETI_GET_FS_LINK(fp[___FRAME_RETI_RA],fs,link)
              else
                ___RETN_GET_FS_LINK(ra,fs,link)

              fp += fs+1;

              next_frame = fp[-link-1];

              if (next_frame == 0)
                ___RESULT = ___FAL;
              else
                {
                  ___hp[0]=___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTTHREAD);
                  ___ADD_VECTOR_ELEM(0,next_frame)
                  ___ADD_VECTOR_ELEM(1,denv)
                  ___hp+=___CONTINUATION_SIZE+1;
                  ___RESULT = ___GET_VECTOR(___CONTINUATION_SIZE);
                }
            }
          else
            {
              /* continuation frame is in the stack */

              ra = ___CAST(___WORD*,frame)[___FRAME_STACK_RA];

              if (ra == ___internal_return)
                ___RETI_GET_FS_LINK(___CAST(___WORD*,frame)[-___RETI_RA],fs,link)
              else
                ___RETN_GET_FS_LINK(ra,fs,link)

              fp = ___CAST(___WORD*,frame)+___FRAME_SPACE(fs);
              frame_ra = fp[-link-1];

              if (frame_ra == ___ps->handler_break)
                {
                  /* first frame of that section */

                  next_frame = fp[___BREAK_FRAME_NEXT];
                }
              else
                {
                  /* not the first frame of that section */

                  *fp = frame_ra;
                  next_frame = ___CAST(___WORD,fp);
                }

              if (next_frame == 0)
                ___RESULT = ___FAL;
              else
                {
                  ___hp[0]=___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTTHREAD);
                  ___ADD_VECTOR_ELEM(0,next_frame)
                  ___ADD_VECTOR_ELEM(1,denv)
                  ___hp+=___CONTINUATION_SIZE+1;
                  ___RESULT = ___GET_VECTOR(___CONTINUATION_SIZE);
                }
            }
          "
          cont)))
        (##check-heap)
        next))

;------------------------------------------------------------------------------

; Closures and subprocedures.

(define-prim (##closure? proc)
  (##declare (not interrupts-enabled))
  (##c-code
   "___RESULT = ___BOOLEAN(___HD_TYP(___HEADER(___ARG1)) != ___PERM);"
   proc))

(define-prim (##closure-code closure))
(define-prim (##closure-ref closure index))
(define-prim (##closure-set! closure index val))

(define-prim (##subprocedure? proc)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   if (___TYP(___ARG1) == ___tSUBTYPED &&
       ___CAST(___label_struct*,___ARG1-___tSUBTYPED)->entry_or_descr == ___ARG1 &&
       ___CAST(___WORD*,___ARG1-___tSUBTYPED)[-___LS] != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM))
     ___RESULT = ___TRU;
   else
     ___RESULT = ___FAL;
   "
   proc))

(define-prim (##subprocedure-id proc)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   if (___TYP(___ARG1) == ___tSUBTYPED)
     {
       ___WORD *start = ___CAST(___WORD*,___ARG1-___tSUBTYPED);
       ___WORD *ptr = start;
       while (*ptr != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM))
         ptr -= ___LS;
       ptr += ___LS;
       ___RESULT = ___FIX( (start-ptr)/___LS );
     }
   else
     ___RESULT = ___FIX(0);
   "
   proc))

(define-prim (##subprocedure-parent proc)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   if (___TYP(___ARG1) == ___tSUBTYPED)
     {
       ___WORD *start = ___CAST(___WORD*,___ARG1-___tSUBTYPED);
       ___WORD *ptr = start;
       while (*ptr != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM))
         ptr -= ___LS;
       ptr += ___LS;
      ___RESULT = ___TAG(ptr,___tSUBTYPED);
     }
   else
     ___RESULT = ___FAL;
   "
   proc))

(define-prim (##procedure-info proc)
  (##declare (not interrupts-enabled))
  (##c-code
   "
   if (___TYP(___ARG1) == ___tSUBTYPED)
     {
       ___WORD *start = ___CAST(___WORD*,___ARG1-___tSUBTYPED);
       ___WORD *ptr = start;
       while (*ptr != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM))
         ptr -= ___LS;
       ___RESULT = ptr[1];
     }
   else
     ___RESULT = ___FAL;
   "
   proc))

;------------------------------------------------------------------------------

; Global variables.

(define-prim (##make-global-var id)
  (##declare (not interrupts-enabled))
  (let ((gv
         (##c-code
          "
          if (___FIELD(___ARG1,___SYMBOL_GLOBAL) == ___FIX(0))
            {
              ___glo_struct *p;
              ___ERR_CODE err = ___alloc_global_var (&p);
              if (err != ___NO_ERR)
                ___RESULT = ___FIX(err);
              else
                {
                  p->val = ___UNB1;
                  p->prm = ___FAL;
                  p->next = 0;
                  if (___ps->glo_list_head == 0)
                    ___ps->glo_list_head = ___CAST(___WORD,p);
                  else
                    ___CAST(___glo_struct*,___ps->glo_list_tail)->next
                      = ___CAST(___WORD,p);
                  ___ps->glo_list_tail = ___CAST(___WORD,p);
                  ___FIELD(___ARG1,___SYMBOL_GLOBAL) = ___CAST(___WORD,p);
                  ___RESULT = ___ARG1;
                }
            }
          else
            ___RESULT = ___ARG1;
          "
          id)))
    (if (##fixnum? gv)
      (begin
        (##raise-heap-overflow-exception)
        (##make-global-var id))
      gv)))

(define-prim (##global-var-ref gv))
(define-prim (##global-var-set! gv val))

(define-prim (##object->global-var->identifier obj)
  (##c-code
   "
   ___WORD p = ___ps->glo_list_head;
   while (p != 0 && ___CAST(___glo_struct*,p)->val != ___ARG1)
     p = ___CAST(___glo_struct*,p)->next;
   ___RESULT = ___FAL;
   if (p != 0)
     {
       int len = ___INT(___VECTORLENGTH(___symbol_table));
       int i;

       for (i=0; i<len; i++)
         {
           ___WORD probe = ___FIELD(___symbol_table,i);

           while (probe != ___NUL)
             {
               if (___FIELD(probe,___SYMBOL_GLOBAL) == p)
                 {
                   ___RESULT = probe;
                   goto end_search;
                 }
               probe = ___FIELD(probe,___SYMKEY_NEXT);
             }
         }
       end_search:;
     }
   "
   obj))

(define-prim (##global-var->identifier gv)
  gv)

;------------------------------------------------------------------------------

; Symbols and keywords.

(define-prim (##make-interned-symbol name)
  (##make-interned-symkey name #t))

(define-prim (##make-interned-keyword name)
  (##make-interned-symkey name #f))

(define-prim (##make-interned-symkey name symbol?)
  (let ((result
         ((c-lambda (scheme-object
                     scheme-object)
                    scheme-object
           "
           unsigned int subtype = (___arg2 != ___FAL)
                                  ? ___sSYMBOL
                                  : ___sKEYWORD;
           ___WORD obj = ___find_symkey_from_schemestring
                           (___arg1,
                            subtype);
           if (obj == ___FAL)
             {
               unsigned long n = ___INT(___STRINGLENGTH(___arg1));
               obj = ___alloc_scmobj (___sSTRING,
                                      n<<___LCS,
                                      ___PERM);
               if (!___FIXNUMP(obj))
                 {
                   memcpy (___BODY_AS(obj,___tSUBTYPED),
                           ___BODY_AS(___arg1,___tSUBTYPED),
                           n<<___LCS);
                   obj = ___new_symkey (obj, subtype);
                 }
             }
           ___result = obj;
           ")
          name
          symbol?)))
    (if (##fixnum? result)
      (begin
        (##raise-heap-overflow-exception)
        (##make-interned-symkey name symbol?))
      result)))

(define ##symbol-counter 0)

(define-prim (##make-uninterned-symbol name)

  ; name must be a nonmutable string

  (let ((new-count (##fixnum.+ ##symbol-counter 1)))
    ; Note: it is unimportant if the increment of ##symbol-counter
    ; is not atomic; it simply means a possible close repetition
    ; of the same hash code
    (set! ##symbol-counter new-count)
    (macro-make-uninterned-symbol
     name
     (##partial-bit-reverse new-count))))

(define ##keyword-counter 0)

(define-prim (##make-uninterned-keyword name)

  ; name must be a nonmutable string

  (let ((new-count (##fixnum.+ ##keyword-counter 1)))
    ; Note: it is unimportant if the increment of ##keyword-counter
    ; is not atomic; it simply means a possible close repetition
    ; of the same hash code
    (set! ##keyword-counter new-count)
    (macro-make-uninterned-keyword
     name
     (##partial-bit-reverse new-count))))

(define-prim (##partial-bit-reverse i)

  (##define-macro (bit n)
    `(##fixnum.arithmetic-shift-left
      (##fixnum.bitwise-and i ,(expt 2 n)) ,(- 28 (* 2 n))))

  (##fixnum.+
   (bit 0)
   (##fixnum.+
    (bit 1)
    (##fixnum.+
     (bit 2)
     (##fixnum.+
      (bit 3)
      (##fixnum.+
       (bit 4)
       (##fixnum.+
        (bit 5)
        (##fixnum.+
         (bit 6)
         (##fixnum.+
          (bit 7)
          (##fixnum.+
           (bit 8)
           (##fixnum.+
            (bit 9)
            (##fixnum.+
             (bit 10)
             (##fixnum.+
              (bit 11)
              (##fixnum.+
               (bit 12)
               (##fixnum.+
                (bit 13)
                (bit 14))))))))))))))))

;------------------------------------------------------------------------------

; Filesystem path manipulation and device "select" routine.

(define-prim (##path-expand path format)
  (##declare (not interrupts-enabled))
  ((c-lambda (nonnull-char-string
              int)
             char-string
    "
    char path[___PATH_MAX_LENGTH+1];
    if (___path_expand (___arg1, path, ___PATH_MAX_LENGTH, ___arg2))
      ___result = path;
    else
      ___result = 0;
    ")
   path
   (case format
     ((relative) 1)
     ((shortest) -1)
     (else 0))))

(define-prim ##path-absolute?
  (c-lambda (nonnull-char-string)
            bool
   "___path_absolute"))

(define-prim ##path-extension
  (c-lambda (nonnull-char-string)
            char-string
   "
   char path[___PATH_MAX_LENGTH+1];
   if (___path_extension (___arg1, path, ___PATH_MAX_LENGTH))
     ___result = path;
   else
     ___result = 0;
   "))

(define-prim ##path-strip-extension
  (c-lambda (nonnull-char-string)
            char-string
   "
   char path[___PATH_MAX_LENGTH+1];
   if (___path_strip_extension (___arg1, path, ___PATH_MAX_LENGTH))
     ___result = path;
   else
     ___result = 0;
   "))

(define-prim ##path-directory
  (c-lambda (nonnull-char-string)
            char-string
   "
   char path[___PATH_MAX_LENGTH+1];
   if (___path_directory (___arg1, path, ___PATH_MAX_LENGTH))
     ___result = path;
   else
     ___result = 0;
   "))

(define-prim ##path-strip-directory
  (c-lambda (nonnull-char-string)
            char-string
   "
   char path[___PATH_MAX_LENGTH+1];
   if (___path_strip_directory (___arg1, path, ___PATH_MAX_LENGTH))
     ___result = path;
   else
     ___result = 0;
   "))

(define-prim ##path-drive
  (c-lambda (nonnull-char-string)
            char-string
   "
   char path[___PATH_MAX_LENGTH+1];
   if (___path_drive (___arg1, path, ___PATH_MAX_LENGTH))
     ___result = path;
   else
     ___result = 0;
   "))

(define-prim ##path-strip-drive
  (c-lambda (nonnull-char-string)
            char-string
   "
   char path[___PATH_MAX_LENGTH+1];
   if (___path_strip_drive (___arg1, path, ___PATH_MAX_LENGTH))
     ___result = path;
   else
     ___result = 0;
   "))

(define-prim (##condvar-select! runq timeout)
  (##declare (not interrupts-enabled))
  (##c-code
   "___RESULT = ___condvar_select (___ARG1, ___ARG2);"
   runq
   timeout))

;------------------------------------------------------------------------------

;;;;;;;;;;;;;******************

(define-prim ##getenv
  (c-lambda (nonnull-char-string)
            char-string
   "___getenv"))

(define-prim ##environ
  (c-lambda ()
            nonnull-char-string-list
   "___environ"))

(define-prim ##shell-command
  (c-lambda (nonnull-char-string)
            int
   "___shell_command"))

;------------------------------------------------------------------------------

; Provide access to low-level I/O operations implemented in "os.c".

(define-prim ##dev-kind
  (c-lambda (scheme-object)
            scheme-object
   "___dev_kind"))

(define-prim ##dev-force-output
  (c-lambda (scheme-object)
            scheme-object
   "___dev_force_output"))

(define-prim ##dev-close
  (c-lambda (scheme-object
             scheme-object)
            scheme-object
   "___dev_close"))

(define-prim ##dev-stream-open-predefined
  (c-lambda (scheme-object
             scheme-object)
            scheme-object
   "___dev_stream_open_predefined"))

(define-prim ##dev-stream-open-path
  (c-lambda (scheme-object
             scheme-object
             scheme-object)
            scheme-object
   "___dev_stream_open_path"))

(define-prim ##dev-stream-open-process
  (c-lambda (scheme-object
             scheme-object
             scheme-object)
            scheme-object
   "___dev_stream_open_process"))

(define-prim ##dev-stream-read
  (c-lambda (scheme-object
             scheme-object
             scheme-object
             scheme-object)
            scheme-object
   "___dev_stream_read"))

(define-prim ##dev-stream-write
  (c-lambda (scheme-object
             scheme-object
             scheme-object
             scheme-object)
            scheme-object
   "___dev_stream_write"))

(define-prim ##dev-stream-width
  (c-lambda (scheme-object)
            scheme-object
   "___dev_stream_width"))

(define-prim ##dev-stream-default-options
  (c-lambda (scheme-object)
            scheme-object
   "___dev_stream_default_options"))

(define-prim ##dev-tty-type-set!
  (c-lambda (scheme-object
             scheme-object
             scheme-object)
            scheme-object
   "___dev_tty_type_set"))

(define-prim ##dev-tty-text-attributes-set!
  (c-lambda (scheme-object
             scheme-object
             scheme-object)
            scheme-object
   "___dev_tty_text_attributes_set"))

(define-prim ##dev-tty-history
  (c-lambda (scheme-object)
            scheme-object
   "___dev_tty_history"))

(define-prim ##dev-tty-history-set!
  (c-lambda (scheme-object
             scheme-object)
            scheme-object
   "___dev_tty_history_set"))

(define-prim ##dev-tty-max-history-length-set!
  (c-lambda (scheme-object
             scheme-object)
            scheme-object
   "___dev_tty_max_history_length_set"))

(define-prim ##dev-tty-paren-balance-duration-set!
  (c-lambda (scheme-object
             scheme-object)
            scheme-object
   "___dev_tty_paren_balance_duration_set"))

(define-prim ##dev-tty-mode-set!
  (c-lambda (scheme-object
             scheme-object)
            scheme-object
   "___dev_tty_mode_set"))

(define-prim ##netdb-get-host-by-name
  (c-lambda (scheme-object)
            scheme-object
   "___netdb_get_host_by_name"))

(define-prim ##dev-tcp-client-open
  (c-lambda (scheme-object
             scheme-object
             scheme-object)
            scheme-object
   "___dev_tcp_client_open"))

(define-prim ##dev-tcp-server-open
  (c-lambda (scheme-object
             scheme-object
             scheme-object)
            scheme-object
   "___dev_tcp_server_open"))

(define-prim ##dev-tcp-server-read
  (c-lambda (scheme-object)
            scheme-object
   "___dev_tcp_server_read"))

(define-prim ##dev-directory-open-path
  (c-lambda (scheme-object)
            scheme-object
   "___dev_directory_open_path"))

(define-prim ##dev-directory-read
  (c-lambda (scheme-object)
            scheme-object
   "___dev_directory_read"))

(define-prim ##port-decode-chars!
  (c-lambda (scheme-object
             scheme-object)
            scheme-object
   "___port_decode_chars"))

(define-prim ##port-encode-chars!
  (c-lambda (scheme-object)
            scheme-object
   "___port_encode_chars"))

;------------------------------------------------------------------------------

; Dynamic compilation and dynamic loading.

(define ##cc-cmd
  ((c-lambda ()
             char-string
    "___result = ___CC_CMD;")))

(define ##ld-cmd
  ((c-lambda ()
             char-string
    "___result = ___LD_CMD;")))

(define ##ld-flip
  ((c-lambda ()
             bool
    "
#ifdef ___LD_FLIP
    ___result = 1;
#else
    ___result = 0;
#endif
    ")))

(define ##dynamic-cc #f)

(if ##cc-cmd
  (set! ##dynamic-cc
    (lambda (root-path output-path cc-options ld-options)
      ((c-lambda (nonnull-char-string
                  nonnull-char-string
                  nonnull-char-string
                  nonnull-char-string
                  nonnull-char-string
                  nonnull-char-string
                  bool)
                 char-string
        "
        char *errmsg;
        if (___dynamic_cc (___arg1,
                           ___arg2,
                           ___arg3,
                           ___arg4,
                           ___arg5,
                           ___arg6,
                           ___arg7,
                           &errmsg))
          ___result = errmsg;
        else
          ___result = 0;
        ")
       ##cc-cmd
       ##ld-cmd
       root-path
       output-path
       cc-options
       ld-options
       ##ld-flip))))

(define-prim ##load-object-file
  (c-lambda (scheme-object)
            scheme-object
   "___load_object_file"))

;------------------------------------------------------------------------------

; Program startup and exit.

(define-prim (##exit #!optional (status 0))
  (##execute-and-clear-jobs! ##exit-jobs)
  (##execute-final-wills!)
  (##cleanup-and-exit-process status))

(define-prim (##cleanup-and-exit-process #!optional (status 0))
  (##declare (not interrupts-enabled))
  (##c-code
   "___cleanup_and_exit_process (___INT(___ARG1)); ___RESULT = ___VOID;"
   status))

(define-prim (##execute-modules exec-vector i)
  (let loop ((i i))
    (let ((len (##vector-length exec-vector)))
      (cond ((##fixnum.= i (##fixnum.- len 1))
             ((##vector-ref exec-vector i)))
            ((##fixnum.< i len)
             ((##vector-ref exec-vector i))
             (loop (##fixnum.+ i 1)))))))

; (##add-exit-job! thunk) can be called to add a job to
; do when the program exits.  (##clear-exit-jobs!) clears the jobs.

(define ##exit-jobs (##make-jobs))

(define-prim (##add-exit-job! thunk)
  (##add-job! ##exit-jobs thunk))

(define-prim (##clear-exit-jobs!)
  (##clear-jobs! ##exit-jobs))

; The kernel is responsible for executing each module of the program
; in sequence:

(define-prim (##execute-program)
  (let ((program-descr
         (##c-code
          "
          ___release_scmobj (___program_descr);
          ___RESULT = ___program_descr;
          ")))
    (let ((exec-vector (##vector-ref program-descr 0)))
      (##execute-modules exec-vector 1))))

(##execute-program)

;==============================================================================
