;==============================================================================

; file: "_nonstd.scm"

; Copyright (C) 1994-2002 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##include "header.scm")

;==============================================================================

; Implementation of exceptions.

(define-procedures-for-error-exception)

;------------------------------------------------------------------------------

; Define type checking procedures.

(define-fail-check-type will)
(define-fail-check-type box)

;------------------------------------------------------------------------------

; Non-standard procedures

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(##define-macro (define-runtime-macro pattern . rest)

  (define (form-size parms) ; this definition must match the one in "_eval.scm"
    (let loop ((l parms) (n 1))
      (if (pair? l)
        (loop (cdr l) (+ n 1))
        (if (null? l) n (- 0 n)))))

  `(##top-cte-add-macro!
     ##interaction-cte
     ',(car pattern)
     (##cons ',(form-size (cdr pattern))
             (lambda ,(cdr pattern) ,@rest))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (exit #!optional (status (macro-absent-obj)))
  (if (##eq? status (macro-absent-obj))
    (##exit)
    (macro-force-vars (status)
      (##exit status))))

(define (error message . arguments)
  (macro-raise (macro-make-error-exception message arguments)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (include filename)
  `(##include ,filename))

;*********************** bug: (define-macro (make-improper z) `(define ,z '(1 . 2)))
(define-runtime-macro (define-macro pattern body . rest)
  `(##define-macro ,pattern ,body ,@rest))

(define-runtime-macro (declare . rest)
  `(##declare ,@rest))

(define-runtime-macro (namespace . rest)
  `(##namespace ,@rest))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (parameterize bindings body . rest)
  (##parameterize-build
   (##cons 'parameterize (##cons bindings (##cons body rest)))))

(define (##parameterize-build src)

  (define (build src bindings body rev-params-vals)
    (cond ((##pair? bindings)
           (let ((binding (##car bindings)))
             (##shape src (##sourcify binding src) 2)
             (let* ((param (##car binding))
                    (val (##cadr binding)))
               (build src
                      (##cdr bindings)
                      body
                      (##cons (##cons (##cons (##gensym) param)
                                      (##cons (##gensym) val))
                              rev-params-vals)))))
          ((##null? bindings)
           (if (##null? rev-params-vals)
             (##cons 'let (##cons '() body))
             (let ((params-vals (##reverse rev-params-vals)))

               (define (bind params-vals)
                 (if (##null? params-vals)
                   (##cons 'let (##cons '() body))
                   (let* ((param-val (##car params-vals))
                          (param (##car param-val))
                          (val (##cdr param-val)))
                     (##list '##parameterize
                             (##car param)
                             (##car val)
                             (##list 'lambda
                                     '()
                                     (bind (##cdr params-vals)))))))

               (##list 'let
                       (let loop ((lst rev-params-vals) (bs '()))
                         (if (##null? lst)
                           bs
                           (let* ((param-val (##car lst))
                                  (param (##car param-val))
                                  (val (##cdr param-val)))
                             (loop (##cdr lst)
                                   (##cons (##list (##car param)
                                                   (##cdr param))
                                           (##cons (##list (##car val)
                                                           (##cdr val))
                                                   bs))))))
                       (##list 'if
                               (##cons 'and
                                       (##map (lambda (param-val)
                                                (let ((param
                                                       (##car param-val)))
                                                  (##list '##parameter?
                                                          (##car param))))
                                              params-vals))
                               (bind params-vals)
                               '(##raise-type-exception-parameter))))))
          (else
           (##raise-expression-parsing-exception
            'ill-formed-binding-list
            src))))

  (build (##sourcify src (##make-source #f #f))
         (##cadr src)
         (##cddr src)
         '()))

(define (##raise-type-exception-parameter)
  (error "PARAMETER expected"));;;;;;;;;;;;;;;;;;;;;

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (cond-expand first-clause . other-clauses)
  (##cond-expand-build
   (##cons 'cond-expand (##cons first-clause other-clauses))))

(define-prim (##cond-expand-build src)

  (define (satisfied? src feature-requirement)
    (cond ((##symbol? feature-requirement)
           (if (##member feature-requirement ##cond-expand-features)
             #t
             #f))
          ((##pair? feature-requirement)
           (let ((first (##car feature-requirement)))
             (cond ((##eq? first 'not)
                    (##shape src (##sourcify feature-requirement src) 2)
                    (##not (satisfied? src (##cadr feature-requirement))))
                   ((or (##eq? first 'and) (##eq? first 'or))
                    (##shape src (##sourcify feature-requirement src) -1)
                    (let loop ((lst (##cdr feature-requirement)))
                      (if (##pair? lst)
                        (let ((x (##car lst)))
                          (if (##eq? (satisfied? src x) (##eq? first 'and))
                            (loop (##cdr lst))
                            (##not (##eq? first 'and))))
                        (##eq? first 'and))))
                   (else
                    (macro-raise
                     (macro-make-expression-parsing-exception
                      'ill-formed-cond-expand
                      src
                      '()))))))
          (else
           (macro-raise
            (macro-make-expression-parsing-exception
             'ill-formed-cond-expand
             src
             '())))))

  (define (build src clauses)
    (if (##pair? clauses)
      (let ((clause (##car clauses)))
        (##shape src (##sourcify clause src) -1)
        (let ((feature-requirement (##car clause)))
          (if (or (and (##eq? feature-requirement 'else)
                       (##null? (##cdr clauses)))
                  (satisfied? src feature-requirement))
            (##cons 'begin (##cdr clause))
            (build src (##cdr clauses)))))
      (macro-raise
       (macro-make-expression-parsing-exception
        'unfulfilled-cond-expand
        src
        '()))))

  (build (##sourcify src (##make-source #f #f))
         (##cdr src)))

(define ##cond-expand-features #f)
(set! ##cond-expand-features
  '(srfi-0 srfi-4 srfi-6 srfi-8 srfi-18 srfi-21 srfi-22 srfi-23))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (receive formals expression first-body . other-body)
  `(##call-with-values
    (lambda () ,expression)
    (lambda ,formals ,first-body ,@other-body)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (define-structure name . field-defs)

  (define (err)
    (##raise-expression-parsing-exception
     'ill-formed-special-form
     (##sourcify
      (##cons 'define-structure (##cons name field-defs))
      (##make-source #f #f))
     'define-structure))

  (define (sym . strings)
    (##string->symbol (##apply ##string-append strings)))

  (if (##symbol? name)
    (let ((name-str (##symbol->string name)))

      (define (generate-accessors fields i tag)
        (if (##null? fields)
          '()
          (let* ((field (##car fields))
                 (field-str (##symbol->string field))
                 (field-ref (sym name-str "-" field-str))
                 (field-set! (sym name-str "-" field-str "-set!")))
            (##cons `(define ,field-set!
                       (letrec ((,field-set!
                                 (lambda (x y)
                                   (##declare (extended-bindings))
                                   (##structure-set! x
                                                     ,i
                                                     y
                                                     ',tag
                                                     ,field-set!))))
                         ,field-set!))
                    (##cons `(define ,field-ref
                               (letrec ((,field-ref
                                         (lambda (x)
                                           (##declare (extended-bindings))
                                           (##structure-ref x
                                                            ,i
                                                            ',tag
                                                            ,field-ref))))
                                 ,field-ref))
                            (generate-accessors (##cdr fields)
                                                (##fixnum.+ i 1)
                                                tag))))))

      (define (generate-definitions rev-all-fields rev-printed-fields)
        (let* ((all-fields (##reverse rev-all-fields))
               (printed-fields (##reverse rev-printed-fields))
               (tag (##list->vector
                      (##cons name
                              (##cons (##length printed-fields)
                                      all-fields)))))
          `(begin
             ,@(generate-accessors all-fields 1 tag)
             (define ,(sym name-str "?")
               (lambda (x)
                 (##declare (extended-bindings))
                 (and (##structure? x)
                      (##eq? (##vector-ref x 0) ',tag))))
             (define ,(sym "make-" name-str)
               (lambda (,@all-fields)
                 (##declare (extended-bindings))
                 (##subtype-set! (##vector ',tag ,@all-fields)
                                 ,(macro-subtype-structure)))))))

      (let loop1 ((l1 field-defs) (l2 '()))
        (if (##pair? l1)
          (let ((rest (##cdr l1)) (field (##car l1)))
            (cond ((##symbol? field)
                   (loop1 rest (##cons field l2)))
                  ((and (or (##null? field) (##pair? field))
                        (##not (##pair? rest)))
                   (let ((printed-fields l2))
                     (let loop2 ((l1 field) (l2 l2))
                       (if (##pair? l1)
                         (let ((rest (##cdr l1)) (field (##car l1)))
                           (cond ((##symbol? field)
                                  (loop2 rest (##cons field l2)))
                                  (else
                                   (err))))
                         (generate-definitions l2 printed-fields)))))
                  (else
                   (err))))
          (generate-definitions l2 l2))))
    (err)))

(define (##structure-ref obj i tag proc)
  (if (and (##structure? obj)
           (##eq? (##vector-ref obj 0) tag))
    (##vector-ref obj i)
    (##raise-type-exception 1 tag proc (##list obj))))

(define (##structure-set! obj i v tag proc)
  (if (and (##structure? obj)
           (##eq? (##vector-ref obj 0) tag))
    (begin (##vector-set! obj i v) (##void))
    (##raise-type-exception 1 tag proc (##list obj v))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (set-gc-report! report?);;;;;;;;;;;;;;;;;;;;;;
  (##set-gc-report! report?))

(define ##gc-report? #f)

(define-prim (##set-gc-report! report?)
  (##declare (not interrupts-enabled))
  (set! ##gc-report? (if report? #t #f)))

(define-prim (##display-gc-report)
  (let ((port ##stderr-port))
    (if (and port
             (let* ((settings
                     (##set-debug-settings! 0 0))
                    (level
                     (macro-debug-settings-level settings)))
               (or (##fixnum.< 1 level)
                   ##gc-report?)))
      (let* ((stats
              (##process-statistics))
             (bytes-allocated
              (##f64vector-ref stats 7))
             (last-gc-real-time
              (##f64vector-ref stats 12))
             (last-gc-heap-size
              (##f64vector-ref stats 13))
             (last-gc-live
              (##f64vector-ref stats 14))
             (last-gc-movable
              (##f64vector-ref stats 15))
             (last-gc-nonmovable
              (##f64vector-ref stats 16)))

        (define (scale x m)
          (##flonum.->exact-int (##flonum.round (##flonum.* x m))))

        (##write-string "*** GC: " port)
        (##write (scale last-gc-real-time 1000.0) port)
        (##write-string " ms, " port)
        (##write (scale bytes-allocated 9.765625e-4) port)
        (##write-string "K alloc, " port)
        (##write (scale last-gc-heap-size 9.765625e-4) port)
        (##write-string "K heap, " port)
        (##write (scale last-gc-live 9.765625e-4) port)
        (##write-string "K live (" port)
        (##write (scale (##flonum./ last-gc-live last-gc-heap-size) 100.0) port)
        (##write-string "% " port)
        (##write (##flonum.->exact-int last-gc-movable) port)
        (##write-string "+" port)
        (##write (##flonum.->exact-int last-gc-nonmovable) port)
        (##write-string ")" port)
        (##newline port)))))

(##add-gc-interrupt-job! ##display-gc-report)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (##void))

(define (void)
  (##void))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (process-times)
  (##process-times))

(define (cpu-time)
  (let ((v (##process-times)))
    (##+ (##f64vector-ref v 0) (##f64vector-ref v 1))))

(define (real-time)
  (let ((v (##process-times)))
    (##f64vector-ref v 2)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (time expr)
  `(##time (lambda () ,expr) ',expr))

(define (##time thunk expr)
  (let ((at-start (##process-statistics)))
    (let ((result (thunk)))
      (let ((at-end (##process-statistics)))

        (define (secs->msecs x)
          (##inexact->exact (##round (##* x 1000))))

        (let* ((out
                (##repl-out))
               (user-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 0)
                      (##f64vector-ref at-start 0))))
               (sys-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 1)
                      (##f64vector-ref at-start 1))))
               (cpu-time
                (##+ user-time sys-time))
               (real-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 2)
                      (##f64vector-ref at-start 2))))
               (gc-user-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 3)
                      (##f64vector-ref at-start 3))))
               (gc-sys-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 4)
                      (##f64vector-ref at-start 4))))
               (gc-real-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 5)
                      (##f64vector-ref at-start 5))))
               (nb-gcs
                (##flonum.->exact-int
                 (##- (##f64vector-ref at-end 6)
                      (##f64vector-ref at-start 6))))
               (minflt
                (##flonum.->exact-int
                 (##- (##f64vector-ref at-end 10)
                      (##f64vector-ref at-start 10))))
               (majflt
                (##flonum.->exact-int
                 (##- (##f64vector-ref at-end 11)
                      (##f64vector-ref at-start 11))))
               (bytes-allocated
                (##flonum.->exact-int
                 (##- (##- (##f64vector-ref at-end 7)
                           (##f64vector-ref at-start 7))
                      (##+ (if (##interp-procedure? thunk)
                             (##f64vector-ref at-end 8) ; thunk call frame space
                             (macro-inexact-+0))
                           (##f64vector-ref at-end 9)))))) ; at-end structure space

          (define (pluralize n msg)
            (##write-string "    " out)
            (if (##= n 0)
              (##write-string "no" out)
              (##write n out))
            (##write-string msg out)
            (if (##not (##= n 1))
              (##write-string "s" out)))

          (##write (##list 'time expr) out)
          (##newline out)

          (##write-string "    " out)
          (##write real-time out)
          (##write-string " ms real time" out)
          (##newline out)

          (##write-string "    " out)
          (##write cpu-time out)
          (##write-string " ms cpu time (" out)
          (##write user-time out)
          (##write-string " user, " out)
          (##write sys-time out)
          (##write-string " system)" out)
          (##newline out)

          (pluralize nb-gcs " collection")
          (if (##not (##= nb-gcs 0))
            (begin
              (##write-string " accounting for " out)
              (##write gc-real-time out)
              (##write-string " ms real time (" out)
              (##write gc-user-time out)
              (##write-string " user, " out)
              (##write gc-sys-time out)
              (##write-string " system)" out)))
          (##newline out)

          (pluralize bytes-allocated " byte")
          (##write-string " allocated" out)
          (##newline out)

          (pluralize minflt " minor fault")
          (##newline out)

          (pluralize majflt " major fault")
          (##newline out)

          result)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##gensym-counter -1)

(define (##gensym #!optional (p (macro-absent-obj)))
  (let ((prefix
         (if (##eq? p (macro-absent-obj))
           'g
           p)))
    (macro-check-symbol prefix 1 (gensym p)
      (let ((new-count
             (##fixnum.modulo
              (##fixnum.+ ##gensym-counter 1)
              1000000)))
        ; Note: it is unimportant if the increment of ##gensym-counter
        ; is not atomic; it simply means a possible close repetition
        ; of the same name
        (set! ##gensym-counter new-count)
        (##make-uninterned-symbol
         (if (##eq? prefix 'g)
           new-count ; ##symbol->string will create the string
           (##string-append (##symbol->string prefix)
                            (##number->string new-count 10))))))))

(define (gensym #!optional (p (macro-absent-obj)))
  (macro-force-vars (p)
    (##gensym p)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(##define-macro (macro-will-size) 3)

(define-prim (##will? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-willtable))
       (##fixnum.= (##vector-length obj) (macro-will-size))))

(define-prim (will? x)
  (macro-force-vars (x)
    (##will? x)))

(define-prim (##make-will testator action)
  (macro-make-will testator action))

(define-prim (make-will testator action)
  (macro-force-vars (action)
    (macro-check-procedure action 2 (make-will testator action)
      (macro-make-will testator action))))

(define-prim (##will-testator will)
  (macro-will-testator will))

(define-prim (will-testator will)
  (macro-force-vars (will)
    (macro-check-will will 1 (will-testator will)
      (macro-will-testator will))))

(define-prim (##will-execute! will)
  (macro-will-execute! will))

(define-prim (will-execute! will)
  (macro-force-vars (will)
    (macro-check-will will 1 (will-execute! will)
      (macro-will-execute! will))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (##box? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-boxvalues))
       (##fixnum.= (##vector-length obj) 1)))

(define-prim (box? obj)
  (macro-force-vars (obj)
    (##box? obj)))

(define-prim (##box obj)
  (##subtype-set! (##vector obj) (macro-subtype-boxvalues)))

(define-prim (box obj)
  (##box obj))

(define-prim (##unbox box)
  (##vector-ref box 0))

(define-prim (unbox box)
  (macro-force-vars (box)
    (macro-check-box box 1 (unbox box)
      (##unbox box))))

(define-prim (##set-box! box val)
  (##vector-set! box 0 val))

(define-prim (set-box! box val)
  (macro-force-vars (box)
    (macro-check-box box 1 (set-box! box val)
      (##set-box! box val))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (command-line)
  ##processed-command-line)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (getenv name)
  (macro-force-vars (name)
    (macro-check-string name 1 (getenv name)
      (##getenv name))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (infix.if expr stat1 . other-stat)
  `(if ,expr
     ,stat1
     ,@other-stat))

(define-runtime-macro (infix.while expr body)
  (let ((loop-name (gensym)))
    `(let ,loop-name ()
       (if ,expr
         (begin
           ,body
           (,loop-name))))))

(define-runtime-macro (infix.for expr1 expr2 expr3 body)
  (let* ((loop-name (gensym))
         (tail1 `((,loop-name)))
         (tail2 (if expr3 `(,body ,expr3 ,@tail1) `(,body ,@tail1)))
         (tail3 (if expr2 `((if ,expr2 (begin ,@tail2))) tail2))
         (tail4 `(let ,loop-name () ,@tail3)))
    (if expr1 `(begin ,expr1 ,tail4) tail4)))

(define-runtime-macro (infix.identifier ident)
  ident)

(define-runtime-macro (infix.compound . body)
  (if (##null? body)
    `(##void)
    `(begin ,@body)))

(define-prim (infix.expression expr)
  expr)

(define-prim (infix.scheme expr)
  expr)

(define-prim (infix.string str)
  str)

(define-prim (infix.char c)
  c)

(define-prim (infix.number n)
  n)

(define-prim (infix.null)
  '())

(define-prim (infix.list x y)
  (##cons x y))

(define-prim (infix.cons x y)
  (##cons x y))

(define-runtime-macro (infix.arrow x y)
  `(infix.call ,y ,x))

(define-prim (infix.call proc . args)
  (##apply proc args))

(define-runtime-macro (infix.dot x y)
  `(infix.call ,y ,x))

(define-runtime-macro (infix.clause x)
  `(pp (list 'clause ,x)));;;;;;;;;;;;;;;;;

(define-prim (infix.index x y)
  (##declare (standard-bindings) (safe))
  (vector-ref x y))

(define-prim (infix.!x x)
  (##not x))

(define-runtime-macro (infix.++x x)
  (cond ((and (list? x) (= (length x) 2) (eq? (car x) 'infix.identifier))
         (let ((temp-name (gensym)))
           `(let ((,temp-name (+ ,x 1)))
              (set! ,(cadr x) ,temp-name)
              ,temp-name)))
        ((and (list? x) (= (length x) 3) (eq? (car x) 'infix.index))
         (let* ((temp1-name (gensym))
                (temp2-name (gensym))
                (temp3-name (gensym)))
           `(let ((,temp1-name ,(cadr x))
                  (,temp2-name ,(caddr x)))
              (let ((,temp3-name (+ (vector-ref ,temp1-name ,temp2-name) 1)))
                (vector-set! ,temp1-name ,temp2-name ,temp3-name)
                ,temp3-name))))
        (else
         (error "unhandled expression type"))))

(define-runtime-macro (infix.x++ x)
  (cond ((and (list? x) (= (length x) 2) (eq? (car x) 'infix.identifier))
         (let ((temp-name (gensym)))
           `(let ((,temp-name ,x))
              (set! ,(cadr x) (+ ,temp-name 1))
              ,temp-name)))
        ((and (list? x) (= (length x) 3) (eq? (car x) 'infix.index))
         (let* ((temp1-name (gensym))
                (temp2-name (gensym))
                (temp3-name (gensym)))
           `(let ((,temp1-name ,(cadr x))
                  (,temp2-name ,(caddr x)))
              (let ((,temp3-name (vector-ref ,temp1-name ,temp2-name)))
                (vector-set! ,temp1-name ,temp2-name (+ ,temp3-name 1))
                ,temp3-name))))
        (else
         (error "unhandled expression type"))))

(define-runtime-macro (infix.--x x)
  `(##void));;;;;;;;;;;;;

(define-runtime-macro (infix.x-- x)
  `(##void));;;;;;;;;;;;;;

(define-prim (infix.~x x)
  (##declare (standard-bindings) (safe) (generic))
  (bitwise-not x))

(define-prim (infix.x%y x y)
  (##declare (standard-bindings) (safe) (generic))
  (remainder x y))

(define-prim (infix.x*y x y)
  (##declare (standard-bindings) (safe) (generic))
  (* x y))

(define-prim (infix.*x x)
  (##declare (safe))
  (x))

(define-prim (infix.x/y x y)
  (##declare (standard-bindings) (safe) (generic))
  (/ x y))

(define-prim (infix.x+y x y)
  (##declare (standard-bindings) (safe) (generic))
  (+ x y))

(define-prim (infix.+x x)
  (##declare (standard-bindings) (safe) (generic))
  (+ x))

(define-prim (infix.x-y x y)
  (##declare (standard-bindings) (safe) (generic))
  (- x y))

(define-prim (infix.-x x)
  (##declare (standard-bindings) (safe) (generic))
  (- x))

(define-prim (infix.x<<y x y)
  (##declare (standard-bindings) (safe) (generic))
  (arithmetic-shift x y))

(define-prim (infix.x>>y x y)
  (##declare (standard-bindings) (safe) (generic))
  (arithmetic-shift x (- y)));;;;;;;;;;;;;;

(define-prim (infix.x<y x y)
  (##declare (standard-bindings) (safe) (generic))
  (< x y))

(define-prim (infix.x<=y x y)
  (##declare (standard-bindings) (safe) (generic))
  (<= x y))

(define-prim (infix.x>y x y)
  (##declare (standard-bindings) (safe) (generic))
  (> x y))

(define-prim (infix.x>=y x y)
  (##declare (standard-bindings) (safe) (generic))
  (>= x y))

(define-prim (infix.x!=y x y)
  (##declare (standard-bindings) (safe) (generic))
  (##not (equal? x y)))

(define-prim (infix.x==y x y)
  (##declare (standard-bindings) (safe) (generic))
  (equal? x y))

(define-prim (infix.x&y x y)
  (##declare (standard-bindings) (safe) (generic))
  (bitwise-and x y))

(define-runtime-macro (infix.&x x)
  `(lambda arg;;;;;;;;;;;;;;;;;;
     (if (##null? arg)
       ,x
       (infix.x=y ,x (infix.scheme (car arg))))))

(define-prim (infix.x^y x y)
  (##declare (standard-bindings) (safe) (generic))
  (bitwise-xor x y))

(define-prim (|infix.x\|y| x y)
  (##declare (standard-bindings) (safe) (generic))
  (bitwise-ior x y))

(define-runtime-macro (infix.x&&y x y)
  `(and ,x ,y))

(define-prim (|infix.x\|\|y| x y)
  `(or ,x ,y))

(define-runtime-macro (infix.x?y:z x y z)
  `(if ,x ,y ,z))

(define-runtime-macro (infix.x%=y x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (infix.x&=y x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (infix.x*=y x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (infix.x+=y x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (infix.x-=y x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (infix.x/=y x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (infix.x<<=y x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (infix.x=y x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (infix.x>>=y x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (infix.x^=y x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (|infix.x\|=y| x y)
  `(##void));;;;;;;;;;;;;;;;

(define-runtime-macro (|infix.x,y| x y)
  `(begin ,x ,y))

;------------------------------------------------------------------------------

; Object encoding/decoding.

(define-prim (##object->encoding obj)
  (let* ((hi (##type-cast obj (macro-type-fixnum)))
         (lo (##type obj)))
    (##+ (##* (if (##fixnum.< hi 0) (##- hi ##bignum.2*min-fixnum) hi)
              4)
         lo)))

(define-prim (##encoding->object encoding)
  (let* ((hi (##quotient encoding 4))
         (lo (##modulo encoding 4))
         (x (if (##fixnum? hi) hi (##+ hi ##bignum.2*min-fixnum))))
    (##type-cast x lo)))

(define-prim (##foreign-address ptr)

  (##declare (not interrupts-enabled))

  (let ((addr (##vector-ref ptr 2)))

    ; Beware! "addr" is an arbitrary C pointer.  If the GC ever
    ; gets hold of it, thinking that it is a reference to a Scheme
    ; datum, it is bound to cause trouble.  Fortunately, because
    ; "addr" is dead after the binding of "hi" and "lo", the
    ; compiler will overwrite the register containing "addr" with
    ; "lo".

    (let* ((hi (##type-cast addr (macro-type-fixnum)))
           (lo (##type addr)))

      (##declare (interrupts-enabled))

      (##+ (##* (if (##fixnum.< hi 0) (##- hi ##bignum.2*min-fixnum) hi)
                4)
           lo))))

;==============================================================================
