/* file: "c_intf.c" */

/* Copyright (C) 1994-2002 by Marc Feeley, All Rights Reserved. */

/* 
 * This module implements the conversion functions for the C
 * interface.
 */

#define ___INCLUDED_FROM_C_INTF
#define ___VERSION 40011
#include "gambit.h"

#include "os.h"
#include "setup.h"
#include "mem.h"
#include "c_intf.h"

___LOCAL ___SCMOBJ ___temp; /* needed by some macros in "gambit.h" */


/*---------------------------------------------------------------------------*/

/* Utilities for 64 bit arithmetic. */

#ifdef ___BUILTIN_64BIT_INT_TYPE

/*
 * If the symbol ___BUILTIN_64BIT_INT_TYPE is defined then the data
 * types ___S64 and ___U64 are builtin integer types.
 */

___EXP_FUNC(___S64,___S64_from_SM32_fn)
   ___P((___SM32 val),
        (val)
___SM32 val;)
{
  return ___S64_from_SM32 (val);
}


___EXP_FUNC(___S64,___S64_from_SM32_UM32_fn)
   ___P((___SM32 hi32,
         ___UM32 lo32),
        (hi32,
         lo32)
___SM32 hi32;
___UM32 lo32;)
{
  return ___S64_from_SM32_UM32 (hi32, lo32);
}


___EXP_FUNC(___S64,___S64_from_LONG_fn)
   ___P((long val),
        (val)
long val;)
{
  return ___S64_from_LONG (val);
}


___EXP_FUNC(long,___S64_to_LONG_fn)
   ___P((___S64 val),
        (val)
___S64 val;)
{
  return ___S64_to_LONG (val);
}


___EXP_FUNC(___BOOL,___S64_fits_in_width_fn)
   ___P((___S64 val,
         int width),
        (val,
         width)
___S64 val;
int width;)
{
  return ___S64_fits_in_width (val, width);
}


___EXP_FUNC(___U64,___U64_from_UM32_fn)
   ___P((___UM32 val),
        (val)
___UM32 val;)
{
  return ___U64_from_UM32 (val);
}


___EXP_FUNC(___U64,___U64_from_UM32_UM32_fn)
   ___P((___UM32 hi32,
         ___UM32 lo32),
        (hi32,
         lo32)
___UM32 hi32;
___UM32 lo32;)
{
  return ___U64_from_UM32_UM32 (hi32, lo32);
}


___EXP_FUNC(___U64,___U64_from_ULONG_fn)
   ___P((unsigned long val),
        (val)
unsigned long val;)
{
  return ___U64_from_ULONG (val);
}


___EXP_FUNC(unsigned long,___U64_to_ULONG_fn)
   ___P((___U64 val),
        (val)
___U64 val;)
{
  return ___U64_to_ULONG (val);
}


___EXP_FUNC(___BOOL,___U64_fits_in_width_fn)
   ___P((___U64 val,
         int width),
        (val,
         width)
___U64 val;
int width;)
{
  return ___U64_fits_in_width (val, width);
}


#else

/*
 * If the symbol ___BUILTIN_64BIT_INT_TYPE is not defined then the
 * data types ___S64 and ___U64 are structures.
 */


___EXP_FUNC(___S64,___S64_from_SM32_fn)
   ___P((___SM32 val),
        (val)
___SM32 val;)
{
  ___S64 r;
  r.lo32 = ___CAST_U32(val);
  r.hi32 = (val < 0) ? -1 : 0;
  return r;
}


___EXP_FUNC(___S64,___S64_from_SM32_UM32_fn)
   ___P((___SM32 hi32,
         ___UM32 lo32),
        (hi32,
         lo32)
___SM32 hi32;
___UM32 lo32;)
{
  ___S64 r;
  r.lo32 = lo32;
  r.hi32 = hi32;
  return r;
}


___EXP_FUNC(___S64,___S64_from_LONG_fn)
   ___P((long val),
        (val)
long val;)
{
#if ___LONG_WIDTH <= 32
  return ___S64_from_SM32 (val);
#else
  return ___S64_from_SM32_UM32 (___CAST_S32(val >> 32), ___CAST_U32(val));
#endif
}


___EXP_FUNC(long,___S64_to_LONG_fn)
   ___P((___S64 val),
        (val)
___S64 val;)
{
#if ___LONG_WIDTH <= 32
  return ___CAST_S32 (val.lo32);
#else
  return (___CAST(long,val.hi32) << 32) + val.lo32;
#endif
}


___EXP_FUNC(___BOOL,___S64_fits_in_width_fn)
   ___P((___S64 val,
         int width),
        (val,
         width)
___S64 val;
int width;)
{
  if (val.hi32 < 0)
    {
      if (width > 32)
        return (val.hi32 >> (width-32-1)) == -1;
      return val.hi32 == -1 && (___CAST_S32(val.lo32) >> (width-1)) == -1;
    }
  else
    {
      if (width > 32)
        return (val.hi32 >> (width-32-1)) == 0;
      return val.hi32 == 0 && (___CAST_S32(val.lo32) >> (width-1)) == 0;
    }
}


___EXP_FUNC(___U64,___U64_from_UM32_fn)
   ___P((___UM32 val),
        (val)
___UM32 val;)
{
  ___U64 r;
  r.lo32 = val;
  r.hi32 = 0;
  return r;
}


___EXP_FUNC(___U64,___U64_from_UM32_UM32_fn)
   ___P((___UM32 hi32,
         ___UM32 lo32),
        (hi32,
         lo32)
___UM32 hi32;
___UM32 lo32;)
{
  ___U64 r;
  r.lo32 = lo32;
  r.hi32 = hi32;
  return r;
}


___EXP_FUNC(___U64,___U64_from_ULONG_fn)
   ___P((unsigned long val),
        (val)
unsigned long val;)
{
#if ___LONG_WIDTH <= 32
  return ___U64_from_UM32 (val);
#else
  return ___U64_from_UM32_UM32 (___CAST_U32(val >> 32), ___CAST_U32(val));
#endif
}


___EXP_FUNC(unsigned long,___U64_to_ULONG_fn)
   ___P((___U64 val),
        (val)
___U64 val;)
{
#if ___LONG_WIDTH <= 32
  return val.lo32;
#else
  return (___CAST(unsigned long,val.hi32) << 32) + val.lo32;
#endif
}


___EXP_FUNC(___BOOL,___U64_fits_in_width_fn)
   ___P((___U64 val,
         int width),
        (val,
         width)
___U64 val;
int width;)
{
  if (width >= 64)
    return 1;
  if (width >= 32)
    return (val.hi32 >> (width-32)) == 0;
  return val.hi32 == 0 && (val.lo32 >> width) == 0;
}


#endif


/*---------------------------------------------------------------------------*/

/* Utilities for UTF-8 encoding of characters. */

/*
 * '___utf8_bytes (c)' returns the number of bytes that are needed to
 * encode the character 'c' using the UTF-8 variable-length encoding.
 * If the character is not legal, 0 is returned.
 */

int ___utf8_bytes
   ___P((___UCS4 c),
        (c)
___UCS4 c;)
{
  if (c <= 0x7f) return 1;
  if (c <= 0x7ff) return 2;
  if (c <= 0xffff)
    {
#ifdef ___REJECT_ILLEGAL_UCS4
      if (c > 0xd7ff && (c <= 0xdfff || c > 0xfffd)) return 0;
#endif
      return 3;
    }
  if (c <= 0x1fffff) return 4;
  if (c <= 0x3ffffff) return 5;
#ifdef ___REJECT_ILLEGAL_UCS4
  if (c > 0x7fffffff) return 0;
#endif
  return 6;
}


/*
 * '___utf8_put (ptr, c)' converts the character 'c' into its UTF-8
 * variable-length encoding.  'ptr' is a pointer to a byte pointer
 * designating the start of the UTF-8 encoding.  On return the byte
 * pointer points to the first byte following the UTF-8 encoding.  If
 * the character is not legal, the pointer is not updated.
 */

void ___utf8_put
   ___P((___UTF8STRING *ptr,
         ___UCS4 c),
        (ptr,
         c)
___UTF8STRING *ptr;
___UCS4 c;)
{
  ___UTF8STRING p = *ptr;
  if (c <= 0x7f)
    {
      *p++ = c;
      *ptr = p;
    }
  else
    {
      int bytes;
      if (c <= 0x7ff) bytes = 2;
      else if (c <= 0xffff)
        {
#ifdef ___REJECT_ILLEGAL_UCS4
          if (c > 0xd7ff && (c <= 0xdfff || c > 0xfffd)) return;
#endif
          bytes = 3;
        }
      else if (c <= 0x1fffff) bytes = 4;
      else if (c <= 0x3ffffff) bytes = 5;
#ifdef ___REJECT_ILLEGAL_UCS4
      else if (c > 0x7fffffff) return;
#endif
      else bytes = 6;
      p += bytes;
      *ptr = p;
      switch (bytes)
        {
          case 6:  *--p = 0x80+(c&0x3f); c >>= 6;
          case 5:  *--p = 0x80+(c&0x3f); c >>= 6;
          case 4:  *--p = 0x80+(c&0x3f); c >>= 6;
          case 3:  *--p = 0x80+(c&0x3f); c >>= 6;
          default: *--p = 0x80+(c&0x3f); c >>= 6;
        }
      *--p = 0xff - (0xff>>bytes) + c;
    }
}


/*
 * '___utf8_get (ptr)' converts a UTF-8 variable-length encoding to
 * the character it encodes.  'ptr' is a pointer to a byte pointer
 * designating the start of the UTF-8 encoding.  If the encoding is
 * legal, the byte pointer will point to the first byte following the
 * UTF-8 encoding and the character is returned.  If the encoding is
 * illegal, the byte pointer is not updated and 0 is returned.
 */

___UCS4 ___utf8_get
   ___P((___UTF8STRING *ptr),
        (ptr)
___UTF8STRING *ptr;)
{
  ___UTF8STRING p = *ptr;
  unsigned char byte = *p++;
  ___UCS4 c;
  int bits;
  if (byte <= 0x7f)
    {
      *ptr = p;
      return byte;
    }
  if (byte <= 0xbf || byte > 0xfd)
    return 0; /* illegal first byte */
  c = byte; /* upper bits are removed later */
  bits = 6;
  while (byte & 0x40)
    {
      unsigned char next = *p++;
      if (next <= 0x7f || next > 0xbf)
        return 0; /* faulty byte found after the first byte */
      c = (c << 6) + (next & 0x3f);
      byte <<= 1;
      bits += 5;
    }
  c &= ((unsigned long)1<<bits)-1;
#ifdef ___REJECT_ILLEGAL_UCS4
  if ((c > 0xd7ff && c <= 0xdfff) ||
      (c > 0xfffd && c <= 0xffff))
    return 0; /* it is not a legal UCS-4 character */
  if (c < 0x80 ||
      c < ((___UCS4)1<<(bits-5)))
    return 0; /* character was not encoded with the shortest sequence */
#endif
  *ptr = p;
  return c;
}


/*---------------------------------------------------------------------------*/

/* Scheme to C conversion */

/* 
 * The following Scheme to C conversion functions may allocate memory
 * in the C heap:
 *
 *    ___SCMOBJ_to_FUNCTION
 *    ___SCMOBJ_to_NONNULLFUNCTION
 *    ___SCMOBJ_to_CHARSTRING
 *    ___SCMOBJ_to_NONNULLCHARSTRING
 *    ___SCMOBJ_to_NONNULLCHARSTRINGLIST
 *    ___SCMOBJ_to_LATIN1STRING
 *    ___SCMOBJ_to_NONNULLLATIN1STRING
 *    ___SCMOBJ_to_NONNULLLATIN1STRINGLIST
 *    ___SCMOBJ_to_UTF8STRING
 *    ___SCMOBJ_to_NONNULLUTF8STRING
 *    ___SCMOBJ_to_NONNULLUTF8STRINGLIST
 *    ___SCMOBJ_to_UCS2STRING
 *    ___SCMOBJ_to_NONNULLUCS2STRING
 *    ___SCMOBJ_to_NONNULLUCS2STRINGLIST
 *    ___SCMOBJ_to_UCS4STRING
 *    ___SCMOBJ_to_NONNULLUCS4STRING
 *    ___SCMOBJ_to_NONNULLUCS4STRINGLIST
 *    ___SCMOBJ_to_VARIANT
 *
 * The memory allocated must be freed by calling the corresponding
 * freeing function:
 *
 *    ___release_function
 *    ___release_string
 *    ___release_string_list
 *    ___release_variant
 */


/* Convert a Scheme integer to a C '___S64'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_S64)
   ___P((___SCMOBJ obj,
         ___S64 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___S64 *x;
int arg_num;)
{
  ___S64 val;

  if (___FIXNUMP(obj))
    {
#if ___SCMOBJ_WIDTH == 32
      val = ___S64_from_SM32 (___INT(obj));
#else
      val = ___INT(obj);
#endif
    }
  else
    {
      if (!___BIGNUMP(obj))
        return ___STOC_S64_ERR+arg_num;
      if (___BIGALENGTH(obj) == ___FIX(1))
        {
          ___BIGADIGIT d0 = ___BIGAFETCH(___BODY_AS(obj,___tSUBTYPED),0);
#if ___BIG_ABASE_WIDTH == 32
          val = ___S64_from_SM32 (___CAST(___BIGADIGITSIGNED,d0));
#else
          val = ___CAST(___BIGADIGITSIGNED,d0);
#endif
        }
#if ___BIG_ABASE_WIDTH == 32
      else if (___BIGALENGTH(obj) == ___FIX(2))
        {
          ___BIGADIGIT d0 = ___BIGAFETCH(___BODY_AS(obj,___tSUBTYPED),0);
          ___BIGADIGIT d1 = ___BIGAFETCH(___BODY_AS(obj,___tSUBTYPED),1);
          val = ___S64_from_SM32_UM32 (___CAST(___BIGADIGITSIGNED,d1), d0);
        }
#endif
      else
        return ___STOC_S64_ERR+arg_num;
    }

  *x = val;
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C '___U64'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_U64)
   ___P((___SCMOBJ obj,
         ___U64 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___U64 *x;
int arg_num;)
{
  ___U64 val;

  if (___FIXNUMP(obj))
    {
      if (___FIXNEGATIVEP(obj))
        return ___STOC_U64_ERR+arg_num;
#if ___SCMOBJ_WIDTH == 32
      val = ___U64_from_UM32 (___CAST_U32(___INT(obj)));
#else
      val = ___INT(obj);
#endif
    }
  else
    {
      if (!___BIGNUMP(obj))
        return ___STOC_U64_ERR+arg_num;
      if (___BIGALENGTH(obj) == ___FIX(1))
        {
          ___BIGADIGIT d0 = ___BIGAFETCH(___BODY_AS(obj,___tSUBTYPED),0);
          if (d0 & ___BIG_ABASE_DIV_2)
            return ___STOC_U64_ERR+arg_num;
#if ___BIG_ABASE_WIDTH == 32
          val = ___U64_from_UM32 (d0);
#else
          val = d0;
#endif
        }
      else if (___BIGALENGTH(obj) == ___FIX(2))
        {
          ___BIGADIGIT d0 = ___BIGAFETCH(___BODY_AS(obj,___tSUBTYPED),0);
          ___BIGADIGIT d1 = ___BIGAFETCH(___BODY_AS(obj,___tSUBTYPED),1);
#if ___BIG_ABASE_WIDTH == 32
          if (d1 & ___BIG_ABASE_DIV_2)
            return ___STOC_U64_ERR+arg_num;
          val = ___U64_from_UM32_UM32 (d1, d0);
#else
          if (d1 != 0)
            return ___STOC_U64_ERR+arg_num;
          val = d0;
#endif
        }
#if ___BIG_ABASE_WIDTH == 32
      else if (___BIGALENGTH(obj) == ___FIX(3))
        {
          ___BIGADIGIT d0 = ___BIGAFETCH(___BODY_AS(obj,___tSUBTYPED),0);
          ___BIGADIGIT d1 = ___BIGAFETCH(___BODY_AS(obj,___tSUBTYPED),1);
          ___BIGADIGIT d2 = ___BIGAFETCH(___BODY_AS(obj,___tSUBTYPED),2);
          if (d2 != 0)
            return ___STOC_U64_ERR+arg_num;
          val = ___U64_from_UM32_UM32 (d1, d0);
        }
#endif
      else
        return ___STOC_U64_ERR+arg_num;
    }

  *x = val;
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C '___S8'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_S8)
   ___P((___SCMOBJ obj,
         ___S8 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___S8 *x;
int arg_num;)
{
  ___S64 val;

  if (___SCMOBJ_to_S64 (obj, &val, arg_num) != ___NO_ERR ||
      !___S64_fits_in_width (val, 8))
    return ___STOC_S8_ERR+arg_num;

  *x = ___S64_to_LONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C '___U8'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_U8)
   ___P((___SCMOBJ obj,
         ___U8 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___U8 *x;
int arg_num;)
{
  ___U64 val;

  if (___SCMOBJ_to_U64 (obj, &val, arg_num) != ___NO_ERR ||
      !___U64_fits_in_width (val, 8))
    return ___STOC_U8_ERR+arg_num;

  *x = ___U64_to_ULONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C '___S16'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_S16)
   ___P((___SCMOBJ obj,
         ___S16 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___S16 *x;
int arg_num;)
{
  ___S64 val;

  if (___SCMOBJ_to_S64 (obj, &val, arg_num) != ___NO_ERR ||
      !___S64_fits_in_width (val, 16))
    return ___STOC_S16_ERR+arg_num;

  *x = ___S64_to_LONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C '___U16'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_U16)
   ___P((___SCMOBJ obj,
         ___U16 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___U16 *x;
int arg_num;)
{
  ___U64 val;

  if (___SCMOBJ_to_U64 (obj, &val, arg_num) != ___NO_ERR ||
      !___U64_fits_in_width (val, 16))
    return ___STOC_U16_ERR+arg_num;

  *x = ___U64_to_ULONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C '___S32'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_S32)
   ___P((___SCMOBJ obj,
         ___S32 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___S32 *x;
int arg_num;)
{
  ___S64 val;

  if (___SCMOBJ_to_S64 (obj, &val, arg_num) != ___NO_ERR ||
      !___S64_fits_in_width (val, 32))
    return ___STOC_S32_ERR+arg_num;

  *x = ___S64_to_LONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C '___U32'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_U32)
   ___P((___SCMOBJ obj,
         ___U32 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___U32 *x;
int arg_num;)
{
  ___U64 val;

  if (___SCMOBJ_to_U64 (obj, &val, arg_num) != ___NO_ERR ||
      !___U64_fits_in_width (val, 32))
    return ___STOC_U32_ERR+arg_num;

  *x = ___U64_to_ULONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C '___F32'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_F32)
   ___P((___SCMOBJ obj,
         ___F32 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___F32 *x;
int arg_num;)
{
  if (!___FLONUMP(obj))
    return ___STOC_F32_ERR+arg_num;

  *x = ___FLONUM_VAL(obj);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C '___F64'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_F64)
   ___P((___SCMOBJ obj,
         ___F64 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___F64 *x;
int arg_num;)
{
  if (!___FLONUMP(obj))
    return ___STOC_F64_ERR+arg_num;

  *x = ___FLONUM_VAL(obj);
  return ___NO_ERR;
}


/* Convert a Scheme character to a C 'char'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_CHAR)
   ___P((___SCMOBJ obj,
         char *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
char *x;
int arg_num;)
{
  ___UCS4 c;

  if (!___CHARP(obj) ||
      (c=ucs4_to_uchar(___INT(obj))) > (1<<___CHAR_WIDTH)-1)
    return ___STOC_CHAR_ERR+arg_num;

  *x = c;
  return ___NO_ERR;
}


/* Convert a Scheme character to a C 'signed char'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_SCHAR)
   ___P((___SCMOBJ obj,
         ___SCHAR *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___SCHAR *x;
int arg_num;)
{
  ___UCS4 c;

  if (!___CHARP(obj) ||
      (c=ucs4_to_uchar(___INT(obj))) > (1<<___CHAR_WIDTH)-1)
    return ___STOC_SCHAR_ERR+arg_num;

  *x = c;
  return ___NO_ERR;
}


/* Convert a Scheme character to a C 'unsigned char'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_UCHAR)
   ___P((___SCMOBJ obj,
         unsigned char *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
unsigned char *x;
int arg_num;)
{
  ___UCS4 c;

  if (!___CHARP(obj) ||
      (c=ucs4_to_uchar(___INT(obj))) > (1<<___CHAR_WIDTH)-1)
    return ___STOC_UCHAR_ERR+arg_num;

  *x = c;
  return ___NO_ERR;
}


/* Convert a Scheme character to a C LATIN-1 encoded character. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_LATIN1)
   ___P((___SCMOBJ obj,
         ___LATIN1 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___LATIN1 *x;
int arg_num;)
{
  ___UCS4 c;

  if (!___CHARP(obj) ||
      (c=___INT(obj)) > 0xff) /* LATIN-1 is 8 bits */
    return ___STOC_LATIN1_ERR+arg_num;

  *x = c;
  return ___NO_ERR;
}


/* Convert a Scheme character to a C UCS-2 encoded character. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_UCS2)
   ___P((___SCMOBJ obj,
         ___UCS2 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UCS2 *x;
int arg_num;)
{
  ___UCS4 c;

  if (!___CHARP(obj) ||
      (c=___INT(obj)) > 0xffff) /* UCS-2 is 16 bits */
    return ___STOC_UCS2_ERR+arg_num;

  *x = c;
  return ___NO_ERR;
}


/* Convert a Scheme character to a C UCS-4 encoded character. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_UCS4)
   ___P((___SCMOBJ obj,
         ___UCS4 *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UCS4 *x;
int arg_num;)
{
  ___UCS4 c;

  if (!___CHARP(obj) ||
      (c=___INT(obj)) > 0x7fffffff) /* UCS-4 is 31 bits */
    return ___STOC_UCS4_ERR+arg_num;

  *x = c;
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C 'short'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_SHORT)
   ___P((___SCMOBJ obj,
         short *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
short *x;
int arg_num;)
{
  ___S64 val;

  if (___SCMOBJ_to_S64 (obj, &val, arg_num) != ___NO_ERR ||
      !___S64_fits_in_width (val, ___SHORT_WIDTH))
    return ___STOC_SHORT_ERR+arg_num;

  *x = ___S64_to_LONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C 'unsigned short'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_USHORT)
   ___P((___SCMOBJ obj,
         unsigned short *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
unsigned short *x;
int arg_num;)
{
  ___U64 val;

  if (___SCMOBJ_to_U64 (obj, &val, arg_num) != ___NO_ERR ||
      !___U64_fits_in_width (val, ___SHORT_WIDTH))
    return ___STOC_USHORT_ERR+arg_num;

  *x = ___U64_to_ULONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C 'int'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_INT)
   ___P((___SCMOBJ obj,
         int *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
int *x;
int arg_num;)
{
  ___S64 val;

  if (___SCMOBJ_to_S64 (obj, &val, arg_num) != ___NO_ERR ||
      !___S64_fits_in_width (val, ___INT_WIDTH))
    return ___STOC_INT_ERR+arg_num;

  *x = ___S64_to_LONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C 'unsigned int'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_UINT)
   ___P((___SCMOBJ obj,
         unsigned int *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
unsigned int *x;
int arg_num;)
{
  ___U64 val;

  if (___SCMOBJ_to_U64 (obj, &val, arg_num) != ___NO_ERR ||
      !___U64_fits_in_width (val, ___INT_WIDTH))
    return ___STOC_UINT_ERR+arg_num;

  *x = ___U64_to_ULONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C 'long'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_LONG)
   ___P((___SCMOBJ obj,
         long *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
long *x;
int arg_num;)
{
  ___S64 val;

  if (___SCMOBJ_to_S64 (obj, &val, arg_num) != ___NO_ERR ||
      !___S64_fits_in_width (val, ___LONG_WIDTH))
    return ___STOC_LONG_ERR+arg_num;

  *x = ___S64_to_LONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme integer to a C 'unsigned long'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_ULONG)
   ___P((___SCMOBJ obj,
         unsigned long *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
unsigned long *x;
int arg_num;)
{
  ___U64 val;

  if (___SCMOBJ_to_U64 (obj, &val, arg_num) != ___NO_ERR ||
      !___U64_fits_in_width (val, ___LONG_WIDTH))
    return ___STOC_ULONG_ERR+arg_num;

  *x = ___U64_to_ULONG (val);
  return ___NO_ERR;
}


/* Convert a Scheme flonum to a C 'float'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_FLOAT)
   ___P((___SCMOBJ obj,
         float *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
float *x;
int arg_num;)
{
  if (!___FLONUMP(obj))
    return ___STOC_FLOAT_ERR+arg_num;

  *x = ___FLONUM_VAL(obj);
  return ___NO_ERR;
}


/* Convert a Scheme flonum to a C 'double'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_DOUBLE)
   ___P((___SCMOBJ obj,
         double *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
double *x;
int arg_num;)
{
  if (!___FLONUMP(obj))
    return ___STOC_DOUBLE_ERR+arg_num;

  *x = ___FLONUM_VAL(obj);
  return ___NO_ERR;
}


/* Convert a Scheme foreign object to a C struct pointer. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_STRUCT)
   ___P((___SCMOBJ obj,
         void **x,
         ___SCMOBJ tag,
         int arg_num),
        (obj,
         x,
         tag,
         arg_num)
___SCMOBJ obj;
void **x;
___SCMOBJ tag;
int arg_num;)
{
  return ___UNKNOWN_ERR;
}


/* Convert a Scheme foreign object to a C union pointer. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_UNION)
   ___P((___SCMOBJ obj,
         void **x,
         ___SCMOBJ tag,
         int arg_num),
        (obj,
         x,
         tag,
         arg_num)
___SCMOBJ obj;
void **x;
___SCMOBJ tag;
int arg_num;)
{
  return ___UNKNOWN_ERR;
}


/* Convert a Scheme foreign object to a C type pointer. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_TYPE)
   ___P((___SCMOBJ obj,
         void **x,
         ___SCMOBJ tag,
         int arg_num),
        (obj,
         x,
         tag,
         arg_num)
___SCMOBJ obj;
void **x;
___SCMOBJ tag;
int arg_num;)
{
  return ___UNKNOWN_ERR;
}


/* Convert a Scheme foreign object to a C pointer. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_POINTER)
   ___P((___SCMOBJ obj,
         void **x,
         ___SCMOBJ tag,
         int arg_num),
        (obj,
         x,
         tag,
         arg_num)
___SCMOBJ obj;
void **x;
___SCMOBJ tag;
int arg_num;)
{
  if (___FALSEP(obj)) /* #f counts as NULL */
    {
      *x = 0;
      return ___NO_ERR;
    }

  if (!___TESTSUBTYPE(obj,___sFOREIGN))
    return ___STOC_POINTER_ERR+arg_num;

  /*******************************fix type rules*/
  if ((tag != ___FAL && /* destination type != void* */
       ___FIELD(obj,___FOREIGN_TAG) != ___FAL && /* source type != void* */
       !___EQP(___FIELD(obj,___FOREIGN_TAG),tag))) /* pointer types != */
    return ___STOC_POINTER_ERR+arg_num;

  *x = ___CAST(void*,___FIELD(obj,___FOREIGN_PTR));
  return ___NO_ERR;
}


/* Convert a Scheme foreign object to a nonnull C pointer. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLPOINTER)
   ___P((___SCMOBJ obj,
         void **x,
         ___SCMOBJ tag,
         int arg_num),
        (obj,
         x,
         tag,
         arg_num)
___SCMOBJ obj;
void **x;
___SCMOBJ tag;
int arg_num;)
{
  if (___SCMOBJ_to_POINTER (obj, x, tag, arg_num) != ___NO_ERR || *x == 0)
    return ___STOC_NONNULLPOINTER_ERR+arg_num;
  return ___NO_ERR;
}


/* Convert a Scheme function to a C function. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_FUNCTION)
   ___P((___SCMOBJ obj,
         void **x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
void **x;
int arg_num;)
{
  if (___FALSEP(obj)) /* #f counts as NULL */
    {
      *x = 0;
      return ___NO_ERR;
    }

  if (___TYP(obj) != ___tSUBTYPED ||
/******************shouldn't we also check for ___sPROCEDURE? */
      ((___label_struct*)___UNTAG_AS(obj,___tSUBTYPED))->entry_or_descr != obj ||
      (___UNTAG_AS(obj,___tSUBTYPED)[-___LS] !=
      ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM)))
    /* 
     * At present, arbitrary Scheme functions cannot be converted to C
     * functions.
     */
    return ___STOC_FUNCTION_ERR+arg_num;

  *x = ___CAST(void*,(___CAST(___label_struct*,___UNTAG_AS(obj,___tSUBTYPED))-1)->host);
  return ___NO_ERR;
}


/* Convert a Scheme function to a nonnull C function. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLFUNCTION)
   ___P((___SCMOBJ obj,
         void **x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
void **x;
int arg_num;)
{
  if (___SCMOBJ_to_FUNCTION (obj, x, arg_num) != ___NO_ERR || *x == 0)
    return ___STOC_NONNULLFUNCTION_ERR+arg_num;
  return ___NO_ERR;
}


/* Convert a Scheme extended boolean to a C boolean. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_BOOL)
   ___P((___SCMOBJ obj,
         ___BOOL *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___BOOL *x;
int arg_num;)
{
  *x = !___FALSEP(obj); /* #f is false, everything else counts as true */
  return ___NO_ERR;
}


/* Convert a Scheme string to a C 'char *'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_CHARSTRING)
   ___P((___SCMOBJ obj,
         char **x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
char **x;
int arg_num;)
{
  unsigned long i, n;
  char *r;

  if (___FALSEP(obj)) /* #f counts as NULL */
    {
      *x = 0;
      return ___NO_ERR;
    }

  if (!___STRINGP(obj))
    return ___STOC_CHARSTRING_ERR+arg_num;

  n = ___INT(___STRINGLENGTH(obj));
  r = ___CAST(char*,___alloc_mem (n+1));

  if (r == 0)
    return ___STOC_HEAP_OVERFLOW_ERR+arg_num;
  for (i=0; i<n; i++)
    {
      ___UCS4 c = ucs4_to_uchar (___INT(___STRINGREF(obj,___FIX(i))));
      if (c > (1<<___CHAR_WIDTH)-1)
        {
          ___free_mem (r);
          return ___STOC_CHARSTRING_ERR+arg_num;
        }
      r[i] = c;
    }
  r[n] = 0;

  *x = r;
  return ___NO_ERR;
}


/* Convert a Scheme string to a nonnull C 'char *'. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLCHARSTRING)
   ___P((___SCMOBJ obj,
         char **x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
char **x;
int arg_num;)
{
  ___ERR_CODE err = ___SCMOBJ_to_CHARSTRING (obj, x, arg_num);
  if (err == ___STOC_CHARSTRING_ERR+arg_num ||
      (err == ___NO_ERR && *x == 0))
    err = ___STOC_NONNULLCHARSTRING_ERR+arg_num;
  return err;
}


/* Convert a Scheme string to a C LATIN-1 encoded character string. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_LATIN1STRING)
   ___P((___SCMOBJ obj,
         ___LATIN1STRING *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___LATIN1STRING *x;
int arg_num;)
{
  unsigned long i, n;
  ___LATIN1STRING r;

  if (___FALSEP(obj)) /* #f counts as NULL */
    {
      *x = 0;
      return ___NO_ERR;
    }

  if (!___STRINGP(obj))
    return ___STOC_CHARSTRING_ERR+arg_num;

  n = ___INT(___STRINGLENGTH(obj));
  r = ___CAST(___LATIN1STRING,___alloc_mem (n+1));
  if (r == 0)
    return ___STOC_HEAP_OVERFLOW_ERR+arg_num;
  for (i=0; i<n; i++)
    {
      ___UCS4 c = ___INT(___STRINGREF(obj,___FIX(i)));
      if (c > 0xff) /* LATIN-1 is 8 bits */
        {
          ___free_mem (r);
          return ___STOC_LATIN1STRING_ERR+arg_num;
        }
      r[i] = c;
    }
  r[n] = 0;

  *x = r;
  return ___NO_ERR;
}


/* Convert a Scheme string to a nonnull C LATIN-1 encoded character string. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLLATIN1STRING)
   ___P((___SCMOBJ obj,
         ___LATIN1STRING *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___LATIN1STRING *x;
int arg_num;)
{
  ___ERR_CODE err = ___SCMOBJ_to_LATIN1STRING (obj, x, arg_num);
  if (err == ___STOC_LATIN1STRING_ERR+arg_num ||
      (err == ___NO_ERR && *x == 0))
    err = ___STOC_NONNULLLATIN1STRING_ERR+arg_num;
  return err;
}


/* Convert a Scheme string to a C UTF-8 encoded character string. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_UTF8STRING)
   ___P((___SCMOBJ obj,
         ___UTF8STRING *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UTF8STRING *x;
int arg_num;)
{
  unsigned long i, bytes, n;
  ___UTF8STRING r;
  ___UTF8STRING p;

  if (___FALSEP(obj)) /* #f counts as NULL */
    {
      *x = 0;
      return ___NO_ERR;
    }

  if (!___STRINGP(obj))
    return ___STOC_UTF8STRING_ERR+arg_num;

  bytes = 0;
  n = ___INT(___STRINGLENGTH(obj));
  for (i=0; i<n; i++)
    {
      ___UCS4 c = ___INT(___STRINGREF(obj,___FIX(i)));
      int c_bytes = ___utf8_bytes (c);
      if (c_bytes == 0)
        return ___STOC_UTF8STRING_ERR+arg_num;
      bytes += c_bytes;
    }
  r = ___CAST(___UTF8STRING,___alloc_mem (bytes+1));
  if (r == 0)
    return ___STOC_HEAP_OVERFLOW_ERR+arg_num;
  p = r;
  for (i=0; i<n; i++)
    ___utf8_put (&p, ___INT(___STRINGREF(obj,___FIX(i))));
  *p = 0;

  *x = r;
  return ___NO_ERR;
}


/* Convert a Scheme string to a nonnull C UTF-8 encoded character string. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLUTF8STRING)
   ___P((___SCMOBJ obj,
         ___UTF8STRING *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UTF8STRING *x;
int arg_num;)
{
  ___ERR_CODE err = ___SCMOBJ_to_UTF8STRING (obj, x, arg_num);
  if (err == ___STOC_UTF8STRING_ERR+arg_num ||
      (err == ___NO_ERR && *x == 0))
    return ___STOC_NONNULLUTF8STRING_ERR+arg_num;
  return err;
}


/* Convert a Scheme string to a C UCS-2 encoded character string. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_UCS2STRING)
   ___P((___SCMOBJ obj,
         ___UCS2STRING *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UCS2STRING *x;
int arg_num;)
{
  unsigned long i, n;
  ___UCS2STRING r;

  if (___FALSEP(obj)) /* #f counts as NULL */
    {
      *x = 0;
      return ___NO_ERR;
    }

  if (!___STRINGP(obj))
    return ___STOC_UCS2STRING_ERR+arg_num;

  n = ___INT(___STRINGLENGTH(obj));
  r = ___CAST(___UCS2STRING,___alloc_mem ((n+1)*sizeof(___UCS2)));
  if (r == 0)
    return ___STOC_HEAP_OVERFLOW_ERR+arg_num;
  for (i=0; i<n; i++)
    {
      ___UCS4 c = ___INT(___STRINGREF(obj,___FIX(i)));
      if (c > 0xffff) /* UCS-2 is 16 bits */
        {
          ___free_mem (r);
          return ___STOC_UCS2STRING_ERR+arg_num;
        }
      r[i] = c;
    }
  r[n] = 0;

  *x = r;
  return ___NO_ERR;
}


/* Convert a Scheme string to a nonnull C UCS-2 encoded character string. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLUCS2STRING)
   ___P((___SCMOBJ obj,
         ___UCS2STRING *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UCS2STRING *x;
int arg_num;)
{
  ___ERR_CODE err = ___SCMOBJ_to_UCS2STRING (obj, x, arg_num);
  if (err == ___STOC_UCS2STRING_ERR+arg_num ||
      (err == ___NO_ERR && *x == 0))
    return ___STOC_NONNULLUCS2STRING_ERR+arg_num;
  return err;
}


/* Convert a Scheme string to a C UCS-4 encoded character string. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_UCS4STRING)
   ___P((___SCMOBJ obj,
         ___UCS4STRING *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UCS4STRING *x;
int arg_num;)
{
  unsigned long i, n;
  ___UCS4STRING r;

  if (___FALSEP(obj)) /* #f counts as NULL */
    {
      *x = 0;
      return ___NO_ERR;
    }

  if (!___STRINGP(obj))
    return ___STOC_UCS4STRING_ERR+arg_num;

  n = ___INT(___STRINGLENGTH(obj));
  r = ___CAST(___UCS4STRING,___alloc_mem ((n+1)*sizeof(___UCS4)));
  if (r == 0)
    return ___STOC_HEAP_OVERFLOW_ERR+arg_num;
  for (i=0; i<n; i++)
    {
      ___UCS4 c = ___INT(___STRINGREF(obj,___FIX(i)));
      if (c > 0x7fffffff) /* UCS-4 is 31 bits */
        {
          ___free_mem (r);
          return ___STOC_UCS4STRING_ERR+arg_num;
        }
      r[i] = c;
    }
  r[n] = 0;

  *x = r;
  return ___NO_ERR;
}


/* Convert a Scheme string to a nonnull C UCS-4 encoded character string. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLUCS4STRING)
   ___P((___SCMOBJ obj,
         ___UCS4STRING *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UCS4STRING *x;
int arg_num;)
{
  ___ERR_CODE err = ___SCMOBJ_to_UCS4STRING (obj, x, arg_num);
  if (err == ___STOC_UCS4STRING_ERR+arg_num ||
      (err == ___NO_ERR && *x == 0))
    return ___STOC_NONNULLUCS4STRING_ERR+arg_num;
  return err;
}


/* Convert a Scheme string to a nonnull C string. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLSTRING)
   ___P((___SCMOBJ obj,
         void **x,
         int arg_num,
         int char_encoding),
        (obj,
         x,
         arg_num,
         char_encoding)
___SCMOBJ obj;
void **x;
int arg_num;
int char_encoding;)
{
  ___ERR_CODE err;

  switch (char_encoding)
    {
    case ___CHAR_ENCODING_LATIN1:
      {
        ___LATIN1STRING cstr;
        if ((err = ___SCMOBJ_to_NONNULLLATIN1STRING (obj, &cstr, arg_num))
            == ___NO_ERR)
          *x = ___CAST(void*,cstr);
        break;
      }
    case ___CHAR_ENCODING_UTF8:
      {
        ___UTF8STRING cstr;
        if ((err = ___SCMOBJ_to_NONNULLUTF8STRING (obj, &cstr, arg_num))
            == ___NO_ERR)
          *x = ___CAST(void*,cstr);
        break;
      }
    case ___CHAR_ENCODING_UCS2:
      {
        ___UCS2STRING cstr;
        if ((err = ___SCMOBJ_to_NONNULLUCS2STRING (obj, &cstr, arg_num))
            == ___NO_ERR)
          *x = ___CAST(void*,cstr);
        break;
      }
    case ___CHAR_ENCODING_UCS4:
      {
        ___UCS4STRING cstr;
        if ((err = ___SCMOBJ_to_NONNULLUCS4STRING (obj, &cstr, arg_num))
            == ___NO_ERR)
          *x = ___CAST(void*,cstr);
        break;
      }
    case ___CHAR_ENCODING_NATIVE:
      {
        char *cstr;
        if ((err = ___SCMOBJ_to_NONNULLCHARSTRING (obj, &cstr, arg_num))
            == ___NO_ERR)
          *x = ___CAST(void*,cstr);
        break;
      }
    default:
      err = ___UNKNOWN_ERR;
      break;
    }

  return err;
}


___LOCAL ___ERR_CODE char_encoding_err_code
   ___P((int char_encoding,
         ___BOOL ctos,
         ___BOOL string_list),
        (char_encoding,
         ctos,
         string_list)
int char_encoding;
___BOOL ctos;
___BOOL string_list;)
{
  ___ERR_CODE *t;

  switch (char_encoding)
    {
    case ___CHAR_ENCODING_LATIN1:
      {
        static ___ERR_CODE tbl[4] = { ___STOC_NONNULLLATIN1STRING_ERR,
                                      ___STOC_NONNULLLATIN1STRINGLIST_ERR,
                                      ___CTOS_NONNULLLATIN1STRING_ERR,
                                      ___CTOS_NONNULLLATIN1STRINGLIST_ERR };
        t = tbl;
        break;
      }
    case ___CHAR_ENCODING_UTF8:
    case ___CHAR_ENCODING_UCS2:
    case ___CHAR_ENCODING_UCS4:
    case ___CHAR_ENCODING_NATIVE:
    default:
      {
        static ___ERR_CODE tbl[4] = { ___STOC_NONNULLCHARSTRING_ERR,
                                      ___STOC_NONNULLCHARSTRINGLIST_ERR,
                                      ___CTOS_NONNULLCHARSTRING_ERR,
                                      ___CTOS_NONNULLCHARSTRINGLIST_ERR };
        t = tbl;
        break;
      }
    }

  return t[ctos*2 + string_list];
}


/* Convert a Scheme list of strings to a nonnull C string list. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLSTRINGLIST)
   ___P((___SCMOBJ obj,
         void **x,
         int arg_num,
         int char_encoding),
        (obj,
         x,
         arg_num,
         char_encoding)
___SCMOBJ obj;
void **x;
int arg_num;
int char_encoding;)
{
  ___ERR_CODE err;
  void **string_list;
  ___SCMOBJ lst1;
  ___SCMOBJ lst2;
  int len;
  int i;

  lst1 = obj;
  lst2 = obj;
  len = 0;

  while (___PAIRP(lst1)) /* compute length, checking for circular lists */
    {
      lst1 = ___CDR(lst1);
      len++;
      if (___EQP(lst1,lst2) || !___PAIRP(lst1))
        break;
      lst1 = ___CDR(lst1);
      lst2 = ___CDR(lst2);
      len++;
    }

  if (!___NULLP(lst1))
    return char_encoding_err_code (char_encoding, 0, 1) + arg_num;

  string_list = ___CAST(void**, ___alloc_mem ((len + 1) * sizeof (void*)));

  if (string_list == 0)
    return ___STOC_HEAP_OVERFLOW_ERR+arg_num;

  err = ___NO_ERR;
  lst1 = obj;
  i = 0;

  while (i < len)
    {
      if ((err = ___SCMOBJ_to_NONNULLSTRING
                   (___CAR(lst1),
                    &string_list[i],
                    arg_num,
                    char_encoding))
          != ___NO_ERR)
        {
          if (err == char_encoding_err_code (char_encoding, 0, 0) + arg_num)
            err = char_encoding_err_code (char_encoding, 0, 1) + arg_num;
          break;
        }

      i++;
      lst1 = ___CDR(lst1);
    }

  string_list[i] = NULL;

  if (err != ___NO_ERR)
    {
      ___release_string_list (string_list);
      return err;
    }

  *x = string_list;

  return ___NO_ERR;
}


/* Convert a Scheme list of strings to a nonnull C 'char *' list. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLCHARSTRINGLIST)
   ___P((___SCMOBJ obj,
         char ***x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
char ***x;
int arg_num;)
{
  void *string_list;
  ___ERR_CODE err;

  if ((err = ___SCMOBJ_to_NONNULLSTRINGLIST
               (obj,
                &string_list,
                arg_num,
                ___CHAR_ENCODING_NATIVE))
      == ___NO_ERR)
    *x = ___CAST(char**,string_list);

  return err;
}


/* Convert a Scheme list of strings to a nonnull C LATIN-1 encoded character string list. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLLATIN1STRINGLIST)
   ___P((___SCMOBJ obj,
         ___LATIN1STRING **x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___LATIN1STRING **x;
int arg_num;)
{
  void *string_list;
  ___ERR_CODE err;

  if ((err = ___SCMOBJ_to_NONNULLSTRINGLIST
               (obj,
                &string_list,
                arg_num,
                ___CHAR_ENCODING_LATIN1))
      == ___NO_ERR)
    *x = ___CAST(___LATIN1STRING*,string_list);

  return err;
}


/* Convert a Scheme list of strings to a nonnull C UTF-8 encoded character string list. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLUTF8STRINGLIST)
   ___P((___SCMOBJ obj,
         ___UTF8STRING **x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UTF8STRING **x;
int arg_num;)
{
  void *string_list;
  ___ERR_CODE err;

  if ((err = ___SCMOBJ_to_NONNULLSTRINGLIST
               (obj,
                &string_list,
                arg_num,
                ___CHAR_ENCODING_UTF8))
      == ___NO_ERR)
    *x = ___CAST(___UTF8STRING*,string_list);

  return err;
}


/* Convert a Scheme list of strings to a nonnull C UCS-2 encoded character string list. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLUCS2STRINGLIST)
   ___P((___SCMOBJ obj,
         ___UCS2STRING **x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UCS2STRING **x;
int arg_num;)
{
  void *string_list;
  ___ERR_CODE err;

  if ((err = ___SCMOBJ_to_NONNULLSTRINGLIST
               (obj,
                &string_list,
                arg_num,
                ___CHAR_ENCODING_UCS2))
      == ___NO_ERR)
    *x = ___CAST(___UCS2STRING*,string_list);

  return err;
}


/* Convert a Scheme list of strings to a nonnull C UCS-4 encoded character string list. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLUCS4STRINGLIST)
   ___P((___SCMOBJ obj,
         ___UCS4STRING **x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___UCS4STRING **x;
int arg_num;)
{
  void *string_list;
  ___ERR_CODE err;

  if ((err = ___SCMOBJ_to_NONNULLSTRINGLIST
               (obj,
                &string_list,
                arg_num,
                ___CHAR_ENCODING_UCS4))
      == ___NO_ERR)
    *x = ___CAST(___UCS4STRING*,string_list);

  return err;
}


/* Convert a Scheme object to a variant. */

___EXP_FUNC(___ERR_CODE,___SCMOBJ_to_VARIANT)
   ___P((___SCMOBJ obj,
         ___VARIANT *x,
         int arg_num),
        (obj,
         x,
         arg_num)
___SCMOBJ obj;
___VARIANT *x;
int arg_num;)
{
  /*
   * Not yet implemented.
   */
  return ___STOC_VARIANT_ERR+arg_num;
}


/*
 * Release a Scheme foreign object, calling the object's release
 * function if it hasn't been done yet.
 */

___EXP_FUNC(___SCMOBJ,___release_foreign)
   ___P((___SCMOBJ obj),
        (obj)
___SCMOBJ obj;)
{
  ___SCMOBJ e;
  ___SCMOBJ (*release_fn) ___P((void *ptr),());
  void *ptr;

  if (!___TESTSUBTYPE(obj,___sFOREIGN))
    return ___FIX(___UNKNOWN_ERR);

  release_fn = ___CAST(___SCMOBJ (*) ___P((void *ptr),()),
                       ___FIELD(obj,___FOREIGN_RELEASE_FN));

  if (release_fn != 0)
    {
      ptr = ___CAST(void*,___FIELD(obj,___FOREIGN_PTR));
      ___FIELD(obj,___FOREIGN_RELEASE_FN) =
        ___CAST(___SCMOBJ,___CAST(___SCMOBJ (*) ___P((void *ptr),()),0));
      ___FIELD(obj,___FOREIGN_PTR) =
        ___CAST(___SCMOBJ,___CAST(void*,0));
      if ((e = release_fn (ptr)) != ___FIX(___NO_ERR))
        return e;
    }

  return ___FIX(___NO_ERR);
}


/* Free storage allocated to a C pointer created by the C-interface. */

___EXP_FUNC(___SCMOBJ,___release_pointer)
   ___P((void *x),
        (x)
void *x;)
{
  /* 
   * Nothing needs to be done because the data pointed to by the
   * pointer is under the control of the "C world".
   */
  return ___FIX(___NO_ERR);
}


/* Free storage allocated to a C function created by the C-interface. */

___EXP_FUNC(___SCMOBJ,___release_function)
   ___P((void *x),
        (x)
void *x;)
{
  /* 
   * At present nothing needs to be done because all C functions
   * created by the C-interface are permanently allocated.
   */
  return ___FIX(___NO_ERR);
}


/* Free storage of a C string created by the C-interface. */

___EXP_FUNC(___SCMOBJ,___release_string)
   ___P((void *x),
        (x)
void *x;)
{
  if (x != 0)
    ___free_mem (x);

  return ___FIX(___NO_ERR);
}


/* Free storage of a C string list created by the C-interface. */

___EXP_FUNC(___SCMOBJ,___release_string_list)
   ___P((void *x),
        (x)
void *x;)
{
  void **string_list = ___CAST(void**,x);
  void *elem;
  int i = 0;

  while ((elem = string_list[i++]) != 0)
    ___free_mem (elem);

  ___free_mem (string_list);

  return ___FIX(___NO_ERR);
}




/* Free storage of a variant created by the C-interface. */

___EXP_FUNC(___SCMOBJ,___release_variant)
   ___P((___VARIANT x),
        (x)
___VARIANT x;)
{
  /*
   * Not yet implemented.
   */
  return ___FIX(___UNKNOWN_ERR);
}


/*---------------------------------------------------------------------------*/

/* C to Scheme conversion */

/* 
 * The C to Scheme conversion functions may allocate memory in the
 * Scheme heap.  However, all objects allocated are still objects with
 * a reference count of 1.  The only special processing that must be
 * performed by the caller of a C to Scheme conversion function is a
 * call to '___release_scmobj' when the caller no longer needs a
 * pointer to the object.  This call is generated automatically by the
 * C-interface.
 */


/* Free storage allocated to a Scheme object created by the C interface. */

___EXP_FUNC(void,___release_scmobj)
   ___P((___SCMOBJ obj),
        (obj)
___SCMOBJ obj;)
{
  if (___MEM_ALLOCATED(obj) &&
      ___HD_TYP(___BODY(obj)[-1]) == ___STILL)
    ___still_obj_refcount_dec (obj);
}


/* Convert a C '___S64' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___S64_to_SCMOBJ)
   ___P((___S64 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___S64 x;
___SCMOBJ *obj;
int arg_num;)
{
  ___SCMOBJ r;

  if (___S64_fits_in_width (x, ___SCMOBJ_WIDTH-___TB))
    {
#if ___SCMOBJ_WIDTH == 32
      r = ___FIX(___S64_to_LONG (x));
#else
      r = ___FIX(x);
#endif
    }
  else
    {
#if ___BIG_ABASE_WIDTH == 32
      int n;
      ___BIGADIGIT d0 = ___S64_lo32 (x);
      ___BIGADIGIT d1 = ___CAST_U32(___S64_hi32 (x));

      if ((d0 & ___BIG_ABASE_DIV_2) == 0)
        n = 1 + (d1 != 0);
      else
        n = 1 + (d1 != ___BIG_ABASE_MIN_1);

      r = ___alloc_scmobj (___sBIGNUM, n<<2, ___STILL);

      if (___FIXNUMP(r))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }

      ___BIGASTORE(___BODY_AS(r,___tSUBTYPED),0,d0);
      if (n == 2)
        ___BIGASTORE(___BODY_AS(r,___tSUBTYPED),1,d1);
#else
      int n = 1;
      ___BIGADIGIT d0 = x;

      r = ___alloc_scmobj (___sBIGNUM, n<<3, ___STILL);

      if (___FIXNUMP(r))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }

      ___BIGASTORE(___BODY_AS(r,___tSUBTYPED),0,d0);
#endif
    }

  *obj = r;
  return ___NO_ERR;
}


/* Convert a C '___U64' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___U64_to_SCMOBJ)
   ___P((___U64 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___U64 x;
___SCMOBJ *obj;
int arg_num;)
{
  ___SCMOBJ r;

  if (___U64_fits_in_width (x, ___SCMOBJ_WIDTH-___TB-1))
    {
#if ___SCMOBJ_WIDTH == 32
      r = ___FIX(___U64_to_ULONG (x));
#else
      r = ___FIX(x);
#endif
    }
  else
    {
#if ___BIG_ABASE_WIDTH == 32
      int n;
      ___BIGADIGIT d0 = ___U64_lo32 (x);
      ___BIGADIGIT d1 = ___U64_hi32 (x);

      if (d1 == 0)
        n = 1 + ((d0 & ___BIG_ABASE_DIV_2) != 0);
      else
        n = 2 + ((d1 & ___BIG_ABASE_DIV_2) != 0);

      r = ___alloc_scmobj (___sBIGNUM, n<<2, ___STILL);

      if (___FIXNUMP(r))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }

      ___BIGASTORE(___BODY_AS(r,___tSUBTYPED),0,d0);
      if (n >= 2)
        {
          ___BIGASTORE(___BODY_AS(r,___tSUBTYPED),1,d1);
          if (n >= 3)
            ___BIGASTORE(___BODY_AS(r,___tSUBTYPED),2,0);
        }
#else
      int n;
      ___BIGADIGIT d0 = x;

      n = 1 + ((d0 & ___BIG_ABASE_DIV_2) != 0);

      r = ___alloc_scmobj (___sBIGNUM, n<<3, ___STILL);

      if (___FIXNUMP(r))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }

      ___BIGASTORE(___BODY_AS(r,___tSUBTYPED),0,d0);
      if (n == 2)
        ___BIGASTORE(___BODY_AS(r,___tSUBTYPED),1,0);
#endif
    }

  *obj = r;
  return ___NO_ERR;
}


/* Convert a C '___S8' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___S8_to_SCMOBJ)
   ___P((___S8 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___S8 x;
___SCMOBJ *obj;
int arg_num;)
{
  /*
   * No error possible because a '___S8' always fits in a Scheme
   * fixnum.
   */
  *obj = ___FIX(x);
  return ___NO_ERR;
}


/* Convert a C '___U8' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___U8_to_SCMOBJ)
   ___P((___U8 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___U8 x;
___SCMOBJ *obj;
int arg_num;)
{
  /*
   * No error possible because a '___U8' always fits in a Scheme
   * fixnum.
   */
  *obj = ___FIX(x);
  return ___NO_ERR;
}


/* Convert a C '___S16' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___S16_to_SCMOBJ)
   ___P((___S16 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___S16 x;
___SCMOBJ *obj;
int arg_num;)
{
  /*
   * No error possible because a '___S16' always fits in a Scheme
   * fixnum.
   */
  *obj = ___FIX(x);
  return ___NO_ERR;
}


/* Convert a C '___U16' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___U16_to_SCMOBJ)
   ___P((___U16 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___U16 x;
___SCMOBJ *obj;
int arg_num;)
{
  /*
   * No error possible because a '___U16' always fits in a Scheme
   * fixnum.
   */
  *obj = ___FIX(x);
  return ___NO_ERR;
}


/* Convert a C '___S32' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___S32_to_SCMOBJ)
   ___P((___S32 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___S32 x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___S64_to_SCMOBJ (___S64_from_SM32 (x), obj, arg_num);
}


/* Convert a C '___U32' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___U32_to_SCMOBJ)
   ___P((___U32 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___U32 x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___U64_to_SCMOBJ (___U64_from_UM32 (x), obj, arg_num);
}


/* Convert a C '___F64' to a Scheme flonum. */

___EXP_FUNC(___ERR_CODE,___F64_to_SCMOBJ)
   ___P((___F64 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___F64 x;
___SCMOBJ *obj;
int arg_num;)
{
  ___SCMOBJ r = ___alloc_scmobj (___sFLONUM, ___FLONUM_SIZE<<___LWS, ___STILL);

  if (___FIXNUMP(r))
    {
      *obj = ___FAL;
      return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
    }

  ___FLONUM_VAL(r) = x;

  *obj = r;
  return ___NO_ERR;
}


/* Convert a C '___F32' to a Scheme flonum. */

___EXP_FUNC(___ERR_CODE,___F32_to_SCMOBJ)
   ___P((___F32 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___F32 x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___F64_to_SCMOBJ (___CAST(___F64,x), obj, arg_num);
}


/* Convert a C 'char' to a Scheme character. */

___EXP_FUNC(___ERR_CODE,___CHAR_to_SCMOBJ)
   ___P((char x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
char x;
___SCMOBJ *obj;
int arg_num;)
{
  /*
   * No error possible because a 'char' always fits in
   * a Scheme character.
   */
  *obj = ___CHR(uchar_to_ucs4 (___CAST(unsigned char,x)));
  return ___NO_ERR;
}


/* Convert a C 'signed char' to a Scheme character. */

___EXP_FUNC(___ERR_CODE,___SCHAR_to_SCMOBJ)
   ___P((___SCHAR x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___SCHAR x;
___SCMOBJ *obj;
int arg_num;)
{
  /*
   * No error possible because a 'signed char' always fits in
   * a Scheme character.
   */
  *obj = ___CHR(uchar_to_ucs4 (___CAST(unsigned char,x)));
  return ___NO_ERR;
}


/* Convert a C 'unsigned char' to a Scheme character. */

___EXP_FUNC(___ERR_CODE,___UCHAR_to_SCMOBJ)
   ___P((unsigned char x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
unsigned char x;
___SCMOBJ *obj;
int arg_num;)
{
  /*
   * No error possible because an 'unsigned char' always fits in
   * a Scheme character.
   */
  *obj = ___CHR(uchar_to_ucs4 (x));
  return ___NO_ERR;
}


/* Convert a C LATIN-1 encoded character to a Scheme character. */

___EXP_FUNC(___ERR_CODE,___LATIN1_to_SCMOBJ)
   ___P((___LATIN1 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___LATIN1 x;
___SCMOBJ *obj;
int arg_num;)
{
  /*
   * No error possible because a LATIN-1 character always fits in
   * a Scheme character.
   */
  *obj = ___CHR(x);
  return ___NO_ERR;
}


/* Convert a C UCS-2 encoded character to a Scheme character. */

___EXP_FUNC(___ERR_CODE,___UCS2_to_SCMOBJ)
   ___P((___UCS2 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UCS2 x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x > ___MAX_CHR) /* check that we are not truncating the character */
    {
      *obj = ___FAL;
      return ___CTOS_UCS2_ERR+arg_num;
    }

  *obj = ___CHR(x);
  return ___NO_ERR;
}


/* Convert a C UCS-4 encoded character to a Scheme character. */

___EXP_FUNC(___ERR_CODE,___UCS4_to_SCMOBJ)
   ___P((___UCS4 x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UCS4 x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x > ___MAX_CHR) /* check that we are not truncating the character */
    {
      *obj = ___FAL;
      return ___CTOS_UCS4_ERR+arg_num;
    }

  *obj = ___CHR(x);
  return ___NO_ERR;
}


/* Convert a C 'short' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___SHORT_to_SCMOBJ)
   ___P((short x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
short x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___S64_to_SCMOBJ (___S64_from_LONG (___CAST(long,x)), obj, arg_num);
}


/* Convert a C 'unsigned short' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___USHORT_to_SCMOBJ)
   ___P((unsigned short x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
unsigned short x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___U64_to_SCMOBJ (___U64_from_ULONG (___CAST(unsigned long,x)), obj, arg_num);
}


/* Convert a C 'int' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___INT_to_SCMOBJ)
   ___P((int x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
int x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___S64_to_SCMOBJ (___S64_from_LONG (___CAST(long,x)), obj, arg_num);
}


/* Convert a C 'unsigned int' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___UINT_to_SCMOBJ)
   ___P((unsigned int x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
unsigned int x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___U64_to_SCMOBJ (___U64_from_ULONG (___CAST(unsigned long,x)), obj, arg_num);
}


/* Convert a C 'long' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___LONG_to_SCMOBJ)
   ___P((long x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
long x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___S64_to_SCMOBJ (___S64_from_LONG (x), obj, arg_num);
}


/* Convert a C 'unsigned long' to a Scheme integer. */

___EXP_FUNC(___ERR_CODE,___ULONG_to_SCMOBJ)
   ___P((unsigned long x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
unsigned long x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___U64_to_SCMOBJ (___U64_from_ULONG (x), obj, arg_num);
}


/* Convert a C 'float' to a Scheme flonum. */

___EXP_FUNC(___ERR_CODE,___FLOAT_to_SCMOBJ)
   ___P((float x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
float x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___F64_to_SCMOBJ (___CAST(___F64,x), obj, arg_num);
}


/* Convert a C 'double' to a Scheme flonum. */

___EXP_FUNC(___ERR_CODE,___DOUBLE_to_SCMOBJ)
   ___P((double x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
double x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___F64_to_SCMOBJ (___CAST(___F64,x), obj, arg_num);
}


/* Convert a C struct pointer to a Scheme foreign object. */

___EXP_FUNC(___ERR_CODE,___STRUCT_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ tag,
         ___SCMOBJ (*release_fn) ___P((void *ptr),()),
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         tag,
         release_fn,
         obj,
         arg_num)
void *x;
___SCMOBJ tag;
___SCMOBJ (*release_fn) ___P((void *ptr),());
___SCMOBJ *obj;
int arg_num;)
{
  return ___UNKNOWN_ERR;
}


/* Convert a C union pointer to a Scheme foreign object. */

___EXP_FUNC(___ERR_CODE,___UNION_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ tag,
         ___SCMOBJ (*release_fn) ___P((void *ptr),()),
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         tag,
         release_fn,
         obj,
         arg_num)
void *x;
___SCMOBJ tag;
___SCMOBJ (*release_fn) ___P((void *ptr),());
___SCMOBJ *obj;
int arg_num;)
{
  return ___UNKNOWN_ERR;
}


/* Convert a C type pointer to a Scheme foreign object. */

___EXP_FUNC(___ERR_CODE,___TYPE_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ tag,
         ___SCMOBJ (*release_fn) ___P((void *ptr),()),
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         tag,
         release_fn,
         obj,
         arg_num)
void *x;
___SCMOBJ tag;
___SCMOBJ (*release_fn) ___P((void *ptr),());
___SCMOBJ *obj;
int arg_num;)
{
  return ___UNKNOWN_ERR;
}


/* Convert a C pointer to a Scheme foreign object. */

___EXP_FUNC(___ERR_CODE,___POINTER_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ tag,
         ___SCMOBJ (*release_fn) ___P((void *ptr),()),
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         tag,
         release_fn,
         obj,
         arg_num)
void *x;
___SCMOBJ tag;
___SCMOBJ (*release_fn) ___P((void *ptr),());
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    *obj = ___FAL; /* #f counts as NULL */
  else
    {
      long i;
      ___SCMOBJ r = ___alloc_scmobj (___sFOREIGN,
                                     ___FOREIGN_SIZE<<___LWS,
                                     ___STILL);
      if (___FIXNUMP(r))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }
      ___FIELD(r,___FOREIGN_TAG) = tag;
      ___FIELD(r,___FOREIGN_RELEASE_FN) = ___CAST(___SCMOBJ,release_fn);
      ___FIELD(r,___FOREIGN_PTR) = ___CAST(___SCMOBJ,x);
      *obj = r;
    }
  return ___NO_ERR;
}


/* Convert a nonnull C pointer to a Scheme foreign object. */

___EXP_FUNC(___ERR_CODE,___NONNULLPOINTER_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ tag,
         ___SCMOBJ (*release_fn) ___P((void *ptr),()),
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         tag,
         release_fn,
         obj,
         arg_num)
void *x;
___SCMOBJ tag;
___SCMOBJ (*release_fn) ___P((void *ptr),());
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    {
      *obj = ___FAL;
      return ___CTOS_NONNULLPOINTER_ERR+arg_num;
    }
  return ___POINTER_to_SCMOBJ (x, tag, release_fn, obj, arg_num);
}


/* Convert a C function to a Scheme function. */

___EXP_FUNC(___ERR_CODE,___FUNCTION_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
void *x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    {
      *obj = ___FAL; /* #f counts as NULL */
      return ___NO_ERR;
    }
  /* 
   * At present, arbitrary C functions cannot be converted to Scheme
   * functions.
   */
  *obj = ___FAL;
  return ___CTOS_FUNCTION_ERR+arg_num;
}


/* Convert a nonnull C function to a Scheme function. */

___EXP_FUNC(___ERR_CODE,___NONNULLFUNCTION_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
void *x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    {
      *obj = ___FAL;
      return ___CTOS_NONNULLFUNCTION_ERR+arg_num;
    }
  return ___FUNCTION_to_SCMOBJ (x, obj, arg_num);
}


/* Convert a C extended boolean to a Scheme boolean. */

___EXP_FUNC(___ERR_CODE,___BOOL_to_SCMOBJ)
   ___P((___BOOL x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___BOOL x;
___SCMOBJ *obj;
int arg_num;)
{
  *obj = x ? ___TRU : ___FAL;
  return ___NO_ERR;
}


/* Convert a C 'char *' to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___CHARSTRING_to_SCMOBJ)
   ___P((char *x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
char *x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    *obj = ___FAL; /* #f counts as NULL */
  else
    {
      ___SCMOBJ r;
      unsigned long i, n = 0;
      while (x[n] != 0)
        n++;
      r = ___alloc_scmobj (___sSTRING, n<<___LCS, ___STILL);
      if (___FIXNUMP(r))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }
      for (i=0; i<n; i++)
        {
          /*
           * No error possible because a 'char' always fits in a
           * Scheme character.
           */
          ___UCS4 c = uchar_to_ucs4 (___CAST(unsigned char,x[i]));
          ___STRINGSET(r,___FIX(i),___CHR(c))
        }
      *obj = r;
    }
  return ___NO_ERR;
}


/* Convert a nonnull C 'char *' to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___NONNULLCHARSTRING_to_SCMOBJ)
   ___P((char *x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
char *x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    {
      *obj = ___FAL;
      return ___CTOS_NONNULLCHARSTRING_ERR+arg_num;
    }
  return ___CHARSTRING_to_SCMOBJ (x, obj, arg_num);
}

/* Convert a C LATIN-1 encoded character string to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___LATIN1STRING_to_SCMOBJ)
   ___P((___LATIN1STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___LATIN1STRING x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    *obj = ___FAL; /* #f counts as NULL */
  else
    {
      ___SCMOBJ r;
      unsigned long i, n = 0;
      while (x[n] != 0)
        n++;
      r = ___alloc_scmobj (___sSTRING, n<<___LCS, ___STILL);
      if (___FIXNUMP(r))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }
      for (i=0; i<n; i++)
        {
          /*
           * No error possible because a LATIN-1 character always fits
           * in a Scheme character.
           */
          ___UCS4 c = x[i];
          ___STRINGSET(r,___FIX(i),___CHR(c))
        }
      *obj = r;
    }
  return ___NO_ERR;
}


/* Convert a nonnull C LATIN-1 encoded character string to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___NONNULLLATIN1STRING_to_SCMOBJ)
   ___P((___LATIN1STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___LATIN1STRING x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    {
      *obj = ___FAL;
      return ___CTOS_NONNULLLATIN1STRING_ERR+arg_num;
    }
  return ___LATIN1STRING_to_SCMOBJ (x, obj, arg_num);
}


/* Convert a C UTF-8 encoded character string to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___UTF8STRING_to_SCMOBJ)
   ___P((___UTF8STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UTF8STRING x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    *obj = ___FAL; /* #f counts as NULL */
  else
    {
      ___SCMOBJ r;
      unsigned long i, n = 0;
      ___UTF8STRING p = x;
      while (___utf8_get (&p) != 0) /* advance until end or error */
        n++;
      r = ___alloc_scmobj (___sSTRING, n<<___LCS, ___STILL);
      if (___FIXNUMP(r))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }
      p = x;
      for (i=0; i<n; i++)
        {
          ___UTF8STRING start = p;
          ___UCS4 c = ___utf8_get (&p);
          if (p == start || c > ___MAX_CHR)
            {
              ___release_scmobj (r);
              *obj = ___FAL;
              return ___CTOS_UTF8STRING_ERR+arg_num;
            }
          ___STRINGSET(r,___FIX(i),___CHR(c))
        }
      *obj = r;
    }
  return ___NO_ERR;
}


/* Convert a nonnull C UTF-8 encoded character string to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___NONNULLUTF8STRING_to_SCMOBJ)
   ___P((___UTF8STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UTF8STRING x;
___SCMOBJ *obj;
int arg_num;)
{
  ___ERR_CODE err;
  if (x == 0)
    {
      *obj = ___FAL;
      return ___CTOS_NONNULLUTF8STRING_ERR+arg_num;
    }
  err = ___UTF8STRING_to_SCMOBJ (x, obj, arg_num);
  if (err == ___CTOS_UTF8STRING_ERR+arg_num)
    err = ___CTOS_NONNULLUTF8STRING_ERR+arg_num;
  return err;
}


/* Convert a C UCS-2 encoded character string to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___UCS2STRING_to_SCMOBJ)
   ___P((___UCS2STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UCS2STRING x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    *obj = ___FAL; /* #f counts as NULL */
  else
    {
      ___SCMOBJ r;
      unsigned long i, n = 0;
      while (x[n] != 0)
        n++;
      r = ___alloc_scmobj (___sSTRING, n<<___LCS, ___STILL);
      if (___FIXNUMP(r))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }
      for (i=0; i<n; i++)
        {
          ___UCS4 c = x[i];
          if (c > ___MAX_CHR)
            {
              ___release_scmobj (r);
              *obj = ___FAL;
              return ___CTOS_UCS2STRING_ERR+arg_num;
            }
          ___STRINGSET(r,___FIX(i),___CHR(c))
        }
      *obj = r;
    }
  return ___NO_ERR;
}


/* Convert a nonnull C UCS-2 encoded character string to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___NONNULLUCS2STRING_to_SCMOBJ)
   ___P((___UCS2STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UCS2STRING x;
___SCMOBJ *obj;
int arg_num;)
{
  ___ERR_CODE err;
  if (x == 0)
    {
      *obj = ___FAL;
      return ___CTOS_NONNULLUCS2STRING_ERR+arg_num;
    }
  err = ___UCS2STRING_to_SCMOBJ (x, obj, arg_num);
  if (err == ___CTOS_UCS2STRING_ERR+arg_num)
    err = ___CTOS_NONNULLUCS2STRING_ERR+arg_num;
  return err;
}

/* Convert a C UCS-4 encoded character string to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___UCS4STRING_to_SCMOBJ)
   ___P((___UCS4STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UCS4STRING x;
___SCMOBJ *obj;
int arg_num;)
{
  if (x == 0)
    *obj = ___FAL; /* #f counts as NULL */
  else
    {
      ___SCMOBJ r;
      unsigned long i, n = 0;
      while (x[n] != 0)
        n++;
      r = ___alloc_scmobj (___sSTRING, n<<___LCS, ___STILL);
      if (___FIXNUMP(r))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }
      for (i=0; i<n; i++)
        {
          ___UCS4 c = x[i];
          if (c > ___MAX_CHR)
            {
              ___release_scmobj (r);
              *obj = ___FAL;
              return ___CTOS_UCS4STRING_ERR+arg_num;
            }
          ___STRINGSET(r,___FIX(i),___CHR(c))
        }
      *obj = r;
    }
  return ___NO_ERR;
}


/* Convert a nonnull C UCS-4 encoded character string to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___NONNULLUCS4STRING_to_SCMOBJ)
   ___P((___UCS4STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UCS4STRING x;
___SCMOBJ *obj;
int arg_num;)
{
  ___ERR_CODE err;
  if (x == 0)
    {
      *obj = ___FAL;
      return ___CTOS_NONNULLUCS4STRING_ERR+arg_num;
    }
  err = ___UCS4STRING_to_SCMOBJ (x, obj, arg_num);
  if (err == ___CTOS_UCS4STRING_ERR+arg_num)
    err = ___CTOS_NONNULLUCS4STRING_ERR+arg_num;
  return err;
}


/* Convert a nonnull C string to a Scheme string. */

___EXP_FUNC(___ERR_CODE,___NONNULLSTRING_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ *obj,
         int arg_num,
         int char_encoding),
        (x,
         obj,
         arg_num,
         char_encoding)
void *x;
___SCMOBJ *obj;
int arg_num;
int char_encoding;)
{
  switch (char_encoding)
    {
    case ___CHAR_ENCODING_LATIN1:
      return ___NONNULLLATIN1STRING_to_SCMOBJ
               (___CAST(___LATIN1STRING,x),
                obj,
                arg_num);
    case ___CHAR_ENCODING_UTF8:
      return ___NONNULLUTF8STRING_to_SCMOBJ
               (___CAST(___UTF8STRING,x),
                obj,
                arg_num);
    case ___CHAR_ENCODING_UCS2:
      return ___NONNULLUCS2STRING_to_SCMOBJ
               (___CAST(___UCS2STRING,x),
                obj,
                arg_num);
    case ___CHAR_ENCODING_UCS4:
      return ___NONNULLUCS4STRING_to_SCMOBJ
               (___CAST(___UCS4STRING,x),
                obj,
                arg_num);
    case ___CHAR_ENCODING_NATIVE:
      return ___NONNULLCHARSTRING_to_SCMOBJ
               (___CAST(char*,x),
                obj,
                arg_num);
    }

  *obj = ___FAL;

  return ___UNKNOWN_ERR;
}


/* Convert a nonnull C string list to a Scheme string list. */

___EXP_FUNC(___ERR_CODE,___NONNULLSTRINGLIST_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ *obj,
         int arg_num,
         int char_encoding),
        (x,
         obj,
         arg_num,
         char_encoding)
void *x;
___SCMOBJ *obj;
int arg_num;
int char_encoding;)
{
  ___SCMOBJ lst;
  void **string_list = ___CAST(void**,x);
  int i = 0;

  while (string_list[i] != 0)
    i++;

  lst = ___NUL;

  while (i-- > 0)
    {
      ___ERR_CODE err;
      ___SCMOBJ str;
      ___SCMOBJ pair;

      if ((err = ___NONNULLSTRING_to_SCMOBJ
                   (string_list[i],
                    &str,
                    arg_num,
                    char_encoding))
          != ___NO_ERR)
        {
          ___release_scmobj (lst);
          *obj = ___FAL;
          if (err == char_encoding_err_code (char_encoding, 1, 0) + arg_num)
            err = char_encoding_err_code (char_encoding, 1, 1) + arg_num;
          return err;
        }

      pair = ___make_pair (str, lst, ___STILL);

      ___release_scmobj (str);
      ___release_scmobj (lst);

      if (___FIXNUMP(pair))
        {
          *obj = ___FAL;
          return ___CTOS_HEAP_OVERFLOW_ERR+arg_num;
        }

      lst = pair;
    }

  *obj = lst;

  return ___NO_ERR;
}


/* Convert a nonnull C 'char *' list to a Scheme list of strings. */

___EXP_FUNC(___ERR_CODE,___NONNULLCHARSTRINGLIST_to_SCMOBJ)
   ___P((char **x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
char **x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___NONNULLSTRINGLIST_to_SCMOBJ
           (___CAST(void*,x),
            obj,
            arg_num,
            ___CHAR_ENCODING_NATIVE);
}


/* Convert a nonnull C LATIN-1 encoded character string list to a Scheme list of strings. */

___EXP_FUNC(___ERR_CODE,___NONNULLLATIN1STRINGLIST_to_SCMOBJ)
   ___P((___LATIN1STRING *x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___LATIN1STRING *x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___NONNULLSTRINGLIST_to_SCMOBJ
           (___CAST(void*,x),
            obj,
            arg_num,
            ___CHAR_ENCODING_LATIN1);
}


/* Convert a nonnull C UTF-8 encoded character string list to a Scheme list of strings. */

___EXP_FUNC(___ERR_CODE,___NONNULLUTF8STRINGLIST_to_SCMOBJ)
   ___P((___UTF8STRING *x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UTF8STRING *x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___NONNULLSTRINGLIST_to_SCMOBJ
           (___CAST(void*,x),
            obj,
            arg_num,
            ___CHAR_ENCODING_UTF8);
}


/* Convert a nonnull C UCS-2 encoded character string list to a Scheme list of strings. */

___EXP_FUNC(___ERR_CODE,___NONNULLUCS2STRINGLIST_to_SCMOBJ)
   ___P((___UCS2STRING *x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UCS2STRING *x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___NONNULLSTRINGLIST_to_SCMOBJ
           (___CAST(void*,x),
            obj,
            arg_num,
            ___CHAR_ENCODING_UCS2);
}


/* Convert a nonnull C UCS-4 encoded character string list to a Scheme list of strings. */

___EXP_FUNC(___ERR_CODE,___NONNULLUCS4STRINGLIST_to_SCMOBJ)
   ___P((___UCS4STRING *x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___UCS4STRING *x;
___SCMOBJ *obj;
int arg_num;)
{
  return ___NONNULLSTRINGLIST_to_SCMOBJ
           (___CAST(void*,x),
            obj,
            arg_num,
            ___CHAR_ENCODING_UCS4);
}


/* Convert a variant to a Scheme object. */

___EXP_FUNC(___ERR_CODE,___VARIANT_to_SCMOBJ)
   ___P((___VARIANT x,
         ___SCMOBJ *obj,
         int arg_num),
        (x,
         obj,
         arg_num)
___VARIANT x;
___SCMOBJ *obj;
int arg_num;)
{
  /*
   * Not yet implemented.
   */
  return ___CTOS_VARIANT_ERR+arg_num;
}


/* Create a "c-define" stack marker. */

___EXP_FUNC(___ERR_CODE,___make_cdef_stack_marker)
   ___P((___SCMOBJ *marker),
        (marker)
___SCMOBJ *marker;)
{
  ___SCMOBJ stack_marker = ___make_vector (1, ___TRU, ___STILL);

  if (___FIXNUMP(stack_marker))
    return ___CDEF_HEAP_OVERFLOW_ERR;/*******************/

  *marker = stack_marker;

  return ___NO_ERR;
}


/* Invalidate a "c-define" stack marker.  Called when a "c-define" returns. */

___EXP_FUNC(void,___kill_cdef_stack_marker)
   ___P((___SCMOBJ marker),
        (marker)
___SCMOBJ marker;)
{
  ___FIELD(marker,0) = ___FAL; /* invalidate the C stack frame */
  ___still_obj_refcount_dec (marker); /* allow GC of stack marker */
}


/*---------------------------------------------------------------------------*/
