;==============================================================================

; file: "header.scm"

; Copyright (C) 1994-2002 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##namespace (""
   not boolean? eqv? eq? equal? pair? cons car cdr set-car! set-cdr!
   caar cadr cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr
   caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr cdaaar cdaadr
   cdadar cdaddr cddaar cddadr cdddar cddddr null? list? list length
   append reverse list-ref memq memv member assq assv assoc symbol?
   symbol->string string->symbol number? complex? real? rational?
   integer? exact? inexact? = < > <= >= zero? positive? negative? odd?
   even? max min + * - / abs quotient remainder modulo gcd lcm numerator
   denominator floor ceiling truncate round rationalize exp log sin cos
   tan asin acos atan sqrt expt make-rectangular make-polar real-part
   imag-part magnitude angle exact->inexact inexact->exact number->string
   string->number char? char=? char<? char>? char<=? char>=? char-ci=?
   char-ci<? char-ci>? char-ci<=? char-ci>=? char-alphabetic?
   char-numeric? char-whitespace? char-upper-case? char-lower-case?
   char->integer integer->char char-upcase char-downcase string?
   make-string string string-length string-ref string-set! string=?
   string<? string>? string<=? string>=? string-ci=? string-ci<?
   string-ci>? string-ci<=? string-ci>=? substring string-append vector?
   make-vector vector vector-length vector-ref vector-set! procedure?
   apply map for-each call-with-current-continuation call-with-input-file
   call-with-output-file input-port? output-port? current-input-port
   current-output-port open-input-file open-output-file close-input-port
   close-output-port eof-object? read read-char peek-char write display
   newline write-char

   eval error pp compile-file-to-c compile-file link-incremental link-flat

   script-arguments

   include
))

(##declare
  (multilisp)
  (extended-bindings)
  (not safe)
  (block)
  (fixnum)
  (inlining-limit 134)
)

;------------------------------------------------------------------------------

; General object representation.

; Type tags.

(##define-macro (macro-type-fixnum)   0)
(##define-macro (macro-type-subtyped) 1)
(##define-macro (macro-type-special)  2)
(##define-macro (macro-type-pair)     3)

; Subtype tags.

(##define-macro (macro-subtype-vector)       0)
(##define-macro (macro-subtype-pair)         1)
(##define-macro (macro-subtype-ratnum)       2)
(##define-macro (macro-subtype-cpxnum)       3)
(##define-macro (macro-subtype-structure)    4)
(##define-macro (macro-subtype-boxvalues)    5)
(##define-macro (macro-subtype-meroon)       6)

(##define-macro (macro-subtype-symbol)       8)
(##define-macro (macro-subtype-keyword)      9)
(##define-macro (macro-subtype-frame)        10)
(##define-macro (macro-subtype-contthread)   11)
(##define-macro (macro-subtype-promise)      12)
(##define-macro (macro-subtype-willtable)    13)
(##define-macro (macro-subtype-procedure)    14)
(##define-macro (macro-subtype-return)       15)

(##define-macro (macro-subtype-foreign)      18)
(##define-macro (macro-subtype-string)       19)
(##define-macro (macro-subtype-s8vector)     20)
(##define-macro (macro-subtype-u8vector)     21)
(##define-macro (macro-subtype-s16vector)    22)
(##define-macro (macro-subtype-u16vector)    23)
(##define-macro (macro-subtype-s32vector)    24)
(##define-macro (macro-subtype-u32vector)    25)
(##define-macro (macro-subtype-f32vector)    26)

; for alignment these 5 must be last:
(##define-macro (macro-subtype-s64vector)    27)
(##define-macro (macro-subtype-u64vector)    28)
(##define-macro (macro-subtype-f64vector)    29)
(##define-macro (macro-subtype-flonum)       30)
(##define-macro (macro-subtype-bignum)       31)

(##define-macro (macro-subtype-ovector? x) `(##fixnum.< ,x 8))
(##define-macro (macro-subtype-bvector? x) `(##fixnum.< 16 ,x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Special objects.

(##define-macro (macro-absent-obj)
  (##namespace ("c#" absent-object))
  `',absent-object)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Structure objects.

(##define-macro (macro-structure-object? obj tag)
  `(let ((obj ,obj))
     (and (##structure? obj)
          (##eq? (##vector-ref obj 0)
                 ,tag))))

(##define-macro (macro-make-structure-object tag . slots)
  `(##subtype-set!
    (##vector ,tag ,@slots)
    (macro-subtype-structure)))

(##define-macro (macro-make-constant-structure-object tag . slots)
  `(##subtype-set!
    '#(,tag ,@slots) ; the constant is side-effected... that's OK in Gambit
    (macro-subtype-structure)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; System procedure classes.

(##define-macro (define-prim form . exprs)

  (define inlinable-procs '(

##type ##type-cast ##subtype ##subtype-set!
##not ##null? ##unbound? ##eq? ##eof-object?
##fixnum? ##flonum? ##special? ##pair? ##subtyped? ##subtyped?
##subtyped.vector? ##subtyped.symbol? ##subtyped.flonum? ##subtyped.bignum?
##procedure? ##promise? ##vector? ##symbol? ##keyword? ##ratnum? ##cpxnum?
##string? ##structure? ##values? ##bignum?
##char? ;;;;  ##closure? ##subprocedure?
##fixnum.max ##fixnum.min
##fixnum.+ ##fixnum.- ##fixnum.*
##fixnum.quotient ##fixnum.remainder ##fixnum.modulo
##fixnum.bitwise-ior ##fixnum.bitwise-xor ##fixnum.bitwise-and ##fixnum.bitwise-not
##fixnum.arithmetic-shift-left
##fixnum.arithmetic-shift-right
##fixnum.logical-shift-right
##fixnum.zero? ##fixnum.positive? ##fixnum.negative?
##fixnum.odd? ##fixnum.even?
##fixnum.= ##fixnum.< ##fixnum.> ##fixnum.<= ##fixnum.>=
##fixnum.->char ##fixnum.<-char
##flonum.->fixnum ##flonum.<-fixnum
##flonum.max ##flonum.min
##flonum.+ ##flonum.- ##flonum.*  ##flonum./
##flonum.abs ##flonum.floor ##flonum.ceiling ##flonum.round
##flonum.exp ##flonum.log
##flonum.sin ##flonum.cos ##flonum.tan
##flonum.asin ##flonum.acos ##flonum.atan
##flonum.sqrt ##flonum.copysign
##flonum.zero? ##flonum.positive? ##flonum.negative?
##flonum.finite? ##flonum.nan? ##flonum.<-fixnum-exact?
##flonum.= ##flonum.< ##flonum.> ##flonum.<= ##flonum.>=
##char=? ##char<? ##char>? ##char<=? ##char>=?
##cons ##set-car! ##set-cdr! ##car ##cdr
##caar ##cadr ##cdar ##cddr
##caaar ##caadr ##cadar ##caddr ##cdaar ##cdadr ##cddar ##cdddr
##caaaar ##caaadr ##caadar ##caaddr ##cadaar ##cadadr ##caddar ##cadddr
##cdaaar ##cdaadr ##cdadar ##cdaddr ##cddaar ##cddadr ##cdddar ##cddddr
;;;##list
##box? ##box ##unbox ##set-box!
;;;##vector
##vector-length ##vector-ref ##vector-set! ##vector-shrink!
;;;##string
##string-length ##string-ref ##string-set! ##string-shrink!
##s8vector? ;;;##s8vector
##s8vector-length ##s8vector-ref ##s8vector-set! ##s8vector-shrink!
##u8vector? ;;;##u8vector
##u8vector-length ##u8vector-ref ##u8vector-set! ##u8vector-shrink!
##s16vector? ;;;##s16vector
##s16vector-length ##s16vector-ref ##s16vector-set! ##s16vector-shrink!
##u16vector? ;;;##u16vector
##u16vector-length ##u16vector-ref ##u16vector-set! ##u16vector-shrink!
##s32vector? ;;;##s32vector
##s32vector-length ##s32vector-ref ##s32vector-set! ##s32vector-shrink!
##u32vector? ;;;##u32vector
##u32vector-length ##u32vector-ref ##u32vector-set! ##u32vector-shrink!
##s64vector? ;;;##s64vector
##s64vector-length ##s64vector-ref ##s64vector-set! ##s64vector-shrink!
##u64vector? ;;;##u64vector
##u64vector-length ##u64vector-ref ##u64vector-set! ##u64vector-shrink!
##f32vector? ;;;##f32vector
##f32vector-length ##f32vector-ref ##f32vector-set! ##f32vector-shrink!
##f64vector? ;;;##f64vector
##f64vector-length ##f64vector-ref ##f64vector-set! ##f64vector-shrink!
;;; ##symbol->string ##keyword->string
##closure-code ##closure-ref ##closure-set!
;;; ##subprocedure-id ##subprocedure-parent
;;; ##procedure-info
##make-promise ##force ##void

##will? ##make-will ##will-testator

##global-var-ref
##global-var-set!

##bignum.negative?
##bignum.adigit-length
##bignum.adigit-set-ones!
##bignum.adigit-inc!
##bignum.adigit-dec!
##bignum.adigit-add!
##bignum.adigit-sub!
##bignum.mdigit-length
##bignum.mdigit-ref
##bignum.mdigit-set!
##bignum.mdigit-mul!
##bignum.mdigit-div!
##bignum.mdigit-quotient
##bignum.mdigit-remainder
##bignum.mdigit-test?

##bignum.adigit-ones?
##bignum.adigit-zero?
##bignum.adigit-negative?
##bignum.adigit-=
##bignum.adigit-<
##bignum.->fixnum
##bignum.<-fixnum
##bignum.adigit-shrink!
##bignum.adigit-set!
##bignum.adigit-cat!
##bignum.adigit-bitwise-and!
##bignum.adigit-bitwise-ior!
##bignum.adigit-bitwise-xor!
##bignum.adigit-bitwise-not!

##bignum.fdigit-length
##bignum.fdigit-ref
##bignum.fdigit-set!

))

  (let ((name
         (if (symbol? form)
           form
           (car form))))
    (let ((val
           (if (symbol? form)
             (if (and (pair? exprs) (null? (cdr exprs)))
               (car exprs)
               (error "Incorrect define-prim"))
             (if (memq name inlinable-procs)
               `(lambda ,(cdr form) ,form)
               (if (null? exprs)
                 (error "define-prim can't inline" name)
                 `(lambda ,(cdr form) ,@exprs))))))
      `(define ,name (let () (##declare (not inline)) ,val)))))

;------------------------------------------------------------------------------

; Macro to define structure accessors.

(##define-macro (macro-slot index struct . val)
  (if (null? val)
    `(##vector-ref ,struct ,index)
    `(##vector-set! ,struct ,index ,@val)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Parameter descriptor objects

; A parameter descriptor is represented by an object vector of length 3
; slot 0 = value of this parameter at the global level
; slot 1 = hash code (non-negative fixnum), must be at same place as symbols
; slot 2 = parameter filter procedure

(##define-macro (macro-make-parameter-descr value hash filter)
  `(##vector ,value ,hash ,filter))

(##define-macro (macro-parameter-descr-value p)         `(macro-slot 0 ,p))
(##define-macro (macro-parameter-descr-value-set! p x)  `(macro-slot 0 ,p ,x))
(##define-macro (macro-parameter-descr-hash p)          `(macro-slot 1 ,p))
(##define-macro (macro-parameter-descr-hash-set! p x)   `(macro-slot 1 ,p ,x))
(##define-macro (macro-parameter-descr-filter p)        `(macro-slot 2 ,p))
(##define-macro (macro-parameter-descr-filter-set! p x) `(macro-slot 2 ,p ,x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Symbol objects

; A symbol is represented by an object vector of length 4
; slot 0 = symbol name (a string or a fixnum <n> for a symbol named "g<n>")
; slot 1 = hash code (non-negative fixnum)
; slot 2 = link to next symbol in symbol table (#f for uninterned)
; slot 3 = pointer to corresponding global variable (0 if none exists)

(##define-macro (macro-make-uninterned-symbol name hash)
  `(##subtype-set!
    (##vector ,name ,hash #f 0)
    (macro-subtype-symbol)))

(##define-macro (macro-symbol-name s)        `(macro-slot 0 ,s))
(##define-macro (macro-symbol-name-set! s x) `(macro-slot 0 ,s ,x))
(##define-macro (macro-symbol-hash s)        `(macro-slot 1 ,s))
(##define-macro (macro-symbol-hash-set! s x) `(macro-slot 1 ,s ,x))
(##define-macro (macro-symbol-next s)        `(macro-slot 2 ,s))
(##define-macro (macro-symbol-next-set! s x) `(macro-slot 2 ,s ,x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Keyword objects

; A keyword is represented by an object vector of length 3
; slot 0 = keyword name (a string or a fixnum <n> for a keyword named "g<n>")
; slot 1 = hash code (non-negative fixnum)
; slot 2 = link to next keyword in keyword table (#f for uninterned)

(##define-macro (macro-make-uninterned-keyword name hash)
  `(##subtype-set!
    (##vector ,name ,hash #f)
    (macro-subtype-keyword)))

(##define-macro (macro-keyword-name k)        `(macro-slot 0 ,k))
(##define-macro (macro-keyword-name-set! k x) `(macro-slot 0 ,k ,x))
(##define-macro (macro-keyword-hash k)        `(macro-slot 1 ,k))
(##define-macro (macro-keyword-hash-set! k x) `(macro-slot 1 ,k ,x))
(##define-macro (macro-keyword-next k)        `(macro-slot 2 ,k))
(##define-macro (macro-keyword-next-set! k x) `(macro-slot 2 ,k ,x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Will objects

; A will is represented by an object vector of length 3
; slot 0 = link to next will in list of non-executable wills
; slot 1 = testator
; slot 2 = action procedure

(##define-macro (macro-make-will testator action)
  `(##make-will ,testator ,action))

(##define-macro (macro-will-next w)            `(macro-slot 0 ,w))
(##define-macro (macro-will-next-set! w x)     `(macro-slot 0 ,w ,x))
(##define-macro (macro-will-testator w)        `(macro-slot 1 ,w))
(##define-macro (macro-will-testator-set! w x) `(macro-slot 1 ,w ,x))
(##define-macro (macro-will-action w)          `(macro-slot 2 ,w))
(##define-macro (macro-will-action-set! w x)   `(macro-slot 2 ,w ,x))

(##define-macro (macro-will-execute! will)
  `(let ((will ,will))
     (##declare (not interrupts-enabled))
     (let ((testator (macro-will-testator will))
           (action (macro-will-action will)))
       (macro-will-testator-set! will #f) ; zap testator
       (macro-will-action-set! will #f)   ; and action procedure
       (let ()
         (##declare (interrupts-enabled))
         (if action
           (action testator))
         (##void)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Continuation objects

; A continuation is represented by an object vector of length 2
; slot 0 = frame object
; slot 1 = dynamic-environment

(##define-macro (macro-continuation-frame c)        `(macro-slot 0 ,c))
(##define-macro (macro-continuation-frame-set! c x) `(macro-slot 0 ,c ,x))
(##define-macro (macro-continuation-denv c)         `(macro-slot 1 ,c))
(##define-macro (macro-continuation-denv-set! c x)  `(macro-slot 1 ,c ,x))

;------------------------------------------------------------------------------

(##define-macro (macro-if-forces forces noforces)
  (if (memq 'force ##compilation-options)
    forces
    noforces))

(##define-macro (macro-force-vars vars expr)
  (if (memq 'force ##compilation-options)
    `(let ,(map (lambda (x) `(,x (##force ,x))) vars) ,expr)
    expr))

(##define-macro (macro-if-checks checks nochecks)
  (if (memq 'check ##compilation-options)
    checks
    nochecks))

(##define-macro (macro-no-force vars expr)
  expr)

(##define-macro (macro-no-check var arg-num form expr)
  expr)

(##define-macro
  (define-prim-fold bool? form zero one two forcing pre-check . post-checks)
  (let* ((name-fn (car form))
         (name-param1 (cadr form))
         (name-param2 (caddr form))
         (name-others 'others)
         (name-result 'result)
         (name-folded-result 'folded-result)
         (param-name-map
          (list (cons name-param1 'param1) (cons name-param2 'param2)))
         (two-wrap
          (if (pair? (cadr two)) (car two) #f))
         (two-call
          (if (pair? (cadr two)) (cadr two) two)))

    (define (param-name sym)
      (cdr (assq sym param-name-map)))

    (define (rewrite expr)
      (let ((x (assq expr param-name-map)))
        (if x
          (cdr x)
          (if (pair? expr)
            (cons (car expr)
                  (map rewrite (cdr expr)))
            expr))))

    (define (parameter-list)
      (if (null? zero)
        ; 1 or more arguments
        (list (param-name name-param1)
              '#!optional
              (list (param-name name-param2) '(macro-absent-obj))
              '#!rest
              name-others)
        ; 0 or more arguments
        (list '#!optional
              (list (param-name name-param1) '(macro-absent-obj))
              (list (param-name name-param2) '(macro-absent-obj))
              '#!rest
              name-others)))

    (define (add-post-checks wrap expr arg-num cont)

      (define (add-wrap x)
        (if wrap
          (list wrap x)
          x))

      (define (inside lst)
        (if (null? lst)
          (if cont
            (cont (add-wrap name-result))
            (add-wrap name-result))
          (let ((check (car lst)))
            (list 'if
                  (list (car check) name-result)
                  (append (list (cadr check))
                          (if (eq? (car check) '##pair?)
                            (let ((x (list '##car name-result)))
                              (list (if arg-num
                                        (list '##fixnum.+ arg-num x)
                                        x)))
                            '())
                          (list #t
                                name-fn
                                (param-name name-param1)
                                (param-name name-param2)
                                name-others))
                  (inside (cdr lst))))))

      (if (or cont (not (null? post-checks)))
        (list 'let
              (list (list name-result expr))
              (inside post-checks))
        (add-wrap expr)))

    (define (add-forcing names expr)
      (list forcing
            names
            expr))

    (define (add-pre-check name arg-num expr)
      (list pre-check
            name
            arg-num
            (cons name-fn
                  (cons (param-name name-param1)
                        (cons (param-name name-param2)
                              name-others)))
            expr))

    (define (exactly-1-arg)
      (add-forcing
       (list (param-name name-param1))
       (add-pre-check
        (param-name name-param1)
        1
        (if (pair? one)
          (add-post-checks
           #f
           (rewrite one)
           #f
           #f)
          (rewrite one)))))

    (define (exactly-2-args)
      (add-post-checks
       two-wrap
       (rewrite two-call)
       #f
       #f))

    (define (at-least-2-args)
      (list 'let
            'loop
            (append (if bool?
                      (list (list name-folded-result #t))
                      '())
                    (list (list name-param1 (param-name name-param1))
                          (list name-param2 (param-name name-param2))
                          (list 'lst name-others)
                          (list 'arg-num 0)))
            (add-post-checks
             two-wrap
             two-call
             'arg-num
             (lambda (result)
               (list 'if
                     '(##null? lst)
                     (if bool?
                       (list 'and
                             result
                             name-folded-result)
                       result)
                     (list 'let
                           '((next (##car lst)))
                           (add-forcing
                            '(next)
                            (add-pre-check
                             'next
                             '(##fixnum.+ arg-num 3)
                             (if bool?
                               (list 'loop
                                     (list 'and
                                           result
                                           name-folded-result)
                                     name-param2
                                     'next
                                     '(##cdr lst)
                                     '(##fixnum.+ arg-num 1))
                               (list 'loop
                                     result
                                     'next
                                     '(##cdr lst)
                                     '(##fixnum.+ arg-num 1)))))))))))

    (define (body)
      (cons 'cond
            (cons (list (list '##not
                              (list '##eq?
                                    (param-name name-param2)
                                    '(macro-absent-obj)))
                        (add-forcing
                         (list (param-name name-param1)
                               (param-name name-param2))
                         (add-pre-check
                          (param-name name-param1)
                          1
                          (add-pre-check
                           (param-name name-param2)
                           2
                           (list 'if
                                 (list '##null? name-others)
                                 (exactly-2-args)
                                 (at-least-2-args))))))
                  (append (if (null? zero)
                            '()
                            (list (list (list '##eq?
                                              (param-name name-param1)
                                              '(macro-absent-obj))
                                        zero)))
                          (list (list 'else
                                      (exactly-1-arg)))))))

    (list 'define-prim
          name-fn
          (list 'lambda
                (parameter-list)
                (body)))))

(##define-macro
  (define-prim-nary form zero one two forcing pre-check . post-checks)
  `(define-prim-fold #f ,form ,zero ,one ,two ,forcing ,pre-check ,@post-checks))

(##define-macro
  (define-prim-nary-bool form zero one two forcing pre-check . post-checks)
  `(define-prim-fold #t ,form ,zero ,one ,two ,forcing ,pre-check ,@post-checks))

(##define-macro (define-check-type-macro type-id predicate . arguments)

  (define (sym . lst)
    (string->symbol (apply string-append (map symbol->string lst))))

  (let ()

    (define macro-check-type  (sym 'macro-check- type-id))
    (define ##fail-check-type (sym '##fail-check- type-id))

    `(##define-macro (,macro-check-type var arg-num form expr)
       (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
       `(macro-if-checks
          (if (,',predicate ,var ,@',arguments)
            ,expr
            ,(if (list? form)
               `(,',##fail-check-type ,arg-num ,(car form) ,@(cdr form))
               `(,',##fail-check-type ,arg-num #t ,(car form) ,@(flat (cdr form)))))
          ,expr))))

(##define-macro (define-check-index-range-macro type-id predicate . arguments)

  (define (sym . lst)
    (string->symbol (apply string-append (map symbol->string lst))))

  (let ()

    (define macro-check-type  (sym 'macro-check- type-id))

    `(##define-macro (,macro-check-type var arg-num ,@arguments form expr)
       (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
       `(macro-if-checks
          (if (##fixnum? ,var)
            (if (,',predicate ,var ,@,(cons 'list arguments))
              ,expr
              ,(if (list? form)
                 `(##raise-range-exception ,arg-num ,(car form) ,@(cdr form))
                 `(##raise-range-exception ,arg-num #t ,(car form) ,@(flat (cdr form)))))
            (if (##bignum? ,var)
              ,(if (list? form)
                 `(##raise-range-exception ,arg-num ,(car form) ,@(cdr form))
                 `(##raise-range-exception ,arg-num #t ,(car form) ,@(flat (cdr form))))
              ,(if (list? form)
                 `(##fail-check-exact-integer ,arg-num ,(car form) ,@(cdr form))
                 `(##fail-check-exact-integer ,arg-num #t ,(car form) ,@(flat (cdr form))))))
          ,expr))))

(##define-macro (define-fail-check-type type-id)

  (define (sym . lst)
    (string->symbol (apply string-append (map symbol->string lst))))

  (let ()

    (define ##fail-check-type (sym '##fail-check- type-id))

    `(define-prim (,##fail-check-type arg-num proc . args)
       (##raise-type-exception arg-num ',type-id proc args))))

(##define-macro (define-structure-type-macros name . fields)

  (define (sym . strings)
    (string->symbol (apply string-append strings)))

  (let ((name-str (symbol->string name))
        (preallocate? (null? fields)))

    (define (generate-accessor-macros fields i)
      (if (null? fields)
        '()
        (let* ((f (car fields))
               (field (if (pair? f) (car f) f))
               (field-str (symbol->string field))
               (macro-field-set! (sym "macro-" name-str "-" field-str "-set!"))
               (macro-field-ref (sym "macro-" name-str "-" field-str))
               (x (cons `(##define-macro (,macro-field-ref x)
                           `(macro-slot ,,i ,x))
                        (generate-accessor-macros (cdr fields)
                                                  (+ i 1)))))
          (if (pair? f)
            (cons `(##define-macro (,macro-field-set! x y)
                     `(macro-slot ,,i ,x ,y))
                  x)
            x))))

    (define (generate-accessor-functions fields)
      (if (null? fields)
        '()
        (let* ((f (car fields))
               (field (if (pair? f) (car f) f))
               (field-str (symbol->string field))
               (macro-field-set! (sym "macro-" name-str "-" field-str "-set!"))
               (field-set! (sym name-str "-" field-str "-set!"))
               (macro-field-ref (sym "macro-" name-str "-" field-str))
               (field-ref (sym name-str "-" field-str))
               (check (sym "macro-check-" name-str))
               (x (cons `(define-prim (,field-ref x)
                           (macro-force-vars (x)
                             (,check
                              x
                              1
                              (,field-ref x)
                              (,macro-field-ref x))))
                        (generate-accessor-functions (cdr fields)))))
          (if (pair? f)
            (cons `(define-prim (,field-set! x y)
                     (macro-force-vars (x)
                       (,check
                        x
                        1
                        (,field-set! x y)
                        (,macro-field-set! x y))))
                  x)
            x))))

    (define (generate-macros fields)
      (let ((field-names
             (map (lambda (x) (if (pair? x) (car x) x)) fields)))
        `(begin
           (##define-macro (,(sym "define-procedures-for-" name-str))
             `(begin
                (define-fail-check-type ,',name)
                ,,(if preallocate?
                   ``(define ,',(sym "##" name-str "-obj")
                       (macro-make-constant-structure-object #(,',name -1)))
                   ``(define ,',(sym "##" name-str "-tag")
                       '#(,',name -1)))
                (define-prim (,',(sym name-str "?") obj)
                  (macro-force-vars (obj)
                    (,',(sym "macro-" name-str "?") obj)))
                ,@',(generate-accessor-functions fields)))
           ,@(generate-accessor-macros fields 1)
           (##define-macro (,(sym "macro-" name-str "?") obj)
             ,(if preallocate?
                ``(##eq? ,obj ,',(sym "##" name-str "-obj"))
                ``(macro-structure-object?
                   ,obj
                   ,',(sym "##" name-str "-tag"))))
           (##define-macro (,(sym "macro-make-" name-str) ,@field-names)
             ,(if preallocate?
               ``,',(sym "##" name-str "-obj")
               ``(macro-make-structure-object
                  ,',(sym "##" name-str "-tag")
                  ,@(list ,@field-names))))
           (define-check-type-macro
             ,name
             ,(sym "macro-" name-str "?")))))

    (generate-macros fields)))

;------------------------------------------------------------------------------

; Macros for "_eval.scm" and "_repl.scm".

(##define-macro (make-global-var id) `(##make-global-var ,id))
(##define-macro (global-var-ref gv) `(##global-var-ref ,gv))
(##define-macro (global-var-set! gv val) `(##global-var-set! ,gv ,val))
(##define-macro (global-var->identifier gv) `(##global-var->identifier ,gv))

(##define-macro (quasi-list->vector x) `(##quasi-list->vector ,x))
(##define-macro (quasi-append x y) `(##quasi-append ,x ,y))
(##define-macro (quasi-cons x y) `(##quasi-cons ,x ,y))

(##define-macro (true? x) x)
(##define-macro (unbound? x) `(##unbound? ,x))

(##define-macro (self-var)     '##self-var)
(##define-macro (selector-var) '##selector-var)
(##define-macro (do-loop-var)  '##do-loop-var)

(##define-macro (rt-error-unbound-global-var code rte variable)
  `(##raise-unbound-global-exception ,code ,rte ,variable))

(##define-macro (rt-error-nonprocedure-oper operator arguments code rte)
  `(##raise-nonprocedure-operator-exception ,operator ,arguments ,code ,rte))

(##define-macro (rt-error-wrong-nb-args procedure arguments)
  `(##raise-wrong-number-of-arguments-exception ,procedure ,arguments))

(##define-macro (rt-error-unknown-keyword-arg procedure arguments)
  `(##raise-unknown-keyword-argument-exception ,procedure ,arguments))

(##define-macro (rt-error-keyword-expected procedure arguments)
  `(##raise-keyword-expected-exception ,procedure ,arguments))

(##define-macro (ct-error-syntax kind source . parameters)
  `(##raise-expression-parsing-exception ,kind ,source ,@parameters))

; Macro to create a node of executable code.

(##define-macro (mk-code code-prc cte src stepper subcodes . lst)
  `(let (($code (##vector #f ,code-prc ,cte ,src ,stepper ,@subcodes ,@lst)))
     ,@(let loop ((l subcodes) (i 5) (r '()))
         (if (pair? l)
           (loop (cdr l)
                 (+ i 1)
                 (cons `(##vector-set! (##vector-ref $code ,i) 0 $code) r))
           (reverse r)))
     $code))

(##define-macro (code-link c)         `(##vector-ref ,c 0))
(##define-macro (code-cprc c)         `(##vector-ref ,c 1))
(##define-macro (code-cte c)          `(##vector-ref ,c 2))
(##define-macro (code-locat c)        `(##vector-ref ,c 3))
(##define-macro (code-locat-set! c l) `(##vector-set! ,c 3 ,l))
(##define-macro (code-stepper c)      `(##vector-ref ,c 4))
(##define-macro (code-length c)       `(##fixnum.- (##vector-length ,c) 5))
(##define-macro (code-ref c n)        `(##vector-ref ,c (##fixnum.+ ,n 5)))
(##define-macro (code-set! c n x)     `(##vector-set! ,c (##fixnum.+ ,n 5) ,x))
(##define-macro (^ n)                 `(##vector-ref $code ,(+ n 5)))

(##define-macro (is-child-code? child parent)
  `(let ((child ,child)
         (parent ,parent))
     (and (##vector? child)
          (##fixnum.< 3 (##vector-length child))
          (##eq? (code-link child) parent))))

(##define-macro (code-run c)
  `(let (($$code ,c))
     ((##vector-ref $$code 1) $$code rte)))

; Macro to create the "code procedure" associated with a code node.

(##define-macro (mk-cprc . body)
  `(let ()
     (##declare (not inline) (not interrupts-enabled) (environment-map))
     (lambda ($code rte)
       (let (($$continue
              (lambda ($code rte)
                (##declare (inline))
                ,@body)))
         (##declare (interrupts-enabled))
         ($$continue $code rte)))))

(##define-macro (mk-gen params . def)
  `(let ()
     (##declare (not inline))
     (lambda (cte tail? src ,@params) ,@def)))

(##define-macro (gen proc src . args)
  `(,proc cte tail? ,src ,@args))

(##define-macro (call-step! vars . body)      `(step! #t 1 ,vars ,@body))
(##define-macro (future-step! vars . body)    `(step! #f 2 ,vars ,@body))
(##define-macro (delay-step! vars . body)     `(step! #f 2 ,vars ,@body))
(##define-macro (lambda-step! vars . body)    `(step! #f 3 ,vars ,@body))
(##define-macro (define-step! vars . body)    `(step! #f 4 ,vars ,@body))
(##define-macro (set!-step! vars . body)      `(step! #f 5 ,vars ,@body))
(##define-macro (reference-step! vars . body) `(step! #f 6 ,vars ,@body))
(##define-macro (constant-step! vars . body)  `(step! #f 7 ,vars ,@body))

(##define-macro (make-no-stepper)
  ''#(#(#f #f #f #f #f #f #f) #f #f #f #f #f #f #f))

(##define-macro (make-step-handlers)
  '(##vector ##step-handler
             ##step-handler
             ##step-handler
             ##step-handler
             ##step-handler
             ##step-handler
             ##step-handler))

(##define-macro (make-main-stepper)
  '(##vector (##vector-copy ##step-handlers)
             #f
             #f
             #f
             #f
             #f
             #f
             #f))

(##define-macro (step! leapable? handler-index vars . body)
  `(let* (($$execute-body
           (lambda ($code rte ,@vars) ,@body))
          ($$handler
           (##vector-ref (code-stepper $code) ,handler-index)))
     (if $$handler
       ($$handler ,leapable?
                  $code
                  rte
                  (lambda ($code rte ,@vars)
                    ($$execute-body $code rte ,@vars))
                  ,@vars)
       ($$execute-body $code rte ,@vars))))

;------------------------------------------------------------------------------

; Macros to manipulate the runtime environment

(##define-macro (mk-rte rte . lst)
  `(##vector ,rte ,@lst))

(##define-macro (mk-rte* rte ns)
  `(let (($rte (##make-vector (##fixnum.+ ,ns 1) (##void))))
     (##vector-set! $rte 0 ,rte)
     $rte))

(##define-macro (rte-up rte)         `(##vector-ref ,rte 0))
(##define-macro (rte-ref rte i)      `(##vector-ref ,rte ,i))
(##define-macro (rte-set! rte i val) `(##vector-set! ,rte ,i ,val))

;==============================================================================

(##include "../lib/_kernel#.scm")
(##include "../lib/_thread#.scm")
(##include "../lib/_system#.scm")
(##include "../lib/_error#.scm")
(##include "../lib/_num#.scm")
(##include "../lib/_io#.scm")
(##include "../lib/_eval#.scm")
(##include "../lib/_repl#.scm")
(##include "../lib/_std#.scm")
(##include "../lib/_nonstd#.scm")
