/* file: "main.c" */

/* Copyright (C) 1994-2002 by Marc Feeley, All Rights Reserved. */

/* This is the driver of the Gambit-C system */

#define ___INCLUDED_FROM_MAIN
#define ___VERSION 40011
#include "gambit.h"

#include "os.h"


___HIDDEN void usage_err
   ___P((char **argv,
         int debug_settings),
        (argv,
         debug_settings)
char **argv;
int debug_settings;)
{
  if (___DEBUG_SETTINGS_LEVEL(debug_settings) != 0)
    {
      char *msgs[4];
      msgs[0] = "Usage: ";
      msgs[1] = argv[0];
      msgs[2] =
        " [-:OPT,OPT...] ...\n"
        " where OPT can be one of:\n"
        "   mSIZE_IN_KILOBYTES      set minimum heap size\n"
        "   hSIZE_IN_KILOBYTES      set maximum heap size\n"
        "   lPERCENT_LIVE_AFTER_GC  set heap live ratio\n"
        "   s                       set standard Scheme mode\n"
        "   d[OPTION...]            set debugging options; options are:\n"
        "                            u/U   signal uncaught exceptions on/off\n"
        "                            s/S   start in single-step mode on/off\n"
        "                            0..9  verbosity level (default is 1)\n"
        "   t[OPTION...]            set terminal options; options are:\n"
        "                            e/E   line-editing mode on/off\n"
        "                            c     use native character encoding\n"
        "                            1     use LATIN1 character encoding\n"
        "                            2     use UCS2 character encoding\n"
        "                            4     use UCS4 character encoding\n"
        "                            8     use UTF8 character encoding\n";
      msgs[3] = 0;
      ___display_error (msgs);
    }
  ___exit_process (1);
}


___EXP_FUNC(int,___main)
  ___P((int argc,
        char **argv,
        ___mod_or_lnk (*linker)(___global_state_struct*)),
       (argc,
        argv,
        linker)
int argc;
char **argv;
___mod_or_lnk (*linker)();)
{
#define LARGEST_INT (int)((~___CAST(unsigned int,0))>>1)

  int options_source;
  long min_heap_len;
  long max_heap_len;
  int live_percent;
  int standard_level;
  int debug_settings;
  int terminal_settings;
  ___ERR_CODE err;
  ___setup_params_struct setup_params;

  /* handle arguments to runtime */

  min_heap_len = 0;
  max_heap_len = 0;
  live_percent = 0;
  standard_level = 0;
  debug_settings = ___DEBUG_SETTINGS_LEVEL_MASK
                   << ___DEBUG_SETTINGS_LEVEL_SHIFT;
  terminal_settings = ___TERMINAL_SETTINGS_ENCODING_MASK
                      << ___TERMINAL_SETTINGS_ENCODING_SHIFT;

  for (options_source=0; options_source<2; options_source++)
    {
      char *arg;

      if (options_source == 0)
        arg = ___getenv (___GAMBCOPT_ENV_VAR);
      else
        {
          arg = argv[1];
          if (argc > 1 && *arg++ == '-' && *arg++ == ':')
            {
              argv[1] = argv[0];
              argc--;
              argv++;
            }
          else
            arg = 0;
        }

      if (arg == 0)
        continue;

      do
        {
          char *s = arg++;
          switch (*s)
            {
            case 'm':
            case 'h':
            case 'l':
              {
                int argval = 0;
                while (*arg >= '0' && *arg <= '9')
                  {
                    int n = *arg - '0';
                    if (argval > (LARGEST_INT>>10)/10 ||
                        (argval == (LARGEST_INT>>10)/10 &&
                         n > ((LARGEST_INT>>10)-argval*10)))
                      usage_err (argv, debug_settings);
                    argval = argval*10 + n;
                    arg++;
                  }
                if (arg == s+1)
                  usage_err (argv, debug_settings);
                else
                  argval = atoi (s+1);
                switch (*s)
                  {
                  case 'm': min_heap_len = argval<<10;
                            break;
                  case 'h': max_heap_len = argval<<10;
                            break;
                  case 'l': live_percent = argval;
                            break;
                  }
                break;
              }

            case 's':
              standard_level = 4;
              break;

            case 'd':
              {
                if (___DEBUG_SETTINGS_LEVEL(debug_settings)
                    == ___DEBUG_SETTINGS_LEVEL_MASK)
                  debug_settings = ___DEBUG_SETTINGS_DEFAULT;

                while (*arg != '\0')
                  {
                    if (*arg >= '0' && *arg <= '9')
                      debug_settings =
                        (debug_settings & ___DEBUG_SETTINGS_LEVEL_MASK)
                        | ((*arg - '0') << ___DEBUG_SETTINGS_LEVEL_SHIFT);
                    else
                      switch (*arg)
                        {
                        case 'u': debug_settings
                                    |= ___DEBUG_SETTINGS_SIGNAL_UNCAUGHT;
                                  break;
                        case 'U': debug_settings
                                    &= ~___DEBUG_SETTINGS_SIGNAL_UNCAUGHT;
                                  break;
                        case 's': debug_settings
                                    |= ___DEBUG_SETTINGS_SINGLE_STEP;
                                  break;
                        case 'S': debug_settings
                                    &= ~___DEBUG_SETTINGS_SINGLE_STEP;
                                  break;
                        default:
                          usage_err (argv, debug_settings);
                        }
                    arg++;
                  }
                break;
              }

            case 't':
              {
#if 0
                if (___TERMINAL_SETTINGS_LEVEL(terminal_settings)
                    == ___TERMINAL_SETTINGS_LEVEL_MASK)
                  terminal_settings = ___TERMINAL_SETTINGS_DEFAULT;
#endif

                while (*arg != '\0')
                  {
                    switch (*arg++)
                      {
                      case 'e': terminal_settings = 0;
                                break;
                      case 'E': terminal_settings = 0;
                                break;
                      case 'c': terminal_settings = 0;
                                break;
                      case '1': terminal_settings = 0;
                                break;
                      case '2': terminal_settings = 0;
                                break;
                      case '4': terminal_settings = 0;
                                break;
                      case '8': terminal_settings = 0;
                                break;
                      default:
                        usage_err (argv, debug_settings);
                      }
                  }
                break;
              }

            default:
              usage_err (argv, debug_settings);
            }
        }

      while (*arg++ == ',') ;

      if (*--arg != '\0')
        usage_err (argv, debug_settings);
    }

#ifdef FORCE_MAX_HEAP
  if (max_heap_len == 0 || max_heap_len > FORCE_MAX_HEAP<<10)
    max_heap_len = FORCE_MAX_HEAP<<10;
#endif

  if (___DEBUG_SETTINGS_LEVEL(debug_settings)
      == ___DEBUG_SETTINGS_LEVEL_MASK)
    debug_settings = 0;

  if (___TERMINAL_SETTINGS_ENCODING(terminal_settings)
      == ___TERMINAL_SETTINGS_ENCODING_MASK)
    terminal_settings = 0;

  /* Setup program, run it and perform any cleanup necessary. */

  ___setup_params_reset (&setup_params);

  setup_params.version           = ___VERSION;
  setup_params.argv              = argv;
  setup_params.min_heap          = min_heap_len;
  setup_params.max_heap          = max_heap_len;
  setup_params.live_percent      = live_percent;
  setup_params.standard_level    = standard_level;
  setup_params.debug_settings    = debug_settings;
  setup_params.terminal_settings = terminal_settings;
  setup_params.linker            = linker;

  err = ___setup (&setup_params);

  if (err != ___NO_ERR)
    {
      char *msgs[2];
      char msg[___ERR_MAX_LENGTH+1];
      ___err_code_to_string (err, msg);
      msgs[0] = msg;
      msgs[1] = 0;
      ___fatal_error (msgs);
    }

  /* Cleanup the program and exit */

  ___cleanup_and_exit_process (0);

  return 0; /* satisfy some compilers */
}
