/* file: "setup.c" */

/* Copyright (C) 1994-2002 by Marc Feeley, All Rights Reserved. */

/* 
 * This module contains the routines that setup the Scheme program for
 * execution.
 */

#define ___INCLUDED_FROM_SETUP
#define ___VERSION 40011
#include "gambit.h"

#include "os.h"
#include "setup.h"
#include "mem.h"
#include "c_intf.h"


/*---------------------------------------------------------------------------*/

/* 
 * Global state structure.
 */

___EXP_DATA(___global_state_struct,___gstate);


/*
 * Global variables needed by this module.
 */

___WORD ___symbol_table;   /* Hash table of all interned symbols.  */
___WORD ___keyword_table;  /* Hash table of all interned keywords. */
___WORD ___command_line;   /* List of command line arguments.      */
___WORD ___program_descr;  /* Program descriptor.                  */

___NEED_GLO(___G__23__23_kernel_2d_handlers) /* from "_kernel.scm" */


/*
 * Parameters passed to ___setup.
 */

___setup_params_struct ___setup_params =
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };


/* 
 * Length of symbol table and keyword table.
 */

#define SYMKEY_TBL_LENGTH 359


/*---------------------------------------------------------------------------*/

/* 
 * Interrupt handling.
 */

/*
 * '___raise_interrupt (code)' is called when an interrupt has
 * occured.  At some later point in time, the Gambit kernel will cause
 * the Scheme procedure ##interrupt-handler to be called with a single
 * integer argument indicating which interrupt has been received.
 * Interrupt codes are defined in "gambit.h".  Currently, the
 * following codes are defined:
 *
 *   ___INTR_USER        user has interrupted the program (e.g. ctrl-C)
 *   ___INTR_HEARTBEAT   heartbeat time interval has elapsed
 *   ___INTR_GC          a garbage collection has finished
 */

___EXP_FUNC(void,___raise_interrupt)
   ___P((int code),
        (code)
int code;)
{
  ___processor_state ___ps = ___PSTATE;

  /* 
   * Note: ___raise_interrupt may be called before the processor state
   * is initialized.  As a consequence, the interrupt(s) received
   * before the initialization of the processor state will be ignored.
   */

#ifdef CALL_GC_FREQUENTLY
  if (code != ___INTR_USER)
    return;
#endif

  ___ps->intr_flag[code] = 1;
  if (___ps->intr_enabled)
    ___ps->stack_trip = ___ps->stack_start;
}


___EXP_FUNC(void,___begin_interrupt_service) ___PVOID
{
  ___processor_state ___ps = ___PSTATE;

  ___ps->stack_trip = ___ps->stack_limit;
}


___EXP_FUNC(___BOOL,___check_interrupt)
   ___P((int code),
        (code)
int code;)
{
  ___processor_state ___ps = ___PSTATE;

  if (___ps->intr_flag[code])
    {
      ___ps->intr_flag[code] = 0;
      return 1;
    }

  return 0;
}


___EXP_FUNC(void,___end_interrupt_service)
   ___P((int code),
        (code)
int code;)
{
  ___processor_state ___ps = ___PSTATE;

  if (___ps->intr_enabled)
    {
#ifdef CALL_HANDLER_AT_EVERY_POLL
      ___ps->stack_trip = ___ps->stack_start;
#else
      while (code < ___NB_INTRS)
        {
          if (___ps->intr_flag[code]) /* don't ignore other interrupts */
            {
              ___ps->stack_trip = ___ps->stack_start;
              break;
            }
          code++;
        }
#endif
    }
}


___EXP_FUNC(void,___disable_interrupts) ___PVOID
{
  ___processor_state ___ps = ___PSTATE;

  ___ps->intr_enabled = 0;

  ___begin_interrupt_service ();
  ___end_interrupt_service (0);
}


___EXP_FUNC(void,___enable_interrupts) ___PVOID
{
  ___processor_state ___ps = ___PSTATE;

  ___ps->intr_enabled = 1;

  ___begin_interrupt_service ();
  ___end_interrupt_service (0);
}


/*---------------------------------------------------------------------------*/

/* 
 * Routines to setup symbol table, keyword table and global variable
 * table.
 */

/* 
 * The hashing functions 'hash_utf8string (str)' and
 * 'hash_schemestring (str)' must compute the same value as the
 * function 'targ-hash' in the file "gsc/_t-c-3.scm".
 * A fixnum error code is returned when there is an error.
 */

#define HASH_STEP(h,c) ((((h)>>8) + (c)) * 331804471) & ___MAX_FIX32

___HIDDEN ___WORD hash_utf8string
   ___P((___UTF8STRING str),
        (str)
___UTF8STRING str;)
{
  ___UM32 h = 0;
  ___UTF8STRING p = str;
  ___UCS4 c;

  while (1)
    {
      ___UTF8STRING start = p;
      c = ___utf8_get (&p);
      if (p == start || c > ___MAX_CHR)
        return ___FIX(___CTOS_UTF8STRING_ERR);
      if (c == 0)
        break;
      h = HASH_STEP(h,c);
    }

  return ___FIX(h);
}


___HIDDEN ___WORD hash_schemestring
   ___P((___WORD str),
        (str)
___WORD str;)
{
  unsigned long i, n = ___INT(___STRINGLENGTH(str));
  ___UM32 h = 0;

  for (i=0; i<n; i++)
    h = HASH_STEP(h,___INT(___STRINGREF(str,___FIX(i))));

  return ___FIX(h);
}


___HIDDEN ___WORD symkey_table
   ___P((unsigned int subtype),
        (subtype)
unsigned int subtype;)
{
  switch (subtype)
    {
    case ___sKEYWORD:
      return ___keyword_table;
    default: /* assume ___sSYMBOL */
      return ___symbol_table;
    }
}


___HIDDEN ___WORD find_symkey_from_utf8string
   ___P((char *str,
         unsigned int subtype),
        (str,
         subtype)
char *str;
unsigned int subtype;)
{
  ___WORD h;
  ___WORD probe;

  h = hash_utf8string (str);

  if (h < ___FIX(0))
    return h;

  probe = ___FIELD(symkey_table (subtype),___INT(h)%SYMKEY_TBL_LENGTH);

  while (probe != ___NUL)
    {
      ___WORD name = ___FIELD(probe,___SYMKEY_NAME);
      unsigned long i;
      unsigned long n = ___INT(___STRINGLENGTH(name));
      ___UTF8STRING p = str;
      for (i=0; i<n; i++)
        if (___utf8_get (&p) !=
            ___CAST(___UCS4,___INT(___STRINGREF(name,___FIX(i)))))
          goto next;
      if (___utf8_get (&p) == 0)
        return probe;
    next:
      probe = ___FIELD(probe,___SYMKEY_NEXT);
    }

  return ___FAL;
}


___WORD ___find_symkey_from_schemestring
   ___P((___WORD str,
         unsigned int subtype),
        (str,
         subtype)
___WORD str;
unsigned int subtype;)
{
  ___WORD tbl = symkey_table (subtype);
  ___WORD h;
  ___WORD probe;

  h = hash_schemestring (str);

  probe = ___FIELD(tbl,___INT(h)%SYMKEY_TBL_LENGTH);

  while (probe != ___NUL)
    {
      ___WORD name = ___FIELD(probe,___SYMKEY_NAME);
      long i = 0;
      long n = ___INT(___STRINGLENGTH(name));
      if (___INT(___STRINGLENGTH(str)) == n)
        {
          for (i=0; i<n; i++)
            if (___STRINGREF(str,___FIX(i)) != ___STRINGREF(name,___FIX(i)))
              goto next;
          return probe;
        }
    next:
      probe = ___FIELD(probe,___SYMKEY_NEXT);
    }

  return ___FAL;
}


___WORD ___new_symkey
   ___P((___WORD name, /* name must be a permanent object */
         unsigned int subtype),
        (name,
         subtype)
___WORD name;
unsigned int subtype;)
{
  ___WORD obj;
  ___WORD tbl;
  ___WORD h;
  int i;

  switch (subtype)
    {
    case ___sKEYWORD:
      obj = ___alloc_scmobj (___sKEYWORD, ___KEYWORD_SIZE<<___LWS, ___PERM);
      break;
    default: /* assume ___sSYMBOL */
      obj = ___alloc_scmobj (___sSYMBOL, ___SYMBOL_SIZE<<___LWS, ___PERM);
      break;
    }

  if (___FIXNUMP(obj))
    return obj;

  tbl = symkey_table (subtype);
  h = hash_schemestring (name);
  i = ___INT(h) % SYMKEY_TBL_LENGTH;

  /* object layout is same for ___sSYMBOL and ___sKEYWORD */

  ___FIELD(obj,___SYMKEY_NAME) = name;
  ___FIELD(obj,___SYMKEY_HASH) = h;
  ___FIELD(obj,___SYMKEY_NEXT) = ___FIELD(tbl,i);

  if (subtype == ___sSYMBOL)
    ___FIELD(obj,___SYMBOL_GLOBAL) = 0;

  ___FIELD(tbl,i) = obj;

  return obj;
}


___HIDDEN ___WORD make_symkey
   ___P((___UTF8STRING str,
         unsigned int subtype),
        (str,
         subtype)
___UTF8STRING str;
unsigned int subtype;)
{
  ___WORD obj = find_symkey_from_utf8string (str, subtype);

  if (___FIXNUMP(obj))
    return obj;

  if (obj == ___FAL)
    {
      ___ERR_CODE err;
      ___WORD name;

      err = ___NONNULLUTF8STRING_to_SCMOBJ (str, &name, 0);

      if (err != ___NO_ERR)
        return ___FIX(err);

      obj = ___new_symkey (name, subtype);
    }

  return obj;
}


___HIDDEN ___WORD make_global
   ___P((___UTF8STRING str,
         int supply,
         ___glo_struct **glo),
        (str,
         supply,
         glo)
___UTF8STRING str;
int supply;
___glo_struct **glo;)
{
  ___WORD sym;
  ___WORD g;
  ___glo_struct *p;

  sym = make_symkey (str, ___sSYMBOL);

  if (___FIXNUMP(sym))
    return sym;

  g = ___FIELD(sym,___SYMBOL_GLOBAL);

  if (g == ___FIX(0))
    {
      ___processor_state ___ps = ___PSTATE;
      ___ERR_CODE err = ___alloc_global_var (&p);
      if (err != ___NO_ERR)
        return ___FIX(err);
      p->val = supply?___UNB2:___UNB1;
      p->prm = ___FAL;
      p->next = 0;
      if (___ps->glo_list_head == 0)
        ___ps->glo_list_head = ___CAST(___WORD,p);
      else
        ___CAST(___glo_struct*,___ps->glo_list_tail)->next =
          ___CAST(___WORD,p);
      ___ps->glo_list_tail = ___CAST(___WORD,p);
      ___FIELD(sym,___SYMBOL_GLOBAL) = ___CAST(___WORD,p);
    }
  else
    {
      p = ___CAST(___glo_struct*,g);
      if (supply && p->val == ___UNB1)
        p->val = ___UNB2;
    }

  *glo = p;

  return ___FIX(___NO_ERR);
}


/*---------------------------------------------------------------------------*/

/*
 * Alignment of objects.
 */

___HIDDEN ___WORD *align
   ___P((___WORD *from,
         long words,
         int need_64bit_alignment),
        (from,
         words,
         need_64bit_alignment)
___WORD *from;
long words;
int need_64bit_alignment;)
{
  ___WORD *to;

#if ___WS == 4
  if (need_64bit_alignment)
    to = ___ALIGNUP((from+1), 8) - 1;
  else
#endif
    to = ___ALIGNUP(from, ___WS);

  if (from != to)
    {
      /* move object up */
      int i;
      for (i=words-1; i>=0; i--)
        to[i] = from[i];
    }

  return to;
}


___HIDDEN ___WORD align_subtyped
   ___P((___WORD *ptr),
        (ptr)
___WORD *ptr;)
{
  ___WORD head = ptr[0];
  int subtype = ___HD_SUBTYPE(head);
  int words = ___HD_WORDS(head);
  return ___TAG(align (ptr, words+1, subtype>=___sS64VECTOR), ___tSUBTYPED);
}


/*---------------------------------------------------------------------------*/

/*
 * Routines to setup a module for execution.
 */

___HIDDEN ___mod_or_lnk linker_to_mod_or_lnk
   ___P((___mod_or_lnk (*linker) (___global_state_struct*)),
        (linker)
___mod_or_lnk (*linker) ();)
{
  ___mod_or_lnk mol = linker (___GSTATE);
  if (mol->module.kind == ___LINKFILE_KIND)
    {
      void **p = mol->linkfile.linker_tbl;
      while (*p != 0)
        {
          *p = linker_to_mod_or_lnk
                 (*___CAST(___mod_or_lnk (**) ___P((___global_state_struct*),()),p));
          p++;
        }
    }
  return mol;
}


___HIDDEN ___WORD for_each_module
   ___P((___mod_or_lnk mol,
         ___WORD (*proc) (___module_struct*)),
        (mol,
         proc)
___mod_or_lnk mol;
___WORD (*proc) ();)
{
  if (mol->module.kind == ___LINKFILE_KIND)
    {
      void **p = mol->linkfile.linker_tbl;
      while (*p != 0)
        {
          ___WORD e = for_each_module (___CAST(___mod_or_lnk,*p++), proc);
          if (e != ___FIX(___NO_ERR))
            return e;
        }
      return 0; /* no error */
    }
  else
    return proc (___CAST(___module_struct*,mol));
}


___HIDDEN void fixref
   ___P((___WORD *p,
         ___WORD *sym_tbl,
         ___WORD *key_tbl,
         ___WORD *cns_tbl,
         ___WORD *sub_tbl),
        (p,
         sym_tbl,
         key_tbl,
         cns_tbl,
         sub_tbl)
___WORD *p;
___WORD *sym_tbl;
___WORD *key_tbl;
___WORD *cns_tbl;
___WORD *sub_tbl;)
{
  ___WORD v = *p;
  switch (___TYP(v))
    {
    case ___tPAIR:
      if (___INT(v)<0)
        *p = sym_tbl[-1-___INT(v)];
      else
        *p = ___TAG(___ALIGNUP(&cns_tbl[(___PAIR_SIZE+1)*___INT(v)],___WS),
                    ___tPAIR);
      break;

    case ___tSUBTYPED:
      if (___INT(v)<0)
        *p = key_tbl[-1-___INT(v)];
      else
        *p = sub_tbl[___INT(v)];
      break;
    }
}


___HIDDEN int module_count;
___HIDDEN ___WORD module_descr;


___HIDDEN ___WORD setup_module_phase1
   ___P((___module_struct *module),
        (module)
___module_struct *module;)
{
  int i, j;
  ___WORD *cns;

  char *name               = module->name;
  int flags                = module->flags;
  ___FAKEWORD *glo_tbl     = module->glo_tbl;
  int sup_count            = module->sup_count;
  ___UTF8STRING *glo_names = module->glo_names;
  ___WORD *sym_tbl         = ___CAST(___WORD*,module->sym_tbl);
  int sym_count            = module->sym_count;
  ___UTF8STRING *sym_names = module->sym_names;
  ___WORD *key_tbl         = ___CAST(___WORD*,module->key_tbl);
  int key_count            = module->key_count;
  ___UTF8STRING *key_names = module->key_names;
  ___WORD *lp              = module->lp;
  ___WORD *lbl_tbl         = ___CAST(___WORD*,module->lbl_tbl);
  int lbl_count            = module->lbl_count;
  ___WORD *ofd_tbl         = module->ofd_tbl;
  int ofd_length           = module->ofd_length;
  ___WORD *cns_tbl         = module->cns_tbl;
  int cns_count            = module->cns_count;
  ___WORD *sub_tbl         = ___CAST(___WORD*,module->sub_tbl);
  int sub_count            = module->sub_count;

  /* 
   * Check that the version of the compiler used to compile the module
   * is compatible with the compiler used to compile the runtime
   * system.
   */

  if (module->version / 10000 < ___VERSION / 10000)
    return ___FIX(___MODULE_VERSION_TOO_OLD_ERR);

  if (module->version / 10000 > ___VERSION / 10000)
    return ___FIX(___MODULE_VERSION_TOO_NEW_ERR);

  /* Align module's pair table */

  if (cns_tbl != 0)
    cns = align (cns_tbl, (___PAIR_SIZE+1)*cns_count, 0);

  /* Setup module's global variable table */

  if (glo_names != 0)
    {
      /* 
       * Create global variables in reverse order so that global
       * variables bound to c-lambdas are created last.
       */
      i = 0;
      while (glo_names[i] != 0)
        i++;
      while (i-- > 0)
        {
          ___glo_struct *glo;
          ___WORD e = make_global (glo_names[i], i<sup_count, &glo);
          if (e != ___FIX(___NO_ERR))
            return e;
          glo_tbl[i] = ___CAST(___FAKEWORD,glo);
        }
    }

  /* Setup module's symbol table */

  if (sym_names != 0)
    {
      i = 0;
      while (sym_names[i] != 0)
        {
          ___WORD sym = make_symkey (sym_names[i], ___sSYMBOL);
          if (___FIXNUMP(sym))
            return sym;
          sym_tbl[i] = sym;
          i++;
        }
    }
  else
    for (i=sym_count-1; i>=0; i--)
      sym_tbl[i] = ___TAG(___ALIGNUP(sym_tbl[i], ___WS), ___tSUBTYPED);

  /* Setup module's keyword table */

  if (key_names != 0)
    {
      i = 0;
      while (key_names[i] != 0)
        {
          ___WORD key = make_symkey (key_names[i], ___sKEYWORD);
          if (___FIXNUMP(key))
            return key;
          key_tbl[i] = key;
          i++;
        }
    }
  else
    for (i=key_count-1; i>=0; i--)
      key_tbl[i] = ___TAG(___ALIGNUP(key_tbl[i], ___WS), ___tSUBTYPED);

  /* Setup module's subtyped object table */

  for (i=sub_count-1; i>=0; i--)
    sub_tbl[i] = align_subtyped (___CAST(___WORD*,sub_tbl[i]));

  /* Fix references in module's pair table */

  for (i=cns_count-1; i>=0; i--)
  {
    fixref (cns+i*(___PAIR_SIZE+1)+1, sym_tbl, key_tbl, cns_tbl, sub_tbl);
    fixref (cns+i*(___PAIR_SIZE+1)+2, sym_tbl, key_tbl, cns_tbl, sub_tbl);
  }

  /* Fix references in module's subtyped object table */

  for (j=sub_count-1; j>=0; j--)
    {
      ___WORD *p = ___UNTAG_AS(sub_tbl[j],___tSUBTYPED);
      ___WORD head = p[0];
      int subtype = ___HD_SUBTYPE(head);
      int words = ___HD_WORDS(head);
      switch (subtype)
        {
        case ___sSYMBOL:
        case ___sKEYWORD:
        case ___sVECTOR:
        case ___sRATNUM:
        case ___sCPXNUM:
          for (i=1; i<=words; i++)
            fixref (p+i, sym_tbl, key_tbl, cns_tbl, sub_tbl);
        }
    }

  /* Align module's out-of-line frame descriptor table */

  if (ofd_tbl != 0)
    ofd_tbl = ___CAST(___WORD*,align (ofd_tbl, ofd_length, 0));

  /* Align module's label table */

  if (lbl_count > 0)
    {
      ___host current_host = 0;
      void **hlbl_ptr = 0;
      ___label_struct *new_lt;
      ___WORD *ofd_alloc;

      module_count++;

      new_lt = ___CAST(___label_struct*,align (lbl_tbl, lbl_count*___LS, 0));
      ofd_alloc = ofd_tbl;

      for (i=0; i<lbl_count; i++)
        {
          ___label_struct *lbl = &new_lt[i];
          ___WORD head = lbl->header;
          if (head == ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM))
            {
              fixref (&lbl->entry_or_descr, sym_tbl, key_tbl, cns_tbl, sub_tbl);
              if (hlbl_ptr != 0)
                hlbl_ptr++; /* skip INTRO label */
            }
          else
            {
              if (flags & 1) /* module uses gcc's computed goto? */
                {
                  if (___CAST_FAKEHOST_TO_HOST(lbl->host) != current_host)
                    {
                      current_host = ___CAST_FAKEHOST_TO_HOST(lbl->host);
                      hlbl_ptr = ___CAST(void**,current_host (0));
                      hlbl_ptr++; /* skip INTRO label */
                    }
                  lbl->host_label = *hlbl_ptr++;
                }
              if (head == ___MAKE_HD((3<<___LWS),___sRETURN,___PERM))
                {
                  ___WORD descr;
                  descr = lbl->entry_or_descr;
                  if (___IFD_GCMAP(descr) == 0) /* out-of-line descriptor? */
                    {
                      int fs;
                      lbl->entry_or_descr = ___CAST(___WORD,ofd_alloc);
                      fs = ___OFD_FS(*ofd_alloc);
                      if (___IFD_KIND(descr) == ___RETI)
                        fs = ___RETI_CFS_TO_FS(fs);
                      ofd_alloc += 1 + ___CEILING_DIV(fs,___WS*8);
                    }
                }
              else
                lbl->entry_or_descr = ___TAG(&lbl->header,___tSUBTYPED);
            }
        }
      *lp = ___TAG(new_lt,___tSUBTYPED);
    }

  return ___FIX(___NO_ERR);
}


___HIDDEN char module_prefix[] = ___MODULE_PREFIX;

#define module_prefix_length (sizeof(module_prefix)-1)


___HIDDEN ___WORD setup_module_phase2
   ___P((___module_struct *module),
        (module)
___module_struct *module;)
{
  ___UTF8STRING *glo_names = module->glo_names;

  if (glo_names != 0)
    {
      ___UTF8STRING name = module->name;
      ___FAKEWORD *glo_tbl = module->glo_tbl;
      int glo_count = module->glo_count;
      int sup_count = module->sup_count;
      int i;
      for (i=sup_count; i<glo_count; i++)
        {
          /* 
           * If the global variable is undefined, add it to the list
           * of undefined variables in the module descriptor.
           */

          ___glo_struct *glo = ___CAST(___glo_struct*,glo_tbl[i]);

          if (glo->val == ___UNB1)
            {
              ___ERR_CODE err;
              ___WORD glo_name;
              ___WORD module_name;
              ___WORD pair1;
              ___WORD pair2;

              err = ___NONNULLUTF8STRING_to_SCMOBJ
                      (glo_names[i], &glo_name, 0);

              if (err != ___NO_ERR)
                return ___FIX(err);

              err = ___NONNULLUTF8STRING_to_SCMOBJ
                      (name+module_prefix_length, &module_name, 0);

              if (err != ___NO_ERR)
                {
                  ___release_scmobj (glo_name);
                  return ___FIX(err);
                }

              pair1 = ___make_pair (glo_name, module_name, ___STILL);

              ___release_scmobj (glo_name);
              ___release_scmobj (module_name);

              if (___FIXNUMP(pair1))
                return pair1;

              pair2 = ___make_pair (pair1, ___FIELD(module_descr,1), ___STILL);

              ___release_scmobj (pair1);

              if (___FIXNUMP(pair2))
                return pair2;

              ___FIELD(module_descr,1) = pair2;

              ___release_scmobj (pair2);
            }
        }
    }

  return ___FIX(___NO_ERR);
}


___HIDDEN ___WORD setup_module_phase3
   ___P((___module_struct *module),
        (module)
___module_struct *module;)
{
  if (module->lbl_count > 0)
    {
      ___FIELD(___FIELD(module_descr,0),module_count) =
        *module->lp+___LS*___WS;
      module_count++;
    }
  return module->init_proc ();
}


___HIDDEN ___WORD setup_modules
   ___P((___mod_or_lnk mol),
        (mol)
___mod_or_lnk mol;)
{
  ___WORD result;

  result = ___make_vector (2, ___NUL, ___STILL);

  if (!___FIXNUMP(result))
    {
      module_descr = result;

      module_count = 0;

      if ((result = for_each_module (mol, setup_module_phase1))
          == ___FIX(___NO_ERR))
        {
          if ((result = for_each_module (mol, setup_module_phase2))
              == ___FIX(___NO_ERR))
            {
              result = ___make_vector (module_count, ___FAL, ___STILL);

              if (!___FIXNUMP(result))
                {
                  ___FIELD(module_descr,0) = result;
                  ___release_scmobj (result);

                  module_count = 0;

                  if ((result = for_each_module (mol, setup_module_phase3))
                      == ___FIX(___NO_ERR))
                    result = module_descr;
                }
            }
        }

      if (___FIXNUMP(result))
        ___release_scmobj (module_descr);
    }

  return result;
}


___HIDDEN char c_id_prefix[] =
#ifdef ___IMPORTED_ID_PREFIX
___IMPORTED_ID_PREFIX
#endif
___C_ID_PREFIX;

#define c_id_prefix_length (sizeof(c_id_prefix)-1)

___HIDDEN char c_id_suffix[] =
#ifdef ___IMPORTED_ID_SUFFIX
___IMPORTED_ID_SUFFIX
#endif
"";

#define c_id_suffix_length (sizeof(c_id_suffix)-1)

___HIDDEN char hex_digits[] = "0123456789abcdef";

#define c_id_subsequent(c) \
(((c)>='A'&&(c)<='Z') || \
 ((c)>='a'&&(c)<='z') || \
 ((c)>='0'&&(c)<='9') || \
 ((c)=='_'))


___HIDDEN int scheme_id_to_c_id
   ___P((char *scm_id,
         char *c_id,
         long max_length),
        (scm_id,
         c_id,
         max_length)
char *scm_id;
char *c_id;
long max_length;)
{
  long i = 0;
  long j = 0;
  unsigned char c, c2;
  long k, n;

  while ((c = scm_id[i++]) != '\0')
    {
      if (c == '_')
        {
          if (j+2 > max_length)
            return 0;
          c_id[j++] = '_';
          c_id[j++] = '_';
        }
      else if (c_id_subsequent(c))
        {
          if (j+1 > max_length)
            return 0;
          c_id[j++] = c;
        }
      else
        {
          c2 = c;
          n = 1;
          while (c2 > 15)
            {
              c2 = c2 >> 4;
              n++;
            }
          if (j+n+2 > max_length)
            return 0;
          c_id[j++] = '_';
          for (k=n-1; k>=0; k--)
            {
              c_id[j+k] = hex_digits[c & 15];
              c = c >> 4;
            }
          j += n;
          c_id[j++] = '_';
        }
    }

  if (j+1 > max_length)
    return 0;

  c_id[j++] = '\0';

  return 1;
}


___WORD ___load_object_file
   ___P((___WORD path),
        (path)
___WORD path;)
{
  int i;
  char temp1[___PATH_MAX_LENGTH+1];
  char temp2[___PATH_MAX_LENGTH+1];
  char *p1;
  char *p2;
  void *f;
  ___mod_or_lnk mol;
  ___ERR_CODE err;
  char *errmsg;
  ___WORD str;
  ___WORD result;

  err = ___SCMOBJ_to_NONNULLCHARSTRING (path, &p1, 1);

  if (err != ___NO_ERR)
    return ___FIX(err);

  for (i=module_prefix_length-1; i>=0; i--)
    temp2[i] = module_prefix[i];

  if (___path_strip_extension (p1,
                               temp1,
                               ___PATH_MAX_LENGTH) == 0 ||
      ___path_strip_directory (temp1,
                               temp2+module_prefix_length,
                               ___PATH_MAX_LENGTH -
                                module_prefix_length) == 0 ||
      scheme_id_to_c_id (temp2,
                         temp1+c_id_prefix_length,
                         ___PATH_MAX_LENGTH -
                          (c_id_prefix_length+c_id_suffix_length)) == 0)
    {
      ___release_string (p1);
      return ___FIX(___IMPL_LIMIT_ERR);
    }

  p2 = temp1;
  for (i=0; i<c_id_prefix_length; i++)
    *p2++ = c_id_prefix[i];
  while (*p2 != '\0')
    p2++;
  for (i=0; i<c_id_suffix_length; i++)
    *p2++ = c_id_suffix[i];
  *p2++ = '\0';

  f = ___dynamic_load (p1, temp1, &errmsg);

  ___release_string (p1);

  if (f == 0)
    {
      err = ___NONNULLCHARSTRING_to_SCMOBJ (errmsg, &str, 0);
      if (err != ___NO_ERR)
        return ___FIX(err);
      ___release_scmobj (str);
      return str;
    }

  mol = linker_to_mod_or_lnk
          (___CAST(___mod_or_lnk (*) ___P((___global_state_struct*),()),f));

  if (mol->linkfile.version < 0) /* was it already setup? */
    return ___FIX(___MODULE_ALREADY_LOADED_ERR);

  result = setup_modules (mol);

  mol->linkfile.version = -1; /* mark link file as 'setup' */

  if (!___FIXNUMP(result))
    ___release_scmobj (result);

  return result;
}


/*---------------------------------------------------------------------------*/

/*
 * Numerical library routines.
 */

#ifdef ___CPU_BIGEND
#define F64_HI8 0
#define F64_HI16 0
#define F64_HI32 0
#define F64_LO32 1
#else
#define F64_HI8 7
#define F64_HI16 3
#define F64_HI32 1
#define F64_LO32 0
#endif


___EXP_FUNC(double,___copysign)
   ___P((double x,
         double y),
        (x,
         y)
double x;
double y;)
{
  ___STORE_U8(&x,
              F64_HI8,
              (___FETCH_U8(&x,F64_HI8)&0x7f)|(___FETCH_U8(&y,F64_HI8)&0x80));

  return x;
}


___EXP_FUNC(___BOOL,___isfinite)
   ___P((double x),
        (x)
double x;)
{
#ifdef ___CRAY_FP_FORMAT

  return 1;

#else

  return ((___FETCH_U16(&x,F64_HI16) ^ 0x7ff0) & 0x7fff) >= 0x10;

#endif
}


___EXP_FUNC(___BOOL,___isnan)
   ___P((double x),
        (x)
double x;)
{
#ifdef ___CRAY_FP_FORMAT

  return 0;

#else

  ___UM32 tmp = (___FETCH_U32(&x,F64_HI32) ^ 0x7ff00000) & 0x7fffffff;

  return tmp < 0x100000 && (tmp | ___FETCH_U32(&x,F64_LO32)) != 0;

#endif
}


___EXP_FUNC(double,___round)
   ___P((double x),
        (x)
double x;)
{
  double f, i, t;
  if (x < 0.0)
    {
      f = modf (-x, &i);
      if (f > 0.5 || (f == 0.5 && modf (i*0.5, &t) != 0.0))
        return -(i+1.0);
      else
        return -i;
    }
  else if (x == 0.0) /* so that round (-0.0) = -0.0 */
    return x;
  else
    {
      f = modf (x, &i);
      if (f > 0.5 || (f == 0.5 && modf (i*0.5, &t) != 0.0))
        return i+1.0;
      else
        return i;
    }
}


#ifndef ___GOOD_ATAN2

___EXP_FUNC(double,___atan2)
   ___P((double y,
         double x),
        (y,
         x)
double y;
double x;)
{
  if (___isnan (x))
    return x;
  else if (___isnan (y))
    return y;
  else if (y == 0.0)
    {
      if (___copysign (1.0, y) > 0.0)
        if (___copysign (1.0, x) > 0.0)
          return 0.0;
        else
          return 3.141592653589793; /* from "header.scm" */
      else
        if (___copysign (1.0, x) > 0.0)
          return -0.0;
        else
          return -3.141592653589793; /* from "header.scm" */
    }
  else if (___isfinite (x) || ___isfinite (y))
    return atan2 (y, x);
  else
    return ___copysign (x/y, x*y); /* returns NAN with appropriate sign */
}

#endif


/*---------------------------------------------------------------------------*/

/* 
 * Initialization of symbol and keyword tables, and global variables.
 */

___HIDDEN void init_symkey_glo1
   ___P((___mod_or_lnk mol),
        (mol)
___mod_or_lnk mol;)
{
  if (mol->module.kind == ___LINKFILE_KIND)
    {
      void **p1 = mol->linkfile.linker_tbl;
      ___FAKEWORD *p2 = mol->linkfile.sym_list;

      while (*p1 != 0)
        init_symkey_glo1 (___CAST(___mod_or_lnk,*p1++));

      while (p2 != 0)
        {
          ___WORD *sym_ptr;
          ___glo_struct *glo;

          sym_ptr = ___CAST(___WORD*,p2);

          p2 = ___CAST(___FAKEWORD*,sym_ptr[0]);
          glo = ___CAST(___glo_struct*,sym_ptr[1+___SYMBOL_GLOBAL]);

          sym_ptr[1+___SYMKEY_HASH] = glo->prm; /* move symbol's hash value */
        }
    }
}


___HIDDEN void init_symkey_glo2
   ___P((___mod_or_lnk mol),
        (mol)
___mod_or_lnk mol;)
{
  if (mol->module.kind == ___LINKFILE_KIND)
    {
      void **p1 = mol->linkfile.linker_tbl;
      ___FAKEWORD *p2 = mol->linkfile.sym_list;
      ___FAKEWORD *p3 = mol->linkfile.key_list;
      ___processor_state ___ps = ___PSTATE;

      while (*p1 != 0)
        init_symkey_glo2 (___CAST(___mod_or_lnk,*p1++));

      while (p2 != 0)
        {
          int i;
          ___WORD sym, str;
          ___WORD *sym_ptr;
          ___glo_struct *glo;

          sym_ptr = ___CAST(___WORD*,p2);

          p2 = ___CAST(___FAKEWORD*,sym_ptr[0]);
          str = align_subtyped (___CAST(___WORD*,sym_ptr[1+___SYMKEY_NAME]));
          glo = ___CAST(___glo_struct*,sym_ptr[1+___SYMBOL_GLOBAL]);

          glo->next = 0;
          if (___ps->glo_list_head == 0)
            ___ps->glo_list_head = ___CAST(___WORD,glo);
          else
            ___CAST(___glo_struct*,___ps->glo_list_tail)->next =
              ___CAST(___WORD,glo);
          ___ps->glo_list_tail = ___CAST(___WORD,glo);

          *sym_ptr = ___MAKE_HD((___SYMBOL_SIZE<<___LWS),___sSYMBOL,___PERM);

          sym = align_subtyped (sym_ptr);

          ___FIELD(sym,___SYMKEY_NAME) = str;
          ___FIELD(sym,___SYMBOL_GLOBAL) = ___CAST(___WORD,glo);

          i = ___INT(___FIELD(sym,___SYMKEY_HASH)) % SYMKEY_TBL_LENGTH;

          ___FIELD(sym,___SYMKEY_NEXT) = ___FIELD(___symbol_table,i);
          ___FIELD(___symbol_table,i) = sym;
        }

      while (p3 != 0)
        {
          ___WORD h;
          int i;
          ___WORD key, str;
          ___WORD *key_ptr;

          key_ptr = ___CAST(___WORD*,p3);

          p3 = ___CAST(___FAKEWORD*,key_ptr[0]);
          str = align_subtyped (___CAST(___WORD*,key_ptr[1+___SYMKEY_NAME]));

          *key_ptr = ___MAKE_HD((___KEYWORD_SIZE<<___LWS),___sKEYWORD,___PERM);

          key = align_subtyped (key_ptr);

          h = hash_schemestring (str);

          ___FIELD(key,___SYMKEY_NAME) = str;
          ___FIELD(key,___SYMKEY_HASH) = h;

          i = ___INT(h) % SYMKEY_TBL_LENGTH;

          ___FIELD(key,___SYMKEY_NEXT) = ___FIELD(___keyword_table,i);
          ___FIELD(___keyword_table,i) = key;
        }
    }
}


/*---------------------------------------------------------------------------*/

/*
 * C to Scheme call handler.
 */

___EXP_FUNC(___ERR_CODE,___call)
   ___P((int nargs,
         ___WORD proc,
         ___WORD stack_marker),
        (nargs,
         proc,
         stack_marker)
int nargs;
___WORD proc;
___WORD stack_marker;)
{
  ___ERR_CODE ___err;
  ___processor_state ___ps = ___PSTATE;
  ___WORD *___fp = ___ps->fp;

  /* 
   * The C function which has called ___call() has put the arguments
   * of the Scheme procedure on the stack above ___fp as shown in the
   * 'on entry' picture below.  The free space under arg1 is for a
   * continuation frame that performs the return of control from
   * Scheme to C.  This frame is set up by ___call() before the
   * destination Scheme procedure is invoked.  The frame is accessed
   * by the return_to_c handler (in "_kernel.scm") which is invoked
   * when the called procedure returns.  The frame contains a heap
   * allocated 'stack marker', allocated by the caller, which
   * indicates if it is still possible to return to the caller
   * (i.e. its activation frame is still on the C stack).  The caller
   * will store #f in the stack marker so that any subsequent attempt
   * to return to that invocation of the caller will be detected and
   * trigger an error.
   *
   *            ON ENTRY:          JUST BEFORE JUMPING TO THE SCHEME PROCEDURE:
   *
   *              STACK                         STACK             HEAP
   *          |     ^      |                |     ^      |
   *          |     |      |                |     |      |
   *          |            |                |            |
   *          |            |                |            |
   *          |   arg N    |                |   arg N    |
   *          |    ...     |       ___fp -->|    ...     |
   *          |   arg 1    |                |   arg 1    |
   *          +------------+                +------------+
   *          |  RESERVED  |                |  RESERVED  |    stack marker
   *          | <PADDING>  |                | <PADDING>  |   +------------+
   *          | undefined  |                |     ---------->|    HEAD    |
   *          | undefined  |                | return adr |   |     #t     |
   *          +------------+                +------------+   +------------+
   * ___fp -->|  RESERVED  |                |  RESERVED  |
   *          |    ...     |                |    ...     |
   *          +------------+                +------------+
   *          |            |                |            |
   */

  ___LD_ARG_REGS /* declare and load GVM argument registers from ___ps */

  ___fp[-1] = ___PSR0;      /* create a frame with the same format as the */
  ___fp[-2] = stack_marker; /* one created for the return to C handler    */
                            /* in "_kernel.scm"                           */

  ___fp -= ___FRAME_SPACE(2) + nargs; /* move ___fp to point to last arg */

  ___POP_ARGS_IN_REGS(nargs) /* load arguments into appropriate registers */

  ___ST_ARG_REGS /* write back GVM argument registers to ___ps */

  ___PSR0 = ___ps->handler_return_to_c;

  ___ps->fp = ___fp;
  ___ps->na = nargs;
  ___ps->pc = ___CAST(___label_struct*,proc-___tSUBTYPED)->entry_or_descr;
  ___PSSELF = proc;

  ___BEGIN_CATCH

  ___WORD ___pc = ___ps->pc;

  while (1)
    {
#define CALL_STEP \
___pc = ___CAST_FAKEHOST_TO_HOST(___CAST(___label_struct*,___pc-___tSUBTYPED) \
                                 ->host)(___ps)
      CALL_STEP;
      CALL_STEP;
      CALL_STEP;
      CALL_STEP;
      CALL_STEP;
      CALL_STEP;
      CALL_STEP;
      CALL_STEP;
    }

  ___END_CATCH

  /* ___err is set by the ___BEGIN_CATCH and ___END_CATCH */

  if (___err != ___UNWIND_C_STACK ||
      stack_marker != ___ps->fp[___FRAME_SPACE(2)-2]) /*need more unwinding?*/
    ___THROW(___err);

  ___ps->fp += ___FRAME_SPACE(2);
  ___PSR0 = ___ps->fp[-1];

  return ___NO_ERR;
}


___EXP_FUNC(void,___propagate_error)
   ___P((___ERR_CODE err),
        (err)
___ERR_CODE err;)
{
  if (err != ___NO_ERR)
    {
      ___processor_state ___ps = ___PSTATE;
      ___THROW (err);
    }
}


/*---------------------------------------------------------------------------*/

/*
 * Setup program and execute it.
 */

___HIDDEN int setup_state = 0; /* 0=pre-setup, 1=post-setup, 2=post-cleanup */


___EXP_FUNC(void,___cleanup) ___PVOID
{
  /*
   * Only do cleanup once after successful setup.
   */

  if (setup_state != 1)
    return;

  setup_state = 2;

  ___cleanup_mem ();
  ___cleanup_os ();
}


___EXP_FUNC(void,___cleanup_and_exit_process)
   ___P((int status),
        (status)
int status;)
{
  ___cleanup ();
  ___exit_process (status);
}


___EXP_FUNC(void,___setup_params_reset)
   ___P((___setup_params_struct *setup_params),
        (setup_params)
___setup_params_struct *setup_params;)
{
  setup_params->version           = 0;
  setup_params->argv              = 0;
  setup_params->min_heap          = 0;
  setup_params->max_heap          = 0;
  setup_params->live_percent      = 0;
  setup_params->gc_hook           = 0;
  setup_params->display_error     = 0;
  setup_params->fatal_error       = 0;
  setup_params->standard_level    = 0;
  setup_params->debug_settings    = 0;
  setup_params->terminal_settings = 0;
  setup_params->linker            = 0;
  setup_params->dummy8            = 0;
  setup_params->dummy7            = 0;
  setup_params->dummy6            = 0;
  setup_params->dummy5            = 0;
  setup_params->dummy4            = 0;
  setup_params->dummy3            = 0;
  setup_params->dummy2            = 0;
  setup_params->dummy1            = 0;
}


___EXP_FUNC(long,___get_min_heap) ___PVOID
{
  return ___setup_params.min_heap;
}


___EXP_FUNC(void,___set_min_heap)
   ___P((long bytes),
        (bytes)
long bytes;)
{
  ___setup_params.min_heap = bytes;
}


___EXP_FUNC(long,___get_max_heap) ___PVOID
{
  return ___setup_params.max_heap;
}


___EXP_FUNC(void,___set_max_heap)
   ___P((long bytes),
        (bytes)
long bytes;)
{
  ___setup_params.max_heap = bytes;
}


___EXP_FUNC(int,___get_live_percent) ___PVOID
{
  return ___setup_params.live_percent;
}


___EXP_FUNC(void,___set_live_percent)
   ___P((int percent),
        (percent)
int percent;)
{
  ___setup_params.live_percent = percent;
}


___EXP_FUNC(int,___get_standard_level) ___PVOID
{
  return ___setup_params.standard_level;
}


___EXP_FUNC(void,___set_standard_level)
   ___P((int level),
        (level)
int level;)
{
  ___setup_params.standard_level = level;
}


___EXP_FUNC(int,___set_debug_settings)
   ___P((int mask,
         int new_settings),
        (mask,
         new_settings)
int mask;
int new_settings;)
{
  return ___setup_params.debug_settings =
         (___setup_params.debug_settings & ~mask) | (new_settings & mask);
}


___EXP_FUNC(int,___set_terminal_settings)
   ___P((int mask,
         int new_settings),
        (mask,
         new_settings)
int mask;
int new_settings;)
{
  return ___setup_params.terminal_settings =
         (___setup_params.terminal_settings & ~mask) | (new_settings & mask);
}


___EXP_FUNC(___ERR_CODE,___setup)
   ___P((___setup_params_struct *setup_params),
        (setup_params)
___setup_params_struct *setup_params;)
{
  ___ERR_CODE ___err;
  ___processor_state ___ps;
  ___mod_or_lnk mol;
  ___WORD ___start;
  ___WORD marker;
  int i;

  /*
   * Check for valid setup_params structure.
   */

  if (setup_params == 0 ||
      setup_params->version != ___VERSION)
    return ___UNKNOWN_ERR;

  /*
   * Only do setup once.
   */

  if (setup_state != 0)
    return ___UNKNOWN_ERR;

  setup_state = 2; /* disallow cleanup, in case setup fails */

  /*
   * Remember setup parameters.
   */

  ___setup_params = *setup_params;

  /* 
   * Setup the operating system module.
   */

  ___setup_os ();

  /* 
   * Setup stack and heap.
   */

  ___setup_mem ();

  setup_state = 1; /* allow cleanup */

  /* 
   * Setup global state to avoid problems on systems that don't
   * support the dynamic loading of files that import functions.
   */

  ___GSTATE->dummy8 = 0;
  ___GSTATE->dummy7 = 0;
  ___GSTATE->dummy6 = 0;
  ___GSTATE->dummy5 = 0;
  ___GSTATE->dummy4 = 0;
  ___GSTATE->dummy3 = 0;
  ___GSTATE->dummy2 = 0;
  ___GSTATE->dummy1 = 0;

#ifdef ___CANT_IMPORT_CLIB_DYNAMICALLY

  ___GSTATE->fabs  = fabs;
  ___GSTATE->floor = floor;
  ___GSTATE->ceil  = ceil;
  ___GSTATE->exp   = exp;
  ___GSTATE->log   = log;
  ___GSTATE->sin   = sin;
  ___GSTATE->cos   = cos;
  ___GSTATE->tan   = tan;
  ___GSTATE->asin  = asin;
  ___GSTATE->acos  = acos;
  ___GSTATE->atan  = atan;
  ___GSTATE->atan2 = atan2;
  ___GSTATE->sqrt  = sqrt;

#endif

#ifdef ___USE_SETJMP
#ifdef ___CANT_IMPORT_SETJMP_DYNAMICALLY

  ___GSTATE->setjmp = setjmp;

#endif
#endif

#ifdef ___CANT_IMPORT_DYNAMICALLY

  ___GSTATE->___copysign
    = ___copysign;

  ___GSTATE->___isfinite
    = ___isfinite;

  ___GSTATE->___isnan
    = ___isnan;

  ___GSTATE->___round
    = ___round;

  ___GSTATE->___S64_from_SM32_fn
    = ___S64_from_SM32_fn;

  ___GSTATE->___S64_from_SM32_UM32_fn
    = ___S64_from_SM32_UM32_fn;

  ___GSTATE->___S64_from_LONG_fn
    = ___S64_from_LONG_fn;

  ___GSTATE->___S64_to_LONG_fn
    = ___S64_to_LONG_fn;

  ___GSTATE->___S64_fits_in_width_fn
    = ___S64_fits_in_width_fn;

  ___GSTATE->___U64_from_UM32_fn
    = ___U64_from_UM32_fn;

  ___GSTATE->___U64_from_UM32_UM32_fn
    = ___U64_from_UM32_UM32_fn;

  ___GSTATE->___U64_from_ULONG_fn
    = ___U64_from_ULONG_fn;

  ___GSTATE->___U64_to_ULONG_fn
    = ___U64_to_ULONG_fn;

  ___GSTATE->___U64_fits_in_width_fn
    = ___U64_fits_in_width_fn;

  ___GSTATE->___SCMOBJ_to_S8
    = ___SCMOBJ_to_S8;

  ___GSTATE->___SCMOBJ_to_U8
    = ___SCMOBJ_to_U8;

  ___GSTATE->___SCMOBJ_to_S16
    = ___SCMOBJ_to_S16;

  ___GSTATE->___SCMOBJ_to_U16
    = ___SCMOBJ_to_U16;

  ___GSTATE->___SCMOBJ_to_S32
    = ___SCMOBJ_to_S32;

  ___GSTATE->___SCMOBJ_to_U32
    = ___SCMOBJ_to_U32;

  ___GSTATE->___SCMOBJ_to_S64
    = ___SCMOBJ_to_S64;

  ___GSTATE->___SCMOBJ_to_U64
    = ___SCMOBJ_to_U64;

  ___GSTATE->___SCMOBJ_to_F32
    = ___SCMOBJ_to_F32;

  ___GSTATE->___SCMOBJ_to_F64
    = ___SCMOBJ_to_F64;

  ___GSTATE->___SCMOBJ_to_CHAR
    = ___SCMOBJ_to_CHAR;

  ___GSTATE->___SCMOBJ_to_SCHAR
    = ___SCMOBJ_to_SCHAR;

  ___GSTATE->___SCMOBJ_to_UCHAR
    = ___SCMOBJ_to_UCHAR;

  ___GSTATE->___SCMOBJ_to_LATIN1
    = ___SCMOBJ_to_LATIN1;

  ___GSTATE->___SCMOBJ_to_UCS2
    = ___SCMOBJ_to_UCS2;

  ___GSTATE->___SCMOBJ_to_UCS4
    = ___SCMOBJ_to_UCS4;

  ___GSTATE->___SCMOBJ_to_SHORT
    = ___SCMOBJ_to_SHORT;

  ___GSTATE->___SCMOBJ_to_USHORT
    = ___SCMOBJ_to_USHORT;

  ___GSTATE->___SCMOBJ_to_INT
    = ___SCMOBJ_to_INT;

  ___GSTATE->___SCMOBJ_to_UINT
    = ___SCMOBJ_to_UINT;

  ___GSTATE->___SCMOBJ_to_LONG
    = ___SCMOBJ_to_LONG;

  ___GSTATE->___SCMOBJ_to_ULONG
    = ___SCMOBJ_to_ULONG;

  ___GSTATE->___SCMOBJ_to_FLOAT
    = ___SCMOBJ_to_FLOAT;

  ___GSTATE->___SCMOBJ_to_DOUBLE
    = ___SCMOBJ_to_DOUBLE;

  ___GSTATE->___SCMOBJ_to_STRUCT
    = ___SCMOBJ_to_STRUCT;

  ___GSTATE->___SCMOBJ_to_UNION
    = ___SCMOBJ_to_UNION;

  ___GSTATE->___SCMOBJ_to_CLASS
    = ___SCMOBJ_to_CLASS;

  ___GSTATE->___SCMOBJ_to_POINTER
    = ___SCMOBJ_to_POINTER;

  ___GSTATE->___SCMOBJ_to_NONNULLPOINTER
    = ___SCMOBJ_to_NONNULLPOINTER;

  ___GSTATE->___SCMOBJ_to_FUNCTION
    = ___SCMOBJ_to_FUNCTION;

  ___GSTATE->___SCMOBJ_to_NONNULLFUNCTION
    = ___SCMOBJ_to_NONNULLFUNCTION;

  ___GSTATE->___SCMOBJ_to_BOOL
    = ___SCMOBJ_to_BOOL;

  ___GSTATE->___SCMOBJ_to_CHARSTRING
    = ___SCMOBJ_to_CHARSTRING;

  ___GSTATE->___SCMOBJ_to_NONNULLCHARSTRING
    = ___SCMOBJ_to_NONNULLCHARSTRING;

  ___GSTATE->___SCMOBJ_to_NONNULLCHARSTRINGLIST
    = ___SCMOBJ_to_NONNULLCHARSTRINGLIST;

  ___GSTATE->___SCMOBJ_to_LATIN1STRING
    = ___SCMOBJ_to_LATIN1STRING;

  ___GSTATE->___SCMOBJ_to_NONNULLLATIN1STRING
    = ___SCMOBJ_to_NONNULLLATIN1STRING;

  ___GSTATE->___SCMOBJ_to_NONNULLLATIN1STRINGLIST
    = ___SCMOBJ_to_NONNULLLATIN1STRINGLIST;

  ___GSTATE->___SCMOBJ_to_UTF8STRING
    = ___SCMOBJ_to_UTF8STRING;

  ___GSTATE->___SCMOBJ_to_NONNULLUTF8STRING
    = ___SCMOBJ_to_NONNULLUTF8STRING;

  ___GSTATE->___SCMOBJ_to_NONNULLUTF8STRINGLIST
    = ___SCMOBJ_to_NONNULLUTF8STRINGLIST;

  ___GSTATE->___SCMOBJ_to_UCS2STRING
    = ___SCMOBJ_to_UCS2STRING;

  ___GSTATE->___SCMOBJ_to_NONNULLUCS2STRING
    = ___SCMOBJ_to_NONNULLUCS2STRING;

  ___GSTATE->___SCMOBJ_to_NONNULLUCS2STRINGLIST
    = ___SCMOBJ_to_NONNULLUCS2STRINGLIST;

  ___GSTATE->___SCMOBJ_to_UCS4STRING
    = ___SCMOBJ_to_UCS4STRING;

  ___GSTATE->___SCMOBJ_to_NONNULLUCS4STRING
    = ___SCMOBJ_to_NONNULLUCS4STRING;

  ___GSTATE->___SCMOBJ_to_NONNULLUCS4STRINGLIST
    = ___SCMOBJ_to_NONNULLUCS4STRINGLIST;

  ___GSTATE->___SCMOBJ_to_NONNULLSTRING
    = ___SCMOBJ_to_NONNULLSTRING;

  ___GSTATE->___SCMOBJ_to_NONNULLSTRINGLIST
    = ___SCMOBJ_to_NONNULLSTRINGLIST;

  ___GSTATE->___SCMOBJ_to_VARIANT
    = ___SCMOBJ_to_VARIANT;

  ___GSTATE->___release_foreign
    = ___release_foreign;

  ___GSTATE->___release_pointer
    = ___release_pointer;

  ___GSTATE->___release_function
    = ___release_function;

  ___GSTATE->___release_string
    = ___release_string;

  ___GSTATE->___release_string_list
    = ___release_string_list;

  ___GSTATE->___release_variant
    = ___release_variant;

  ___GSTATE->___S8_to_SCMOBJ
    = ___S8_to_SCMOBJ;

  ___GSTATE->___U8_to_SCMOBJ
    = ___U8_to_SCMOBJ;

  ___GSTATE->___S16_to_SCMOBJ
    = ___S16_to_SCMOBJ;

  ___GSTATE->___U16_to_SCMOBJ
    = ___U16_to_SCMOBJ;

  ___GSTATE->___S32_to_SCMOBJ
    = ___S32_to_SCMOBJ;

  ___GSTATE->___U32_to_SCMOBJ
    = ___U32_to_SCMOBJ;

  ___GSTATE->___S64_to_SCMOBJ
    = ___S64_to_SCMOBJ;

  ___GSTATE->___U64_to_SCMOBJ
    = ___U64_to_SCMOBJ;

  ___GSTATE->___F32_to_SCMOBJ
    = ___F32_to_SCMOBJ;

  ___GSTATE->___F64_to_SCMOBJ
    = ___F64_to_SCMOBJ;

  ___GSTATE->___CHAR_to_SCMOBJ
    = ___CHAR_to_SCMOBJ;

  ___GSTATE->___SCHAR_to_SCMOBJ
    = ___SCHAR_to_SCMOBJ;

  ___GSTATE->___UCHAR_to_SCMOBJ
    = ___UCHAR_to_SCMOBJ;

  ___GSTATE->___LATIN1_to_SCMOBJ
    = ___LATIN1_to_SCMOBJ;

  ___GSTATE->___UCS2_to_SCMOBJ
    = ___UCS2_to_SCMOBJ;

  ___GSTATE->___UCS4_to_SCMOBJ
    = ___UCS4_to_SCMOBJ;

  ___GSTATE->___SHORT_to_SCMOBJ
    = ___SHORT_to_SCMOBJ;

  ___GSTATE->___USHORT_to_SCMOBJ
    = ___USHORT_to_SCMOBJ;

  ___GSTATE->___INT_to_SCMOBJ
    = ___INT_to_SCMOBJ;

  ___GSTATE->___UINT_to_SCMOBJ
    = ___UINT_to_SCMOBJ;

  ___GSTATE->___LONG_to_SCMOBJ
    = ___LONG_to_SCMOBJ;

  ___GSTATE->___ULONG_to_SCMOBJ
    = ___ULONG_to_SCMOBJ;

  ___GSTATE->___FLOAT_to_SCMOBJ
    = ___FLOAT_to_SCMOBJ;

  ___GSTATE->___DOUBLE_to_SCMOBJ
    = ___DOUBLE_to_SCMOBJ;

  ___GSTATE->___STRUCT_to_SCMOBJ
    = ___STRUCT_to_SCMOBJ;

  ___GSTATE->___UNION_to_SCMOBJ
    = ___UNION_to_SCMOBJ;

  ___GSTATE->___CLASS_to_SCMOBJ
    = ___CLASS_to_SCMOBJ;

  ___GSTATE->___POINTER_to_SCMOBJ
    = ___POINTER_to_SCMOBJ;

  ___GSTATE->___NONNULLPOINTER_to_SCMOBJ
    = ___NONNULLPOINTER_to_SCMOBJ;

  ___GSTATE->___FUNCTION_to_SCMOBJ
    = ___FUNCTION_to_SCMOBJ;

  ___GSTATE->___NONNULLFUNCTION_to_SCMOBJ
    = ___NONNULLFUNCTION_to_SCMOBJ;

  ___GSTATE->___BOOL_to_SCMOBJ
    = ___BOOL_to_SCMOBJ;

  ___GSTATE->___CHARSTRING_to_SCMOBJ
    = ___CHARSTRING_to_SCMOBJ;

  ___GSTATE->___NONNULLCHARSTRING_to_SCMOBJ
    = ___NONNULLCHARSTRING_to_SCMOBJ;

  ___GSTATE->___NONNULLCHARSTRINGLIST_to_SCMOBJ
    = ___NONNULLCHARSTRINGLIST_to_SCMOBJ;

  ___GSTATE->___LATIN1STRING_to_SCMOBJ
    = ___LATIN1STRING_to_SCMOBJ;

  ___GSTATE->___NONNULLLATIN1STRING_to_SCMOBJ
    = ___NONNULLLATIN1STRING_to_SCMOBJ;

  ___GSTATE->___NONNULLLATIN1STRINGLIST_to_SCMOBJ
    = ___NONNULLLATIN1STRINGLIST_to_SCMOBJ;

  ___GSTATE->___UTF8STRING_to_SCMOBJ
    = ___UTF8STRING_to_SCMOBJ;

  ___GSTATE->___NONNULLUTF8STRING_to_SCMOBJ
    = ___NONNULLUTF8STRING_to_SCMOBJ;

  ___GSTATE->___NONNULLUTF8STRINGLIST_to_SCMOBJ
    = ___NONNULLUTF8STRINGLIST_to_SCMOBJ;

  ___GSTATE->___UCS2STRING_to_SCMOBJ
    = ___UCS2STRING_to_SCMOBJ;

  ___GSTATE->___NONNULLUCS2STRING_to_SCMOBJ
    = ___NONNULLUCS2STRING_to_SCMOBJ;

  ___GSTATE->___NONNULLUCS2STRINGLIST_to_SCMOBJ
    = ___NONNULLUCS2STRINGLIST_to_SCMOBJ;

  ___GSTATE->___UCS4STRING_to_SCMOBJ
    = ___UCS4STRING_to_SCMOBJ;

  ___GSTATE->___NONNULLUCS4STRING_to_SCMOBJ
    = ___NONNULLUCS4STRING_to_SCMOBJ;

  ___GSTATE->___NONNULLUCS4STRINGLIST_to_SCMOBJ
    = ___NONNULLUCS4STRINGLIST_to_SCMOBJ;

  ___GSTATE->___NONNULLSTRING_to_SCMOBJ
    = ___NONNULLSTRING_to_SCMOBJ;

  ___GSTATE->___NONNULLSTRINGLIST_to_SCMOBJ
    = ___NONNULLSTRINGLIST_to_SCMOBJ;

  ___GSTATE->___VARIANT_to_SCMOBJ
    = ___VARIANT_to_SCMOBJ;

  ___GSTATE->___release_foreign
    = ___release_foreign;

  ___GSTATE->___make_cdef_stack_marker
    = ___make_cdef_stack_marker;

  ___GSTATE->___kill_cdef_stack_marker
    = ___kill_cdef_stack_marker;

  ___GSTATE->___release_scmobj
    = ___release_scmobj;

  ___GSTATE->___alloc_scmobj
    = ___alloc_scmobj;

  ___GSTATE->___make_pair
    = ___make_pair;

  ___GSTATE->___make_vector
    = ___make_vector;

  ___GSTATE->___make_string
    = ___make_string;

  ___GSTATE->___still_obj_refcount_inc
    = ___still_obj_refcount_inc;

  ___GSTATE->___still_obj_refcount_dec
    = ___still_obj_refcount_dec;

  ___GSTATE->___get_min_heap
    = ___get_min_heap;

  ___GSTATE->___set_min_heap
    = ___set_min_heap;

  ___GSTATE->___get_max_heap
    = ___get_max_heap;

  ___GSTATE->___set_max_heap
    = ___set_max_heap;

  ___GSTATE->___get_live_percent
    = ___get_live_percent;

  ___GSTATE->___set_live_percent
    = ___set_live_percent;

  ___GSTATE->___get_standard_level
    = ___get_standard_level;

  ___GSTATE->___set_standard_level
    = ___set_standard_level;

  ___GSTATE->___set_debug_settings
    = ___set_debug_settings;

  ___GSTATE->___set_terminal_settings
    = ___set_terminal_settings;

  ___GSTATE->___cleanup
    = ___cleanup;

  ___GSTATE->___cleanup_and_exit_process
    = ___cleanup_and_exit_process;

  ___GSTATE->___call
    = ___call;

  ___GSTATE->___propagate_error
    = ___propagate_error;

  ___GSTATE->___raise_interrupt
    = ___raise_interrupt;

#endif

  /* 
   * Get processor state.
   */

  ___ps = ___PSTATE;

  /*
   * Setup multithreading structures.
   */

  ___ps->current_thread  = ___FAL;
  ___ps->run_queue = ___FAL;

  /*
   * Setup registers.
   */

  for (i=0; i<___NB_GVM_REGS; i++)
    ___ps->r[i] = ___VOID;

  /* 
   * Setup exception handling.
   */


#ifdef ___USE_SETJMP

  ___ps->catcher = 0;

#endif

  /* 
   * Setup interrupt system.
   */

  ___disable_interrupts ();

  for (i=0; i<___NB_INTRS; i++)
    ___ps->intr_flag[i] = 0;

  ___begin_interrupt_service ();
  ___end_interrupt_service (0);

  /* 
   * Create empty global variable list, symbol table and keyword
   * table.
   */

  ___ps->glo_list_head = 0;
  ___ps->glo_list_tail = 0;

  ___symbol_table = ___make_vector (SYMKEY_TBL_LENGTH, ___NUL, ___PERM);

  if (___FIXNUMP(___symbol_table))
    {
      ___cleanup ();
      return ___INT(___symbol_table);
    }

  ___keyword_table = ___make_vector (SYMKEY_TBL_LENGTH, ___NUL, ___PERM);

  if (___FIXNUMP(___keyword_table))
    {
      ___cleanup ();
      return ___INT(___keyword_table);
    }

  /*
   * Setup program's linker structure.
   */

  mol = linker_to_mod_or_lnk (___setup_params.linker);

  /* 
   * Initialize symbol table, keyword table, global variables
   * and primitives.
   */

  init_symkey_glo1 (mol);
  init_symkey_glo2 (mol);

  /* 
   * Setup each module.
   */

  ___program_descr = setup_modules (mol);

  if (___FIXNUMP(___program_descr))
    {
      ___cleanup ();
      return ___INT(___program_descr);
    }

  /* 
   * Create list of command line arguments (accessible through ##command-line).
   */

  {
    char **argv = ___setup_params.argv;
    char *dummy_argv[2];

    if (argv[0] == 0) /* use dummy program name if none supplied */
      {
        dummy_argv[0] = "";
        dummy_argv[1] = 0;
        argv = dummy_argv;
      }

#define ___COMMAND_LINE_CE_SELECT(latin1,utf8,ucs2,ucs4,native) native

    if ((___err = ___NONNULLSTRINGLIST_to_SCMOBJ
                    (argv,
                     &___command_line,
                     0,
                     ___CE(___COMMAND_LINE_CE_SELECT)))
        != ___NO_ERR)
      {
        ___cleanup ();
        return ___err;
      }
  }

  /* 
   * Setup kernel handlers.
   */

  ___start = ___G__23__23_kernel_2d_handlers.prm;

#define ___PH_LBL0 1

  /* 
   * The label numbers must match those in the procedure
   * "##kernel-handlers" in the file "_kernel.scm".
   */

  ___internal_return             = ___LBL(1);
  ___ps->handler_break           = ___LBL(2);
  ___ps->handler_stack_limit     = ___LBL(3);
  ___ps->handler_heap_limit      = ___LBL(4);
  ___ps->handler_not_proc        = ___LBL(5);
  ___ps->handler_not_proc_glo    = ___LBL(6);
  ___ps->handler_wrong_nargs     = ___LBL(7);
  ___ps->handler_get_rest        = ___LBL(8);
  ___ps->handler_get_key         = ___LBL(9);
  ___ps->handler_get_key_rest    = ___LBL(10);
  ___ps->handler_force           = ___LBL(11);
  ___ps->handler_clam_conv_error = ___LBL(12);
  ___ps->handler_cdef_conv_error = ___LBL(13);
  ___ps->handler_return_to_c     = ___LBL(14);

  /*
   * Call kernel to start executing program.
   */

  ___ps->r[0] = ___ps->handler_break;

  ___BEGIN_CATCH

  if ((___err = ___make_cdef_stack_marker (&marker)) == ___NO_ERR)
    {
      ___err = ___call (0, ___FIELD(___FIELD(___program_descr,0),0), marker);
      ___kill_cdef_stack_marker (marker);
    }

  ___END_CATCH

  if (___err != ___NO_ERR)
    {
      ___cleanup ();
      return ___err;
    }

  return ___NO_ERR;
}


/*---------------------------------------------------------------------------*/
