;==============================================================================

; file: "_io.scm"

; Copyright (C) 1994-2002 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##include "header.scm")

;==============================================================================

; Implementation of exceptions.

(define-procedures-for-datum-parsing-exception)

(define-prim (##raise-datum-parsing-exception kind readenv . parameters)
  (macro-raise (macro-make-datum-parsing-exception kind readenv parameters)))

;------------------------------------------------------------------------------

; Error handling.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define-prim (##trap-psettings prim . args)
  (##raise-os-exception "Invalid port settings" #f #t prim args))

;;;;;;;;;;;;;;;;
(define-prim (##trap-open-file prim . args)
  (##raise-os-exception "Can't open file" #f #t prim args))

;------------------------------------------------------------------------------

; Implementation of write environments.

(define-prim (##make-writeenv
              style
              port
              readtable
              marktable
              force?
              width
              close-parens
              level
              limit)
  (macro-make-writeenv
   style
   port
   readtable
   marktable
   force?
   width
   close-parens
   level
   limit))

;------------------------------------------------------------------------------

; Implementation of read environments.

(define-prim (##make-readenv
              port
              readtable
              wrapper
              unwrapper)
  (macro-make-readenv
   port
   readtable
   wrapper
   unwrapper
   '()
   0))

(define-prim (##readenv-current-filepos re)
  (##readenv-relative-filepos re 0))

(define-prim (##readenv-relative-filepos re offset)
  (let* ((port
          (macro-readenv-port re))
         (line
          (macro-port-char-rlines port))
         (char-count
          (##fixnum.- (##fixnum.+ (macro-port-char-rchars port)
                                  (macro-port-char-rlo port))
                      offset))
         (col
          (##fixnum.- char-count
                      (macro-port-char-rcurline port))))
    (##make-filepos line col char-count)))

;------------------------------------------------------------------------------

; Implementation of port settings.

(define-prim (##make-psettings
              direction
              allowed-settings
              #!optional
              (settings (macro-absent-obj)))
  (let ((psettings
         (##vector direction
                   (##vector (macro-default-char-encoding)
                             (macro-default-eol-encoding)
                             (macro-default-buffering))
                   (##vector (macro-default-char-encoding)
                             (macro-default-eol-encoding)
                             (macro-default-buffering))
                   (macro-default-append)
                   (macro-default-create)
                   (macro-default-truncate)
                   (macro-default-permissions)
                   (macro-default-output-width)
                   (macro-default-stderr-redir)
                   (macro-default-pseudo-term)
                   (macro-default-coalesce)
                   (macro-default-keep-alive)
                   (macro-default-backlog)
                   (macro-default-reuse-address))))
    (##parse-psettings!
     allowed-settings
     (if (##eq? settings (macro-absent-obj)) '() settings)
     psettings)))

(define-prim (##parse-psettings! allowed-settings settings psettings)

  (define (char-encoding value)
    (cond ((##eq? value 'ascii)
           (macro-char-encoding-ascii))
          ((##eq? value 'latin1)
           (macro-char-encoding-latin1))
          ((##eq? value 'utf8)
           (macro-char-encoding-utf8))
          ((##eq? value 'ucs2)
           (macro-char-encoding-ucs2))
          ((##eq? value 'ucs2le)
           (macro-char-encoding-ucs2le))
          ((##eq? value 'ucs2be)
           (macro-char-encoding-ucs2be))
          ((##eq? value 'ucs4)
           (macro-char-encoding-ucs4))
          ((##eq? value 'ucs4le)
           (macro-char-encoding-ucs4le))
          ((##eq? value 'ucs4be)
           (macro-char-encoding-ucs4be))
          ((##eq? value 'native)
           (macro-char-encoding-native))
          (else
           #f)))

  (define (eol-encoding value)
    (cond ((##eq? value 'lf)
           (macro-eol-encoding-lf))
          ((##eq? value 'cr)
           (macro-eol-encoding-cr))
          ((##eq? value 'cr-lf)
           (macro-eol-encoding-crlf))
          ((##eq? value 'lf-cr)
           (macro-eol-encoding-lfcr))
          (else
           #f)))

  (define (buffering value)
    (cond ((##eq? value #t)
           (macro-buffering))
          ((##eq? value 'line)
           (macro-line-buffering))
          ((##eq? value #f)
           (macro-no-buffering))
          (else
           #f)))

  (define (direction value)
    (cond ((##eq? value 'input)
           (macro-direction-in))
          ((##eq? value 'output)
           (macro-direction-out))
          ((##eq? value 'input-output)
           (macro-direction-inout))
          (else
           #f)))

  (define (append-flag value)
    (if value
      (macro-append)
      (macro-no-append)))

  (define (create-flag value)
    (cond ((##eq? value #f)
           (macro-no-create))
          ((##eq? value 'maybe)
           (macro-maybe-create))
          ((##eq? value #t)
           (macro-create))
          (else
           #f)))

  (define (truncate-flag value)
    (if value
      (macro-truncate)
      (macro-no-truncate)))

  (define (permissions value)
    (if (and (##fixnum? value)
             (##not (##fixnum.< value 0))
             (##fixnum.< value #o1000))
      value
      #f))

  (define (output-width value)
    (if (and (##fixnum? value)
             (##fixnum.< 0 value))
      value
      #f))

  (define (stderr-redir value)
    (cond ((##eq? value #t)
           (macro-stderr-to-stdout))
          ((##eq? value #f)
           (macro-stderr-unchanged))
          (else
           #f)))

  (define (pseudo-term value)
    (cond ((##eq? value #t)
           (macro-pseudo-term))
          ((##eq? value #f)
           (macro-no-pseudo-term))
          (else
           #f)))

  (define (coalesce value)
    (cond ((##eq? value #t)
           (macro-coalesce))
          ((##eq? value #f)
           (macro-no-coalesce))
          (else
           #f)))

  (define (keep-alive value)
    (cond ((##eq? value #t)
           (macro-keep-alive))
          ((##eq? value #f)
           (macro-no-keep-alive))
          (else
           #f)))

  (define (backlog value)
    (if (and (##fixnum? value)
             (##not (##fixnum.< value 0)))
      value
      #f))

  (define (reuse-address value)
    (cond ((##eq? value #t)
           (macro-reuse-address))
          ((##eq? value #f)
           (macro-no-reuse-address))
          (else
           #f)))

  (let loop ((lst settings))
    (macro-force-vars (lst)
      (cond ((##pair? lst)
             (let ((name (##car lst))
                   (rest1 (##cdr lst)))
               (macro-force-vars (name rest1)
                 (if (and (##memq name allowed-settings)
                          (##pair? rest1))

                   (let ((value (##car rest1))
                         (rest2 (##cdr rest1)))
                     (macro-force-vars (value)

                       (cond ((##eq? name 'input-char-encoding:)
                              (let ((x (char-encoding value)))
                                (if x
                                  (begin
                                    (macro-options-char-encoding-set!
                                     (macro-psettings-roptions psettings)
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'output-char-encoding:)
                              (let ((x (char-encoding value)))
                                (if x
                                  (begin
                                    (macro-options-char-encoding-set!
                                     (macro-psettings-woptions psettings)
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'char-encoding:)
                              (let ((x (char-encoding value)))
                                (if x
                                  (begin
                                    (macro-options-char-encoding-set!
                                     (macro-psettings-roptions psettings)
                                     x)
                                    (macro-options-char-encoding-set!
                                     (macro-psettings-woptions psettings)
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'input-eol-encoding:)
                              (let ((x (eol-encoding value)))
                                (if x
                                  (begin
                                    (macro-options-eol-encoding-set!
                                     (macro-psettings-roptions psettings)
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'output-eol-encoding:)
                              (let ((x (eol-encoding value)))
                                (if x
                                  (begin
                                    (macro-options-eol-encoding-set!
                                     (macro-psettings-woptions psettings)
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'eol-encoding:)
                              (let ((x (eol-encoding value)))
                                (if x
                                  (begin
                                    (macro-options-eol-encoding-set!
                                     (macro-psettings-roptions psettings)
                                     x)
                                    (macro-options-eol-encoding-set!
                                     (macro-psettings-woptions psettings)
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'input-buffering:)
                              (let ((x (buffering value)))
                                (if x
                                  (begin
                                    (macro-options-buffering-set!
                                     (macro-psettings-roptions psettings)
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'output-buffering:)
                              (let ((x (buffering value)))
                                (if x
                                  (begin
                                    (macro-options-buffering-set!
                                     (macro-psettings-woptions psettings)
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'buffering:)
                              (let ((x (buffering value)))
                                (if x
                                  (begin
                                    (macro-options-buffering-set!
                                     (macro-psettings-roptions psettings)
                                     x)
                                    (macro-options-buffering-set!
                                     (macro-psettings-woptions psettings)
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'direction:)
                              (let ((x (direction value)))
                                (if x
                                  (begin
                                    (macro-psettings-direction-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'append:)
                              (let ((x (append-flag value)))
                                (if x
                                  (begin
                                    (macro-psettings-append-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'create:)
                              (let ((x (create-flag value)))
                                (if x
                                  (begin
                                    (macro-psettings-create-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'truncate:)
                              (let ((x (truncate-flag value)))
                                (if x
                                  (begin
                                    (macro-psettings-truncate-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'permissions:)
                              (let ((x (permissions value)))
                                (if x
                                  (begin
                                    (macro-psettings-permissions-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'output-width:)
                              (let ((x (output-width value)))
                                (if x
                                  (begin
                                    (macro-psettings-output-width-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'stderr-redirection:)
                              (let ((x (stderr-redir value)))
                                (if x
                                  (begin
                                    (macro-psettings-stderr-redir-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'pseudo-terminal:)
                              (let ((x (pseudo-term value)))
                                (if x
                                  (begin
                                    (macro-psettings-pseudo-term-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'coalesce:)
                              (let ((x (coalesce value)))
                                (if x
                                  (begin
                                    (macro-psettings-coalesce-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'keep-alive:)
                              (let ((x (keep-alive value)))
                                (if x
                                  (begin
                                    (macro-psettings-keep-alive-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'backlog:)
                              (let ((x (backlog value)))
                                (if x
                                  (begin
                                    (macro-psettings-backlog-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             ((##eq? name 'reuse-address:)
                              (let ((x (reuse-address value)))
                                (if x
                                  (begin
                                    (macro-psettings-reuse-address-set!
                                     psettings
                                     x)
                                    (loop rest2))
                                  #f)))

                             (else
                              #f))))

                   #f))))

            ((##null? lst)
             psettings)

            (else
             #f)))))

(define-prim (##psettings->roptions psettings default-options)
  (##psettings-options->options
   (macro-psettings-roptions psettings)
   (##fixnum.modulo default-options 16384)))

(define-prim (##psettings->woptions psettings default-options)
  (##psettings-options->options
   (macro-psettings-woptions psettings)
   (##fixnum.quotient default-options 16384)))

(define-prim (##psettings-options->options options default-options)
  (let ((buffering
         (macro-options-buffering options))
        (eol-encoding
         (macro-options-eol-encoding options))
        (char-encoding
         (macro-options-char-encoding options)))
    (##fixnum.+
     (##fixnum.* (macro-char-encoding-shift)
                 (if (##fixnum.= char-encoding (macro-default-char-encoding))
                   (##fixnum.modulo
                    (##fixnum.quotient default-options
                                       (macro-char-encoding-shift))
                    (macro-char-encoding-range))
                   char-encoding))
     (##fixnum.+
      (##fixnum.* (macro-eol-encoding-shift)
                  (if (##fixnum.= eol-encoding (macro-default-eol-encoding))
                    (##fixnum.modulo
                     (##fixnum.quotient default-options
                                        (macro-eol-encoding-shift))
                     (macro-eol-encoding-range))
                    eol-encoding))
      (##fixnum.+
       (##fixnum.* (macro-open-state-shift)
                   (##fixnum.modulo
                    (##fixnum.quotient default-options
                                       (macro-open-state-shift))
                    (macro-open-state-range)))
       (##fixnum.* (macro-buffering-shift)
                   (if (##fixnum.= buffering (macro-default-buffering))
                     (##fixnum.modulo
                      (##fixnum.quotient default-options
                                         (macro-buffering-shift))
                      (macro-buffering-range))
                     buffering)))))))

(define-prim (##psettings->device-flags psettings)
  (let ((direction
         (macro-psettings-direction psettings))
        (append
         (macro-psettings-append psettings))
        (create
         (macro-psettings-create psettings))
        (truncate
         (macro-psettings-truncate psettings)))
  (##fixnum.+
   (##fixnum.* 16
               direction)
   (##fixnum.+
    (##fixnum.* 8
                (if (##not (##fixnum.= append (macro-default-append)))
                  append
                  (macro-no-append)))
    (##fixnum.+
     (##fixnum.* 2
                 (cond ((##not (##fixnum.= create (macro-default-create)))
                        create)
                       ((##fixnum.= direction (macro-direction-out))
                        (macro-maybe-create))
                       (else
                        (macro-no-create))))
     (cond ((##not (##fixnum.= truncate (macro-default-truncate)))
            truncate)
           ((##fixnum.= direction (macro-direction-out))
            (macro-truncate))
           (else
            (macro-no-truncate))))))))

(define-prim (##psettings->permissions psettings)
  (let ((permissions (macro-psettings-permissions psettings)))
    (if (##not (##fixnum.= permissions (macro-default-permissions)))
      permissions
      #o777)))

(define-prim (##psettings->output-width psettings)
  (let ((output-width (macro-psettings-output-width psettings)))
    (if (##not (##fixnum.= output-width (macro-default-output-width)))
      output-width
      80)))

;------------------------------------------------------------------------------

; Implementation of port type checking.

(define ##port-tag
  '#(port -1))

(define-prim (##port? obj)
  (macro-port? obj))

(define-prim (port? obj)
  (macro-force-vars (obj)
    (macro-port? obj)))

(define-prim (##input-port? obj)
  (macro-input-port? obj))

(define-prim (input-port? obj)
  (macro-force-vars (obj)
    (macro-input-port? obj)))

(define-prim (##output-port? obj)
  (macro-output-port? obj))

(define-prim (output-port? obj)
  (macro-force-vars (obj)
    (macro-output-port? obj)))

(define-fail-check-type port)
(define-fail-check-type input-port)
(define-fail-check-type output-port)
(define-fail-check-type character-input-port)
(define-fail-check-type character-output-port)
(define-fail-check-type byte-input-port)
(define-fail-check-type byte-output-port)

;------------------------------------------------------------------------------

; Implementation of device ports.

(define-prim (##make-device-port device-name rdevice wdevice psettings)

  (define char-buf-len 512) ; character buffer length
  (define byte-buf-len 1024) ; byte buffer length

  (let* ((mutex
          (macro-make-port-mutex))
         (rkind
          (if rdevice
            (##dev-kind rdevice)
            (macro-none-kind)))
         (wkind
          (if wdevice
            (##dev-kind wdevice)
            (macro-none-kind)))
         (roptions
          (if (##fixnum.= rkind (macro-none-kind))
            0
            (##psettings->roptions psettings
                                   (##dev-stream-default-options rdevice))))
         (rtimeout
          #t)
         (rtimeout-thunk
          #f)
         (woptions
          (if (##fixnum.= wkind (macro-none-kind))
            0
            (##psettings->woptions psettings
                                   (##dev-stream-default-options wdevice))))
         (wtimeout
          #t)
         (wtimeout-thunk
          #f)
         (char-rbuf
          (and (##not (##fixnum.= rkind (macro-none-kind)))
               (##make-string (if (macro-unbuffered? roptions)
                                1
                                char-buf-len))))
         (char-rlo
          0)
         (char-rhi
          0)
         (char-rchars
          0)
         (char-rlines
          0)
         (char-rcurline
          0)
         (char-rbuf-fill
          ##char-rbuf-fill)
         (char-peek-eof?
          #f)
         (char-wbuf
          (and (##not (##fixnum.= wkind (macro-none-kind)))
               (##make-string (if (macro-unbuffered? woptions)
                                1
                                char-buf-len))))
         (char-wlo
          0)
         (char-whi
          0)
         (char-wchars
          0)
         (char-wlines
          0)
         (char-wcurline
          0)
         (char-wbuf-drain
          ##char-wbuf-drain)
         (byte-rbuf
          (and (##not (##fixnum.= rkind (macro-none-kind)))
               (##make-u8vector byte-buf-len)))
         (byte-rlo
          0)
         (byte-rhi
          0)
         (byte-rbuf-fill
          ##byte-rbuf-fill)
         (byte-wbuf
          (and (##not (##fixnum.= wkind (macro-none-kind)))
               (##make-u8vector byte-buf-len)))
         (byte-wlo
          0)
         (byte-whi
          0)
         (byte-wbuf-drain
          ##byte-wbuf-drain)
         (rdevice-condvar
          (and (##not (##fixnum.= rkind (macro-none-kind)))
               (##make-rdevice-condvar rdevice)))
         (wdevice-condvar
          (and (##not (##fixnum.= wkind (macro-none-kind)))
               (##make-wdevice-condvar wdevice))))

     (define (name port)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (macro-port-device-name port))

     (define (read-datum port re)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (##read-datum-or-eof re))

     (define (write-datum port obj we)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (##wr we obj))

     (define (newline port)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (##write-char #\newline port))

     (define (force-output port prim arg1 arg2);;;;;;;;;;;;force-output should block!!!!

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (macro-port-mutex-lock! port) ; get exclusive access to port

       (let ((code (force-output-aux port)))
         (macro-port-mutex-unlock! port)
         (if (##fixnum.< code 0)
           (if (##fixnum.= code ##err-code-EAGAIN)
             #f
             (##raise-os-exception #f code prim arg1 arg2))
           (##void))))

     (define (force-output-aux port)

       ; It is assumed that the thread has exclusive access to the port.

       (##declare (not interrupts-enabled))

       (let ((code (drain-output port)))
         (if (##fixnum? code)
           code
           (let* ((wdevice-condvar (macro-port-wdevice-condvar port))
                  (wdevice (macro-condvar-name wdevice-condvar)))
             (##dev-force-output wdevice)))))

     (define (drain-output port)

       ; It is assumed that the thread has exclusive access to the port.

       (##declare (not interrupts-enabled))

       (let ((code ((macro-port-char-wbuf-drain port) port)))
         (if (##fixnum? code)
           code
           ((macro-port-byte-wbuf-drain port) port))))

     (define (close port prim arg1)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (macro-port-mutex-lock! port) ; get exclusive access to port

       (let ((result (close-aux1 port prim)))
         (macro-port-mutex-unlock! port)
         (if (##fixnum? result)
           (##raise-os-exception #f result prim arg1)
           result)))

     (define (close-aux1 port prim)

       ; It is assumed that the thread has exclusive access to the port.

       (##declare (not interrupts-enabled))

       (if (or (##fixnum.= (macro-port-wkind port) (macro-none-kind))
               (##eq? prim close-input-port))
         (close-aux2 port prim)
         (let ((code (drain-output port)))
           (if (and (##fixnum? code)
                    (##not (##fixnum.= code ##err-code-EAGAIN)))

             code

             ; The close operation may have failed to force the output.
             ; However the close operation is not allowed to block, so
             ; we just continue and close the device.  The user can make
             ; sure that the output is forced by calling force-output
             ; (which can block) before calling close-port.

             (close-aux2 port prim)))))

     (define (close-aux2 port prim)

       ; It is assumed that the thread has exclusive access to the port.

       (##declare (not interrupts-enabled))

       (##close-device
        port
        (macro-port-rdevice-condvar port)
        (macro-port-wdevice-condvar port)
        prim))

     (define (set-rtimeout port timeout thunk)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (macro-port-mutex-lock! port) ; get exclusive access to port

       (macro-port-rtimeout-set! port timeout)
       (macro-port-rtimeout-thunk-set! port thunk)
       (##condvar-signal-no-reschedule!
        (macro-port-rdevice-condvar port)
        #t)
       (macro-port-mutex-unlock! port)
       (##void))

     (define (set-wtimeout port timeout thunk)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (macro-port-mutex-lock! port) ; get exclusive access to port

       (macro-port-wtimeout-set! port timeout)
       (macro-port-wtimeout-thunk-set! port thunk)
       (##condvar-signal-no-reschedule!
        (macro-port-wdevice-condvar port)
        #t)
       (macro-port-mutex-unlock! port)
       (##void))

     (define (output-width port)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (macro-port-mutex-lock! port) ; get exclusive access to port

       (let* ((wdevice-condvar (macro-port-wdevice-condvar port))
              (wdevice (macro-condvar-name wdevice-condvar))
              (result (##dev-stream-width wdevice)))
         (macro-port-mutex-unlock! port)
         (if (##fixnum.< result 0)
           (##raise-os-exception #f result output-port-width port)
           result)))

     (macro-make-device-port
      mutex
      rkind
      wkind
      name
      read-datum
      write-datum
      newline
      force-output
      close
      roptions
      rtimeout
      rtimeout-thunk
      set-rtimeout
      woptions
      wtimeout
      wtimeout-thunk
      set-wtimeout
      char-rbuf
      char-rlo
      char-rhi
      char-rchars
      char-rlines
      char-rcurline
      char-rbuf-fill
      char-peek-eof?
      char-wbuf
      char-wlo
      char-whi
      char-wchars
      char-wlines
      char-wcurline
      char-wbuf-drain
      output-width
      byte-rbuf
      byte-rlo
      byte-rhi
      byte-rbuf-fill
      byte-wbuf
      byte-wlo
      byte-whi
      byte-wbuf-drain
      rdevice-condvar
      wdevice-condvar
      device-name)))

(define-prim (##make-rdevice-condvar rdevice)
  (let ((cv (##make-condvar rdevice)))
    (macro-btq-owner-set! cv 0) ; mark as read device
    cv))

(define-prim (##make-wdevice-condvar wdevice)
  (let ((cv (##make-condvar wdevice)))
    (macro-btq-owner-set! cv 2) ; mark as write device
    cv))

(define-prim (##make-device-port-from-single-device
         device-name
         device
         psettings)
  (let ((direction (macro-psettings-direction psettings)))
    (cond ((##fixnum.= direction (macro-direction-in))
           (##make-device-port device-name
                               device
                               #f
                               psettings))
          ((##fixnum.= direction (macro-direction-out))
           (##make-device-port device-name
                               #f
                               device
                               psettings))
          (else
           (##make-device-port device-name
                               device
                               device
                               psettings)))))

(define-prim (##close-device port rdevice-condvar wdevice-condvar prim)

  (##declare (not interrupts-enabled))

  (let ((rdevice
         (if (##fixnum.= (macro-port-rkind port) (macro-none-kind))
           #f
           (macro-condvar-name rdevice-condvar)))
        (wdevice
         (if (##fixnum.= (macro-port-wkind port) (macro-none-kind))
           #f
           (macro-condvar-name wdevice-condvar))))
    (if (and (##eq? rdevice wdevice)
             (##eq? prim close-port))
      (let ((code1
             (##dev-close rdevice (macro-direction-inout))))
        (if (##fixnum.< code1 0)
          code1
          (##void)))
      (let ((code2
             (if (and rdevice
                      (##not (##eq? prim close-output-port)))
               (##dev-close rdevice (macro-direction-in))
               0)))
        (if (##fixnum.< code2 0)
          code2
          (let ((code3
                 (if (and wdevice
                          (##not (##eq? prim close-input-port)))
                   (##dev-close wdevice (macro-direction-out))
                   0)))
            (if (##fixnum.< code3 0)
              code3
              (##void))))))))

(define-prim (##char-rbuf-fill port block?)

  ; This procedure returns one of the following values:
  ;  - #t if characters were added to the char buffer,
  ;  - #f if no character could be added to the char buffer (because
  ;    end-of-file was reached),
  ;  - fixnum indicating an error code (in particular, only if block?
  ;    is false or there was a read timeout and the timeout thunk
  ;    returned #f, ##err-code-EAGAIN is returned to indicate that no
  ;    character was currently available).

  ; It is assumed that the thread has exclusive access to the port.

  (##declare (not interrupts-enabled))

  (let loop ()

    ; convert bytes from the byte buffer into characters in the char buffer

    (let ((code1 (##port-decode-chars! port #f)))

      (cond ((##fixnum? code1)

             ; an error occurred, return the error code to caller

             code1)

            (code1

             ; characters were added to char buffer

             #t)

            (else

             ; no characters were added to char buffer, so try to get
             ; some more bytes

             (let ((code2 ((macro-port-byte-rbuf-fill port) port block?)))

               (cond ((##fixnum? code2)

                      ; an error occurred, return the error code to caller

                      code2)

                     (code2

                      ; bytes were added to byte buffer, so try again
                      ; to extract characters from the byte buffer

                      (loop))

                     (else

                      ; no bytes were added to byte buffer
                      ; (end-of-file was reached)

                      ; The following call to ##port-decode-chars! will
                      ; check that the byte buffer is empty.  If the
                      ; buffer is not empty an error code is returned
                      ; to indicate that the remaining bytes can't
                      ; form a character, otherwise #f is returned.

                      (##port-decode-chars! port #t)))))))))

(define-prim (##byte-rbuf-fill port block?)

  ; This procedure returns one of the following values:
  ;  - #t if bytes were added to the byte buffer,
  ;  - #f if no byte could be added to the byte buffer (because
  ;    end-of-file was reached),
  ;  - fixnum indicating an error code (in particular, only if block?
  ;    is false or there was a read timeout and the timeout thunk
  ;    returned #f, ##err-code-EAGAIN is returned to indicate that no
  ;    byte was currently available).

  ; It is assumed that the thread has exclusive access to the port.

  (##declare (not interrupts-enabled))

  (let loop ()

    ; read into byte buffer from rhi to end of buffer

    (let* ((byte-rbuf
            (macro-port-byte-rbuf port))
           (byte-rhi
            (macro-port-byte-rhi port))
           (n
            (##dev-stream-read
             (macro-condvar-name (macro-port-rdevice-condvar port))
             byte-rbuf
             byte-rhi
             (##u8vector-length byte-rbuf))))

      (if (##fixnum.< n 0)

        ; the read caused an error

        (cond ((##fixnum.= n ##err-code-EAGAIN)

               ; the read would block, so if that is an option, wait
               ; and then try again otherwise return the error code to
               ; the caller

               (if block?
                 (begin
                   (macro-port-mutex-unlock! port)
                   (let ((continue?
                          (or (##device-condvar-wait!
                               (macro-port-rdevice-condvar port)
                               (macro-port-rtimeout port))
                              ((macro-port-rtimeout-thunk port)))))
                     (macro-port-mutex-lock! port) ; regain access to port
                     (if continue?
                       (loop)
                       n)))
                 n))

              ((##fixnum.= n ##err-code-EINTR)

               ; the read was interrupted, so try again

               (loop))

              (else

               ; return the error code to the caller

               n))

        ; the read completed successfully

        (if (##fixnum.= n 0) ; was end-of-file reached?
          #f
          (begin
            (macro-port-byte-rhi-set! port
              (##fixnum.+ (macro-port-byte-rhi port) n))
            #t))))))

(define-prim (##char-wbuf-drain port)

  ; This procedure returns #f when the char buffer was successfully
  ; drained or it returns an error code (fixnum).  In particular, only
  ; if there was a write timeout and the timeout thunk returned #f,
  ; ##err-code-EAGAIN is returned to indicate that no char could be
  ; written at this time.

  ; It is assumed that the thread has exclusive access to the port.

  (##declare (not interrupts-enabled))

  (let loop ()

    ; convert characters from char buffer into bytes in the byte buffer

    (let ((code1 (##port-encode-chars! port)))

      (cond ((##fixnum? code1)

             ; an error occurred, return the error code to caller

             code1)

            (code1

             ; the byte buffer is full, so drain it and continue
             ; draining char buffer

             (let ((code2 ((macro-port-byte-wbuf-drain port) port)))

               (if (##fixnum? code2)

                 ; an error occurred, return the error code to caller

                 code2

                 ; the byte buffer was successfully drained, continue
                 ; draining char buffer

                 (loop))))

            (else

             ; the byte buffer is not full and the char buffer is
             ; empty

             #f)))))

(define-prim (##byte-wbuf-drain port)

  ; This procedure returns #f when the byte buffer was successfully
  ; drained or it returns an error code (fixnum).  In particular, only
  ; if there was a write timeout and the timeout thunk returned #f,
  ; ##err-code-EAGAIN is returned to indicate that no byte could be
  ; written at this time.

  ; It is assumed that the thread has exclusive access to the port.

  (##declare (not interrupts-enabled))

  (let loop ()

    (let ((byte-wlo (macro-port-byte-wlo port))
          (byte-whi (macro-port-byte-whi port)))
      (if (##fixnum.< byte-wlo byte-whi)

        ; the byte buffer is not empty, write content of byte buffer
        ; from wlo to whi

        (let ((n
               (##dev-stream-write
                (macro-condvar-name (macro-port-wdevice-condvar port))
                (macro-port-byte-wbuf port)
                byte-wlo
                byte-whi)))

          (if (##fixnum.< n 0)

            ; the write caused an error

            (cond ((##fixnum.= n ##err-code-EAGAIN)

                   ; the write would block, so wait and then try again

                   (macro-port-mutex-unlock! port)
                   (let ((continue?
                          (or (##device-condvar-wait!
                               (macro-port-wdevice-condvar port)
                               (macro-port-wtimeout port))
                              ((macro-port-wtimeout-thunk port)))))
                     (macro-port-mutex-lock! port) ; regain access to port
                     (if continue?
                       (loop)
                       n)))

                  ((##fixnum.= n ##err-code-EINTR)

                   ; the write was interrupted, so try again

                   (loop))

                  (else

                   ; return the error code to the caller

                   n))

            ; some bytes (possibly zero) were written, advance
            ; wlo and try to write more

            (begin
              (macro-port-byte-wlo-set! port
                (##fixnum.+ (macro-port-byte-wlo port) n))
              (loop))))

        ; the byte buffer is empty, reset wlo and whi

        (begin
          (macro-port-byte-wlo-set! port 0)
          (macro-port-byte-whi-set! port 0)
          #f)))))

;------------------------------------------------------------------------------

; Implementation of vector, string and u8vector ports.

(##define-macro (define-prim-vector-port-procedures
                  name
                  empty-vect
                  vect-zap!
                  drain-output
                  allowed-settings)

  (define (sym . lst)
    (string->symbol (apply string-append (map symbol->string lst))))

  (let ()

    (define vect-input-port
      (sym name '-input-port))

    (define vect-output-port
      (sym name '-output-port))

    (define macro-check-vect-output-port
      (sym 'macro-check- name '-output-port))

    (define ##fail-check-vect    (sym '##fail-check- name))
    (define ##make-vect          (sym '##make- name))
    (define ##vect?              (sym '## name '?))
    (define ##vect-ref           (sym '## name '-ref))
    (define ##vect-set!          (sym '## name '-set!))
    (define ##vect-length        (sym '## name '-length))
    (define ##subvect            (sym '##sub name))
    (define ##subvect-move!      (sym '##sub name '-move!))
    (define ##subvect->fifo      (sym '##sub name '->fifo))
    (define ##fifo->vect         (sym '##fifo-> name))
    (define ##open-vect-and-call (sym '##open- name '-and-call))
    (define ##open-input-vect    (sym '##open-input- name))
    (define ##open-output-vect   (sym '##open-output- name))
    (define ##open-vect          (sym '##open- name))
    (define ##make-vect-port     (sym '##make- name '-port))
    (define ##get-output-vect    (sym '##get-output- name))

    (define open-vect            (sym 'open- name))
    (define open-input-vect      (sym 'open-input- name))
    (define open-output-vect     (sym 'open-output- name))
    (define get-output-vect      (sym 'get-output- name))

    (define call-with-input-vect  (sym 'call-with-input- name))
    (define call-with-output-vect (sym 'call-with-output- name))
    (define with-input-from-vect  (sym 'with-input-from- name))
    (define with-output-to-vect   (sym 'with-output-to- name))

    (define define-vect-port-methods
      (sym 'define- name '-port-methods))

    (define macro-port-vect-rbuf
      (sym 'macro-port- name '-rbuf))
    (define macro-port-vect-rbuf-set!
      (sym 'macro-port- name '-rbuf-set!))
    (define macro-port-vect-rlo
      (sym 'macro-port- name '-rlo))
    (define macro-port-vect-rlo-set!
      (sym 'macro-port- name '-rlo-set!))
    (define macro-port-vect-rhi
      (sym 'macro-port- name '-rhi))
    (define macro-port-vect-rhi-set!
      (sym 'macro-port- name '-rhi-set!))
    (define macro-port-vect-rbuf-fill
      (sym 'macro-port- name '-rbuf-fill))
    (define macro-port-vect-rbuf-fill-set!
      (sym 'macro-port- name '-rbuf-fill-set!))
    (define macro-port-vect-wbuf
      (sym 'macro-port- name '-wbuf))
    (define macro-port-vect-wbuf-set!
      (sym 'macro-port- name '-wbuf-set!))
    (define macro-port-vect-wlo
      (sym 'macro-port- name '-wlo))
    (define macro-port-vect-wlo-set!
      (sym 'macro-port- name '-wlo-set!))
    (define macro-port-vect-whi
      (sym 'macro-port- name '-whi))
    (define macro-port-vect-whi-set!
      (sym 'macro-port- name '-whi-set!))
    (define macro-port-vect-wbuf-drain
      (sym 'macro-port- name '-wbuf-drain))
    (define macro-port-vect-wbuf-drain-set!
      (sym 'macro-port- name '-wbuf-drain-set!))
    (define macro-port-vect-fifo
      (sym 'macro-port- name '-fifo))
    (define macro-port-vect-fifo-set!
      (sym 'macro-port- name '-fifo-set!))
    (define macro-port-vect-rcondvar
      (sym 'macro-port- name '-rcondvar))
    (define macro-port-vect-rcondvar-set!
      (sym 'macro-port- name '-rcondvar-set!))
    (define macro-port-vect-wcondvar
      (sym 'macro-port- name '-wcondvar))
    (define macro-port-vect-wcondvar-set!
      (sym 'macro-port- name '-wcondvar-set!))
    (define macro-port-vect-buffering-limit
      (sym 'macro-port- name '-buffering-limit))
    (define macro-port-vect-buffering-limit-set!
      (sym 'macro-port- name '-buffering-limit-set!))

    (define vect-rbuf-fill
      (sym name '-rbuf-fill))

    (define vect-wbuf-drain
      (sym name '-wbuf-drain))

    `(begin

       (define-fail-check-type ,vect-input-port)
       (define-fail-check-type ,vect-output-port)

       (##define-macro (,define-vect-port-methods)
         `(begin

            (define (,',vect-rbuf-fill port block?)

              ; This procedure returns one of the following values:
              ;  - #t if something was added to the read buffer,
              ;  - #f if nothing could be added to the read buffer
              ;    (because end-of-file was reached),
              ;  - fixnum indicating an error code (in particular,
              ;    only if block? is false or there was a read timeout
              ;    and the timeout thunk returned #f, ##err-code-EAGAIN
              ;    is returned to indicate that nothing is currently
              ;    available to be read).

              ; It is assumed that the thread has exclusive access to the port.

              (##declare (not interrupts-enabled))

              (let loop ()
                (let ((vect-rbuf (,',macro-port-vect-rbuf port))
                      (vect-wbuf (,',macro-port-vect-wbuf port)))
                  (if (##not (##eq? vect-rbuf vect-wbuf))
                    (let ((vect-rhi (,',macro-port-vect-rhi port))
                          (len (,',##vect-length vect-rbuf)))
                      (cond ((##fixnum.< vect-rhi len)
                             (,',macro-port-vect-rhi-set! port len)
                             #t)
                            (else
                             (let ((new-vect-rbuf
                                    (macro-fifo-advance!
                                     (,',macro-port-vect-fifo port))))
                               (,',macro-port-vect-wlo-set!
                                port
                                (##fixnum.- (,',macro-port-vect-wlo port) len))
                               (,',macro-port-vect-rbuf-set!
                                port
                                new-vect-rbuf)
                               (,',macro-port-vect-rlo-set! port 0)
                               (,',macro-port-vect-rhi-set! port 0)
                               (##condvar-signal-no-reschedule!
                                (,',macro-port-vect-wcondvar port)
                                #t)
                               (loop)))))
                    (let ((vect-rhi (,',macro-port-vect-rhi port))
                          (vect-whi (,',macro-port-vect-whi port)))
                      (cond ((##fixnum.< vect-rhi vect-whi)
                             (,',macro-port-vect-rhi-set! port vect-whi)
                             #t)
                            ((macro-closed? (macro-port-woptions port))
                             #f)
                            (block?
                             (let ((continue?
                                    (or (##mutex-signal-and-condvar-wait!
                                         (macro-port-mutex port)
                                         (,',macro-port-vect-rcondvar port)
                                         (macro-port-rtimeout port))
                                        ((macro-port-rtimeout-thunk port)))))
                               (macro-port-mutex-lock! port)
                               (if continue?
                                 (loop)
                                 ##err-code-EAGAIN)))
                            (else
                             ##err-code-EAGAIN)))))))

            (define (,',vect-wbuf-drain port)

              ; This procedure returns #f when the write buffer was
              ; successfully drained or it returns an error code
              ; (fixnum).  In particular, only if there was a write
              ; timeout and the timeout thunk returned #f,
              ; ##err-code-EAGAIN is returned to indicate that nothing
              ; could be written at this time.

              ; It is assumed that the thread has exclusive access to the port.

              (##declare (not interrupts-enabled))

              (let loop ()
                (let ((buffering-limit
                       (,',macro-port-vect-buffering-limit port)))
                  (if (and buffering-limit
                           (let ((unread
                                  (##fixnum.- (,',macro-port-vect-wlo port)
                                              (,',macro-port-vect-rlo port))))
                             (##fixnum.< buffering-limit unread)))
                    (let ((continue?
                           (or (##mutex-signal-and-condvar-wait!
                                (macro-port-mutex port)
                                (,',macro-port-vect-wcondvar port)
                                (macro-port-wtimeout port))
                               ((macro-port-wtimeout-thunk port)))))
                      (macro-port-mutex-lock! port)
                      (if continue?
                        (loop)
                        ##err-code-EAGAIN))
                    (let ((new-vect-wbuf (,',##make-vect chunk-size)))
                      (,',macro-port-vect-wlo-set!
                       port
                       (##fixnum.+
                        (,',macro-port-vect-wlo port)
                        (,',##vect-length (,',macro-port-vect-wbuf port))))
                      (,',macro-port-vect-whi-set! port 0)
                      (,',macro-port-vect-wbuf-set! port new-vect-wbuf)
                      (macro-fifo-insert-at-tail!
                       (,',macro-port-vect-fifo port)
                       new-vect-wbuf)
                      (##condvar-signal-no-reschedule!
                       (,',macro-port-vect-rcondvar port)
                       #t)
                      #f)))))

            (define (name port)

              ; It is assumed that the thread **does not** have exclusive
              ; access to the port.

              (##declare (not interrupts-enabled))

              '(,',name))

            (define (force-output port prim arg1 arg2)

              ; It is assumed that the thread **does not** have exclusive
              ; access to the port.

              (##declare (not interrupts-enabled))

              (macro-port-mutex-lock! port) ; get exclusive access to port

              (##condvar-signal-no-reschedule!
               (,',macro-port-vect-rcondvar port)
               #t)

              ,',(if drain-output
                   `(let ((code (,drain-output port)))
                      (macro-port-mutex-unlock! port)
                      (if (##fixnum? code)
                        (if (##fixnum.= code ##err-code-EAGAIN)
                          #f
                          (##raise-os-exception #f code prim arg1 arg2))
                        (##void)))
                   `(begin
                      (macro-port-mutex-unlock! port)
                      (##void))))

            (define (close port prim arg1)

              ; It is assumed that the thread **does not** have exclusive
              ; access to the port.

              (##declare (not interrupts-enabled))

              ((macro-port-force-output port)
               port
               prim
               arg1
               (macro-absent-obj))

              (macro-port-mutex-lock! port) ; get exclusive access to port

              (if (##not (##eq? prim close-output-port))
                (begin
                  (macro-port-roptions-set!
                   port
                   (macro-close! (macro-port-roptions port)))
                  (##condvar-signal-no-reschedule!
                   (,',macro-port-vect-wcondvar port)
                   #t)))

              (if (##not (##eq? prim close-input-port))
                (begin
                  (macro-port-woptions-set!
                   port
                   (macro-close! (macro-port-woptions port)))
                  (##condvar-signal-no-reschedule!
                   (,',macro-port-vect-rcondvar port)
                   #t)))

              (##void))

            (define (set-rtimeout port timeout thunk)

              ; It is assumed that the thread **does not** have exclusive
              ; access to the port.

              (##declare (not interrupts-enabled))

              (macro-port-mutex-lock! port) ; get exclusive access to port

              (macro-port-rtimeout-set! port timeout)
              (macro-port-rtimeout-thunk-set! port thunk)
              (##condvar-signal-no-reschedule!
               (,',macro-port-vect-rcondvar port)
               #t)
              (macro-port-mutex-unlock! port)
              (##void))

            (define (set-wtimeout port timeout thunk)

              ; It is assumed that the thread **does not** have exclusive
              ; access to the port.

              (##declare (not interrupts-enabled))

              (macro-port-mutex-lock! port) ; get exclusive access to port

              (macro-port-wtimeout-set! port timeout)
              (macro-port-wtimeout-thunk-set! port thunk)
              (##condvar-signal-no-reschedule!
               (,',macro-port-vect-wcondvar port)
               #t)
              (macro-port-mutex-unlock! port)
              (##void))))

       (define-prim (,##subvect->fifo vect start end chunk-size)
         (let ((fifo (macro-make-fifo)))
           (let loop ((lo start))
             (let ((hi (##fixnum.+ lo chunk-size)))
               (if (##fixnum.< hi end)
                 (begin
                   (macro-fifo-insert-at-tail! fifo (,##subvect vect lo hi))
                   (loop hi))
                 (begin
                   (macro-fifo-insert-at-tail! fifo (,##subvect vect lo end))
                   fifo))))))

       (define-prim (,##fifo->vect fifo start end)
         (let* ((len (##fixnum.max (##fixnum.- end start) 0))
                (vect (,##make-vect len)))
           (let loop ((elems (macro-fifo-next fifo))
                      (hi end)
                      (lo start)
                      (i 0))
             (if (##fixnum.< lo hi)
               (let* ((chunk
                       (macro-fifo-elem elems))
                      (chunk-len
                       (,##vect-length chunk))
                      (n
                       (##fixnum.min (##fixnum.- chunk-len lo)
                                     (##fixnum.- hi lo))))
                 (,##subvect-move! chunk lo (##fixnum.+ lo n) vect i)
                 (loop (macro-fifo-next elems)
                       (##fixnum.- hi chunk-len)
                       (##fixnum.- (##fixnum.+ lo n) chunk-len)
                       (##fixnum.+ i n)))
               vect))))

       (define-prim (,##open-vect-and-call
                     direction
                     prim
                     receiver
                     v
                     settings
                     arg)
         (let ((vect
                (if (##eq? v (macro-absent-obj))
                  ',empty-vect
                  v)))
           (if (##not (,##vect? vect))
             (,##fail-check-vect 1 prim v settings arg)
             (let ((psettings
                    (##make-psettings
                     direction
                     ',allowed-settings
                     settings)))
               (if (##not psettings)
                 (##trap-psettings prim v settings arg)
                 (receiver
                  (,##make-vect-port vect 0 (,##vect-length vect) psettings)
                  arg))))))

       (define-prim (,##open-vect
                     #!optional
                     (vect (macro-absent-obj))
                     (settings (macro-absent-obj)))
         (,##open-vect-and-call
          (macro-direction-inout)
          ,open-vect
          (lambda (port dummy) port)
          vect
          settings
          (macro-absent-obj)))

       (define-prim (,open-vect
                     #!optional
                     (vect (macro-absent-obj))
                     (settings (macro-absent-obj)))
         (macro-force-vars (vect settings)
           (,##open-vect vect settings)))

       (define-prim (,##open-input-vect
                     vect
                     #!optional
                     (settings (macro-absent-obj)))
         (,##open-vect-and-call
          (macro-direction-in)
          ,open-input-vect
          (lambda (port dummy) port)
          vect
          settings
          (macro-absent-obj)))

       (define-prim (,open-input-vect
                     vect
                     #!optional
                     (settings (macro-absent-obj)))
         (macro-force-vars (vect settings)
           (,##open-input-vect vect settings)))

       (define-prim (,##open-output-vect
                     #!optional
                     (settings (macro-absent-obj)))
         (,##open-vect-and-call
          (macro-direction-out)
          ,open-output-vect
          (lambda (port dummy) port)
          (macro-absent-obj)
          settings
          (macro-absent-obj)))

       (define-prim (,open-output-vect
                     #!optional
                     (settings (macro-absent-obj)))
         (macro-force-vars (settings)
           (,##open-output-vect settings)))

       (define-prim (,##get-output-vect port)

         (##declare (not interrupts-enabled))

         ((macro-port-force-output port)
          port
          ,get-output-vect
          port
          (macro-absent-obj))

         (macro-port-mutex-lock! port) ; get exclusive access to port

         (let* ((vect-fifo
                 (,macro-port-vect-fifo port))
                (result
                 (,##fifo->vect
                  vect-fifo
                  (,macro-port-vect-rlo port)
                  (##fixnum.+ (,macro-port-vect-wlo port)
                              (,macro-port-vect-whi port))))
                (new-vect-buf
                 (macro-fifo-advance-to-tail! vect-fifo)))

           ; zap the entries of the buffer to avoid leaks

           ,(if vect-zap!
              `(let loop ((i
                           (if (##eq?
                                (,macro-port-vect-rbuf port)
                                new-vect-buf)
                             (,macro-port-vect-rlo port)
                             0)))
                 (if (##fixnum.< i (,macro-port-vect-whi port))
                   (begin
                     (,vect-zap! new-vect-buf i)
                     (loop (##fixnum.+ i 1)))))
              #f)

           (,macro-port-vect-rbuf-set! port new-vect-buf)
           (,macro-port-vect-rlo-set! port 0)
           (,macro-port-vect-rhi-set! port 0)

           (,macro-port-vect-wbuf-set! port new-vect-buf)
           (,macro-port-vect-wlo-set! port 0)
           (,macro-port-vect-whi-set! port 0)

           (macro-port-mutex-unlock! port)

           result))

       (define-prim (,get-output-vect port)
         (macro-force-vars (port)
           (,macro-check-vect-output-port
            port
            1
            (,get-output-vect port)
            (,##get-output-vect port))))

       (define-prim (,call-with-input-vect
                     vect
                     arg2
                     #!optional
                     (arg3 (macro-absent-obj)))

         (define (call settings proc arg-num)
           (macro-check-procedure
            proc
            arg-num
            (,call-with-input-vect vect arg2 arg3)
            (,##open-vect-and-call
             (macro-direction-in)
             ,call-with-input-vect
             (lambda (port proc) (proc port))
             vect
             settings
             proc)))

         (macro-force-vars (vect arg2 arg3)
           (cond ((##eq? arg3 (macro-absent-obj))
                  (call arg3 arg2 2))
                 (else
                  (call arg2 arg3 3)))))

       (define-prim (,call-with-output-vect
                     arg1
                     #!optional
                     (arg2 (macro-absent-obj)))

         (define (call settings proc arg-num)
           (macro-check-procedure
            proc
            arg-num
            (,call-with-output-vect arg1 arg2)
            (,##open-vect-and-call
             (macro-direction-out)
             ,call-with-output-vect
             (lambda (port proc) (proc port) (,##get-output-vect port))
             (macro-absent-obj)
             settings
             proc)))

         (macro-force-vars (arg1 arg2)
           (cond ((##eq? arg2 (macro-absent-obj))
                  (call arg2 arg1 1))
                 (else
                  (call arg1 arg2 2)))))

       (define-prim (,with-input-from-vect
                     vect
                     arg2
                     #!optional
                     (arg3 (macro-absent-obj)))

         (define (call settings thunk arg-num)
           (macro-check-procedure
            thunk
            arg-num
            (,with-input-from-vect vect arg2 arg3)
            (,##open-vect-and-call
             (macro-direction-in)
             ,with-input-from-vect
             (lambda (port thunk)
               (macro-dynamic-bind input-port port thunk))
             vect
             settings
             thunk)))

         (macro-force-vars (vect arg2 arg3)
           (cond ((##eq? arg3 (macro-absent-obj))
                  (call arg3 arg2 2))
                 (else
                  (call arg2 arg3 3)))))

       (define-prim (,with-output-to-vect
                     arg1
                     #!optional
                     (arg2 (macro-absent-obj)))

         (define (call settings thunk arg-num)
           (macro-check-procedure
            thunk
            arg-num
            (,with-output-to-vect arg1 arg2)
            (,##open-vect-and-call
             (macro-direction-out)
             ,with-output-to-vect
             (lambda (port thunk)
               (let ((results ; may get bound to a multiple-values object
                      (macro-dynamic-bind output-port port thunk)))
                 (,##get-output-vect port)))
             (macro-absent-obj)
             settings
             thunk)))

         (macro-force-vars (arg1 arg2)
           (cond ((##eq? arg2 (macro-absent-obj))
                  (call arg2 arg1 1))
                 (else
                  (call arg1 arg2 2))))))))

(define-prim (##vect-port-options options kind buffering)
  (##psettings-options->options
   options
   (##fixnum.+
    (##fixnum.* (macro-open-state-shift)
                (if (##fixnum.= kind (macro-none-kind))
                  (macro-open-state-closed)
                  (macro-open-state-open)))
    (##fixnum.* (macro-buffering-shift)
                buffering))))

;------------------------------------------------------------------------------

; Implementation of vector ports.

(define-prim-vector-port-procedures
  vector
  #()
  (lambda (vect i) (##vector-set! vect i #f))
  #f
  (input-buffering:
   output-buffering:
   buffering:
   direction:))

(define-prim (##make-vector-port src start end psettings)

  (define chunk-size 16)

  (let* ((direction
          (macro-psettings-direction psettings))
         (len
          (##fixnum.max (##fixnum.- end start) 0))
         (vector-fifo
          (##subvector->fifo src start end chunk-size))
         (mutex
          (macro-make-port-mutex))
         (rkind
          (if (##fixnum.= direction (macro-direction-out))
            (macro-none-kind)
            (macro-vector-kind)))
         (wkind
          (if (##fixnum.= direction (macro-direction-in))
            (macro-none-kind)
            (macro-vector-kind)))
         (roptions
          (##vect-port-options
           (macro-psettings-roptions psettings)
           rkind
           (macro-buffering)))
         (rtimeout
          #t)
         (rtimeout-thunk
          #f)
         (woptions
          (##vect-port-options
           (macro-psettings-woptions psettings)
           wkind
           (macro-buffering)))
         (wtimeout
          #t)
         (wtimeout-thunk
          #f)
         (vector-rbuf
          (macro-fifo-elem (macro-fifo-next vector-fifo)))
         (vector-rlo
          0)
         (vector-rhi
          (##vector-length vector-rbuf))
         (vector-wbuf
          (macro-fifo-elem (macro-fifo-tail vector-fifo)))
         (vector-whi
          (##vector-length vector-wbuf))
         (vector-wlo
          (##fixnum.- len vector-whi))
         (vector-rcondvar
          (##make-condvar #f))
         (vector-wcondvar
          (##make-condvar #f))
         (vector-buffering-limit
          #f))

     (define (read-datum port re)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (macro-port-mutex-lock! port) ; get exclusive access to port

       (let loop ()

         (let ((vector-rlo (macro-port-vector-rlo port))
               (vector-rhi (macro-port-vector-rhi port)))
           (if (##fixnum.< vector-rlo vector-rhi)

             ; the next object is in the object read buffer

             (let* ((vector-rbuf
                     (macro-port-vector-rbuf port))
                    (obj
                     (##vector-ref vector-rbuf vector-rlo)))

               ; frequent simple case, just advance rlo and zap vector
               ; to avoid retaining objects uselessly

               (##vector-set! vector-rbuf vector-rlo #f)
               (macro-port-vector-rlo-set! port (##fixnum.+ vector-rlo 1))
               (macro-port-mutex-unlock! port)
               obj)

             ; try to get more objects into the object read
             ; buffer, and try again if successful otherwise
             ; signal an error or return end-of-file object

             (let ((code ((macro-port-vector-rbuf-fill port) port #t)))

               (cond ((##fixnum? code)

                      ; the conversion or read caused an error

                      (macro-port-mutex-unlock! port)
                      (if (##fixnum.= code ##err-code-EAGAIN)
                        #f ; the read timeout thunk returned #f
                        (##raise-os-exception #f code read port)))

                     (code

                      ; some objects were added to object buffer

                      (loop))

                     (else

                      ; no objects were added to object buffer

                      (macro-port-mutex-unlock! port)
                      #!eof)))))))

     (define (write-datum port obj we)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (macro-port-mutex-lock! port) ; get exclusive access to port

       (let loop ()

         (let ((vector-wbuf (macro-port-vector-wbuf port))
               (vector-whi+1 (##fixnum.+ (macro-port-vector-whi port) 1)))
           (if (##not (##fixnum.< (##vector-length vector-wbuf) vector-whi+1))

             ; there is enough space in the object write buffer, so add
             ; object and increment whi

             (let ()

               (##vector-set! vector-wbuf (##fixnum.- vector-whi+1 1) obj)

               ; advance whi

               (macro-port-vector-whi-set! port vector-whi+1)

               ; force output if port is set for unbuffered output

               (if (macro-unbuffered? (macro-port-woptions port))
                 (begin
                   (macro-port-mutex-unlock! port)
                   ((macro-port-force-output port)
                    port
                    write
                    obj
                    port))
                 (begin
                   (macro-port-mutex-unlock! port)
                   (##void))))

             ; make some space in the object buffer and try again

             (let ((code ((macro-port-vector-wbuf-drain port) port)))
               (if (##fixnum? code)
                 (begin
                   (macro-port-mutex-unlock! port)
                   (if (##fixnum.= code ##err-code-EAGAIN)
                     #f
                     (##raise-os-exception #f code write obj port)))
                 (loop)))))))

     (define (newline port)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (##void))

     (define-vector-port-methods)

     (macro-make-object-port
      mutex
      rkind
      wkind
      name
      read-datum
      write-datum
      newline
      force-output
      close
      roptions
      rtimeout
      rtimeout-thunk
      set-rtimeout
      woptions
      wtimeout
      wtimeout-thunk
      set-wtimeout
      vector-rbuf
      vector-rlo
      vector-rhi
      vector-rbuf-fill
      vector-wbuf
      vector-wlo
      vector-whi
      vector-wbuf-drain
      vector-fifo
      vector-rcondvar
      vector-wcondvar
      vector-buffering-limit)))

;------------------------------------------------------------------------------

; Implementation of string ports.

(define-prim-vector-port-procedures
  string
  ""
  #f
  #f
  (input-buffering:
   output-buffering:
   buffering:
   direction:
   output-width:))

(define-prim (##make-string-port src start end psettings)

  (define chunk-size 32)

  (let* ((direction
          (macro-psettings-direction psettings))
         (len
          (##fixnum.max (##fixnum.- end start) 0))
         (string-fifo
          (##substring->fifo src start end chunk-size))
         (mutex
          (macro-make-port-mutex))
         (rkind
          (if (##fixnum.= direction (macro-direction-out))
            (macro-none-kind)
            (macro-string-kind)))
         (wkind
          (if (##fixnum.= direction (macro-direction-in))
            (macro-none-kind)
            (macro-string-kind)))
         (roptions
          (##vect-port-options
           (macro-psettings-roptions psettings)
           rkind
           (macro-buffering)))
         (rtimeout
          #t)
         (rtimeout-thunk
          #f)
         (woptions
          (##vect-port-options
           (macro-psettings-woptions psettings)
           wkind
           (macro-buffering)))
         (wtimeout
          #t)
         (wtimeout-thunk
          #f)
         (string-rbuf
          (macro-fifo-elem (macro-fifo-next string-fifo)))
         (string-rlo
          0)
         (string-rhi
          (##string-length string-rbuf))
         (char-rchars
          0)
         (char-rlines
          0)
         (char-rcurline
          0)
         (char-peek-eof?
          #f)
         (string-wbuf
          (macro-fifo-elem (macro-fifo-tail string-fifo)))
         (string-whi
          (##string-length string-wbuf))
         (string-wlo
          (##fixnum.- len string-whi))
         (char-wchars
          0)
         (char-wlines
          0)
         (char-wcurline
          0)
         (string-rcondvar
          (##make-condvar #f))
         (string-wcondvar
          (##make-condvar #f))
         (string-width
          (##psettings->output-width psettings))
         (string-buffering-limit
          #f))

     (define (read-datum port re)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (##read-datum-or-eof re))

     (define (write-datum port obj we)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (##wr we obj))

     (define (newline port)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (##write-char #\newline port))

     (define (output-width port)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (macro-port-string-width port))

     (define-string-port-methods)

     (macro-make-character-port
      mutex
      rkind
      wkind
      name
      read-datum
      write-datum
      newline
      force-output
      close
      roptions
      rtimeout
      rtimeout-thunk
      set-rtimeout
      woptions
      wtimeout
      wtimeout-thunk
      set-wtimeout
      string-rbuf
      string-rlo
      string-rhi
      char-rchars
      char-rlines
      char-rcurline
      string-rbuf-fill
      char-peek-eof?
      string-wbuf
      string-wlo
      string-whi
      char-wchars
      char-wlines
      char-wcurline
      string-wbuf-drain
      output-width
      string-fifo
      string-rcondvar
      string-wcondvar
      string-width
      string-buffering-limit)))

;------------------------------------------------------------------------------

; Implementation of u8vector ports.

(define-prim-vector-port-procedures
  u8vector
  #u8()
  #f
  (lambda (port) ((macro-port-char-wbuf-drain port) port))
  (input-char-encoding:
   output-char-encoding:
   char-encoding:
   input-eol-encoding:
   output-eol-encoding:
   eol-encoding:
   input-buffering:
   output-buffering:
   buffering:
   direction:
   output-width:))

(define-prim (##make-u8vector-port src start end psettings)

  (define char-buf-len 16) ; character buffer length
  (define chunk-size 32)

  (let* ((direction
          (macro-psettings-direction psettings))
         (len
          (##fixnum.max (##fixnum.- end start) 0))
         (u8vector-fifo
          (##subu8vector->fifo src start end chunk-size))
         (mutex
          (macro-make-port-mutex))
         (rkind
          (if (##fixnum.= direction (macro-direction-out))
            (macro-none-kind)
            (macro-u8vector-kind)))
         (wkind
          (if (##fixnum.= direction (macro-direction-in))
            (macro-none-kind)
            (macro-u8vector-kind)))
         (roptions
          (##vect-port-options
           (macro-psettings-roptions psettings)
           rkind
           (macro-no-buffering)))
         (rtimeout
          #t)
         (rtimeout-thunk
          #f)
         (woptions
          (##vect-port-options
           (macro-psettings-woptions psettings)
           wkind
           (macro-no-buffering)))
         (wtimeout
          #t)
         (wtimeout-thunk
          #f)
         (char-rbuf
          (and (##not (##fixnum.= rkind (macro-none-kind)))
               (##make-string (if (macro-unbuffered? roptions)
                                1
                                char-buf-len))))
         (char-rlo
          0)
         (char-rhi
          0)
         (char-rchars
          0)
         (char-rlines
          0)
         (char-rcurline
          0)
         (char-rbuf-fill
          ##char-rbuf-fill)
         (char-peek-eof?
          #f)
         (char-wbuf
          (and (##not (##fixnum.= wkind (macro-none-kind)))
               (##make-string (if (macro-unbuffered? woptions)
                                1
                                char-buf-len))))
         (char-wlo
          0)
         (char-whi
          0)
         (char-wchars
          0)
         (char-wlines
          0)
         (char-wcurline
          0)
         (char-wbuf-drain
          ##char-wbuf-drain)
         (u8vector-rbuf
          (macro-fifo-elem (macro-fifo-next u8vector-fifo)))
         (u8vector-rlo
          0)
         (u8vector-rhi
          (##u8vector-length u8vector-rbuf))
         (u8vector-wbuf
          (macro-fifo-elem (macro-fifo-tail u8vector-fifo)))
         (u8vector-whi
          (##u8vector-length u8vector-wbuf))
         (u8vector-wlo
          (##fixnum.- len u8vector-whi))
         (u8vector-rcondvar
          (##make-condvar #f))
         (u8vector-wcondvar
          (##make-condvar #f))
         (u8vector-width
          (##psettings->output-width psettings))
         (u8vector-buffering-limit
          #f))

     (define (read-datum port re)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (##read-datum-or-eof re))

     (define (write-datum port obj we)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (##wr we obj))

     (define (newline port)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (##write-char #\newline port))

     (define (output-width port)

       ; It is assumed that the thread **does not** have exclusive
       ; access to the port.

       (##declare (not interrupts-enabled))

       (macro-port-u8vector-width port))

     (define-u8vector-port-methods)

     (macro-make-byte-port
      mutex
      rkind
      wkind
      name
      read-datum
      write-datum
      newline
      force-output
      close
      roptions
      rtimeout
      rtimeout-thunk
      set-rtimeout
      woptions
      wtimeout
      wtimeout-thunk
      set-wtimeout
      char-rbuf
      char-rlo
      char-rhi
      char-rchars
      char-rlines
      char-rcurline
      char-rbuf-fill
      char-peek-eof?
      char-wbuf
      char-wlo
      char-whi
      char-wchars
      char-wlines
      char-wcurline
      char-wbuf-drain
      output-width
      u8vector-rbuf
      u8vector-rlo
      u8vector-rhi
      u8vector-rbuf-fill
      u8vector-wbuf
      u8vector-wlo
      u8vector-whi
      u8vector-wbuf-drain
      u8vector-fifo
      u8vector-rcondvar
      u8vector-wcondvar
      u8vector-width
      u8vector-buffering-limit)))

;------------------------------------------------------------------------------

; Implementation of generic object port procedures.

(define-prim (##port-of-kind? obj kind)
  (##declare (not interrupts-enabled))
  (and (macro-port? obj)
       (##fixnum.= (##fixnum.bitwise-and (##port-kind obj) kind) kind)))

(define-prim (##port-kind port)
  (##declare (not interrupts-enabled))
  (let ((rkind (macro-port-rkind port)))
    (if (##fixnum.= rkind (macro-none-kind))
      (macro-port-wkind port)
      rkind)))

(define-prim (##port-device port)
  (##declare (not interrupts-enabled))
  (if (##fixnum.= (macro-port-rkind port) (macro-none-kind))
    (let ((wdevice-condvar (macro-port-wdevice-condvar port)))
      (macro-condvar-name wdevice-condvar))
    (let ((rdevice-condvar (macro-port-rdevice-condvar port)))
      (macro-condvar-name rdevice-condvar))))

(define-prim (##port-name port)
  (##declare (not interrupts-enabled))
  ((macro-port-name port) port))

(define-prim (##read port)

  (##declare (not interrupts-enabled))

  (if (macro-character-input-port? port)
    (let* ((noop
            (lambda (re x) x)) ; do not wrap datum
           (re
            (##make-readenv
             port
             (macro-current-readtable)
             noop
             noop)))
      ((macro-port-read-datum port) port re))
    ((macro-port-read-datum port) port #f)))

(define-prim (read
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-input-port)
             p)))
      (macro-check-input-port port 1 (read p)
        (##read port)))))

(define-prim (##write-generic-to-character-port style port rt force? limit obj)

  (##declare (not interrupts-enabled))

  (let* ((mt
          (and (##readtable-sharing-allowed? rt)
               (##make-marktable)))
         (width
          (##output-port-width port)))

    (if mt
      (let ((we1 (##make-writeenv 'mark port rt mt force? width 0 0 limit)))
        ((macro-port-write-datum port) port obj we1)))

    (let ((we2 (##make-writeenv style port rt mt force? width 0 0 limit)))
      ((macro-port-write-datum port) port obj we2)
      (##fixnum.- limit (macro-writeenv-limit we2)))))

(define-prim (##write obj port)
  (if (macro-character-output-port? port)
    (begin
      (##write-generic-to-character-port
       'write
       port
       (macro-current-readtable)
       (macro-if-forces #t #f)
       ##max-fixnum
       obj)
      (##void))
    ((macro-port-write-datum port) port obj #f)))

(define-prim (write
              obj
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (obj p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-output-port)
             p)))
      (macro-check-output-port port 2 (write obj p)
        (##write obj port)))))

(define-prim (##display obj port)
  (if (macro-character-output-port? port)
    (begin
      (##write-generic-to-character-port
       'display
       port
       (macro-current-readtable)
       (macro-if-forces #t #f)
       ##max-fixnum
       obj)
      (##void))
    ((macro-port-write-datum port) port obj #f)))

(define-prim (display
              obj
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (obj p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-output-port)
             p)))
      (macro-check-output-port port 2 (display obj p)
        (##display obj port)))))

(define-prim (##pretty-print obj port)
  (if (macro-character-output-port? port)
    (begin
      (##write-generic-to-character-port
       'pretty-print
       port
       (macro-current-readtable)
       (macro-if-forces #t #f)
       ##max-fixnum
       obj)
      (##newline port))
    ((macro-port-write-datum port) port obj #f)))

(define-prim (pretty-print
              obj
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (obj p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-output-port)
             p)))
      (macro-check-output-port port 2 (pretty-print obj p)
        (##pretty-print obj port)))))

(define-prim (pp
              obj
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (obj p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-output-port)
             p)))
      (macro-check-output-port port 2 (pp obj p)
        (##pretty-print
         (if (##procedure? obj)
           (##decompile obj)
           obj)
         port)))))

(define-prim (##newline port)
  (##declare (not interrupts-enabled))
  ((macro-port-newline port) port))

(define-prim (newline
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-output-port)
             p)))
      (macro-check-output-port port 1 (newline p)
        (##newline port)))))

(define-prim (##force-output port)
  (##declare (not interrupts-enabled))
  ((macro-port-force-output port)
   port
   force-output
   port
   (macro-absent-obj)))

(define-prim (force-output
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-output-port)
             p)))
      (macro-check-output-port port 1 (force-output p)
        (##force-output port)))))

(define-prim (##close-input-port port)
  (##declare (not interrupts-enabled))
  ((macro-port-close port) port close-input-port port))

(define-prim (close-input-port port)
  (macro-force-vars (port)
    (macro-check-input-port port 1 (close-input-port port)
      (##close-input-port port))))

(define-prim (##close-output-port port)
  (##declare (not interrupts-enabled))
  ((macro-port-close port) port close-output-port port))

(define-prim (close-output-port port)
  (macro-force-vars (port)
    (macro-check-output-port port 1 (close-output-port port)
      (##close-output-port port))))

(define-prim (##close-port port)
  (##declare (not interrupts-enabled))
  ((macro-port-close port) port close-port port))

(define-prim (close-port port)
  (macro-force-vars (port)
    (macro-check-port port 1 (close-port port)
      (##close-port port))))

(define-prim (##input-port-timeout-set! port absrel-timeout thunk)
  (##declare (not interrupts-enabled))
  (let ((timeout (##absrel-timeout->timeout absrel-timeout)))
    ((macro-port-set-rtimeout port) port timeout thunk)))

(define-prim (input-port-timeout-set!
              port
              absrel-timeout
              #!optional
              (t (macro-absent-obj)))
  (macro-force-vars (port absrel-timeout t)
    (let ((thunk
           (if (##eq? t (macro-absent-obj))
             (lambda () #f)
             t)))
      (macro-check-input-port
       port
       1
       (input-port-timeout-set! port absrel-timeout t)
       (macro-check-absrel-time-or-false
        absrel-timeout
        2
        (input-port-timeout-set! port absrel-timeout t)
        (macro-check-procedure
         thunk
         3
         (input-port-timeout-set! port absrel-timeout t)
         (##input-port-timeout-set! port absrel-timeout thunk)))))))

(define-prim (##output-port-timeout-set! port absrel-timeout thunk)
  (##declare (not interrupts-enabled))
  (let ((timeout (##absrel-timeout->timeout absrel-timeout)))
    ((macro-port-set-wtimeout port) port timeout thunk)))

(define-prim (output-port-timeout-set!
              port
              absrel-timeout
              #!optional
              (t (macro-absent-obj)))
  (macro-force-vars (port absrel-timeout t)
    (let ((thunk
           (if (##eq? t (macro-absent-obj))
             (lambda () #f)
             t)))
      (macro-check-output-port
       port
       1
       (output-port-timeout-set! port absrel-timeout t)
       (macro-check-absrel-time-or-false
        absrel-timeout
        2
        (output-port-timeout-set! port absrel-timeout t)
        (macro-check-procedure
         thunk
         3
         (output-port-timeout-set! port absrel-timeout t)
         (##output-port-timeout-set! port absrel-timeout thunk)))))))

(define-prim (##input-port-line-set! port line)
  (##declare (not interrupts-enabled))
  (macro-port-char-rlines-set! port (##fixnum.- line 1)))

(define-prim (##input-port-line port)
  (##declare (not interrupts-enabled))
  (##fixnum.+ (macro-port-char-rlines port) 1))

(define-prim (input-port-line port)
  (macro-force-vars (port)
    (macro-check-character-input-port port 1 (input-port-line port)
      (##input-port-line port))))

(define-prim (##input-port-column-set! port col)
  (##declare (not interrupts-enabled))
  (macro-port-char-rcurline-set!
   port
   (##fixnum.+ (##fixnum.- (##fixnum.+ (macro-port-char-rchars port)
                                       (macro-port-char-rlo port))
                           col)
               1)))

(define-prim (##input-port-column port)
  (##declare (not interrupts-enabled))
  (##fixnum.+ (##fixnum.- (##fixnum.+ (macro-port-char-rchars port)
                                      (macro-port-char-rlo port))
                          (macro-port-char-rcurline port))
              1))

(define-prim (input-port-column port)
  (macro-force-vars (port)
    (macro-check-character-input-port port 1 (input-port-column port)
      (##input-port-column port))))

(define-prim (##output-port-line-set! port line)
  (##declare (not interrupts-enabled))
  (macro-port-char-wlines-set! port (##fixnum.- line 1)))

(define-prim (##output-port-line port)
  (##declare (not interrupts-enabled))
  (##fixnum.+ (macro-port-char-wlines port) 1))

(define-prim (output-port-line port)
  (macro-force-vars (port)
    (macro-check-character-output-port port 1 (output-port-line port)
      (##output-port-line port))))

(define-prim (##output-port-column-set! port col)
  (##declare (not interrupts-enabled))
  (macro-port-char-wcurline-set!
   port
   (##fixnum.+ (##fixnum.- (##fixnum.+ (macro-port-char-wchars port)
                                       (macro-port-char-whi port))
                           col)
               1)))

(define-prim (##output-port-column port)
  (##declare (not interrupts-enabled))
  (##fixnum.+ (##fixnum.- (##fixnum.+ (macro-port-char-wchars port)
                                      (macro-port-char-whi port))
                          (macro-port-char-wcurline port))
              1))

(define-prim (output-port-column port)
  (macro-force-vars (port)
    (macro-check-character-output-port port 1 (output-port-column port)
      (##output-port-column port))))

(define-prim (##output-port-width port)
  (##declare (not interrupts-enabled))
  ((macro-port-output-width port) port))

(define-prim (output-port-width port)
  (macro-force-vars (port)
    (macro-check-character-output-port port 1 (output-port-width port)
      (##output-port-width port))))

(define-prim (##object->limited-string obj width)
  (if (##fixnum.< width 0)
    (##string)
    (let* ((port
            (##open-output-string))
           (we
            (##make-writeenv
             'write
             port
             (macro-current-readtable)
             #f
             (macro-if-forces #t #f)
             0
             0
             0
             (##fixnum.+ width 1))))
      (##wr we obj)
      (let* ((str (##get-output-string port))
             (len (##string-length str)))
        (if (##fixnum.< width len)
          (begin
            (if (##fixnum.< 0 width)
              (begin
                (##string-set! str (##fixnum.- width 1) #\.)
                (if (##fixnum.< 1 width)
                  (begin
                    (##string-set! str (##fixnum.- width 2) #\.)
                    (if (##fixnum.< 2 width)
                      (##string-set! str (##fixnum.- width 3) #\.))))))
            (##string-shrink! str width)
            str)
          str)))))

;------------------------------------------------------------------------------

; Implementation of generic char port procedures.

(define-prim (##char-ready? port)

  (##declare (not interrupts-enabled))

  (macro-port-mutex-lock! port) ; get exclusive access to port

  (if (macro-port-char-peek-eof? port)

    (begin
      (macro-port-char-peek-eof?-set! port #f)
      (macro-port-mutex-unlock! port)
      #t)

    (let ((char-rlo (macro-port-char-rlo port))
          (char-rhi (macro-port-char-rhi port)))
      (if (##fixnum.< char-rlo char-rhi)
        (begin
          (macro-port-mutex-unlock! port)
          #t)
        (let ((code ((macro-port-char-rbuf-fill port) port #f)))
          (macro-port-mutex-unlock! port)
          (if (##fixnum? code)
            (if (##fixnum.= code ##err-code-EAGAIN)
              #f ; a call to read-char would block
              (##raise-os-exception #f code char-ready? port))
            #t)))))) ; a call to read-char would not block

(define-prim (char-ready?
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-input-port)
             p)))
      (macro-check-character-input-port port 1 (char-ready? p)
        (##char-ready? port)))))

(define-prim (##peek-char port)

  (##declare (not interrupts-enabled))

  (macro-port-mutex-lock! port) ; get exclusive access to port

  (let loop ()

    (let ((char-rlo (macro-port-char-rlo port))
          (char-rhi (macro-port-char-rhi port)))
      (if (##fixnum.< char-rlo char-rhi)

        ; the next character is in the character read buffer

        (let ((c (##string-ref (macro-port-char-rbuf port) char-rlo)))
          (macro-port-mutex-unlock! port)
          c)

        (if (macro-port-char-peek-eof? port)

          (begin
            (macro-port-mutex-unlock! port)
            #!eof)

          ; try to get more characters into the character read
          ; buffer, and try again if successful otherwise
          ; signal an error or return end-of-file object

          (let ((code ((macro-port-char-rbuf-fill port) port #t)))

            (cond ((##fixnum? code)

                   ; the conversion or read caused an error

                   (macro-port-mutex-unlock! port)
                   (if (##fixnum.= code ##err-code-EAGAIN)
                     #f ; the read timeout thunk returned #f
                     (##raise-os-exception #f code peek-char port)))

                  (code

                   ; some characters were added to char buffer

                   (loop))

                  (else

                   ; no characters were added to char buffer

                   (macro-port-char-peek-eof?-set! port #t)
                   (macro-port-mutex-unlock! port)
                   #!eof))))))))

(define-prim (peek-char
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-input-port)
             p)))
      (macro-check-character-input-port port 1 (peek-char p)
        (##peek-char port)))))

(define-prim (##read-char port)

  (##declare (not interrupts-enabled))

  (macro-port-mutex-lock! port) ; get exclusive access to port

  (let loop ()

    (let ((char-rlo (macro-port-char-rlo port))
          (char-rhi (macro-port-char-rhi port)))
      (if (##fixnum.< char-rlo char-rhi)

        ; the next character is in the character read buffer

        (let ((c (##string-ref (macro-port-char-rbuf port) char-rlo)))
          (if (##not (##char=? c #\newline))

            ; frequent simple case, just advance rlo

            (begin
              (macro-port-char-rlo-set! port (##fixnum.+ char-rlo 1))
              (macro-port-mutex-unlock! port)
              c)

            ; end-of-line processing requires updating counters

            (let ((char-rlo+1 (##fixnum.+ char-rlo 1)))

              ; advance rlo

              (macro-port-char-rlo-set! port char-rlo+1)

              ; keep track of number of characters read

              (let ((char-rchars (macro-port-char-rchars port)))
                (macro-port-char-rcurline-set! port
                  (##fixnum.+ char-rchars char-rlo+1)))

              ; keep track of number of lines read

              (let ((char-rlines (macro-port-char-rlines port)))
                (macro-port-char-rlines-set! port
                  (##fixnum.+ char-rlines 1)))

              (macro-port-mutex-unlock! port)
              #\newline)))

        (if (macro-port-char-peek-eof? port)

          (begin
            (macro-port-char-peek-eof?-set! port #f)
            (macro-port-mutex-unlock! port)
            #!eof)

          ; try to get more characters into the character read
          ; buffer, and try again if successful otherwise
          ; signal an error or return end-of-file object

          (let ((code ((macro-port-char-rbuf-fill port) port #t)))

            (cond ((##fixnum? code)

                   ; the conversion or read caused an error

                   (macro-port-mutex-unlock! port)
                   (if (##fixnum.= code ##err-code-EAGAIN)
                     #f ; the read timeout thunk returned #f
                     (##raise-os-exception #f code read-char port)))

                  (code

                   ; some characters were added to char buffer

                   (loop))

                  (else

                   ; no characters were added to char buffer

                   (macro-port-mutex-unlock! port)
                   #!eof))))))))

(define-prim (read-char
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-input-port)
             p)))
      (macro-check-character-input-port port 1 (read-char p)
        (##read-char port)))))

(define-prim (##read-byte port)
  (##declare (not interrupts-enabled))
  (let ((c (##read-char port)));;;;;;;;;;;;;;
    (if (##char? c)
      (##fixnum.<-char c)
      c)))

(define-prim (read-byte
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-input-port)
             p)))
      (macro-check-byte-input-port port 1 (read-byte p)
        (##read-byte port)))))

(define-prim (##read-line port separator include-separator?)

  (define max-chunk-length 512)

  (define (read-chunk i)
    (let loop ((i i))
      (if (##fixnum.< i max-chunk-length)
        (let ((c (macro-read-char port)))
          (if (##char? c)
            (if (##eq? c separator)
              (if include-separator?
                (let ((s (##make-string (##fixnum.+ i 1))))
                  (##string-set! s i c)
                  s)
                (##make-string i))
              (let ((s (loop (##fixnum.+ i 1))))
                (##string-set! s i c)
                s))
            (##make-string i)))
        (##make-string i))))

  (let ((first (macro-read-char port)))
    (if (##char? first)
      (if (##eq? first separator)
        (if include-separator?
          (##string first)
          (##string))
        (let ((chunk1 (read-chunk 1)))
          (##string-set! chunk1 0 first)
          (if (or (##fixnum.< (##string-length chunk1)
                              max-chunk-length)
                  (##eq? (##string-ref
                          chunk1
                          (##fixnum.- max-chunk-length 1))
                         separator))
            chunk1
            (let loop ((chunks (##list chunk1)))
              (let* ((new-chunk (read-chunk 0))
                     (new-chunks (##cons new-chunk chunks)))
                (if (or (##fixnum.< (##string-length new-chunk)
                                    max-chunk-length)
                        (##eq? (##string-ref
                                new-chunk
                                (##fixnum.- max-chunk-length 1))
                               separator))
                  (##append-strings (##reverse new-chunks))
                  (loop new-chunks)))))))
      first)))

(define-prim (read-line
              #!optional
              (p (macro-absent-obj))
              (sep (macro-absent-obj))
              (inc-sep? (macro-absent-obj)))
  (macro-force-vars (p sep inc-sep?)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-input-port)
             p))
          (separator
           (if (##eq? sep (macro-absent-obj))
             #\newline
             sep))
          (include-separator?
           (if (##eq? inc-sep? (macro-absent-obj))
             #f
             inc-sep?)))
      (macro-check-character-input-port port 1 (read-line p sep inc-sep?)
        (##read-line port separator include-separator?)))))

(define-prim (##read-all port reader)
  (let ((fifo (macro-make-fifo)))
    (let loop ()
      (let ((obj (reader port)))
        (if (##eof-object? obj)
          (macro-fifo->list fifo)
          (begin
            (macro-fifo-insert-at-tail! fifo obj)
            (loop)))))))

(define-prim (read-all
              #!optional
              (p (macro-absent-obj))
              (r (macro-absent-obj)))
  (macro-force-vars (p r)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-input-port)
             p))
          (reader
           (if (##eq? r (macro-absent-obj))
             ##read
             r)))
      (macro-check-input-port port 1 (read-all p r)
        (##read-all port reader)))))

(define-prim (##read-subu8vector u8vect start end port)

  ; It is assumed that the thread **does not** have exclusive access
  ; to the port.

  (##declare (not interrupts-enabled))

  (macro-port-mutex-lock! port) ; get exclusive access to port

  (if (or (##fixnum.< (macro-port-char-rlo port)
                      (macro-port-char-rhi port))
          (macro-port-char-peek-eof? port))

    (begin
      (macro-port-mutex-unlock! port)
      (error "character buffer is not empty!"));;;;;;;;;;;;;;;;;;;

    (let loop1 ((n 0))
      (let ((remaining (##fixnum.- end (##fixnum.+ start n))))
        (if (##not (##fixnum.< 0 remaining))
          (begin
            (macro-port-mutex-unlock! port)
            n)
          (let* ((byte-rlo
                  (macro-port-byte-rlo port))
                 (byte-rhi
                  (macro-port-byte-rhi port))
                 (available-in-byte-rbuf
                  (##fixnum.- byte-rhi byte-rlo)))
            (if (##fixnum.< 0 available-in-byte-rbuf)
              (let* ((to-transfer
                      (##fixnum.min remaining available-in-byte-rbuf))
                     (limit
                      (##fixnum.+ byte-rlo to-transfer))
                     (byte-rbuf
                      (macro-port-byte-rbuf port)))
                (macro-port-byte-rlo-set! port limit)
                (let loop2 ((i (##fixnum.+ start n))
                            (j byte-rlo))
                  (if (##fixnum.< j limit)
                    (begin
                      (##u8vector-set! u8vect i (##u8vector-ref byte-rbuf j))
                      (loop2 (##fixnum.+ i 1)
                             (##fixnum.+ j 1)))))
                (loop1 (##fixnum.+ n to-transfer)))
              (let ((code ((macro-port-byte-rbuf-fill port) port #t)))
                (cond ((##fixnum? code)

                       ; an error occurred, signal an error if no
                       ; bytes were previously transfered from byte
                       ; buffer and (in the case of a read timeout)
                       ; the timeout thunk returned #f

                       (macro-port-mutex-unlock! port)

                       (if (or (##fixnum.< 0 n)
                               (##fixnum.= code ##err-code-EAGAIN))
                         n
                         (##raise-os-exception
                          #f
                          code
                          read-subu8vector
                          u8vect
                          start
                          end
                          port)))

                      (code

                       ; bytes were added to byte buffer, so try again
                       ; to transfer bytes from the byte buffer

                       (loop1 n))

                      (else

                       ; no bytes were added to byte buffer
                       ; (end-of-file was reached)

                       (macro-port-mutex-unlock! port)

                       (if (##fixnum.< 0 n)
                         n
                         #!eof)))))))))))

(define-prim (read-subu8vector
              u8vect
              start
              end
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (u8vect start end p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-input-port)
             p)))
      (macro-check-u8vector u8vect 1 (read-subu8vector u8vect start end p)
        (macro-check-index-range-incl
         start
         2
         0
         (##u8vector-length u8vect)
         (read-subu8vector u8vect start end p)
         (macro-check-index-range-incl
          end
          3
          start
          (##u8vector-length u8vect)
          (read-subu8vector u8vect start end p)
          (macro-check-byte-input-port
           port
           4
           (read-subu8vector u8vect start end p)
           (##read-subu8vector u8vect start end port))))))))

(define-prim (##write-char c port)

  (##declare (not interrupts-enabled))

  (macro-port-mutex-lock! port) ; get exclusive access to port

  (let loop ()

    (let ((char-wbuf (macro-port-char-wbuf port))
          (char-whi+1 (##fixnum.+ (macro-port-char-whi port) 1)))
      (if (##not (##fixnum.< (##string-length char-wbuf) char-whi+1))

        ; there is enough space in the character write buffer, so add
        ; character and increment whi

        (let ()

          (##string-set! char-wbuf (##fixnum.- char-whi+1 1) c)

          ; advance whi

          (macro-port-char-whi-set! port char-whi+1)

          (if (##not (##char=? c #\newline))

            ; force output if port is set for unbuffered output

            (if (macro-unbuffered? (macro-port-woptions port))
              (begin
                (macro-port-mutex-unlock! port)
                ((macro-port-force-output port)
                 port
                 write-char
                 c
                 port))
              (begin
                (macro-port-mutex-unlock! port)
                (##void)))

            ; end-of-line processing requires updating counters

            (begin

              ; keep track of number of characters written

              (let ((char-wchars (macro-port-char-wchars port)))
                (macro-port-char-wcurline-set! port
                  (##fixnum.+ char-wchars char-whi+1)))

              ; keep track of number of lines written

              (let ((char-wlines (macro-port-char-wlines port)))
                (macro-port-char-wlines-set! port
                  (##fixnum.+ char-wlines 1)))

              ; force output if port is set for unbuffered or
              ; line-buffered output

              (if (##not (macro-buffered?
                          (macro-port-woptions port)))
                (begin
                  (macro-port-mutex-unlock! port)
                  ((macro-port-force-output port)
                   port
                   write-char
                   c
                   port))
                (begin
                  (macro-port-mutex-unlock! port)
                  (##void))))))

        ; make some space in the character buffer and try again

        (let ((code3 ((macro-port-char-wbuf-drain port) port)))
          (if (##fixnum? code3)
            (begin
              (macro-port-mutex-unlock! port)
              (if (##fixnum.= code3 ##err-code-EAGAIN)
                #f
                (##raise-os-exception #f code3 write-char c port)))
            (loop)))))))

(define-prim (write-char
              c
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (c p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-output-port)
             p)))
      (macro-check-char c 1 (write-char c p)
        (macro-check-character-output-port port 2 (write-char c p)
          (##write-char c port))))))

(define-prim (##write-byte b port)
  (##declare (not interrupts-enabled))
  (##write-char (##fixnum.->char b) port));;;;;;;;;;;;;;;;;;;;;;

(define-prim (write-byte
              b
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (b p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-output-port)
             p)))
      (macro-check-exact-unsigned-int8 b 1 (write-byte b p)
        (macro-check-byte-output-port port 2 (write-byte b p)
          (##write-byte b port))))))

(define-prim (##write-substring str start end port)
  (##declare (not interrupts-enabled))
  (let loop ((i start))
    (if (##fixnum.< i end)
      (begin
        (macro-write-char (##string-ref str i) port)
        (let ()
          (##declare (interrupts-enabled))
          (loop (##fixnum.+ i 1)))))))

(define-prim (##write-string str port)
  (##declare (not interrupts-enabled))
  (##write-substring str 0 (##string-length str) port))

(define-prim (##write-subu8vector u8vect start end port)

  ; It is assumed that the thread **does not** have exclusive access
  ; to the port.

  (##declare (not interrupts-enabled))

  (macro-port-mutex-lock! port) ; get exclusive access to port

  (let ((code
         (and (##fixnum.< (macro-port-char-wlo port)
                          (macro-port-char-whi port))
              ((macro-port-char-wbuf-drain port) port))))
    (if (##fixnum? code)

      (begin
        (macro-port-mutex-unlock! port)
        (if (##fixnum.= code ##err-code-EAGAIN)
          0
          (##raise-os-exception
           #f
           code
           write-subu8vector
           u8vect
           start
           end
           port)))

      (let loop1 ((n 0))
        (let ((remaining (##fixnum.- end (##fixnum.+ start n))))
          (if (##not (##fixnum.< 0 remaining))
            (begin
              (macro-port-mutex-unlock! port)
              n)
            (let* ((byte-whi
                    (macro-port-byte-whi port))
                   (byte-wbuf
                    (macro-port-byte-wbuf port))
                   (available-in-byte-wbuf
                    (##fixnum.- (##u8vector-length byte-wbuf) byte-whi)))
              (if (##fixnum.< 0 available-in-byte-wbuf)
                (let* ((to-transfer
                        (##fixnum.min remaining available-in-byte-wbuf))
                       (limit
                        (##fixnum.+ byte-whi to-transfer)))
                  (macro-port-byte-whi-set! port limit)
                  (let loop2 ((i (##fixnum.+ start n))
                              (j byte-whi))
                    (if (##fixnum.< j limit)
                      (begin
                        (##u8vector-set! byte-wbuf j (##u8vector-ref u8vect i))
                        (loop2 (##fixnum.+ i 1)
                               (##fixnum.+ j 1)))))
                  (loop1 (##fixnum.+ n to-transfer)))
                (let ((code ((macro-port-byte-wbuf-drain port) port)))
                  (if (##fixnum? code)
                    (begin

                      ; an error occurred, signal an error if no bytes
                      ; were previously transfered from byte buffer
                      ; and (in the case of a write timeout) the
                      ; timeout thunk returned #f

                      (macro-port-mutex-unlock! port)

                      (if (or (##fixnum.< 0 n)
                              (##fixnum.= code ##err-code-EAGAIN))
                        n
                        (##raise-os-exception
                         #f
                         code
                         write-subu8vector
                         u8vect
                         start
                         end
                         port)))

                    ; the byte buffer was successfully drained, so try
                    ; again to transfer bytes to the byte buffer

                    (loop1 n)))))))))))

(define-prim (write-subu8vector
              u8vect
              start
              end
              #!optional
              (p (macro-absent-obj)))
  (macro-force-vars (u8vect start end p)
    (let ((port
           (if (##eq? p (macro-absent-obj))
             (macro-current-output-port)
             p)))
      (macro-check-u8vector u8vect 1 (write-subu8vector u8vect start end p)
        (macro-check-index-range-incl
         start
         2
         0
         (##u8vector-length u8vect)
         (write-subu8vector u8vect start end p)
         (macro-check-index-range-incl
          end
          3
          start
          (##u8vector-length u8vect)
          (write-subu8vector u8vect start end p)
          (macro-check-byte-output-port
           port
           4
           (write-subu8vector u8vect start end p)
           (##write-subu8vector u8vect start end port))))))))

;------------------------------------------------------------------------------

; Implementation of tty device ports.

(define-fail-check-type tty-device-port)

(define-prim (##tty? port)
  (##declare (not interrupts-enabled))
  (macro-tty-device-port? port))

(define-prim (tty? port)
  (macro-force-vars (port)
    (##tty? port)))

(define-prim (##tty-type-set! port term-type emacs-bindings)
  (##dev-tty-type-set! (##port-device port) term-type emacs-bindings))

(define-prim (tty-type-set! port term-type emacs-bindings)
  (macro-force-vars (port term-type emacs-bindings)
    (macro-check-tty-device-port
     port
     1
     (tty-type-set! port term-type emacs-bindings)
     (macro-check-string
      term-type
      2
      (tty-type-set! port term-type emacs-bindings)
      (##tty-type-set! port term-type emacs-bindings)))))

(define-prim (##tty-text-attributes-set! port input output)
  (##dev-tty-text-attributes-set! (##port-device port) input output))

(define-prim (tty-text-attributes-set! port input output)
  (macro-force-vars (port input output)
    (macro-check-tty-device-port
     port
     1
     (tty-text-attributes-set! port input output)
     (macro-check-fixnum-range
      input
      2
      0
      1024
      (tty-text-attributes-set! port input output)
      (macro-check-fixnum-range
       output
       3
       0
       1024
       (tty-text-attributes-set! port input output)
       (##tty-text-attributes-set! port input output))))))

(define-prim (##tty-history port)
  (##dev-tty-history (##port-device port)))

(define-prim (tty-history port)
  (macro-force-vars (port)
    (macro-check-tty-device-port
     port
     1
     (tty-history port)
     (##tty-history port))))

(define-prim (##tty-history-set! port history)
  (##dev-tty-history-set! (##port-device port) history))

(define-prim (tty-history-set! port history)
  (macro-force-vars (port history)
    (macro-check-tty-device-port
     port
     1
     (tty-history-set! port history)
     (macro-check-string
      history
      2
      (tty-history-set! port history)
      (##tty-history-set! port history)))))

(define-prim (##tty-max-history-length-set! port max-length)
  (##dev-tty-max-history-length-set! (##port-device port) max-length))

(define-prim (tty-max-history-length-set! port max-length)
  (macro-force-vars (port max-length)
    (macro-check-tty-device-port
     port
     1
     (tty-max-history-length-set! port max-length)
     (macro-check-index
      max-length
      2
      (tty-max-history-length-set! port max-length)
      (##tty-max-history-length-set! port max-length)))))

(define-prim (##tty-paren-balance-duration-set! port duration)
  (##dev-tty-paren-balance-duration-set! (##port-device port) duration))

(define-prim (tty-paren-balance-duration-set! port duration)
  (macro-force-vars (port duration)
    (macro-check-tty-device-port
     port
     1
     (tty-paren-balance-duration-set! port duration)
     (macro-check-real
      duration
      2
      (tty-paren-balance-duration-set! port duration)
      (##tty-paren-balance-duration-set!
       port
       (macro-real->inexact duration))))))

(define-prim (##tty-mode-set! port mode)
  (##dev-tty-mode-set! (##port-device port) mode))

(define-prim (tty-mode-set! port mode)
  (macro-force-vars (port mode)
    (macro-check-tty-device-port
     port
     1
     (tty-mode-set! port mode)
     (##tty-mode-set! port mode))))

;------------------------------------------------------------------------------

; Implementation of process device ports.

(define-fail-check-type process-port)

(define-prim (##make-process-psettings
              #!optional
              (settings (macro-absent-obj)))
  (##make-psettings
   (macro-direction-inout)
   '(stderr-redirection:
     pseudo-terminal:
     output-width:
     input-char-encoding:
     output-char-encoding:
     char-encoding:
     input-eol-encoding:
     output-eol-encoding:
     eol-encoding:
     input-buffering:
     output-buffering:
     buffering:)
   settings))

(define-prim (##open-process path-and-arguments environment settings)

  (define (psettings->options psettings)
    (let ((stderr-redir
           (macro-psettings-stderr-redir psettings))
          (pseudo-term
           (macro-psettings-pseudo-term psettings)))
      (##fixnum.+
       (##fixnum.* 2 pseudo-term)
       stderr-redir)))

  (let ((psettings
         (##make-process-psettings settings)))
    (if (##not psettings)
      (##trap-psettings open-process path-and-arguments environment settings)
      (let ((device
             (##dev-stream-open-process
              path-and-arguments
              environment
              (psettings->options psettings))))
        (cond ((##fixnum? device)
               (##raise-os-exception
                #f
                device
                open-process
                path-and-arguments
                environment
                settings))
              ((##not device)
               (macro-trap-open-file
                (open-process path-and-arguments environment settings)))
              (else
               (##make-device-port-from-single-device
                (##cons 'process path-and-arguments)
                device
                psettings)))))))

(define-prim (open-process
              path-and-arguments
              #!optional
              (environment (macro-absent-obj))
              (settings (macro-absent-obj)))
  (macro-force-vars (path-and-arguments settings)
    (let* ((lst1 (##copy-string-list path-and-arguments))
           (lst2 (if (or (##eq? environment (macro-absent-obj))
                         (##not environment))
                   #f
                   (##copy-string-list environment))))
      (cond ((or (##null? lst1)
                 (##fixnum? lst1))
             (##fail-check-string-list 1
                                       open-process
                                       path-and-arguments
                                       environment
                                       settings))
            ((##fixnum? lst2)
             (##fail-check-string-list 2
                                       open-process
                                       path-and-arguments
                                       environment
                                       settings))
            (else
             (##open-process lst1 lst2 settings))))))

;------------------------------------------------------------------------------

; Implementation of hostent objects.

(define-fail-check-type hostent)

(define ##hostent-tag
  '#(hostent -1))

(define-prim (##hostent? obj)
  (macro-hostent? obj))

(define-prim (hostent? obj)
  (macro-force-vars (obj)
    (macro-hostent? obj)))

(define-prim (hostent-official-name hostent)
  (macro-force-vars (hostent)
    (macro-check-hostent hostent 1 (hostent-official-name hostent)
      (macro-hostent-official-name hostent))))

(define-prim (hostent-aliases hostent)
  (macro-force-vars (hostent)
    (macro-check-hostent hostent 1 (hostent-aliases hostent)
      (macro-hostent-aliases hostent))))

(define-prim (hostent-addresses hostent)
  (macro-force-vars (hostent)
    (macro-check-hostent hostent 1 (hostent-addresses hostent)
      (macro-hostent-addresses hostent))))

(define-prim (##get-host-by-name name)
  (let ((result (##netdb-get-host-by-name name)))
    (if (##fixnum? result)
      (##raise-os-exception #f result get-host-by-name name)
      (begin
        (##vector-set! result 0 ##hostent-tag)
        (##subtype-set! result (macro-subtype-structure))))))

(define-prim (get-host-by-name name)
  (macro-force-vars (name)
    (macro-check-string name 1 (get-host-by-name name)
      (##get-host-by-name name))))

;------------------------------------------------------------------------------

; Implementation of TCP client device ports.

(define-fail-check-type tcp-client-port)

(define-prim (##make-tcp-psettings
              client?
              #!optional
              (settings (macro-absent-obj)))
  (define allowed-server-settings
    '(reuse-address:
      backlog:
      keep-alive:
      coalesce:
      output-width:
      input-char-encoding:
      output-char-encoding:
      char-encoding:
      input-eol-encoding:
      output-eol-encoding:
      eol-encoding:
      input-buffering:
      output-buffering:
      buffering:))

  (##make-psettings
   (macro-direction-inout)
   (if client?
     (##cddr allowed-server-settings)
     allowed-server-settings)
   settings))

(define-prim (##make-tcp-client-port device psettings)
  (##make-device-port-from-single-device
   '(tcp-client)
   device
   psettings))

(define-prim (##open-tcp-client server-addr port-num settings)

  (define (psettings->options psettings)
    (let ((keep-alive
           (macro-psettings-keep-alive psettings))
          (coalesce
           (macro-psettings-coalesce psettings)))
      (##fixnum.+
       (##fixnum.* 2 coalesce)
       keep-alive)))

  (let ((psettings
         (##make-tcp-psettings #t settings)))
    (if (##not psettings)
      (##trap-psettings open-tcp-client server-addr port-num settings)
      (let ((device
             (##dev-tcp-client-open
              server-addr
              port-num
              (psettings->options psettings))))
        (if (##fixnum? device)
          (##raise-os-exception
           #f
           device
           open-tcp-client
           server-addr
           port-num
           settings)
          (##make-tcp-client-port device psettings))))))

(define-prim (open-tcp-client
              server-addr
              port-num
              #!optional
              (settings (macro-absent-obj)))
  (macro-force-vars (server-addr port-num settings)
    (if (##not (or (##string? server-addr)
                   (and (##u8vector? server-addr)
                        (or (##fixnum.= (##u8vector-length server-addr) 4)
                            (##fixnum.= (##u8vector-length server-addr) 16)))))
      (error);;;;;;;;;;;;;;;;;
      (macro-check-fixnum-range
       port-num
       2
       0
       65536
       (open-tcp-client server-addr port-num settings)
       (if (##string? server-addr)
         (let ((hostent (##netdb-get-host-by-name server-addr)))
           (if (##fixnum? hostent)
             (##raise-os-exception
              #f
              hostent
              open-tcp-client
              server-addr
              port-num
              settings)
             (##open-tcp-client
              (##car (macro-hostent-addresses hostent))
              port-num
              settings)))
         (##open-tcp-client
          server-addr
          port-num
          settings))))))

;------------------------------------------------------------------------------

; Implementation of TCP server ports.

(define-fail-check-type tcp-server-port)

(define-prim (##make-tcp-server-port rdevice client-psettings)
  (let ((mutex
         (macro-make-port-mutex))
        (rkind
         (macro-object-kind))
        (wkind
         (macro-none-kind))
        (roptions
         0)
        (rtimeout
         #t)
        (rtimeout-thunk
         #f)
        (woptions
         0)
        (wtimeout
         #t)
        (wtimeout-thunk
         #f)
        (rdevice-condvar
         (and rdevice
              (##make-rdevice-condvar rdevice))))

        (define (name port)

          ; It is assumed that the thread **does not** have exclusive
          ; access to the port.

          (##declare (not interrupts-enabled))

          '(tcp-server))

        (define (read-datum port re)

          ; It is assumed that the thread **does not** have exclusive
          ; access to the port.

          (##declare (not interrupts-enabled))

          (macro-port-mutex-lock! port) ; get exclusive access to port

          (let loop ()
            (let ((client-device
                   (##dev-tcp-server-read
                    (macro-condvar-name
                     (macro-port-tcp-server-rdevice-condvar port)))))
              (if (##fixnum? client-device)

                (cond ((##fixnum.= client-device ##err-code-EAGAIN)

                       ; the read would block, so  wait and then try again

                       (macro-port-mutex-unlock! port)
                       (let ((continue?
                              (or (##device-condvar-wait!
                                   (macro-port-tcp-server-rdevice-condvar port)
                                   (macro-port-rtimeout port))
                                  ((macro-port-rtimeout-thunk port)))))
                         (macro-port-mutex-lock! port) ; regain access to port
                         (if continue?
                           (loop)
                           (##raise-os-exception
                            #f
                            client-device
                            read
                            port))))

                      ((##fixnum.= client-device ##err-code-EINTR)

                       ; the read was interrupted, so try again

                       (loop))

                      (else

                       ; signal an error

                       (macro-port-mutex-unlock! port)
                       (##raise-os-exception #f client-device read port)))

                (begin
                  (macro-port-mutex-unlock! port)
                  (##make-tcp-client-port
                   client-device
                   (macro-port-tcp-server-client-psettings port)))))))

        (define write-datum #f)

        (define newline #f)

        (define force-output #f)

        (define (set-rtimeout port timeout thunk)

          ; It is assumed that the thread **does not** have exclusive
          ; access to the port.

          (##declare (not interrupts-enabled))

          (macro-port-mutex-lock! port) ; get exclusive access to port

          (macro-port-rtimeout-set! port timeout)
          (macro-port-rtimeout-thunk-set! port thunk)
          (##condvar-signal-no-reschedule!
           (macro-port-tcp-server-rdevice-condvar port)
           #t)
          (macro-port-mutex-unlock! port)
          (##void))

        (define set-wtimeout #f)

        (define (close port prim arg1)

          ; It is assumed that the thread **does not** have exclusive
          ; access to the port.

          (##declare (not interrupts-enabled))

          (macro-port-mutex-lock! port) ; get exclusive access to port

          (let ((result
                 (##close-device
                  port
                  (macro-port-tcp-server-rdevice-condvar port)
                  #f
                  prim)))
            (macro-port-mutex-unlock! port)
            (if (##fixnum? result)
              (##raise-os-exception #f result prim arg1)
              result)))

        (macro-make-object-port
         mutex
         rkind
         wkind
         name
         read-datum
         write-datum
         newline
         force-output
         close
         roptions
         rtimeout
         rtimeout-thunk
         set-rtimeout
         woptions
         wtimeout
         wtimeout-thunk
         set-wtimeout
         rdevice-condvar
         client-psettings)))

(define-prim (##open-tcp-server port-num settings)

  (define (psettings->options psettings)
    (macro-psettings-reuse-address psettings))

  (let ((psettings
         (##make-tcp-psettings #f settings)))
    (if (##not psettings)
      (##trap-psettings open-tcp-server port-num settings)
      (let ((rdevice
             (##dev-tcp-server-open
              port-num
              (macro-psettings-backlog psettings)
              (psettings->options psettings))))
        (if (##fixnum? rdevice)
          (##raise-os-exception
           #f
           rdevice
           open-tcp-server
           port-num
           settings)
          (##make-tcp-server-port rdevice psettings))))))

(define-prim (open-tcp-server
              port-num
              #!optional
              (settings (macro-absent-obj)))
  (macro-force-vars (port-num settings)
    (macro-check-fixnum-range
     port-num
     1
     0
     65536
     (open-tcp-server port-num settings)
     (##open-tcp-server port-num settings))))

;------------------------------------------------------------------------------

; Implementation of directory ports.

(define-fail-check-type directory-port)

(define-prim (##make-directory-port rdevice path)
  (let ((mutex
         (macro-make-port-mutex))
        (rkind
         (macro-object-kind))
        (wkind
         (macro-none-kind))
        (roptions
         0)
        (rtimeout
         #t)
        (rtimeout-thunk
         #f)
        (woptions
         0)
        (wtimeout
         #t)
        (wtimeout-thunk
         #f)
        (rdevice-condvar
         (and rdevice
              (##make-rdevice-condvar rdevice))))

        (define (name port)

          ; It is assumed that the thread **does not** have exclusive
          ; access to the port.

          (##declare (not interrupts-enabled))

          (macro-port-directory-path port))

        (define (read-datum port re)

          ; It is assumed that the thread **does not** have exclusive
          ; access to the port.

          (##declare (not interrupts-enabled))

          (macro-port-mutex-lock! port) ; get exclusive access to port

          (let loop ()
            (let ((datum
                   (##dev-directory-read
                    (macro-condvar-name
                     (macro-port-directory-rdevice-condvar port)))))
              (if (##fixnum? datum)

                (cond ((##fixnum.= datum ##err-code-EAGAIN)

                       ; the read would block, so  wait and then try again

                       (macro-port-mutex-unlock! port)
                       (let ((continue?
                              (or (##device-condvar-wait!
                                   (macro-port-directory-rdevice-condvar port)
                                   (macro-port-rtimeout port))
                                  ((macro-port-rtimeout-thunk port)))))
                         (macro-port-mutex-lock! port) ; regain access to port
                         (if continue?
                           (loop)
                           (##raise-os-exception #f datum read port))))

                      ((##fixnum.= datum ##err-code-EINTR)

                       ; the read was interrupted, so try again

                       (loop))

                      (else

                       ; signal an error

                       (macro-port-mutex-unlock! port)
                       (##raise-os-exception #f datum read port)))

                (begin
                  (macro-port-mutex-unlock! port)
                  datum)))))

        (define write-datum #f)

        (define newline #f)

        (define force-output #f)

        (define (set-rtimeout port timeout thunk)

          ; It is assumed that the thread **does not** have exclusive
          ; access to the port.

          (##declare (not interrupts-enabled))

          (macro-port-mutex-lock! port) ; get exclusive access to port

          (macro-port-rtimeout-set! port timeout)
          (macro-port-rtimeout-thunk-set! port thunk)
          (##condvar-signal-no-reschedule!
           (macro-port-directory-rdevice-condvar port)
           #t)
          (macro-port-mutex-unlock! port)
          (##void))

        (define set-wtimeout #f)

        (define (close port prim arg1)

          ; It is assumed that the thread **does not** have exclusive
          ; access to the port.

          (##declare (not interrupts-enabled))

          (macro-port-mutex-lock! port) ; get exclusive access to port

          (let ((result
                 (##close-device
                  port
                  (macro-port-directory-rdevice-condvar port)
                  #f
                  prim)))
            (macro-port-mutex-unlock! port)
            (if (##fixnum? result)
              (##raise-os-exception #f result prim arg1)
              result)))

        (macro-make-object-port
         mutex
         rkind
         wkind
         name
         read-datum
         write-datum
         newline
         force-output
         close
         roptions
         rtimeout
         rtimeout-thunk
         set-rtimeout
         woptions
         wtimeout
         wtimeout-thunk
         set-wtimeout
         rdevice-condvar
         path)))

(define-prim (##open-directory path)
  (let ((absolute-path (##path-expand path 'absolute)))
    (and absolute-path
         (let ((rdevice (##dev-directory-open-path absolute-path)))
           (and rdevice
                (if (##fixnum? rdevice)
                  (##raise-os-exception #f rdevice open-directory path)
                  (##make-directory-port rdevice path)))))))

(define-prim (open-directory path)
  (macro-force-vars (path)
    (macro-check-string path 1 (open-directory path)
      (let ((port (##open-directory path)))
        (if port
          port
          (macro-trap-open-file (open-directory path)))))))

;------------------------------------------------------------------------------

(define-prim (##make-path-psettings
              direction
              #!optional
              (settings (macro-absent-obj)))
  (##make-psettings
   direction
   '(direction:
     append:
     create:
     truncate:
     permissions:
     output-width:
     input-char-encoding:
     output-char-encoding:
     char-encoding:
     input-eol-encoding:
     output-eol-encoding:
     eol-encoding:
     input-buffering:
     output-buffering:
     buffering:)
   settings))

(define-prim (##open-path
              direction
              prim
              path
              #!optional
              (settings (macro-absent-obj)))
  (let ((psettings (##make-path-psettings direction settings)))
    (if (##not psettings)
      (##trap-psettings prim path settings)
      (let ((absolute-path (##path-expand path 'absolute)))
        (and absolute-path
             (let ((device
                    (##dev-stream-open-path
                     absolute-path
                     (##psettings->device-flags psettings)
                     (##psettings->permissions psettings))))
               (and device
                    (if (##fixnum? device)
                      (##raise-os-exception #f device prim path settings)
                      (##make-device-port-from-single-device
                       absolute-path
                       device
                       psettings)))))))))

(define-prim (##open-file
              path
              #!optional
              (settings (macro-absent-obj)))
  (##open-path (macro-direction-inout)
               open-file
               path
               settings))

(define-prim (open-file
              path
              #!optional
              (settings (macro-absent-obj)))
  (macro-force-vars (path settings)
    (macro-check-string path 1 (open-file path settings)
      (let ((port (##open-file path settings)));;;;;;;;;;;;exception might be raised inside!!!!
        (if port
          port
          (macro-trap-open-file (open-file path settings)))))))

(define-prim (##open-input-file
              path
              #!optional
              (settings (macro-absent-obj)))
  (##open-path (macro-direction-in)
               open-input-file
               path
               settings))

(define-prim (open-input-file
              path
              #!optional
              (settings (macro-absent-obj)))
  (macro-force-vars (path settings)
    (macro-check-string path 1 (open-input-file path settings)
      (let ((port (##open-input-file path settings)))
        (if port
          port
          (macro-trap-open-file (open-input-file path settings)))))))

(define-prim (##open-output-file
              path
              #!optional
              (settings (macro-absent-obj)))
  (##open-path (macro-direction-out)
               open-output-file
               path
               settings))

(define-prim (open-output-file
              path
              #!optional
              (settings (macro-absent-obj)))
  (macro-force-vars (path settings)
    (macro-check-string path 1 (open-output-file path settings)
      (let ((port (##open-output-file path settings)))
        (if port
          port
          (macro-trap-open-file (open-output-file path settings)))))))

(define-prim (call-with-input-file
              path
              arg2
              #!optional
              (arg3 (macro-absent-obj)))

  (define (call proc arg-num settings)
    (macro-check-procedure
     proc
     arg-num
     (call-with-input-file path arg2 arg3)
     (let ((port (##open-input-file path settings)))
       (if port
         (let ((results ; may get bound to a multiple-values object
                (proc port)))
           (##close-input-port port)
           results)
         (macro-trap-open-file (call-with-input-file path arg2 arg3))))))

  (macro-force-vars (path arg2 arg3)
    (macro-check-string path 1 (call-with-input-file path arg2 arg3)
      (if (##eq? arg3 (macro-absent-obj))
        (call arg2 2 arg3)
        (call arg3 3 arg2)))))

(define-prim (call-with-output-file
              path
              arg2
              #!optional
              (arg3 (macro-absent-obj)))

  (define (call proc arg-num settings)
    (macro-check-procedure
     proc
     arg-num
     (call-with-output-file path arg2 arg3)
     (let ((port (##open-output-file path settings)))
       (if port
         (let ((results ; may get bound to a multiple-values object
                (proc port)))
           (##close-output-port port)
           results)
         (macro-trap-open-file (call-with-output-file path arg2 arg3))))))

  (macro-force-vars (path arg2 arg3)
    (macro-check-string path 1 (call-with-output-file path arg2 arg3)
      (if (##eq? arg3 (macro-absent-obj))
        (call arg2 2 arg3)
        (call arg3 3 arg2)))))

(define-prim (with-input-from-file
              path
              arg2
              #!optional
              (arg3 (macro-absent-obj)))

  (define (call thunk arg-num settings)
    (macro-check-procedure
     thunk
     arg-num
     (with-input-from-file path arg2 arg3)
     (let ((port (##open-input-file path settings)))
       (if port
         (let ((results ; may get bound to a multiple-values object
                (macro-dynamic-bind input-port port thunk)))
           (##close-input-port port)
           results)
         (macro-trap-open-file (with-input-from-file path arg2 arg3))))))

  (macro-force-vars (path arg2 arg3)
    (macro-check-string path 1 (with-input-from-file path arg2 arg3)
      (if (##eq? arg3 (macro-absent-obj))
        (call arg2 2 arg3)
        (call arg3 3 arg2)))))

(define-prim (with-output-to-file
              path
              arg2
              #!optional
              (arg3 (macro-absent-obj)))

  (define (call thunk arg-num settings)
    (macro-check-procedure
     thunk
     arg-num
     (with-output-to-file path arg2 arg3)
     (let ((port (##open-output-file path settings)))
       (if port
         (let ((results ; may get bound to a multiple-values object
                (macro-dynamic-bind output-port port thunk)))
           (##close-output-port port)
           results)
         (macro-trap-open-file (with-output-to-file path arg2 arg3))))))

  (macro-force-vars (path arg2 arg3)
    (macro-check-string path 1 (with-output-to-file path arg2 arg3)
      (if (##eq? arg3 (macro-absent-obj))
        (call arg2 2 arg3)
        (call arg3 3 arg2)))))

;------------------------------------------------------------------------------

(define-prim (with-input-from-port port thunk)
  (macro-force-vars (port thunk)
    (macro-check-input-port port 1 (with-input-from-port port thunk)
      (macro-check-procedure thunk 2 (with-input-from-port port thunk)
        (macro-dynamic-bind input-port port thunk)))))

(define-prim (with-output-to-port port thunk)
  (macro-force-vars (port thunk)
    (macro-check-output-port port 1 (with-output-to-port port thunk)
      (macro-check-procedure thunk 2 (with-output-to-port port thunk)
        (macro-dynamic-bind output-port port thunk)))))

;------------------------------------------------------------------------------

(define-prim (##open-predefined
              direction
              name
              index
              #!optional
              (settings (macro-absent-obj)))
  (let* ((psettings
          (##make-path-psettings direction settings))
         (device
          (##dev-stream-open-predefined
           index
           (##psettings->device-flags psettings))))
    (if (##fixnum? device)
      (##exit 1)
      (and device
           (##make-device-port-from-single-device
            name
            device
            psettings)))))

(define ##stdin   #f)
(define ##stdout  #f)
(define ##stderr  #f)
(define ##console-port #f)
(define ##ide-port     #f)

(set! ##stdin   (##open-predefined (macro-direction-in)    '(stdin)   -1))
(set! ##stdout  (##open-predefined (macro-direction-out)   '(stdout)  -2))
(set! ##stderr  (##open-predefined (macro-direction-out)   '(stderr)  -3))
(set! ##console-port (##open-predefined (macro-direction-inout) '(console) -4))
(set! ##ide-port     (##open-predefined (macro-direction-inout) '(ide)     -5))

(set! ##stdin-port ##stdin);;;;;;;;;;;;;;;;
(set! ##stdout-port ##stdout);;;;;;;;;;;;;;;;
(set! ##stderr-port ##stderr);;;;;;;;;;;;;;;;

(define-prim (##force-output-on-predefined)
;(pp '##force-output-on-predefined ##stdout-port);;;;;;;;;;;;;;;;;;;;
  (let ((port ##stdout-port)) (and port (##force-output port)))
  (let ((port ##stderr-port)) (and port (##force-output port))))

(##add-exit-job! ##force-output-on-predefined)

;------------------------------------------------------------------------------

(##define-macro (macro-peek-next-char-or-eof re) ; possibly returns end-of-file
  `(macro-peek-char (macro-readenv-port ,re)))

(##define-macro (macro-read-next-char-or-eof re) ; possibly returns end-of-file
  `(macro-read-char (macro-readenv-port ,re)))

(define-prim (##make-filepos line col char-count)
  (if (and (##fixnum.< line (macro-max-lines))
           (##not (##fixnum.< (macro-max-fixnum32-div-max-lines) col)))
    (##fixnum.+ line (##fixnum.* col (macro-max-lines)))
    (##fixnum.- 0 char-count)))

(define-prim (##filepos-line filepos)
  (if (##fixnum.< filepos 0)
    0
    (##fixnum.modulo filepos (macro-max-lines))))

(define-prim (##filepos-col filepos)
  (if (##fixnum.< filepos 0)
    (##fixnum.- 0 filepos)
    (##fixnum.quotient filepos (macro-max-lines))))

;------------------------------------------------------------------------------

; Implementation of readtables.

(define-fail-check-type readtable)

(define-prim (##readtable? obj)
  (macro-readtable? obj))

(define-prim (readtable? obj)
  (macro-force-vars (obj)
    (macro-readtable? obj)))

(define-prim (case-conversion?-set! rt conversion?)
  (macro-force-vars (rt conversion?)
    (macro-check-readtable rt 1 (case-conversion?-set! rt conversion?)
      (begin
        (##readtable-case-conversion?-set! rt conversion?)
        (##void)))))

(define-prim (keywords-allowed?-set! rt allowed?)
  (macro-force-vars (rt allowed?)
    (macro-check-readtable rt 1 (keywords-allowed?-set! rt allowed?)
      (begin
        (##readtable-keywords-allowed?-set! rt allowed?)
        (##void)))))

;------------------------------------------------------------------------------

(define ##last-serial-number 0)

(define ##serial-number-table (##make-table))

(define-prim (##object->serial-number obj)
  (##declare (not interrupts-enabled))
  (let ((sn (##table-ref ##serial-number-table obj #f)))
    (or sn
        (let* ((x (##fixnum.+ ##last-serial-number 1))
               (n (if (##fixnum.< x 0) 0 x)))
          (set! ##last-serial-number n)
          (##table-set! ##serial-number-table obj n)
          n))))

(define-prim (##serial-number->object sn)
  (##declare (not interrupts-enabled))
  (##table-reverse-ref ##serial-number-table sn #f))

(define-prim (serial-number->object sn)
  (##serial-number->object sn));;;;;;;;;;;;check exact integer, raise exception if not found

;------------------------------------------------------------------------------

(define-prim (##make-marktable)
  (##declare (not interrupts-enabled))
  (##vector -1 '()))

(define-prim (##marktable-mark! table obj)
  (##declare (not interrupts-enabled))
  (let ((alist (##vector-ref table 1)))
    (let ((x (##assq obj alist)));;;;;;;;;;;;;
      (if x
        (begin
          (##set-cdr! x #t)
          #f)
        (begin
          (##vector-set! table 1 (##cons (##cons obj #f) alist))
          #t)))))

(define-prim (##marktable-lookup! table obj stamp?)
  (##declare (not interrupts-enabled))
  (let ((alist (##vector-ref table 1)))
    (let ((x (##assq obj alist)));;;;;;;;;;;;;;;;
      (if x
        (let ((id (##cdr x)))
          (if (and stamp? (##eq? id #t))
            (let ((n (##fixnum.+ (##vector-ref table 0) 1)))
              (##vector-set! table 0 n)
              (##set-cdr! x n)
              x)
            id))
        #f))))

(define-prim (##marktable-save table)
  (##declare (not interrupts-enabled))
  (##vector-ref table 0))

(define-prim (##marktable-restore! table n)
  (##declare (not interrupts-enabled))
  (##vector-set! table 0 n)
  (let ((alist (##vector-ref table 1)))
    (let loop ((lst alist))
      (if (##pair? lst)
        (let* ((x (##car lst))
               (id (##cdr x)))
          (if (and (##fixnum? id)
                   (##fixnum.< n id))
            (##set-cdr! x #t))
          (loop (##cdr lst)))))))

;------------------------------------------------------------------------------

(define-prim (##default-wr we obj)
  (let ((limit (macro-writeenv-limit we)))
    (if (##fixnum.< 0 limit)
      (cond ((##symbol? obj)
             (##wr-symbol we obj))
            ((##keyword? obj)
             (##wr-keyword we obj))
            ((##pair? obj)
             (##wr-pair we obj))
            ((##complex? obj)
             (##wr-complex we obj))
            ((##char? obj)
             (##wr-char we obj))
            ((##string? obj)
             (##wr-string we obj))
            ((##vector? obj)
             (##wr-vector we obj))
            ((##foreign? obj)
             (##wr-foreign we obj))
            ((##procedure? obj)
             (##wr-procedure we obj))
            ((##will? obj)
             (##wr-will we obj))
            ((##promise? obj)
             (##wr-promise we obj))
            ((##s8vector? obj)
             (##wr-s8vector we obj))
            ((##u8vector? obj)
             (##wr-u8vector we obj))
            ((##s16vector? obj)
             (##wr-s16vector we obj))
            ((##u16vector? obj)
             (##wr-u16vector we obj))
            ((##s32vector? obj)
             (##wr-s32vector we obj))
            ((##u32vector? obj)
             (##wr-u32vector we obj))
            ((##s64vector? obj)
             (##wr-s64vector we obj))
            ((##u64vector? obj)
             (##wr-u64vector we obj))
            ((##f32vector? obj)
             (##wr-f32vector we obj))
            ((##f64vector? obj)
             (##wr-f64vector we obj))
            ((##structure? obj)
             (##wr-structure we obj))
            ((##thread? obj)
             (##wr-thread we obj))
            ((##continuation? obj)
             (##wr-continuation we obj))
            ((##frame? obj)
             (##wr-frame we obj))
            ((##return? obj)
             (##wr-return we obj))
            ((##meroon? obj)
             (##wr-meroon we obj))
            ((##box? obj)
             (##wr-box we obj))
            (else
             (##wr-other we obj))))))

(define ##wr #f)
(set! ##wr ##default-wr)

(define-prim (##wr-str we s)
  (##wr-substr we s 0 (##string-length s)))

(define-prim (##wr-substr we s i j)
  (let ((limit (macro-writeenv-limit we)))
    (if (##fixnum.< 0 limit)
      (let ((len (##fixnum.- j i))
            (port (macro-writeenv-port we)))
        (if (##fixnum.< limit len)
          (begin
            (##write-substring s i (##fixnum.+ i limit) port)
            (macro-writeenv-limit-set! we 0))
          (begin
            (##write-substring s i j port)
            (macro-writeenv-limit-set! we (##fixnum.- limit len))))))))

(define-prim (##wr-ch we c)
  (let ((limit (macro-writeenv-limit we)))
    (if (##fixnum.< 0 limit)
      (begin
        (##write-char c (macro-writeenv-port we))
        (macro-writeenv-limit-set! we (##fixnum.- limit 1))))))

(define-prim (##wr-spaces we n)
  (let loop ((i n))
    (if (##fixnum.< 0 i)
      (let ((x (if (##fixnum.< 40 i) 40 i)))
        (##wr-substr we "                                        " 0 x)
        (loop (##fixnum.- i x))))))

(define-prim (##wr-indent we col)
  (##wr-ch we #\newline)
  (##wr-spaces we (##fixnum.- col 1)))

(define-prim (##wr-sn we obj type name)
  (##wr-str we "#<")
  (##wr-str we type)
  (##wr-ch we #\space)
  (##wr-str we (##number->string (##object->serial-number obj) 10))
  (if (##not (##eq? name (##void)))
    (begin
      (##wr-ch we #\space)
      (##wr-no-display we name)))
  (##wr-ch we #\>))

(define-prim (##wr-no-display we obj)
  (if (##eq? (macro-writeenv-style we) 'display)
    (begin
      (macro-writeenv-style-set! we 'write)
      (##wr we obj)
      (macro-writeenv-style-set! we 'display))
    (##wr we obj)))

(define-prim (##wr-mark we obj)
  (let ((mt (macro-writeenv-marktable we)))
    (if mt
      (##marktable-mark! mt obj)
      #t)))

(define-prim (##wr-stamp we obj)
  (let ((mt (macro-writeenv-marktable we)))
    (if mt
      (let ((id (##marktable-lookup! mt obj #t)))
        (if id
          (begin
            (##wr-ch we #\#)
            (if (##fixnum? id)
              (begin
                (##wr-str we (##number->string id 10))
                (##wr-ch we #\#)
                #f)
              (begin
                (##wr-str we (##number->string (##cdr id) 10))
                (##wr-ch we #\=)
                #t)))
          #t))
      #t)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Write methods for each object type

(define-prim (##wr-symbol we obj)
  (let ((uninterned? (##uninterned-symbol? obj)))
    (if (##eq? (macro-writeenv-style we) 'mark)
      (if uninterned?
        (##wr-mark we obj))
      (if (or (##not uninterned?)
              (##wr-stamp we obj))
        (begin
          (if uninterned?
            (##wr-str we "#:"))
          (let ((str (##symbol->string obj)))
            (if (or (##eq? (macro-writeenv-style we) 'display)
                    (##not (##escape-symbol? we str)))
              (##wr-str we str)
              (##wr-escaped-string we str #\|))))))))

(define-prim (##escape-symbol? we str)
  (let ((n (##string-length str)))
    (or (##fixnum.= n 0)
        (and (##fixnum.= n 1)
             (##char=? (##string-ref str 0) #\.))
        (and (##char=? (##string-ref str 0) #\#)
             (or (##fixnum.= n 1)
                 (##not (##char=? (##string-ref str 1) #\#))))
        (##string->number str 10 #t)
        (and (##fixnum.< 1 n)
             (let ((keywords-allowed?
                    (##readtable-keywords-allowed?
                     (macro-writeenv-readtable we))))
               (and keywords-allowed?
                    (##char=? (##string-ref
                               str
                               (if (##eq? keywords-allowed? 'prefix)
                                 0
                                 (##fixnum.- n 1)))
                              #\:))))
        (##escape-symkey? we str))))

(define-prim (##escape-symkey? we str);;;;;;;;;;;;;;;;;;;;;;;;;;
  (let ((n (##string-length str)))
    (let loop ((i (##fixnum.- n 1)))
      (if (##fixnum.< i 0)
        #f
        (let ((c (##string-ref str i)))
          (or (##readtable-char-delimiter? (macro-writeenv-readtable we) c)
              (##not (##char=? c (##readtable-convert-case
                                  (macro-writeenv-readtable we)
                                  c)))
              (loop (##fixnum.- i 1))))))))

(define-prim (##wr-keyword we obj)
  (let ((uninterned? (##uninterned-keyword? obj)))
    (if (##eq? (macro-writeenv-style we) 'mark)
      (if uninterned?
        (##wr-mark we obj))
      (if (or (##not uninterned?)
              (##wr-stamp we obj))
        (begin
          (if uninterned?
            (##wr-str we "#:"))
          (let* ((str
                  (##keyword->string obj))
                 (keywords-allowed?
                  (##readtable-keywords-allowed?
                   (macro-writeenv-readtable we))))
            (if (##eq? keywords-allowed? 'prefix)
              (##wr-ch we #\:))
            (if (or (##eq? (macro-writeenv-style we) 'display)
                    (##not (##escape-keyword? we str)))
              (##wr-str we str)
              (##wr-escaped-string we str #\|))
            (if (##not (##eq? keywords-allowed? 'prefix))
              (##wr-ch we #\:))))))))

(define-prim (##escape-keyword? we str)
  (let ((n (##string-length str)))
    (or (##fixnum.= n 0)
        (and (##char=? (##string-ref str 0) #\#)
             (or (##fixnum.= n 1)
                 (##not (##char=? (##string-ref str 1) #\#)))
             (let ((keywords-allowed?
                    (##readtable-keywords-allowed?
                     (macro-writeenv-readtable we))))
               (##not (##eq? keywords-allowed? 'prefix))))
        (##escape-symkey? we str))))

(define-prim (##wr-pair we obj)

  (define (force-if-required we x)
    (if (macro-writeenv-force? we)
      (##force x)
      x))

  (define (read-macro-prefix we head tail)
    (and (##pair? tail)
         (##null? (force-if-required we (##cdr tail)))
         (let ((mt (macro-writeenv-marktable we)))
           (##not (and mt
                       (##marktable-lookup! mt tail #f))))
         (cond ((##eq? head 'quote)
                "'")
               ((##eq? head 'quasiquote)
                "`")
               ((##eq? head 'unquote)
                ; We have to check that the next character written after
                ; the comma won't be an "@" because the reader would
                ; interpret this as a ",@" readmacro.  The algorithm is
                ; slow but correct and modular.
                (let ((limit (macro-writeenv-limit we)))
                  (if (##fixnum.< 1 limit) ; speed up ",,,,,,xxx" case
                    (let* ((mt
                            (macro-writeenv-marktable we))
                           (state
                            (and mt (##marktable-save mt)))
                           (port
                            (##open-output-string))
                           (we2
                            (##make-writeenv
                             (macro-writeenv-style we)
                             port
                             (macro-writeenv-readtable we)
                             mt
                             (macro-writeenv-force? we)
                             (macro-writeenv-width we)
                             (macro-writeenv-close-parens we)
                             (macro-writeenv-level we)
                             1)))
                      (##wr we2 (##car tail))
                      (if mt (##marktable-restore! mt state))
                      (let ((str (##get-output-string port)))
                        (if (or (##fixnum.< (##string-length str) 1)
                                (##char=? (##string-ref str 0) #\@))
                          ", " ; force a space after the comma
                          ",")))
                    ",")))
               ((##eq? head 'unquote-splicing)
                ",@")
               ((##eq? head
                       (##readtable-sharp-quote-keyword
                        (macro-writeenv-readtable we)))
                "#'")
               (else
                #f))))

  (define (wr-list-possibly-with-read-macro-prefix we obj plain-pretty-print?)
    (let* ((head
            (force-if-required we (##car obj)))
           (tail
            (force-if-required we (##cdr obj)))
           (prefix
            (read-macro-prefix we head tail)))
      (if prefix
        (begin
          (##wr-str we prefix)
          (##wr we (##car tail)))
        (wr-list-using-format
         we
         obj
         (if (##eq? (macro-writeenv-style we) 'pretty-print)
           (if plain-pretty-print?
             plain-format
             (get-format we head tail))
           space-format)))))

  (define space-format
    '#(0 #f 0 #f -1))

  (define plain-format
    '#(0 #f 1 #f 0))

  (define (get-format we head tail)
    (if (##symbol? head)
      (let ((x
             (##assq head
                     (##readtable-pretty-print-formats
                      (macro-writeenv-readtable we)))))
        (cond (x
               (##cdr x))
              ((##eq? head 'let) ; check for named let
               (if (and (##pair? tail)
                        (##symbol? (force-if-required we (##car tail))))
                 '#(2 #t 3 #f 1)
                 '#(1 #t 2 #f 1)))
              (else
               (if (##fixnum.< (##string-length (##symbol->string head)) 8)
                 '#(1 #f 0 #f 1)
                 plain-format))))
      plain-format))

  (define (wr-list-using-format we obj format)

    (##wr-ch we #\()

    (let ((level
           (macro-writeenv-level we)))
      (if (##not (##fixnum.< level
                             (##readtable-max-write-level
                              (macro-writeenv-readtable we))))
        (##wr-str we "...")
        (let* ((close-parens
                (macro-writeenv-close-parens we))
               (new-close-parens
                (##fixnum.+ close-parens 1)))
          (macro-writeenv-level-set! we (##fixnum.+ level 1))
          (let ((start-col
                 (##output-port-column (macro-writeenv-port we))))
            (let loop ((lst obj)
                       (i 0)
                       (col start-col))

              (define (wr-elem elem)
                (let ((new-col
                       (cond ((##fixnum.= i 0)
                              col)
                             ((or (##fixnum.< (##vector-ref format 4) 0)
                                  (##fixnum.< i (##vector-ref format 0)))
                              (##wr-ch we #\space)
                              col)
                             ((##fixnum.= i (##vector-ref format 0))
                              (##wr-ch we #\space)
                              (##output-port-column (macro-writeenv-port we)))
                             ((##fixnum.= i (##vector-ref format 2))
                              (let ((new-col
                                     (##fixnum.+ start-col
                                                 (##vector-ref format 4))))
                                (##wr-indent we new-col)
                                new-col))
                             (else
                              (##wr-indent we col)
                              col))))
                  (if (##pair? elem)
                    (wr-list
                     we
                     elem
                     (if (##fixnum.< i (##vector-ref format 2))
                       (##vector-ref format 1)
                       (##vector-ref format 3)))
                    (##wr we elem))
                  new-col))

              (define (wr-str str)
                (let ((style (macro-writeenv-style we)))
                  (macro-writeenv-style-set! we 'display)
                  (wr-elem str)
                  (macro-writeenv-style-set! we style)))

              (if (##fixnum.< 0 (macro-writeenv-limit we))
                (cond ((##pair? lst)
                       (if (##not (##fixnum.< i
                                              (##readtable-max-write-length
                                               (macro-writeenv-readtable we))))
                         (wr-str "...")
                         (let ((mt (macro-writeenv-marktable we)))
                           (if (and (##fixnum.< 0 i)
                                    mt
                                    (##marktable-lookup! mt lst #f))
                             (begin
                               (wr-str ".")
                               (macro-writeenv-close-parens-set!
                                we
                                new-close-parens)
                               (wr-elem lst))
                             (let* ((head
                                     (force-if-required we (##car lst)))
                                    (tail
                                     (force-if-required we (##cdr lst)))
                                    (prefix
                                     (read-macro-prefix we head tail)))
                               (if prefix
                                 (begin
                                   (wr-str ".")
                                   (wr-str prefix)
                                   (macro-writeenv-close-parens-set!
                                    we
                                    new-close-parens)
                                   (##wr we (##car tail)))
                                 (begin
                                   (macro-writeenv-close-parens-set!
                                    we
                                    (if (##null? tail)
                                      new-close-parens
                                      0))
                                   (loop tail
                                         (##fixnum.+ i 1)
                                         (wr-elem head)))))))))
                      ((##not (##null? lst))
                       (wr-str ".")
                       (macro-writeenv-close-parens-set! we new-close-parens)
                       (wr-elem lst))))))
          (macro-writeenv-level-set! we level)
          (macro-writeenv-close-parens-set! we close-parens))))

    (##wr-ch we #\)))

  (define (wr-list we obj plain-pretty-print?)
    (if (or (##not (##eq? (macro-writeenv-style we) 'pretty-print))
            (##not (##wr-one-line-pretty-print
                    we
                    obj
                    (lambda (we obj)
                      (wr-list-possibly-with-read-macro-prefix
                       we
                       obj
                       plain-pretty-print?)))))
      (wr-list-possibly-with-read-macro-prefix
       we
       obj
       plain-pretty-print?)))

  (case (macro-writeenv-style we)
    ((mark)
     (if (##wr-mark we obj)
       (begin
         (##wr we (##car obj))
         (##wr we (##cdr obj)))))
    ((display)
     (##wr we (##car obj))
     (##wr we (##cdr obj)))
    (else
     (if (##wr-stamp we obj)
       (wr-list we obj #f)))))

(define-prim (##wr-one-line-pretty-print we obj wr-obj)
  (let* ((col
          (##output-port-column (macro-writeenv-port we)))
         (available-space-for-obj
          (##fixnum.- (##fixnum.- (macro-writeenv-width we)
                                  (macro-writeenv-close-parens we))
                      col)))
    (and (##fixnum.< available-space-for-obj (macro-writeenv-limit we))
         (let* ((mt
                 (macro-writeenv-marktable we))
                (state
                 (and mt (##marktable-save mt)))
                (port
                 (##open-output-string))
                (we2
                 (##make-writeenv
                  'write
                  port
                  (macro-writeenv-readtable we)
                  mt
                  (macro-writeenv-force? we)
                  (macro-writeenv-width we)
                  (macro-writeenv-close-parens we)
                  (macro-writeenv-level we)
                  (##fixnum.+ available-space-for-obj 1))))
           (wr-obj we2 obj)
           (let ((str (##get-output-string port)))
             (if (##fixnum.< available-space-for-obj (##string-length str))
               (begin
                 (if mt (##marktable-restore! mt state))
                 #f)
               (begin
                 (##wr-str we str)
                 #t)))))))

(define-prim (##wr-complex we obj)
  (if (##not (##eq? (macro-writeenv-style we) 'mark))
    (##wr-str we (##number->string obj 10))))

(define-prim (##wr-char we obj)
  (case (macro-writeenv-style we)
    ((mark)
     #f)
    ((display)
     (##wr-ch we obj))
    (else
     (let ((x (##assq-cdr obj
                          (##readtable-named-char-table
                           (macro-writeenv-readtable we)))))
       (if x
         (begin
           (##wr-str we "#\\")
           (##wr-str we (##car x)))
         (let ((n (##fixnum.<-char obj)))
           (if (or (##fixnum.< n 32)
                   (##fixnum.= n 127)
                   (##fixnum.< (##readtable-max-unescaped-char
                                (macro-writeenv-readtable we))
                               n))
             (begin
               (##wr-str we "#\\#x")
               (##wr-str we (##number->string n 16)))
             (begin
               (##wr-str we "#\\")
               (##wr-ch we obj)))))))))

(define-prim (##wr-string we obj)
  (case (macro-writeenv-style we)
    ((mark)
     (##wr-mark we obj))
    ((display)
     (##wr-str we obj))
    (else
     (if (##wr-stamp we obj)
       (##wr-escaped-string we obj #\")))))

(define-prim (##wr-escaped-string we s special-escape)
  (##wr-ch we special-escape)
  (let loop ((i 0) (j 0) (escape-digit-limit #f))
    (if (##fixnum.< j (##string-length s))
      (let* ((c
              (##string-ref s j))
             (x
              (cond ((##char=? c special-escape)
                     c)
                    ((##assq-cdr c
                                 (##readtable-escaped-char-table
                                  (macro-writeenv-readtable we)))
                     => ##car)
                    (else
                     #f)))
             (n
              (##fixnum.<-char c))
             (j+1
              (##fixnum.+ j 1)))
        (if (if (or (##fixnum.< n 32) (##fixnum.= n 127))
              (##readtable-escape-ctrl-chars?
               (macro-writeenv-readtable we))
              (or x
                  (##fixnum.< (##readtable-max-unescaped-char
                               (macro-writeenv-readtable we))
                              n)
                  (and escape-digit-limit
                       (##fixnum.< n 128)
                       (##not (##char=? c #\#)) ; avoid treating "#" like "0"
                       (##fixnum.< (##u8vector-ref ##char-to-digit-table n)
                                   escape-digit-limit))))
          (begin
            (##wr-substr we s i j)
            (##wr-ch we #\\)
            (cond (x
                   (##wr-ch we x)
                   (loop j+1 j+1 #f))
                  ((##fixnum.< n 256)
                   (##wr-str we (##number->string n 8))
                   (loop j+1 j+1 (if (##fixnum.< n 64) 8 #f)))
                  (else
                   (##wr-ch we #\x)
                   (##wr-str we (##number->string n 16))
                   (loop j+1 j+1 16))))
          (loop i j+1 #f)))
      (begin
        (##wr-substr we s i j)
        (##wr-ch we special-escape)))))

(define-prim (##wr-vector we obj)
  (##wr-vector-aux1 we obj (##vector-length obj) ##vector-ref "#("))

(define-prim (##wr-vector-aux1 we obj len vect-ref prefix)
  (case (macro-writeenv-style we)
    ((mark)
     (if (##wr-mark we obj)
       (##wr-vector-aux2 we obj len vect-ref)))
    ((display)
     (##wr-vector-aux2 we obj len vect-ref))
    (else
     (if (##wr-stamp we obj)
       (##wr-vector-aux3 we obj len vect-ref prefix)))))

(define-prim (##wr-vector-aux2 we obj len vect-ref)
  (let loop ((i 0))
    (if (##fixnum.< i len)
      (begin
        (##wr we (vect-ref obj i))
        (loop (##fixnum.+ i 1))))))

(define-prim (##wr-vector-aux3 we obj len vect-ref prefix)

  (define (force-if-required we x)
    (if (macro-writeenv-force? we)
      (##force x)
      x))

  (define (wr-vect we obj len vect-ref prefix)

    (##wr-str we prefix)

    (let ((level
           (macro-writeenv-level we)))
      (if (##not (##fixnum.< level
                             (##readtable-max-write-level
                              (macro-writeenv-readtable we))))
        (##wr-str we "...")
        (let* ((close-parens
                (macro-writeenv-close-parens we))
               (new-close-parens
                (##fixnum.+ close-parens 1)))
          (macro-writeenv-level-set! we (##fixnum.+ level 1))
          (let ((start-col
                 (##output-port-column (macro-writeenv-port we))))
            (let loop ((i 0))
              (if (##fixnum.< 0 (macro-writeenv-limit we))
                (if (##fixnum.< i len)
                  (let ()
                    (if (##fixnum.< 0 i)
                      (if (##eq? (macro-writeenv-style we) 'pretty-print)
                        (##wr-indent we start-col)
                        (##wr-ch we #\space)))
                    (if (##not (##fixnum.< i
                                           (##readtable-max-write-length
                                            (macro-writeenv-readtable we))))
                      (##wr-str we "...")
                      (let ((elem
                             (force-if-required we (vect-ref obj i)))
                            (new-i
                             (##fixnum.+ i 1)))
                        (macro-writeenv-close-parens-set!
                         we
                         (if (##fixnum.= new-i len)
                           new-close-parens
                           0))
                        (##wr we elem)
                        (loop new-i))))))))
          (macro-writeenv-level-set! we level)
          (macro-writeenv-close-parens-set! we close-parens))))

    (##wr-ch we #\)))

  (if (or (##not (##eq? (macro-writeenv-style we) 'pretty-print))
          (##not (##wr-one-line-pretty-print
                  we
                  obj
                  (lambda (we obj)
                    (wr-vect we obj len vect-ref prefix)))))
    (wr-vect we obj len vect-ref prefix)))

(define-prim (##wr-foreign we obj)
  (if (##not (##eq? (macro-writeenv-style we) 'mark))
    (let ((typ (##vector-ref obj 0)))
      (##wr-str we "#<")
      (if typ
        (##wr-no-display we typ)
        (##wr-str we "foreign"))
      (##wr-ch we #\space)
      (##wr-str we (##number->string (##object->serial-number obj) 10))
      (##wr-str we " 0x")
      (##wr-str we (##number->string (##foreign-address obj) 16));;;;;;;;;;;;;;;;;;;;;;
      (##wr-ch we #\>))))

(define-prim (##wr-procedure we obj)
  (if (##eq? (##readtable-sharing-allowed?
              (macro-writeenv-readtable we))
             'serialize)
    (if (##closure? obj)
      (if (##eq? (macro-writeenv-style we) 'mark)
        (if (##wr-mark we obj)
          (##wr-vector-aux2 we obj (##vector-length obj) ##vector-ref))
        (if (##wr-stamp we obj)
          (##wr-vector-aux3 we obj (##vector-length obj) ##vector-ref "#closure(")));;;;;;;;;;
      (if (##not (##eq? (macro-writeenv-style we) 'mark))
        (##wr-vector-aux3 we obj 1 (lambda (obj i) (##object->global-var->identifier obj)) "#procedure(")))
    (if (##not (##eq? (macro-writeenv-style we) 'mark))
      (let ((name (##object->global-var->identifier obj)))
        (if name
          (##wr-sn we obj "procedure" name)
          (if (##subprocedure? obj)
            (begin
              (##wr-str we "#<subprocedure ")
              (##wr-str we (##number->string (##subprocedure-id obj) 10))
              (##wr-str we " in ")
              (##wr we (##subprocedure-parent obj))
              (##wr-ch we #\>))
            (##wr-sn we obj "procedure" (##void))))))))

(define-prim (##wr-will we obj)
  (if (##not (##eq? (macro-writeenv-style we) 'mark))
    (##wr-sn we obj "will" (##void))))

(define-prim (##wr-promise we obj)
  (if (macro-writeenv-force? we)
    (##wr we (##force obj))
    (if (##not (##eq? (macro-writeenv-style we) 'mark))
      (##wr-sn we obj "promise" (##void)))))

(define-prim (##wr-s8vector we obj)
  (##wr-vector-aux1 we obj (##s8vector-length obj) ##s8vector-ref "#s8("))

(define-prim (##wr-u8vector we obj)
  (##wr-vector-aux1 we obj (##u8vector-length obj) ##u8vector-ref "#u8("))

(define-prim (##wr-s16vector we obj)
  (##wr-vector-aux1 we obj (##s16vector-length obj) ##s16vector-ref "#s16("))

(define-prim (##wr-u16vector we obj)
  (##wr-vector-aux1 we obj (##u16vector-length obj) ##u16vector-ref "#u16("))

(define-prim (##wr-s32vector we obj)
  (##wr-vector-aux1 we obj (##s32vector-length obj) ##s32vector-ref "#s32("))

(define-prim (##wr-u32vector we obj)
  (##wr-vector-aux1 we obj (##u32vector-length obj) ##u32vector-ref "#u32("))

(define-prim (##wr-s64vector we obj)
  (##wr-vector-aux1 we obj (##s64vector-length obj) ##s64vector-ref "#s64("))

(define-prim (##wr-u64vector we obj)
  (##wr-vector-aux1 we obj (##u64vector-length obj) ##u64vector-ref "#u64("))

(define-prim (##wr-f32vector we obj)
  (##wr-vector-aux1 we obj (##f32vector-length obj) ##f32vector-ref "#f32("))

(define-prim (##wr-f64vector we obj)
  (##wr-vector-aux1 we obj (##f64vector-length obj) ##f64vector-ref "#f64("))

(define-prim (##wr-structure we obj)
  (let* ((tag (##vector-ref obj 0))
         (len (##vector-ref tag 1)))
    (if (##eq? (macro-writeenv-style we) 'mark)
      (if (##wr-mark we obj)
        (##wr-vector-aux2
         we
         obj
         (##fixnum.- len 1)
         (lambda (obj i) (##vector-ref obj (##fixnum.+ i 1)))))
      (if (##wr-stamp we obj)
        (cond ((macro-port? obj)
               (##wr-sn
                we
                obj
                (if (##input-port? obj)
                  (if (##output-port? obj) "input-output-port" "input-port")
                  "output-port")
                (##port-name obj)))
              ((macro-mutex? obj)
               (##wr-sn
                we
                obj
                "mutex"
                (macro-mutex-name obj)))
              ((macro-condvar? obj)
               (##wr-sn
                we
                obj
                "condition-variable"
                (macro-condvar-name obj)))
              ((macro-tgroup? obj)
               (##wr-sn
                we
                obj
                "thread-group"
                (macro-tgroup-name obj)))
;;;;;;;;;;;;;;;;;;;;others????
;;;;;;;;;;;;;;;;;;;;pretty-print?
              ((##fixnum.= len -1)
               (##wr-sn
                we
                obj
                (##symbol->string (##vector-ref tag 0))
                (##void)))
              (else
               (##wr-str we "#s(")
               (##wr-no-display we (##vector-ref tag 0))
               (let loop2 ((i 1))
                 (let ((limit (macro-writeenv-limit we)))
                   (if (##fixnum.< 0 limit)
                     (if (##fixnum.< len i)
                       (##wr-ch we #\))
                       (let ((field (##vector-ref tag (##fixnum.+ i 1)))
                             (value (##vector-ref obj i)))
                         (##wr-str we " (")
                         (##wr-no-display we field)
                         (##wr-ch we #\space)
                         (##wr-no-display we value)
                         (##wr-ch we #\))
                         (loop2 (##fixnum.+ i 1)))))))))))))

(define-prim (##wr-thread we obj)
  (if (##not (##eq? (macro-writeenv-style we) 'mark))
    (##wr-sn we
             obj
             "thread"
             (macro-thread-name obj))))

(define-prim (##wr-continuation we obj)
  (if (##not (##eq? (macro-writeenv-style we) 'mark))
    (##wr-sn we obj "continuation" (##void))))

(define-prim (##wr-frame we obj)
  (if (##not (##eq? (macro-writeenv-style we) 'mark))
    (##wr-sn we obj "frame" (##void))))

(define-prim (##wr-return we obj)
  (if (##not (##eq? (macro-writeenv-style we) 'mark))
    (begin
      (##wr-str we "#<return ")
      (##wr-str we (##number->string (##subprocedure-id obj) 10))
      (##wr-str we " in ")
      (##wr we (##subprocedure-parent obj))
      (##wr-ch we #\>))))

(define-prim (##wr-meroon we obj)
  (if (##not (##eq? (macro-writeenv-style we) 'mark))
    (##wr-sn we obj "meroon" (##void))))

(define-prim (##wr-box we obj)
  (if (##eq? (macro-writeenv-style we) 'mark)
    (if (##wr-mark we obj)
      (##wr we (##unbox obj)))
    (if (or (##eq? (macro-writeenv-style we) 'display)
            (##wr-stamp we obj))
      (begin
        (##wr-str we "#&")
        (##wr we (##unbox obj))))))

(define-prim (##wr-other we obj)
  (if (##not (##eq? (macro-writeenv-style we) 'mark))
    (cond ((##eq? obj #t)
           (##wr-str we "#t"))
          ((##eq? obj #f)
           (##wr-str we "#f"))
          ((##eq? obj '())
           (if (##not (##eq? (macro-writeenv-style we) 'display))
             (##wr-str we "()")))
          ((##unbound? obj)
           (##wr-str we "#<unbound>"))
          ((##eq? obj (##void))
           (##wr-str we "#<void>"))
          (else
           (let ((x
                  (##assq-cdr obj
                              (##readtable-sharp-bang-table
                               (macro-writeenv-readtable we)))))
             (if x
               (begin
                 (##wr-str we "#!")
                 (##wr-str we (##car x)))
               (##wr-str we "#<unknown>")))))))

;------------------------------------------------------------------------------

(define-prim (##current-input-port)
  (macro-current-input-port))

(define-prim (current-input-port)
  (macro-current-input-port))

(define-prim (##current-output-port)
  (macro-current-output-port))

(define-prim (current-output-port)
  (macro-current-output-port))

(define-prim (current-error-port)
  (macro-current-error-port))

(define-prim (current-readtable)
  (macro-current-readtable))

(define ##main-readtable #f)

;------------------------------------------------------------------------------

; IEEE Scheme procedures:

(define-prim (##eof-object? x)
  (##eq? x #!eof))

(define-prim (eof-object? x)
  (macro-force-vars (x)
    (##eof-object? x)))

;------------------------------------------------------------------------------

; R4RS Scheme procedures:

(define-prim (transcript-on path)
  (macro-check-string path 1 (transcript-on path)
    (##void)))

(define-prim (transcript-off)
  (##void))

;------------------------------------------------------------------------------

(define-prim (file-exists? path)
  (macro-force-vars (path)
    (macro-check-string path 1 (file-exists? path)
      (let ((port (##open-input-file path)))
        (if port
          (begin
            (##close-input-port port)
            #t)
          #f)))))

;------------------------------------------------------------------------------

; The reader.

(##declare (inlining-limit 300))

(##define-macro (* . args)                `(##fixnum.* ,@args))
(##define-macro (+ . args)                `(##fixnum.+ ,@args))
(##define-macro (- . args)                `(##fixnum.- ,@args))
(##define-macro (< . args)                `(##fixnum.< ,@args))
(##define-macro (assoc . args)            `(##assoc ,@args))
(##define-macro (assq . args)             `(##assq ,@args))
(##define-macro (car . args)              `(##car ,@args))
(##define-macro (cdr . args)              `(##cdr ,@args))
(##define-macro (char-downcase . args)    `(##char-downcase ,@args))
(##define-macro (char-upcase . args)      `(##char-upcase ,@args))
(##define-macro (char<? . args)           `(##char<? ,@args))
(##define-macro (char=? . args)           `(##char=? ,@args))
(##define-macro (char? . args)            `(##char? ,@args))
(##define-macro (cons . args)             `(##cons ,@args))
(##define-macro (eq? . args)              `(##eq? ,@args))
(##define-macro (exact? . args)           `(##exact? ,@args))
(##define-macro (for-each . args)         `(##for-each ,@args))
(##define-macro (integer? . args)         `(##integer? ,@args))
(##define-macro (list . args)             `(##list ,@args))
(##define-macro (make-string . args)      `(##make-string ,@args))
(##define-macro (make-vector . args)      `(##make-vector ,@args))
(##define-macro (modulo . args)           `(##modulo ,@args))
(##define-macro (not . args)              `(##not ,@args))
(##define-macro (null? . args)            `(##null? ,@args))
(##define-macro (pair? . args)            `(##pair? ,@args))
(##define-macro (quotient . args)         `(##quotient ,@args))
(##define-macro (real? . args)            `(##real? ,@args))
(##define-macro (reverse . args)          `(##reverse ,@args))
(##define-macro (box . args)              `(##box ,@args))
(##define-macro (set-box! . args)         `(##set-box! ,@args))
(##define-macro (set-car! . args)         `(##set-car! ,@args))
(##define-macro (set-cdr! . args)         `(##set-cdr! ,@args))
(##define-macro (string->number . args)   `(##string->number ,@args))
(##define-macro (string->symbol-object . args)   `(##make-interned-symkey ,@args #t))
(##define-macro (string->uninterned-symbol-object . args)   `(##make-uninterned-symbol ,@args))
(##define-macro (string-length . args)    `(##string-length ,@args))
(##define-macro (string-append . args)    `(##string-append ,@args))
(##define-macro (string-ref . args)       `(##string-ref ,@args))
(##define-macro (string-set! . args)      `(##string-set! ,@args))
(##define-macro (string-ci=? . args)      `(##string-ci=? ,@args))
(##define-macro (substring . args)        `(##substring ,@args))
(##define-macro (symbol->string . args)   `(##symbol->string ,@args))
(##define-macro (vector . args)           `(##vector ,@args))
(##define-macro (vector-ref . args)       `(##vector-ref ,@args))
(##define-macro (vector-set! . args)      `(##vector-set! ,@args))
(##define-macro (vector? . args)          `(##vector? ,@args))

(##define-macro (make-s8vect n)           `(##make-s8vector ,n))
(##define-macro (s8vect-set! . args)      `(##s8vector-set! ,@args))
(##define-macro (make-u8vect n)           `(##make-u8vector ,n))
(##define-macro (u8vect-set! . args)      `(##u8vector-set! ,@args))
(##define-macro (make-s16vect n)          `(##make-s16vector ,n))
(##define-macro (s16vect-set! . args)     `(##s16vector-set! ,@args))
(##define-macro (make-u16vect n)          `(##make-u16vector ,n))
(##define-macro (u16vect-set! . args)     `(##u16vector-set! ,@args))
(##define-macro (make-s32vect n)          `(##make-s32vector ,n))
(##define-macro (s32vect-set! . args)     `(##s32vector-set! ,@args))
(##define-macro (make-u32vect n)          `(##make-u32vector ,n))
(##define-macro (u32vect-set! . args)     `(##u32vector-set! ,@args))
(##define-macro (make-s64vect n)          `(##make-s64vector ,n))
(##define-macro (s64vect-set! . args)     `(##s64vector-set! ,@args))
(##define-macro (make-u64vect n)          `(##make-u64vector ,n))
(##define-macro (u64vect-set! . args)     `(##u64vector-set! ,@args))
(##define-macro (make-f32vect n)          `(##make-f32vector ,n))
(##define-macro (f32vect-set! . args)     `(##f32vector-set! ,@args))
(##define-macro (make-f64vect n)          `(##make-f64vector ,n))
(##define-macro (f64vect-set! . args)     `(##f64vector-set! ,@args))

(##define-macro (ucs4->character . args)  `(##fixnum.->char ,@args))
(##define-macro (character->ucs4 . args)  `(##fixnum.<-char ,@args))
(##define-macro (in-char-range? n)        `(##not (##< ##max-char ,n)))

(##define-macro (string->keyword-object . args) `(##make-interned-symkey ,@args #f))
(##define-macro (string->uninterned-keyword-object . args) `(##make-uninterned-keyword ,@args))

(##define-macro (in-integer-range? n lo hi)
  `(and (##not (##< ,n ,lo)) (##not (##< ,hi ,n))))

(##define-macro (false-obj) #f)

; Tables for reader and writer.

(define ##standard-escaped-char-table
  '((#\\ . #\\)
    (#\a . #\bel)
    (#\b . #\backspace)
    (#\t . #\tab)
    (#\n . #\newline)
    (#\v . #\vt)
    (#\f . #\page)
    (#\r . #\return)))

(define ##standard-named-char-table
  '(("newline"   . #\newline) ; here to take precedence over linefeed
    ("space"     . #\space)
    ("nul"       . #\nul)
    ("bel"       . #\bel)
    ("backspace" . #\backspace)
    ("tab"       . #\tab)
    ("linefeed"  . #\linefeed)
    ("vt"        . #\vt)
    ("page"      . #\page)
    ("return"    . #\return)
    ("rubout"    . #\rubout)))

(define ##standard-sharp-bang-table
  '(("optional" . #!optional)
    ("rest"     . #!rest)
    ("key"      . #!key)
    ("eof"      . #!eof)
    (""         . #!)))

(define ##standard-pretty-print-formats
  '((lambda         . #(1 #t 2 #f 1))
    (if             . #(1 #f 0 #f 1))
    (set!           . #(1 #f 0 #f 1))
    (cond           . #(1 #t 0 #t 1))
    (case           . #(1 #f 2 #t 1))
    (and            . #(1 #f 0 #f 1))
    (or             . #(1 #f 0 #f 1))
;    (let            . #(1 #t 2 #f 1)) ; let is handled in pretty printer
    (let*           . #(1 #t 2 #f 1))
    (letrec         . #(1 #t 2 #f 1))
    (begin          . #(0 #f 1 #f 1))
    (do             . #(1 #t 3 #f 1))
    (define         . #(1 #f 2 #f 1))
    (##define-macro . #(1 #f 2 #f 1))
    (define-macro   . #(1 #f 2 #f 1))
    (declare        . #(0 #f 1 #f 1))
    (##declare      . #(0 #f 1 #f 1))
    (declare        . #(0 #f 1 #f 1))
))

;==============================================================================

; For compatibility between the interpreter and compiler, this section
; must be the same as the corresponding section in the file
; "gsc/_source.scm" (except that ## and ** are exchanged).

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; A chartable structure is a vector-like data structure which is
; indexed using a character.

(define (##make-chartable default)
  (vector (make-vector 128 default) default '()))

(define (##chartable-ref ct c)
  (let ((i (character->ucs4 c)))
    (if (< i 128)
      (vector-ref (vector-ref ct 0) i)
      (let ((x (assq i (vector-ref ct 2))))
        (if x
          (cdr x)
          (vector-ref ct 1))))))

(define (##chartable-set! ct c val)
  (let ((i (character->ucs4 c)))
    (if (< i 128)
      (vector-set! (vector-ref ct 0) i val)
      (let ((x (assq i (vector-ref ct 2))))
        (if x
          (set-cdr! x val)
          (vector-set! ct 2 (cons (cons i val) (vector-ref ct 2))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; A readtable structure contains parsing information for the reader.
; It indicates what action must be taken when a given character is
; encountered.

(define ##readtable-tag
  '#(readtable -1))

(define (##make-readtable
         case-conversion?
         keywords-allowed?
         escaped-char-table
         named-char-table
         sharp-bang-table
         char-delimiter?-table
         char-handler-table
         char-sharp-handler-table
         max-unescaped-char
         escape-ctrl-chars?
         sharing-allowed?
         eval-allowed?
         max-write-level
         max-write-length
         pretty-print-formats
         sharp-quote-keyword
         infix-type?)
  (##subtype-set!
    (vector
     ##readtable-tag
     case-conversion?
     keywords-allowed?
     escaped-char-table
     named-char-table
     sharp-bang-table
     char-delimiter?-table
     char-handler-table
     char-sharp-handler-table
     max-unescaped-char
     escape-ctrl-chars?
     sharing-allowed?
     eval-allowed?
     max-write-level
     max-write-length
     pretty-print-formats
     sharp-quote-keyword
     infix-type?)
    (macro-subtype-structure)))

(define (##readtable-case-conversion? rt)
  (vector-ref rt 1))

(define (##readtable-case-conversion?-set! rt x)
  (vector-set! rt 1 x))

(define (##readtable-keywords-allowed? rt)
  (vector-ref rt 2))

(define (##readtable-keywords-allowed?-set! rt x)
  (vector-set! rt 2 x))

(define (##readtable-escaped-char-table rt)
  (vector-ref rt 3))

(define (##readtable-escaped-char-table-set! rt x)
  (vector-set! rt 3 x))

(define (##readtable-named-char-table rt)
  (vector-ref rt 4))

(define (##readtable-named-char-table-set! rt x)
  (vector-set! rt 4 x))

(define (##readtable-sharp-bang-table rt)
  (vector-ref rt 5))

(define (##readtable-sharp-bang-table-set! rt x)
  (vector-set! rt 5 x))

(define (##readtable-char-delimiter?-table rt)
  (vector-ref rt 6))

(define (##readtable-char-delimiter?-table-set! rt x)
  (vector-set! rt 6 x))

(define (##readtable-char-handler-table rt)
  (vector-ref rt 7))

(define (##readtable-char-handler-table-set! rt x)
  (vector-set! rt 7 x))

(define (##readtable-char-sharp-handler-table rt)
  (vector-ref rt 8))

(define (##readtable-char-sharp-handler-table-set! rt x)
  (vector-set! rt 8 x))

(define (##readtable-max-unescaped-char rt)
  (vector-ref rt 9))

(define (##readtable-max-unescaped-char-set! rt x)
  (vector-set! rt 9 x))

(define (##readtable-escape-ctrl-chars? rt)
  (vector-ref rt 10))

(define (##readtable-escape-ctrl-chars?-set! rt x)
  (vector-set! rt 10 x))

(define (##readtable-sharing-allowed? rt)
  (vector-ref rt 11))

(define (##readtable-sharing-allowed?-set! rt x)
  (vector-set! rt 11 x))

(define (##readtable-eval-allowed? rt)
  (vector-ref rt 12))

(define (##readtable-eval-allowed?-set! rt x)
  (vector-set! rt 12 x))

(define (##readtable-max-write-level rt)
  (vector-ref rt 13))

(define (##readtable-max-write-level-set! rt x)
  (vector-set! rt 13 x))

(define (##readtable-max-write-length rt)
  (vector-ref rt 14))

(define (##readtable-max-write-length-set! rt x)
  (vector-set! rt 14 x))

(define (##readtable-pretty-print-formats rt)
  (vector-ref rt 15))

(define (##readtable-pretty-print-formats-set! rt x)
  (vector-set! rt 15 x))

(define (##readtable-sharp-quote-keyword rt)
  (vector-ref rt 16))

(define (##readtable-sharp-quote-keyword-set! rt x)
  (vector-set! rt 16 x))

(define (##readtable-infix-type? rt)
  (vector-ref rt 17))

(define (##readtable-infix-type?-set! rt x)
  (vector-set! rt 17 x))

(define (##readtable-char-delimiter? rt c)
  (##chartable-ref (##readtable-char-delimiter?-table rt) c))

(define (##readtable-char-delimiter?-set! rt c delimiter?)
  (##chartable-set! (##readtable-char-delimiter?-table rt) c delimiter?))

(define (##readtable-char-handler rt c)
  (##chartable-ref (##readtable-char-handler-table rt) c))

(define (##readtable-char-handler-set! rt c handler)
  (##chartable-set! (##readtable-char-handler-table rt) c handler))

(define (##readtable-char-sharp-handler rt c)
  (##chartable-ref (##readtable-char-sharp-handler-table rt) c))

(define (##readtable-char-sharp-handler-set! rt c handler)
  (##chartable-set! (##readtable-char-sharp-handler-table rt) c handler))

(define (##readtable-char-class-set! rt c delimiter? handler)
  (##readtable-char-delimiter?-set! rt c delimiter?)
  (##readtable-char-handler-set! rt c handler))

(define (##readtable-convert-case rt c)
  (let ((case-conversion? (##readtable-case-conversion? rt)))
    (if case-conversion?
      (if (eq? case-conversion? 'upcase)
        (char-upcase c)
        (char-downcase c))
      c)))

(define (##readtable-string-convert-case! rt s)
  (let ((case-conversion? (##readtable-case-conversion? rt)))
    (if case-conversion?
      (if (eq? case-conversion? 'upcase)
        (let loop ((i (- (string-length s) 1)))
          (if (not (< i 0))
            (begin
              (string-set! s i (char-upcase (string-ref s i)))
              (loop (- i 1)))))
        (let loop ((i (- (string-length s) 1)))
          (if (not (< i 0))
            (begin
              (string-set! s i (char-downcase (string-ref s i)))
              (loop (- i 1)))))))))

(define (##readtable-parse-keyword rt s intern? create?)
  (let ((keywords-allowed? (##readtable-keywords-allowed? rt)))
    (and keywords-allowed?
         (let ((len (string-length s)))
           (and (< 1 len)
                (if (eq? keywords-allowed? 'prefix)
                  (and (char=? (string-ref s 0) #\:)
                       (if create?
                         (let ((key-str (substring s 1 len)))
                           (if intern?
                             (string->keyword-object key-str)
                             (string->uninterned-keyword-object key-str)))
                         #t))
                  (and (char=? (string-ref s (- len 1)) #\:)
                       (if create?
                         (let ((key-str (substring s 0 (- len 1))))
                           (if intern?
                             (string->keyword-object key-str)
                             (string->uninterned-keyword-object key-str)))
                         #t))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Procedures to read datums.

; (##read-datum-or-eof re) attempts to read a datum in the read
; environment "re", skipping all whitespace and comments in the
; process.  The "filepos" field of the read environment indicates the
; position where the enclosing datum starts (e.g. list or vector).  If
; a datum is read it is returned (wrapped if the read environment asks
; for it); if the end-of-file is reached the end-of-file object is
; returned (never wrapped); otherwise an error is signaled.  The read
; environment's "pos" field is only modified if a datum was read, in
; which case it is the position where the datum starts.

(define (##read-datum-or-eof re)
  (let* ((old-pos (macro-readenv-filepos re))
         (obj (##read-datum-or-label-or-none re)))
    (if (eq? obj (##none-marker))
      (let ((c (macro-peek-next-char-or-eof re)))
        (if (char? c)
          (begin
            (macro-readenv-filepos-set! re (##readenv-current-filepos re))
            (macro-read-next-char-or-eof re) ; make sure reader makes progress
            (##raise-datum-parsing-exception 'datum-or-eof-expected re)
            (macro-readenv-filepos-set! re old-pos) ; restore pos
            (##read-datum-or-eof re)) ; skip error
          (begin
            (macro-read-next-char-or-eof re) ; make sure reader makes progress
            #!eof))) ; end-of-file was reached so return end-of-file object
      (begin
        (##read-check-labels! re)
        obj))))

; (##read-datum-or-label re) attempts to read a datum in the read
; environment "re", skipping all whitespace and comments in the
; process.  The "filepos" field of the read environment indicates the
; position where the enclosing datum starts (e.g. list or vector).  If
; a datum is read it is returned (wrapped if the read environment asks
; for it); if a label reference is read (i.e. "#n#") a "label-marker"
; is returned; if the end-of-file is reached or no datum can be read
; an error is signaled.  The read environment's "filepos" field is
; only modified if a datum was read, in which case it is the position
; where the datum starts.

(define (##read-datum-or-label re)
  (let* ((old-pos (macro-readenv-filepos re))
         (obj (##read-datum-or-label-or-none re)))
    (if (eq? obj (##none-marker))
      (begin
        (macro-readenv-filepos-set! re (##readenv-current-filepos re))
        (let ((c (macro-read-next-char-or-eof re))) ; force progress
          (##raise-datum-parsing-exception 'datum-expected re)
          (if (not (char? c))
            (macro-readenv-wrap re #f) ; return something
            (begin
              (macro-readenv-filepos-set! re old-pos) ; restore pos
              (##read-datum-or-label re))))) ; skip error
      obj)))

; (##read-datum-or-label-or-none re) attempts to read a datum in the
; read environment "re", skipping all whitespace and comments in the
; process.  The "filepos" field of the read environment indicates the
; position where the enclosing datum starts (e.g. list or vector).  If
; a datum is read it is returned (wrapped if the read environment asks
; for it); if a label reference is read (i.e. "#n#") a "label-marker"
; is returned; if the end-of-file is reached or no datum can be read
; the "none-marker" is returned.  The read environment's "filepos"
; field is only modified if a datum was read, in which case it is the
; position where the datum starts.

(define (##read-datum-or-label-or-none re)
  (let* ((old-pos (macro-readenv-filepos re))
         (obj (##read-datum-or-label-or-none-or-dot re)))
    (if (eq? obj (##dot-marker))
      (begin
        (macro-readenv-filepos-set! re (##readenv-relative-filepos re 1))
        (##raise-datum-parsing-exception 'improperly-placed-dot re)
        (macro-readenv-filepos-set! re old-pos) ; restore pos
        (##read-datum-or-label-or-none re)) ; skip error
      obj)))

; (##read-datum-or-label-or-none-or-dot re) attempts to read a datum
; in the read environment "re", skipping all whitespace and comments
; in the process.  The "filepos" field of the read environment
; indicates the position where the enclosing datum starts (e.g. list
; or vector).  If a datum is read it is returned (wrapped if the read
; environment asks for it); if a label reference is read (i.e. "#n#")
; a "label-marker" is returned; if a lone dot is read the "dot-marker"
; is returned; if the end-of-file is reached or no datum can be read
; the "none-marker" is returned.  The read environment's "filepos"
; field is only modified if a datum was read, in which case it is the
; position where the datum starts.

(define (##read-datum-or-label-or-none-or-dot re)
  (let ((next (macro-peek-next-char-or-eof re)))
    (if (char? next)
      ((##readtable-char-handler (macro-readenv-readtable re) next) re next)
      (##none-marker))))

; Special objects returned by ##read-datum-or-label-or-none-or-dot.

(define (##none-marker) '#(none))     ; indicates no following datum
(define (##dot-marker) '#(dot))       ; indicates an isolated dot
(define ##label-marker-tag '#(label)) ; indicates a label of the form "#n#"

(define (##label-marker? obj)
  (and (vector? obj)
       (< 0 (vector-length obj))
       (eq? (vector-ref obj 0) ##label-marker-tag)))

(define (##label-marker-enter! re n)
  (let* ((labels (macro-readenv-labels re))
         (x (assoc n labels)))
    (if x
      (cdr x)
      (let ((lm (vector ##label-marker-tag #f '())))
        (macro-readenv-labels-set! re (cons (cons n lm) labels))
        lm))))

(define (##label-marker-reference re n)
  (let* ((lm (##label-marker-enter! re n))
         (handlers (vector-ref lm 2)))
    (if handlers
      lm
      (vector-ref lm 1))))

(define (##label-marker-define re n obj)
  (if (##label-marker? obj)
    (##raise-datum-parsing-exception 'illegal-label-definition re n)
    (let* ((lm (##label-marker-enter! re n))
           (handlers (vector-ref lm 2)))
      (if handlers
        (begin
          (vector-set! lm 1 obj)
          (vector-set! lm 2 #f)
          (let loop ((lst handlers))
            (if (pair? lst)
              (let ((handler (car lst)))
                (handler obj)
                (loop (cdr lst))))))
        (##raise-datum-parsing-exception 'duplicate-label-definition re n)))))

(define (##label-marker-fixup-handler-add! lm handler)
  (let ((handlers (vector-ref lm 2)))
    (if handlers
      (vector-set! lm 2 (cons handler handlers))
      (handler (vector-ref lm 1)))))

(define (##read-check-labels! re)
  (let loop1 ((lst (macro-readenv-labels re)))
    (if (pair? lst)
      (let* ((x (car lst))
             (lm (cdr x)))
        (let ((handlers (vector-ref lm 2)))
          (if handlers
            (begin
              (##raise-datum-parsing-exception
               'missing-label-definition
               re
               (car x))
              (let loop2 ((lst handlers))
                (if (pair? lst)
                  (let ((handler (car lst)))
                    (handler #f)
                    (loop2 (cdr lst))))))))
        (loop1 (cdr lst))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Procedure to read a list of datums (possibly an improper list).

(define (##build-list re allow-improper? start-pos close)
  (let ((obj (##read-datum-or-label-or-none re)))
    (if (eq? obj (##none-marker))
      (begin
        (##read-next-char-expecting re close)
        '())
      (begin
        (macro-readenv-filepos-set! re start-pos) ; restore pos
        (let ((lst (cons obj '())))
          (if (##label-marker? obj)
            (##label-marker-fixup-handler-add!
             obj
             (lambda (resolved-obj)
               (set-car! lst resolved-obj))))
          (let loop ((end lst))
            (let ((obj
                   (if allow-improper?
                     (##read-datum-or-label-or-none-or-dot re)
                     (##read-datum-or-label-or-none re))))
              (cond ((eq? obj (##none-marker))
                     (##read-next-char-expecting re close)
                     lst)
                    ((eq? obj (##dot-marker))
                     (let ((obj (##read-datum-or-label re)))
                       (macro-readenv-filepos-set! re start-pos) ; restore pos
                       (set-cdr! end obj)
                       (if (##label-marker? obj)
                         (##label-marker-fixup-handler-add!
                          obj
                          (lambda (resolved-obj)
                            (set-cdr! end resolved-obj))))
                       (let ((x (##read-datum-or-label-or-none re))) ; skip whitespace!
                         (if (eq? x (##none-marker))
                           (##read-next-char-expecting re close)
                           (begin
                             (macro-readenv-filepos-set! re start-pos) ; restore pos
                             (##raise-datum-parsing-exception 'incomplete-form re)))
                         lst)))
                    (else
                     (macro-readenv-filepos-set! re start-pos) ; restore pos
                     (let ((tail (cons obj '())))
                       (if (##label-marker? obj)
                         (##label-marker-fixup-handler-add!
                          obj
                          (lambda (resolved-obj)
                            (set-car! tail resolved-obj))))
                       (set-cdr! end tail)
                       (loop tail)))))))))))

(define (##read-next-char-expecting re c) ; only accepts c as the next char
  (let ((next (macro-peek-next-char-or-eof re)))
    (if (char? next)
      (if (char=? next c)
        (macro-read-next-char-or-eof re)
        (##raise-datum-parsing-exception 'incomplete-form re))
      (##raise-datum-parsing-exception 'incomplete-form-eof-reached re))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Procedure to read a vector or byte vector.

(define (##build-vector re kind start-pos close)

  (define (exact-integer-check n lo hi)
    (and (integer? n)
         (exact? n)
         (in-integer-range? n lo hi)))

  (define (inexact-real-check n)
    (and (real? n)
         (not (exact? n))))

  (let loop ((i 0))
    (macro-readenv-filepos-set! re start-pos) ; restore pos
    (let ((x (##read-datum-or-label-or-none re)))
      (if (eq? x (##none-marker))
        (begin
          (##read-next-char-expecting re close)
          (case kind
            ((vector)    (make-vector i))
            ((s8vector)  (make-s8vect i))
            ((u8vector)  (make-u8vect i))
            ((s16vector) (make-s16vect i))
            ((u16vector) (make-u16vect i))
            ((s32vector) (make-s32vect i))
            ((u32vector) (make-u32vect i))
            ((s64vector) (make-s64vect i))
            ((u64vector) (make-u64vect i))
            ((f32vector) (make-f32vect i))
            ((f64vector) (make-f64vect i))))
        (if (eq? kind 'vector)
          (let ((vect (loop (+ i 1))))
            (vector-set! vect i x)
            (if (##label-marker? x)
              (##label-marker-fixup-handler-add!
               x
               (lambda (resolved-obj)
                 (vector-set! vect i resolved-obj))))
            vect)
          (let ((ux
                 (and (not (##label-marker? x))
                      (macro-readenv-unwrap re x))))
            (case kind
              ((s8vector)
               (if (exact-integer-check ux -128 127)
                 (let ((vect (loop (+ i 1))))
                   (s8vect-set! vect i ux)
                   vect)
                 (begin
                   (##raise-datum-parsing-exception 's8-expected re)
                   (loop i))))
              ((u8vector)
               (if (exact-integer-check ux 0 255)
                 (let ((vect (loop (+ i 1))))
                   (u8vect-set! vect i ux)
                   vect)
                 (begin
                   (##raise-datum-parsing-exception 'u8-expected re)
                   (loop i))))
              ((s16vector)
               (if (exact-integer-check ux -32768 32767)
                 (let ((vect (loop (+ i 1))))
                   (s16vect-set! vect i ux)
                   vect)
                 (begin
                   (##raise-datum-parsing-exception 's16-expected re)
                   (loop i))))
              ((u16vector)
               (if (exact-integer-check ux 0 65535)
                 (let ((vect (loop (+ i 1))))
                   (u16vect-set! vect i ux)
                   vect)
                 (begin
                   (##raise-datum-parsing-exception 'u16-expected re)
                   (loop i))))
              ((s32vector)
               (if (exact-integer-check ux -2147483648 2147483647)
                 (let ((vect (loop (+ i 1))))
                   (s32vect-set! vect i ux)
                   vect)
                 (begin
                   (##raise-datum-parsing-exception 's32-expected re)
                   (loop i))))
              ((u32vector)
               (if (exact-integer-check ux 0 4294967295)
                 (let ((vect (loop (+ i 1))))
                   (u32vect-set! vect i ux)
                   vect)
                 (begin
                   (##raise-datum-parsing-exception 'u32-expected re)
                   (loop i))))
              ((s64vector)
               (if (exact-integer-check ux -9223372036854775808 9223372036854775807)
                 (let ((vect (loop (+ i 1))))
                   (s64vect-set! vect i ux)
                   vect)
                 (begin
                   (##raise-datum-parsing-exception 's64-expected re)
                   (loop i))))
              ((u64vector)
               (if (exact-integer-check ux 0 18446744073709551615)
                 (let ((vect (loop (+ i 1))))
                   (u64vect-set! vect i ux)
                   vect)
                 (begin
                   (##raise-datum-parsing-exception 'u64-expected re)
                   (loop i))))
              ((f32vector)
               (if (inexact-real-check ux)
                 (let ((vect (loop (+ i 1))))
                   (f32vect-set! vect i ux)
                   vect)
                 (begin
                   (##raise-datum-parsing-exception 'inexact-real-expected re)
                   (loop i))))
              ((f64vector)
               (if (inexact-real-check ux)
                 (let ((vect (loop (+ i 1))))
                   (f64vect-set! vect i ux)
                   vect)
                 (begin
                   (##raise-datum-parsing-exception 'inexact-real-expected re)
                   (loop i)))))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Procedures to read delimited tokens.

(define (##build-delimited-string re c i)
  (let loop ((i i))
    (let ((next (macro-peek-next-char-or-eof re)))
      (if (or (not (char? next))
              (##readtable-char-delimiter? (macro-readenv-readtable re) next))
        (make-string i c)
        (begin
          (macro-read-next-char-or-eof re) ; skip "next"
          (let ((s (loop (+ i 1))))
            (string-set! s i next)
            s))))))

(define (##build-delimited-number/keyword/symbol re c i intern?)

  (define (string->sym str)
    (if intern?
      (string->symbol-object str)
      (string->uninterned-symbol-object str)))

  (define (string->key str)
    (if intern?
      (string->keyword-object str)
      (string->uninterned-keyword-object str)))

  (cond ((and (char=? c #\|)
              (eq? i 1))
         (let* ((str
                 (##build-escaped-string-up-to re #\|))
                (keywords-allowed?
                 (##readtable-keywords-allowed?
                  (macro-readenv-readtable re))))
           (if (and keywords-allowed?
                    (not (eq? keywords-allowed? 'prefix))
                    (eq? (macro-peek-next-char-or-eof re) #\:))
             (begin
               (macro-read-next-char-or-eof re) ; skip #\:
               (string->key str))
             (string->sym str))))
        ((and (char=? c #\:)
              (eq? i 1)
              (let ((keywords-allowed?
                     (##readtable-keywords-allowed?
                      (macro-readenv-readtable re))))
                (eq? keywords-allowed? 'prefix))
              (eq? (macro-peek-next-char-or-eof re) #\|))
         (macro-read-next-char-or-eof re) ; skip #\|
         (let ((str
                (##build-escaped-string-up-to re #\|)))
           (string->key str)))
        (else
         (let ((str
                (##build-delimited-string re c i)))
           (or (and intern? (string->number str 10))
               (begin
                 (##readtable-string-convert-case!
                  (macro-readenv-readtable re)
                  str)
                 (or (##readtable-parse-keyword
                      (macro-readenv-readtable re)
                      str
                      intern?
                      #t)
                     (string->sym str))))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##build-escaped-string-up-to re close)

  (define (char-octal? c)
    (and (not (char<? c #\0)) (not (char<? #\7 c))))

  (define (char-hexadecimal? c)
    (or (and (not (char<? c #\0)) (not (char<? #\9 c)))
        (and (not (char<? c #\a)) (not (char<? #\f c)))
        (and (not (char<? c #\A)) (not (char<? #\F c)))))

  (define (ucs4 n)
    (if (in-char-range? n)
      (ucs4->character n)
      (begin
        (##raise-datum-parsing-exception 'character-out-of-range re)
        #\nul)))

  (define (read-escape-octal c)
    (let ((str (let loop ((i 1))
                 (let ((next (macro-peek-next-char-or-eof re)))
                   (if (and (< i 3)
                            (char? next)
                            (char-octal? next))
                     (begin
                       (macro-read-next-char-or-eof re) ; skip "next"
                       (let ((s (loop (+ i 1))))
                         (string-set! s i next)
                         s))
                     (make-string i))))))
      (string-set! str 0 c)
      (ucs4 (string->number str 8))))

  (define (read-escape-hexadecimal)
    (let ((next (macro-peek-next-char-or-eof re)))
      (if (and (char? next)
               (char-hexadecimal? next))
        (begin
          (macro-read-next-char-or-eof re) ; skip "next"
          (let ((str (let loop ((i 1))
                       (let ((next2 (macro-peek-next-char-or-eof re)))
                         (if (and (char? next2)
                                  (char-hexadecimal? next2))
                           (begin
                             (macro-read-next-char-or-eof re) ; skip "next2"
                             (let ((s (loop (+ i 1))))
                               (string-set! s i next2)
                               s))
                           (make-string i))))))
            (string-set! str 0 next)
            (ucs4 (string->number str 16))))
        (begin
          (##raise-datum-parsing-exception 'invalid-hex-escape re)
          #\nul))))

  (define (read-escape)
    (let ((next (macro-read-next-char-or-eof re)))
      (cond ((not (char? next))
             ; read-chunk will report the end-of-file error
             #\nul)
            ((char-octal? next)
             (read-escape-octal next))
            ((char=? next #\x)
             (read-escape-hexadecimal))
            ((char=? next close)
             close)
            (else
             (let ((x (assq next
                            (##readtable-escaped-char-table
                             (macro-readenv-readtable re)))))
               (if x
                 (cdr x)
                 (begin
                   (##raise-datum-parsing-exception 'invalid-escaped-character re next)
                   #\nul)))))))

  (define max-chunk-length 512)

  (define (read-chunk)
    (let loop ((i 0))
      (if (< i max-chunk-length)
        (let ((c (macro-read-next-char-or-eof re)))
          (cond ((not (char? c))
                 (##raise-datum-parsing-exception 'incomplete-form-eof-reached re)
                 (make-string i))
                ((char=? c close)
                 (make-string i))
                ((char=? c #\\)
                 (let* ((c (read-escape))
                        (s (loop (+ i 1))))
                   (string-set! s i c)
                   s))
                (else
                 (let ((s (loop (+ i 1))))
                   (string-set! s i c)
                   s))))
        (make-string i))))

  (let ((chunk1 (read-chunk)))
    (if (< (string-length chunk1) max-chunk-length)
      chunk1
      (let loop ((chunks (list chunk1)))
        (let* ((new-chunk (read-chunk))
               (new-chunks (cons new-chunk chunks)))
          (if (< (string-length new-chunk) max-chunk-length)
            (##append-strings (reverse new-chunks))
            (loop new-chunks)))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##build-decimal-integer re c i)
  (let loop ((i i))
    (let ((next (macro-peek-next-char-or-eof re)))
      (if (or (not (char? next))
              (let ((n (character->ucs4 next)))
                (not (and (< 47 n) (< n 58)))))
        (make-string i c)
        (begin
          (macro-read-next-char-or-eof re) ; skip "next"
          (let ((s (loop (+ i 1))))
            (string-set! s i next)
            s))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##build-read-macro re start-pos kind)
  (let ((obj (##read-datum-or-label re)))
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let* ((cell2 (cons obj '()))
           (cell1 (cons (macro-readenv-wrap re kind) cell2)))
      (if (##label-marker? obj)
        (##label-marker-fixup-handler-add!
         obj
         (lambda (resolved-obj)
           (set-car! cell2 resolved-obj))))
      (macro-readenv-wrap re cell1))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Procedures to handle comments.

(define (##skip-extended-comment re open1 open2 close1 close2)
  (let loop1 ((level 0))
    (let ((c (macro-read-next-char-or-eof re)))
      (if (not (char? c))
        (##raise-datum-parsing-exception 'incomplete-form-eof-reached re)
        (let loop2 ((level level) (c c))
          (if (or (char=? c open1) (char=? c close1))
            (let ((x (macro-read-next-char-or-eof re)))
              (if (not (char? x))
                (##raise-datum-parsing-exception 'incomplete-form-eof-reached re)
                (if (char=? c open1)
                  (if (char=? x open2)
                    (loop1 (+ level 1))
                    (loop2 level x))
                  (if (char=? x close2)
                    (if (< 0 level)
                      (loop1 (- level 1))
                      #f) ; comment has ended
                    (loop2 level x)))))
            (loop1 level)))))))

(define (##skip-single-line-comment re)
  (let loop ()
    (let ((next (macro-peek-next-char-or-eof re)))
      (if (char? next)
        (begin
          (macro-read-next-char-or-eof re) ; skip "next"
          (if (not (char=? next #\newline))
            (loop)))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Procedure to read datums starting with '#'.

(define (##read-sharp re c)
  (let* ((old-pos (macro-readenv-filepos re))
         (start-pos (##readenv-current-filepos re)))
    (macro-read-next-char-or-eof re) ; skip #\#
    (let ((next (macro-peek-next-char-or-eof re)))
      (if (char? next)
        ((##readtable-char-sharp-handler (macro-readenv-readtable re) next)
         re
         next
         start-pos)
        (begin
          (macro-readenv-filepos-set! re start-pos) ; set pos to error
          (##raise-datum-parsing-exception 'incomplete-form-eof-reached re)
          (macro-readenv-filepos-set! re old-pos) ; restore pos
          (##read-datum-or-label-or-none-or-dot re)))))) ; skip error

(define (##read-sharp-vector re next start-pos)
  (macro-read-next-char-or-eof re) ; skip char after #\#
  (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
  (let ((vect (##build-vector re 'vector start-pos #\))))
    (macro-readenv-wrap re vect)))

(define (##read-sharp-char re next start-pos)
  (let ((old-pos (macro-readenv-filepos re)))
    (macro-read-next-char-or-eof re) ; skip char after #\#
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let ((c (macro-read-next-char-or-eof re)))
      (if (not (char? c))
        (begin
          (##raise-datum-parsing-exception 'incomplete-form-eof-reached re)
          (macro-readenv-filepos-set! re old-pos) ; restore pos
          (##read-datum-or-label-or-none-or-dot re)) ; skip error
        (let ((next (macro-peek-next-char-or-eof re)))
          (if (or (not (char? next))
                  (##readtable-char-delimiter?
                   (macro-readenv-readtable re)
                   next))
            (macro-readenv-wrap re c)
            (let ((name (##build-delimited-string re c 1)))
              (let ((x (##read-assoc-string-ci=?
                        name
                        (##readtable-named-char-table
                         (macro-readenv-readtable re)))))
                (if x
                  (macro-readenv-wrap re (cdr x))
                  (let ((n (and (char=? (string-ref name 0) #\#)
                                (string->number name 10))))
                    (if (or (not n)
                            (not (integer? n))
                            (not (exact? n)))
                      (begin
                        (##raise-datum-parsing-exception
                         'invalid-character-name
                         re
                         name)
                        (macro-readenv-filepos-set! re old-pos) ; restore pos
                        (##read-datum-or-label-or-none-or-dot re)) ; skip error
                      (if (not (in-char-range? n))
                        (begin
                          (##raise-datum-parsing-exception
                           'character-out-of-range
                           re)
                          (macro-readenv-filepos-set! re old-pos) ; restore pos
                          (##read-datum-or-label-or-none-or-dot re)) ; skip error
                        (macro-readenv-wrap re (ucs4->character n))))))))))))))

(define (##read-sharp-comment re next start-pos)
  (let ((old-pos (macro-readenv-filepos re)))
    (macro-readenv-filepos-set! re start-pos) ; in case error in comment
    (macro-read-next-char-or-eof re) ; skip char after #\#
    (##skip-extended-comment re #\# next next #\#)
    (macro-readenv-filepos-set! re old-pos) ; restore pos
    (##read-datum-or-label-or-none-or-dot re))) ; read what follows comment

(define (##read-sharp-bang re next start-pos)
  (let ((old-pos (macro-readenv-filepos re)))
    (macro-read-next-char-or-eof re) ; skip char after #\#
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let ((name (##build-delimited-string re #\space 0)))
      (let ((x (##read-assoc-string-ci=?
                name
                (##readtable-sharp-bang-table
                 (macro-readenv-readtable re)))))
        (if x
          (macro-readenv-wrap re (cdr x))
          (begin
            (##raise-datum-parsing-exception 'invalid-sharp-bang-name re name)
            (macro-readenv-filepos-set! re old-pos) ; restore pos
            (##read-datum-or-label-or-none-or-dot re))))))) ; skip error

(define (##read-sharp-sharp re next start-pos)
  (macro-read-next-char-or-eof re) ; skip char after #\#
  (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
  (let ((obj (##build-delimited-number/keyword/symbol re #\# 2 #t)))
    (macro-readenv-wrap re obj)))

(define (##read-sharp-colon re next start-pos)
  (let ((old-pos (macro-readenv-filepos re)))
    (macro-read-next-char-or-eof re) ; skip char after #\#
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let ((c (macro-read-next-char-or-eof re)))
      (if (char? c)
        (let ((obj (##build-delimited-number/keyword/symbol re c 1 #f)))
          (macro-readenv-wrap re obj))
        (begin
          (##raise-datum-parsing-exception 'incomplete-form-eof-reached re)
          (macro-readenv-filepos-set! re old-pos) ; restore pos
          (##read-datum-or-label-or-none-or-dot re)))))) ; skip error

(define (##read-sharp-quotation re next start-pos)
  (macro-read-next-char-or-eof re) ; skip char after #\#
  (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
  (##build-read-macro
   re
   start-pos
   (##readtable-sharp-quote-keyword (macro-readenv-readtable re))))

(define (##read-sharp-ampersand re next start-pos)
  (macro-read-next-char-or-eof re) ; skip char after #\#
  (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
  (let ((obj (##read-datum-or-label re)))
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let ((b (box obj)))
      (if (##label-marker? obj)
        (##label-marker-fixup-handler-add!
         obj
         (lambda (resolved-obj)
           (set-box! b resolved-obj))))
      (macro-readenv-wrap re b))))

(define (##read-sharp-dot re next start-pos)
  (if (not (##readtable-eval-allowed? (macro-readenv-readtable re)))
    (begin
      (##raise-datum-parsing-exception 'invalid-token re "#.")
      (##read-datum-or-label-or-none-or-dot re)) ; skip error
    (begin
      (macro-read-next-char-or-eof re) ; skip char after #\#
      (let* ((expr
              (##read-expr-from-port
               (macro-readenv-port re)))
             (val
              (##eval expr)))
        (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
        (macro-readenv-wrap re val)))))

(define (##read-sharp-digit re next start-pos)
  (let ((old-pos (macro-readenv-filepos re)))
    (macro-read-next-char-or-eof re) ; skip char after #\#
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let ((s (##build-decimal-integer re next 2)))
      (string-set! s 0 #\0)
      (let* ((n (string->number s 10))
             (c (macro-peek-next-char-or-eof re)))
        (cond ((or (and (not (eq? c #\#))
                        (not (eq? c #\=)))
                   (not (##readtable-sharing-allowed?
                         (macro-readenv-readtable re))))
               (##wrap-op1 re
                           start-pos
                           'serial-number->object
                           n))
              ((eq? c #\#)
               (macro-read-next-char-or-eof re) ; skip #\#
               (##label-marker-reference re n))
              (else
               (macro-read-next-char-or-eof re) ; skip #\=
               (let ((obj (##read-datum-or-label re)))
                 (##label-marker-define re n obj)
                 obj)))))))

(define (##wrap re pos datum)
  (let ((old-pos (macro-readenv-filepos re)))
    (macro-readenv-filepos-set! re pos)
    (let ((x (macro-readenv-wrap re datum)))
      (macro-readenv-filepos-set! re old-pos) ; restore pos
      x)))

(define (##wrap-op re pos op args)
  (##wrap re pos (cons (##wrap re pos op) args)))

(define (##wrap-op0 re pos op)
  (##wrap-op re pos op '()))

(define (##wrap-op1 re pos op arg1)
  (##wrap-op re pos op (list arg1)))

(define (##wrap-op2 re pos op arg1 arg2)
  (##wrap-op re pos op (list arg1 arg2)))

(define (##wrap-op3 re pos op arg1 arg2 arg3)
  (##wrap-op re pos op (list arg1 arg2 arg3)))

(define (##wrap-op4 re pos op arg1 arg2 arg3 arg4)
  (##wrap-op re pos op (list arg1 arg2 arg3 arg4)))

(define (##read-sharp-other re next start-pos)
  (let ((old-pos (macro-readenv-filepos re)))
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let* ((s
            (##build-delimited-string re #\# 1))
           (num
            (string->number s 10)))
      (if num

        (macro-readenv-wrap re num)

        (let ()

          (define (build-vect re kind)
            (let ((c (macro-read-next-char-or-eof re)))
              (if (eq? c #\()
                (macro-readenv-wrap re (##build-vector re kind start-pos #\)))
                (begin
                  (##raise-datum-parsing-exception 'open-paren-expected re)
                  (macro-readenv-filepos-set! re old-pos) ; restore pos
                  (##read-datum-or-label-or-none-or-dot re))))) ; skip error

          (cond ((string-ci=? s "#f")
                 (macro-readenv-wrap re (false-obj)))
                ((string-ci=? s "#t")
                 (macro-readenv-wrap re #t))
                ((string-ci=? s "#s8")
                 (build-vect re 's8vector))
                ((string-ci=? s "#u8")
                 (build-vect re 'u8vector))
                ((string-ci=? s "#s16")
                 (build-vect re 's16vector))
                ((string-ci=? s "#u16")
                 (build-vect re 'u16vector))
                ((string-ci=? s "#s32")
                 (build-vect re 's32vector))
                ((string-ci=? s "#u32")
                 (build-vect re 'u32vector))
                ((string-ci=? s "#s64")
                 (build-vect re 's64vector))
                ((string-ci=? s "#u64")
                 (build-vect re 'u64vector))
                ((string-ci=? s "#f32")
                 (build-vect re 'f32vector))
                ((string-ci=? s "#f64")
                 (build-vect re 'f64vector))
                ((string-ci=? s "#closure");;;;;;;;;;;;;;;;;;
                 (macro-readenv-wrap
                  re
                  (##list
                   (macro-readenv-wrap re 'closure)
                   (build-vect re 'vector))))
                ((string-ci=? s "#procedure");;;;;;;;;;;;;;;;;;
                 (macro-readenv-wrap
                  re
                  (##list
                   (macro-readenv-wrap re 'procedure)
                   (build-vect re 'vector))))
                (else
                 (##raise-datum-parsing-exception 'invalid-token re s)
                 (macro-readenv-filepos-set! re old-pos) ; restore pos
                 (##read-datum-or-label-or-none-or-dot re)))))))) ; skip error

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##read-whitespace re c)
  (macro-read-next-char-or-eof re) ; skip whitespace character
  (##read-datum-or-label-or-none-or-dot re)) ; read what follows whitespace

(define (##read-single-line-comment re c)
  (##skip-single-line-comment re) ; skip comment
  (##read-datum-or-label-or-none-or-dot re)) ; read what follows comment

(define (##read-escaped-string re c)
  (let ((start-pos (##readenv-current-filepos re)))
    (macro-read-next-char-or-eof re) ; skip #\"
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let ((str (##build-escaped-string-up-to re c)))
      (macro-readenv-wrap re str))))

(define (##read-quotation re c)
  (let ((start-pos (##readenv-current-filepos re)))
    (macro-read-next-char-or-eof re) ; skip #\'
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (##build-read-macro re start-pos 'quote)))

(define (##read-quasiquotation re c)
  (let ((start-pos (##readenv-current-filepos re)))
    (macro-read-next-char-or-eof re) ; skip #\`
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (##build-read-macro re start-pos 'quasiquote)))

(define (##read-unquotation re c)
  (let ((start-pos (##readenv-current-filepos re)))
    (macro-read-next-char-or-eof re) ; skip #\,
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let ((next (macro-peek-next-char-or-eof re)))
      (if (eq? next #\@)
        (begin
          (macro-read-next-char-or-eof re) ; skip #\@
          (##build-read-macro re start-pos 'unquote-splicing))
        (##build-read-macro re start-pos 'unquote)))))

(define (##read-list re c)
  (let ((start-pos (##readenv-current-filepos re)))
    (macro-read-next-char-or-eof re) ; skip #\( or #\[ or #\{ or #\<
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let ((close
           (cond ((char=? c #\[) #\])
                 ((char=? c #\{) #\})
                 ((char=? c #\<) #\>)
                 (else           #\)))))
      (let ((lst (##build-list re #t start-pos close)))
        (macro-readenv-wrap re lst)))))

(define (##read-none re c)
  (##none-marker))

(define (##read-illegal re c)
  (let* ((old-pos (macro-readenv-filepos re))
         (start-pos (##readenv-current-filepos re)))
    (macro-read-next-char-or-eof re) ; skip illegal character
    (macro-readenv-filepos-set! re start-pos) ; set pos to illegal char
    (##raise-datum-parsing-exception 'illegal-character re c)
    (macro-readenv-filepos-set! re old-pos) ; restore pos
    (##read-datum-or-label-or-none-or-dot re))) ; skip error

(define (##read-dot re c)
  (let ((start-pos (##readenv-current-filepos re)))
    (macro-read-next-char-or-eof re) ; skip #\.
    (let ((next (macro-peek-next-char-or-eof re)))
      (if (or (not (char? next))
              (##readtable-char-delimiter? (macro-readenv-readtable re) next))
        (##dot-marker)
        (begin
          (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
          (let ((obj (##build-delimited-number/keyword/symbol re c 1 #t)))
            (macro-readenv-wrap re obj)))))))

(define (##read-number/keyword/symbol re c)
  (let ((start-pos (##readenv-current-filepos re)))
    (macro-read-next-char-or-eof re) ; skip "c"
    (macro-readenv-filepos-set! re start-pos) ; set pos to start of datum
    (let ((obj (##build-delimited-number/keyword/symbol re c 1 #t)))
      (macro-readenv-wrap re obj))))

(define (##read-assoc-string-ci=? x lst)
  (let loop ((lst lst))
    (if (pair? lst)
      (let ((couple (car lst)))
        (let ((y (car couple)))
          (if (string-ci=? x y)
            couple
            (loop (cdr lst)))))
      #f)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Infix parser.

(define (##read-infix re c)

  (##define-macro (define-infix-token name . params)
    `(define ,name
       '#(,@params)))

  (##define-macro (define-infix-op
                   name
                   precedence
                   associativity
                   . scheme-names)
    `(define-infix-token ,name
       ,(+ (* precedence 2) (if (eq? associativity 'lr) 0 1))
       ,@scheme-names))

  (define (op? x)
    (and (vector? x)
         (< 1 (vector-length x))))

  (define (precedence op)
    (quotient (vector-ref op 0) 2))

  (define (left-to-right? op)
    (= (modulo (vector-ref op 0) 2) 0))

  (define (binary-or-ternary? op)
    (vector-ref op 1))

  (define (ternary? tok)
    (cond ((eq? tok op.?)
           |op.:|)
          (else
           #f)))

  (define (unary-prefix? op)
    (and (< 2 (vector-length op))
         (vector-ref op 2)))

  (define (unary-postfix? op)
    (and (< 3 (vector-length op))
         (vector-ref op 3)))

  (define-infix-token |token.(|  -1)
  (define-infix-token |token.)|  -2)
  (define-infix-token |token.->| -3)
  (define-infix-token |token..|  -4)
  (define-infix-token |token...| -5)
  (define-infix-token |token.;|  -6)
  (define-infix-token |token.[|  -7)
  (define-infix-token |token.\\| -8)
  (define-infix-token |token.]|  -9)
  (define-infix-token |token.{|  -10)
  (define-infix-token |token.}|  -11)

  (define-infix-op op.!      2  rl #f             infix.!x           )
  (define-infix-op op.++     2  rl #f             infix.++x infix.x++)
  (define-infix-op op.--     2  rl #f             infix.--x infix.x--)
  (define-infix-op op.~      2  rl #f             infix.~x           )
  (define-infix-op op.%      3  lr infix.x%y                         )
  (define-infix-op op.*      3  lr infix.x*y      infix.*x           )
  (define-infix-op op./      3  lr infix.x/y                         )
  (define-infix-op op.+      4  lr infix.x+y      infix.+x           )
  (define-infix-op op.-      4  lr infix.x-y      infix.-x           )
  (define-infix-op op.<<     5  lr infix.x<<y                        )
  (define-infix-op op.>>     5  lr infix.x>>y                        )
  (define-infix-op op.<      6  lr infix.x<y                         )
  (define-infix-op op.<=     6  lr infix.x<=y                        )
  (define-infix-op op.>      6  lr infix.x>y                         )
  (define-infix-op op.>=     6  lr infix.x>=y                        )
  (define-infix-op op.!=     7  lr infix.x!=y                        )
  (define-infix-op op.==     7  lr infix.x==y                        )
  (define-infix-op op.&      8  lr infix.x&y      infix.&x           )
  (define-infix-op op.^      9  lr infix.x^y                         )
  (define-infix-op |op.\||   10 lr |infix.x\|y|                      )
  (define-infix-op op.&&     11 lr infix.x&&y                        )
  (define-infix-op |op.\|\|| 12 lr |infix.x\|\|y|                    )
  (define-infix-op op.?      13 rl infix.x?y:z                       )
  (define-infix-op |op.:|    14 rl infix.x:y                         )
  (define-infix-op op.%=     15 rl infix.x%=y                        )
  (define-infix-op op.&=     15 rl infix.x&=y                        )
  (define-infix-op op.*=     15 rl infix.x*=y                        )
  (define-infix-op op.+=     15 rl infix.x+=y                        )
  (define-infix-op op.-=     15 rl infix.x-=y                        )
  (define-infix-op op./=     15 rl infix.x/=y                        )
  (define-infix-op op.<<=    15 rl infix.x<<=y                       )
  (define-infix-op op.=      15 rl infix.x=y                         )
  (define-infix-op op.>>=    15 rl infix.x>>=y                       )
  (define-infix-op op.^=     15 rl infix.x^=y                        )
  (define-infix-op |op.\|=|  15 rl |infix.x\|=y|                     )
  (define-infix-op op.:=     16 rl infix.x:=y                        )
  (define-infix-op |op.,|    17 lr |infix.x,y|                       )
  (define-infix-op op.:-     18 rl infix.x:-y                        )

  (define max-precedence 18)

  (define (decimal-digit? c)
    (and (not (char<? c #\0)) (not (char<? #\9 c))))

  (define (alphabetic? c)
    (or (and (not (char<? c #\a)) (not (char<? #\z c)))
        (and (not (char<? c #\A)) (not (char<? #\Z c)))))

  (define (identifier-starter? c)
    (or (char=? c #\_)
        (alphabetic? c)))

  (define (parse-number re c1 c2)
    (let ((str
           (let loop ((i 2) (state (if (char=? c1 #\.) 1 0)))
             (let ((next (macro-peek-next-char-or-eof re)))
               (if (or (not (char? next))
                       (not (or (decimal-digit? next)
                                (and (= state 0)
                                     (char=? next #\.))
                                (and (= state 1)
                                     (or (char=? next #\e)
                                         (char=? next #\E)))
                                (and (= state 2)
                                     (or (char=? next #\+)
                                         (char=? next #\-))))))
                 (make-string i c2)
                 (begin
                   (macro-read-next-char-or-eof re) ; skip "next"
                   (let ((s
                          (loop (+ i 1)
                                (if (or (= state 2)
                                        (not (decimal-digit? next)))
                                  (+ state 1)
                                  state))))
                     (string-set! s i next)
                     s)))))))
      (string-set! str 0 c1)
      (let ((last (string-ref str (- (string-length str) 1))))
        (if (or (char=? last #\.) (decimal-digit? last))
          (string->number str)
          (begin
            (invalid-infix-syntax-number re)
            (macro-inexact-+0))))))

  (define (parse-character re c)
    (let ((str
           (##build-escaped-string-up-to re c)))
      (if (= (string-length str) 1)
        (string-ref str 0)
        (begin
          (invalid-infix-syntax-character re)
          #\nul))))

  (define (parse-identifier re c)
    (let ((str
           (let loop ((i 1))
             (let ((next (macro-peek-next-char-or-eof re)))
               (if (or (not (char? next))
                       (and (not (alphabetic? next))
                            (not (decimal-digit? next))
                            (not (char=? next #\_))))
                 (make-string i c)
                 (begin
                   (macro-read-next-char-or-eof re) ; skip "next"
                   (let ((s (loop (+ i 1))))
                     (string-set! s i next)
                     s)))))))
      (string->symbol str)))

  (define (parse-token re)
    (parse-token-starting-with re (macro-peek-next-char-or-eof re)))

  (define (parse-token-starting-with re c)
    (cond ((not (char? c))
           (##none-marker))
          ((eq? (##readtable-char-handler (macro-readenv-readtable re) c)
                ##read-whitespace)
           (macro-read-next-char-or-eof re) ; skip whitespace character
           (parse-token re))
          (else
           (let ((start-pos (##readenv-current-filepos re)))

             (macro-readenv-filepos-set! re start-pos) ; restore pos

             (if (or (char=? c #\+)
                     (char=? c #\-)
                     (char=? c #\*)
                     (char=? c #\/)
                     (char=? c #\%)
                     (char=? c #\!)
                     (char=? c #\~)
                     (char=? c #\&)
                     (char=? c #\|)
                     (char=? c #\^)
                     (char=? c #\<)
                     (char=? c #\>)
                     (char=? c #\=)
                     (char=? c #\{)
                     (char=? c #\})
                     (char=? c #\()
                     (char=? c #\))
                     (char=? c #\[)
                     (char=? c #\])
                     (char=? c #\;)
                     (char=? c #\,)
                     (char=? c #\:)
                     (char=? c #\.)
                     (char=? c #\?)
                     (char=? c #\\)
                     (char=? c #\")
                     (char=? c #\')
                     (decimal-digit? c)
                     (identifier-starter? c))
               (begin
                 (macro-read-next-char-or-eof re) ; skip c
                 (let ((next (macro-peek-next-char-or-eof re)))
                   (let ((x (if (char? next) next #\space)))

                     (define (token tok)
                       tok)

                     (define (one-char-token tok)
                       (token tok))

                     (define (two-char-token tok)
                       (macro-read-next-char-or-eof re) ; skip last
                       (token tok))

                     (cond ((char=? c #\+)
                            (cond ((char=? x #\+)
                                   (two-char-token op.++))
                                  ((char=? x #\=)
                                   (two-char-token op.+=))
                                  (else
                                   (one-char-token op.+))))
                           ((char=? c #\-)
                            (cond ((char=? x #\-)
                                   (two-char-token op.--))
                                  ((char=? x #\=)
                                   (two-char-token op.-=))
                                  ((char=? x #\>)
                                   (two-char-token |token.->|))
                                  (else
                                   (one-char-token op.-))))
                           ((char=? c #\*)
                            (cond ((char=? x #\=)
                                   (two-char-token op.*=))
                                  (else
                                   (one-char-token op.*))))
                           ((char=? c #\/)
                            (cond ((char=? x #\/)
                                   (macro-read-next-char-or-eof re); skip #\/
                                   (##skip-single-line-comment re)
                                   (parse-token re))
                                  ((char=? x #\=)
                                   (two-char-token op./=))
                                  (else
                                   (one-char-token op./))))
                           ((char=? c #\%)
                            (cond ((char=? x #\=)
                                   (two-char-token op.%=))
                                  (else
                                   (one-char-token op.%))))
                           ((char=? c #\!)
                            (cond ((char=? x #\=)
                                   (two-char-token op.!=))
                                  (else
                                   (one-char-token op.!))))
                           ((char=? c #\~)
                            (one-char-token op.~))
                           ((char=? c #\&)
                            (cond ((char=? x #\&)
                                   (two-char-token op.&&))
                                  ((char=? x #\=)
                                   (two-char-token op.&=))
                                  (else
                                   (one-char-token op.&))))
                           ((char=? c #\|)
                            (cond ((char=? x #\|)
                                   (two-char-token |op.\|\||))
                                  ((char=? x #\=)
                                   (two-char-token |op.\|=|))
                                  (else
                                   (one-char-token |op.\||))))
                           ((char=? c #\^)
                            (cond ((char=? x #\=)
                                   (two-char-token op.^=))
                                  (else
                                   (one-char-token op.^))))
                           ((char=? c #\<)
                            (cond ((char=? x #\<)
                                   (macro-read-next-char-or-eof re); skip #\<
                                   (let ((next2
                                          (macro-peek-next-char-or-eof re)))
                                     (let ((x2 (if (char? next2)
                                                 next2
                                                 #\space)))
                                       (cond ((char=? x2 #\=)
                                              (two-char-token op.<<=))
                                             (else
                                              (one-char-token op.<<))))))
                                  ((char=? x #\=)
                                   (two-char-token op.<=))
                                  (else
                                   (one-char-token op.<))))
                           ((char=? c #\>)
                            (cond ((char=? x #\>)
                                   (macro-read-next-char-or-eof re); skip #\>
                                   (let ((next2
                                          (macro-peek-next-char-or-eof re)))
                                     (let ((x2 (if (char? next2)
                                                 next2
                                                 #\space)))
                                       (cond ((char=? x2 #\=)
                                              (two-char-token op.>>=))
                                             (else
                                              (one-char-token op.>>))))))
                                  ((char=? x #\=)
                                   (two-char-token op.>=))
                                  (else
                                   (one-char-token op.>))))
                           ((char=? c #\=)
                            (cond ((char=? x #\=)
                                   (two-char-token op.==))
                                  (else
                                   (one-char-token op.=))))
                           ((char=? c #\{)
                            (one-char-token |token.{|))
                           ((char=? c #\})
                            (one-char-token |token.}|))
                           ((char=? c #\()
                            (one-char-token |token.(|))
                           ((char=? c #\))
                            (one-char-token |token.)|))
                           ((char=? c #\[)
                            (one-char-token |token.[|))
                           ((char=? c #\])
                            (one-char-token |token.]|))
                           ((char=? c #\;)
                            (one-char-token |token.;|))
                           ((char=? c #\,)
                            (one-char-token |op.,|))
                           ((char=? c #\:)
                            (cond ((char=? x #\=)
                                   (two-char-token op.:=))
                                  ((char=? x #\-)
                                   ; In the C syntax "1?2:-3" is parsed as
                                   ; "1 ? 2 : -3".  In order to support the
                                   ; ":-" operator cleanly, the source code
                                   ; must have whitespace between the ":"
                                   ; and "-", otherwise it will be parsed
                                   ; as "1 ? 2 :- 3" (which is a syntax error).
                                   (two-char-token op.:-))
                                  (else
                                   (one-char-token |op.:|))))
                           ((char=? c #\.)
                            (cond ((decimal-digit? x)
                                   (macro-read-next-char-or-eof re) ; skip x
                                   (token (parse-number re c x)))
                                  ((char=? x #\.)
                                   (two-char-token |token...|))
                                  (else
                                   (one-char-token |token..|))))
                           ((char=? c #\?)
                            (one-char-token op.?))
                           ((char=? c #\\)
                            (one-char-token |token.\\|))
                           ((char=? c #\")
                            (token (##build-escaped-string-up-to re c)))
                           ((char=? c #\')
                            (token (parse-character re c)))
                           ((decimal-digit? c)
                            (token (parse-number re #\0 c)))
                           (else
                            (token (parse-identifier re c)))))))
               (##none-marker))))))

  (define (get-token re maybe-tok)
    (or maybe-tok (parse-token re)))

  (define (expect re maybe-tok expected)
    (let ((tok (get-token re maybe-tok)))
      (if (eq? tok expected)
        #f
        (begin
          (invalid-infix-syntax re)
          tok))))

  (define (read-arguments-tail re maybe-tok cont)
    (let ((tok (get-token re maybe-tok)))
      (if (eq? tok |token.)|)
        (cont re
              #f
              '())
        (let loop ((re re) (tok tok) (args '()))
          (cond ((expression-starter? re tok)
                 (read-expression
                  re
                  tok
                  max-precedence
                  'no-comma
                  (lambda (re maybe-tok expr)
                    (let ((new-args (cons expr args)))
                      (let ((tok (get-token re maybe-tok)))
                        (cond ((eq? tok |op.,|)
                               (loop re
                                     (get-token re #f)
                                     new-args))
                              (else
                               (cont re
                                     (expect re tok |token.)|)
                                     (reverse new-args)))))))))
                (else
                 (invalid-infix-syntax re)
                 (cont re
                       tok
                       (reverse args))))))))

  (define (read-list re maybe-tok start-pos cont)
    (let loop ((re re) (maybe-tok maybe-tok) (first? #t) (cont cont))
      (let ((tok (get-token re maybe-tok)))
        (cond ((expression-starter? re tok)
               (read-expression
                re
                tok
                max-precedence
                'no-comma-and-no-bar
                (lambda (re maybe-tok expr1)
                  (let ((tok (get-token re maybe-tok)))
                    (cond ((eq? tok |op.,|)
                           (loop re
                                 #f
                                 #f
                                 (lambda (re maybe-tok expr2)
                                   (cont re
                                         maybe-tok
                                         (##wrap-op2 re
                                                     start-pos
                                                     'infix.list
                                                     expr1
                                                     expr2)))))
                          ((eq? tok |op.\||)
                           (read-expression
                            re
                            #f
                            max-precedence
                            #f
                            (lambda (re maybe-tok expr2)
                              (cont re
                                    (expect re maybe-tok |token.]|)
                                    (##wrap-op2 re
                                                start-pos
                                                'infix.cons
                                                expr1
                                                expr2)))))
                          (else
                           (cont re
                                 (expect re tok |token.]|)
                                 (##wrap-op2 re
                                             start-pos
                                             'infix.list
                                             expr1
                                             (##wrap-op0 re
                                                         start-pos
                                                         'infix.null)))))))))
              (else
               (cont re
                     (expect re tok |token.]|)
                     (##wrap-op0 re
                                 start-pos
                                 'infix.null)))))))

  (define (expression-starter? re tok)
    ; this function must be kept in sync with "read-expression"
    (or (eq? tok op.*)
        (eq? tok op.+)
        (eq? tok op.-)
        (eq? tok op.&)
        (level2-starter? re tok)))

  (define (level2-starter? re tok)
    (or (eq? tok op.!)
        (eq? tok op.++)
        (eq? tok op.--)
        (eq? tok op.~)
        (primary-starter? re tok)))

  (define (primary-starter? re tok)
    (or (eq? tok |token.(|)
        (eq? tok |token.[|)
        (eq? tok |token.\\|)
        (symbol? tok)
        (string? tok)
        (char? tok)
        (number? tok)
        (pair? tok)
        (infix-type? re tok)))

  (define (read-expression re maybe-tok level restriction cont)
    (let* ((tok
            (get-token re maybe-tok))
           (start-pos
            (macro-readenv-filepos re)))
      (cond ((and (= level 2)
                  (op? tok)
                  (unary-prefix? tok))
             =>
             (lambda (scheme-name)
               (let ((tok (get-token re #f)))
                 (if (and (eq? tok |op.!|)
                          (not (level2-starter? re tok)))
                   (cont re
                         tok
                         (##wrap-op0 re
                                     start-pos
                                     'infix.cut))
                   (read-expression
                    re
                    tok
                    2
                    restriction
                    (lambda (re maybe-tok expr)
                      (cont re
                            maybe-tok
                            (##wrap-op1 re
                                        start-pos
                                        scheme-name
                                        expr))))))))
            ((and (= level 2)
                  (eq? tok 'new))
             (read-identifier
              re
              #f
              (lambda (re maybe-tok identifier)
                (read-arguments-tail
                 re
                 (expect re #f |token.(|)
                 (lambda (re maybe-tok args)
                   (cont re
                         maybe-tok
                         (##wrap-op re
                                    start-pos
                                    'infix.new
                                    (cons identifier
                                          args))))))))
            ((< 0 level)
             (read-expression
              re
              tok
              (- level 1)
              restriction
              (lambda (re maybe-tok expr1)
                (let ((tok (get-token re maybe-tok)))
                  (cond ((= level 1)
                         (let loop ((re re)
                                    (last-tok tok)
                                    (last-expr1 expr1))
                           (cond ((and (op? last-tok)
                                       (unary-postfix? last-tok))
                                  =>
                                  (lambda (scheme-name)
                                    (loop re
                                          (get-token re #f)
                                          (##wrap-op1 re
                                                      start-pos
                                                      scheme-name
                                                      last-expr1))))
                                 ((eq? last-tok |token.(|)
                                  (read-arguments-tail
                                   re
                                   #f
                                   (lambda (re maybe-tok args)
                                     (loop re
                                           (get-token re maybe-tok)
                                           (##wrap-op re
                                                      start-pos
                                                      'infix.call
                                                      (cons last-expr1
                                                            args))))))
                                 ((eq? last-tok |token.[|)
                                  (read-expression
                                   re
                                   #f
                                   max-precedence
                                   #f
                                   (lambda (re maybe-tok expr2)
                                     (loop re
                                           (get-token
                                            re
                                            (expect re maybe-tok |token.]|))
                                           (##wrap-op2 re
                                                       start-pos
                                                       'infix.index
                                                       last-expr1
                                                       expr2)))))
                                 ((or (eq? last-tok |token.->|)
                                      (eq? last-tok |token..|))
                                  (let ((next
                                         (macro-peek-next-char-or-eof re)))
                                    (if (or (not (eq? last-tok |token..|))
                                            (and (char? next)
                                                 (identifier-starter? next)))
                                      (read-expression
                                       re
                                       (parse-token-starting-with re next)
                                       (- level 1)
                                       restriction
                                       (lambda (re maybe-tok expr2)
                                         (loop re
                                               (get-token re maybe-tok)
                                               (##wrap-op2 re
                                                           start-pos
                                                           (if (eq? last-tok
                                                                    |token.->|)
                                                             'infix.arrow
                                                             'infix.dot)
                                                           last-expr1
                                                           expr2))))
                                      (cont re
                                            last-tok
                                            last-expr1))))
                                 (else
                                  (cont re
                                        last-tok
                                        last-expr1)))))
                        ((and (op? tok)
                              (= level (precedence tok))
                              (not
                               (cond ((eq? restriction 'no-comma)
                                      (eq? tok |op.,|))
                                     ((eq? restriction 'no-comma-and-no-bar)
                                      (or (eq? tok |op.,|) (eq? tok |op.\||)))
                                     ((eq? restriction 'no-colon)
                                      (eq? tok |op.:|))
                                     (else
                                      #f)))
                              (binary-or-ternary? tok))
                         =>
                         (lambda (scheme-name)
                           (cond ((ternary? tok)
                                  =>
                                  (lambda (end-tok)
                                    (read-expression
                                     re
                                     #f
                                     max-precedence
                                     'no-colon ; assumes that end-tok = |op.:|
                                     (lambda (re maybe-tok expr2)
                                       (read-expression
                                        re
                                        (expect re maybe-tok end-tok)
                                        level
                                        restriction
                                        (lambda (re maybe-tok expr3)
                                          (cont re
                                                maybe-tok
                                                (##wrap-op3 re
                                                            start-pos
                                                            scheme-name
                                                            expr1
                                                            expr2
                                                            expr3))))))))
                                 ((left-to-right? tok)
                                  (let loop ((re re)
                                             (last-scheme-name scheme-name)
                                             (last-expr1 expr1))
                                    (read-expression
                                     re
                                     #f
                                     (- level 1)
                                     restriction
                                     (lambda (re maybe-tok expr2)
                                       (let ((expr1
                                              (##wrap-op2 re
                                                          start-pos
                                                          last-scheme-name
                                                          last-expr1
                                                          expr2)))
                                         (let ((tok (get-token re maybe-tok)))
                                           (cond ((and (op? tok)
                                                       (= level
                                                          (precedence tok))
                                                       (binary-or-ternary?
                                                        tok))
                                                  =>
                                                  (lambda (scheme-name)
                                                    (loop re
                                                          scheme-name
                                                          expr1)))
                                                 (else
                                                  (cont re
                                                        tok
                                                        expr1)))))))))
                                 (else
                                  (read-expression
                                   re
                                   #f
                                   level
                                   restriction
                                   (lambda (re maybe-tok expr2)
                                     (cont re
                                           maybe-tok
                                           (##wrap-op2 re
                                                       start-pos
                                                       scheme-name
                                                       expr1
                                                       expr2))))))))
                        (else
                         (cont re
                               tok
                               expr1)))))))
            ((infix-type? re tok)
             (let ((type (macro-readenv-wrap re tok)))
               (read-procedure
                re
                (expect re #f |token.(|)
                start-pos
                type
                cont)))
            ((symbol? tok)
             (let ((identifier (macro-readenv-wrap re tok)))
               (cont re
                     #f
                     (##wrap-op1 re
                                 start-pos
                                 'infix.identifier
                                 identifier))))
            ((pair? tok) ; This special token represents two
                         ; consecutive tokens.  This trick is used
                         ; because the parser needs a lookahead of 2
                         ; tokens to distinguish definitions from
                         ; anonymous procedures, and to distinguish
                         ; label definitions from expressions.
             (if (cdr tok)
               (let ((expr (car tok)))
                 (cont re
                       (cdr tok)
                       expr))
               (let ((type (car tok)))
                 (read-procedure
                  re
                  #f
                  start-pos
                  type
                  cont))))
            ((or (string? tok)
                 (char? tok)
                 (number? tok))
             (let ((literal (macro-readenv-wrap re tok)))
               (cont re
                     #f
                     (##wrap-op1 re
                                 start-pos
                                 'infix.literal
                                 literal))))
            ((eq? tok |token.(|)
             (let ((tok (get-token re #f)))

               (define (check-closing re maybe-tok x)
                 (cont re
                       (expect re maybe-tok |token.)|)
                       x))

               (if (eq? tok |token.{|)
                 (let ((start-pos (macro-readenv-filepos re)))
                   (read-compound-statement
                    re
                    #f
                    start-pos
                    check-closing))
                 (read-expression
                  re
                  tok
                  max-precedence
                  #f
                  check-closing))))
            ((eq? tok |token.[|)
             (read-list
              re
              #f
              start-pos
              cont))
            ((eq? tok |token.\\|)
             (let ((expr (##read-datum-or-label re)))
               (let* ((obj2
                       (cons expr
                             '()))
                      (obj1
                       (##wrap-op re
                                  start-pos
                                  'infix.scheme
                                  obj2)))
                 (if (##label-marker? expr)
                   (##label-marker-fixup-handler-add!
                    expr
                    (lambda (resolved-obj)
                      (set-car! obj2 resolved-obj))))
                 (cont re
                       #f
                       obj1))))
            (else
             (invalid-infix-syntax re)
             (cont re
                   tok
                   (##wrap-op1 re
                               start-pos
                               'infix.scheme
                               (##wrap re start-pos #f)))))))

  (define (read-paren-expression re maybe-tok cont)
    (let ((tok
           (get-token re (expect re maybe-tok |token.(|))))

      (define (check-closing re maybe-tok x)
        (cont re
              (expect re maybe-tok |token.)|)
              x))

      (let ((start-pos (macro-readenv-filepos re)))
        (if (infix-type? re tok)
          (read-definition-or-expression-or-clause
           re
           tok
           start-pos
           #f
           #f
           check-closing)
          (read-expression-or-clause
           re
           tok
           start-pos
           #f
           #f
           check-closing)))))

  (define (read-expression-or-clause
           re
           maybe-tok
           start-pos
           restriction
           semi?
           cont)
    (read-expression
     re
     maybe-tok
     max-precedence
     restriction
     (if semi?
       (lambda (re maybe-tok expr)
         (let ((tok
                (get-token re maybe-tok)))
           (if (eq? tok |token..|)
             (cont re
                   #f
                   (##wrap-op1 re
                               start-pos
                               'infix.clause
                               expr))
             (cont re
                   (expect re tok |token.;|)
                   expr))))
       cont)))

  (define (read-definition-or-expression-or-clause
           re
           tok
           start-pos
           restriction
           semi?
           cont)
    (let* ((type
            (macro-readenv-wrap re tok))
           (tok
            (get-token re #f)))
      (if (eq? tok |token.(|)
        (read-expression-or-clause
         re
         (cons type #f) ; special combined token
         start-pos
         restriction
         semi?
         cont)
        (read-identifier
         re
         tok
         (lambda (re maybe-tok identifier)
           (read-definition
            re
            #f
            semi?
            start-pos
            type
            identifier
            cont))))))

  (define (read-identifier re maybe-tok cont)
    (let ((tok (get-token re maybe-tok)))
      (if (or (not (symbol? tok))
              (infix-type? re tok))
        (begin
          (invalid-infix-syntax re)
          (cont re
                tok
                (macro-readenv-wrap re 'error)))
        (cont re
              #f
              (macro-readenv-wrap re tok)))))

  (define (statement-starter? re tok)
    ; this function must be kept in sync with "read-statement"
    (or (eq? tok |token.{|)
        (eq? tok |token.;|)
        (infix-type? re tok)
        (expression-starter? re tok)))

  (define (read-statement re maybe-tok cont)
    (let* ((tok
            (get-token re maybe-tok))
           (start-pos
            (macro-readenv-filepos re)))
      (cond ((eq? tok |token.{|)
             (read-compound-statement
              re
              #f
              start-pos
              cont))
            ((eq? tok 'if)
             (read-paren-expression
              re
              #f
              (lambda (re maybe-tok expr)
                (read-statement
                 re
                 maybe-tok
                 (lambda (re maybe-tok stat1)
                   (let ((tok
                          (get-token re maybe-tok)))
                     (if (eq? tok 'else)
                       (read-statement
                        re
                        #f
                        (lambda (re maybe-tok stat2)
                          (cont re
                                maybe-tok
                                (##wrap-op3 re
                                            start-pos
                                            'infix.if
                                            expr
                                            stat1
                                            stat2))))
                       (cont re
                             tok
                             (##wrap-op2 re
                                         start-pos
                                         'infix.if
                                         expr
                                         stat1)))))))))
            ((eq? tok 'while)
             (read-paren-expression
              re
              #f
              (lambda (re maybe-tok expr)
                (read-statement
                 re
                 maybe-tok
                 (lambda (re maybe-tok stat)
                   (cont re
                         maybe-tok
                         (##wrap-op2 re
                                     start-pos
                                     'infix.while
                                     expr
                                     stat)))))))
            ((eq? tok 'do)
             (read-statement
              re
              #f
              (lambda (re maybe-tok stat)
                (read-paren-expression
                 re
                 (expect re maybe-tok 'while)
                 (lambda (re maybe-tok expr)
                   (cont re
                         (expect re maybe-tok |token.;|)
                         (##wrap-op2 re
                                     start-pos
                                     'infix.do-while
                                     stat
                                     expr)))))))
            ((eq? tok 'for)
             (let ()

               (define (get-stat1 re maybe-tok)
                 (read-statement
                  re
                  (expect re maybe-tok |token.(|)
                  (lambda (re maybe-tok stat1)
                    (get-expr2 re
                               maybe-tok
                               stat1))))

               (define (get-expr2 re maybe-tok stat1)
                 (let ((tok
                        (get-token re maybe-tok)))
                   (if (expression-starter? re tok)
                     (read-expression
                      re
                      tok
                      max-precedence
                      #f
                      (lambda (re maybe-tok expr2)
                        (get-expr3 re
                                   maybe-tok
                                   stat1
                                   expr2)))
                     (get-expr3 re
                                tok
                                stat1
                                (##wrap re start-pos #f)))))

               (define (get-expr3 re maybe-tok stat1 expr2)
                 (let ((tok
                        (get-token re (expect re maybe-tok |token.;|))))
                   (if (expression-starter? re tok)
                     (read-expression
                      re
                      tok
                      max-precedence
                      #f
                      (lambda (re maybe-tok expr3)
                        (get-body re
                                  maybe-tok
                                  stat1
                                  expr2
                                  expr3)))
                     (get-body re
                               tok
                               stat1
                               expr2
                               (##wrap re start-pos #f)))))

               (define (get-body re maybe-tok stat1 expr2 expr3)
                 (read-statement
                  re
                  (expect re maybe-tok |token.)|)
                  (lambda (re maybe-tok stat)
                    (cont re
                          maybe-tok
                          (##wrap-op4 re
                                      start-pos
                                      'infix.for
                                      stat1
                                      expr2
                                      expr3
                                      stat)))))

               (get-stat1 re
                          #f)))
            ((eq? tok 'switch)
             (read-paren-expression
              re
              #f
              (lambda (re maybe-tok expr)
                (read-statement
                 re
                 maybe-tok
                 (lambda (re maybe-tok stat)
                   (cont re
                         maybe-tok
                         (##wrap-op2 re
                                     start-pos
                                     'infix.switch
                                     expr
                                     stat)))))))
            ((eq? tok 'break)
             (cont re
                   (expect re #f |token.;|)
                   (##wrap-op0 re
                               start-pos
                               'infix.break)))
            ((eq? tok 'continue)
             (cont re
                   (expect re #f |token.;|)
                   (##wrap-op0 re
                               start-pos
                               'infix.continue)))
            ((eq? tok 'return)
             (let ((tok
                    (get-token re #f)))
               (if (expression-starter? re tok)
                 (read-expression
                  re
                  tok
                  max-precedence
                  #f
                  (lambda (re maybe-tok expr)
                    (cont re
                          (expect re maybe-tok |token.;|)
                          (##wrap-op1 re
                                      start-pos
                                      'infix.return
                                      expr))))
                 (cont re
                       (expect re tok |token.;|)
                       (##wrap-op0 re
                                   start-pos
                                   'infix.return)))))
            ((eq? tok 'goto)
             (read-expression
              re
              #f
              max-precedence
              #f
              (lambda (re maybe-tok expr)
                (cont re
                      (expect re maybe-tok |token.;|)
                      (##wrap-op1 re
                                  start-pos
                                  'infix.goto
                                  expr)))))
            ((eq? tok 'case)
             (read-expression
              re
              #f
              max-precedence
              'no-colon
              (lambda (re maybe-tok expr)
                (read-statement
                 re
                 (expect re maybe-tok |op.:|)
                 (lambda (re maybe-tok stat)
                   (cont re
                         maybe-tok
                         (##wrap-op2 re
                                     start-pos
                                     'infix.case
                                     expr
                                     stat)))))))
            ((eq? tok |token.;|)
             (cont re
                   #f
                   (##wrap-op0 re
                               start-pos
                               'infix.compound)))
            ((infix-type? re tok)
             (read-definition-or-expression-or-clause
              re
              tok
              start-pos
              'no-colon
              #t
              cont))
            ((and (symbol? tok)
                  (not (eq? tok 'new)))
             (let* ((identifier
                     (macro-readenv-wrap re tok))
                    (tok
                     (get-token re #f)))
               (if (eq? tok |op.:|)
                 (read-statement
                  re
                  #f
                  (lambda (re maybe-tok stat)
                    (cont re
                          maybe-tok
                          (##wrap-op2 re
                                      start-pos
                                      'infix.label
                                      identifier
                                      stat))))
                 (read-expression-or-clause
                  re
                  (cons ; special combined token
                   (##wrap-op1 re
                               start-pos
                               'infix.identifier
                               identifier)
                   tok)
                  start-pos
                  'no-colon
                  #t
                  cont))))
            (else
             (read-expression-or-clause
              re
              tok
              start-pos
              'no-colon
              #t
              cont)))))

  (define (read-definition re maybe-tok semi? start-pos type identifier cont)

    (define (get-dimensions re maybe-tok rev-dims)
      (let ((tok (get-token re maybe-tok)))
        (cond ((eq? tok |token.[|)
               (read-expression
                re
                #f
                max-precedence
                #f
                (lambda (re maybe-tok dim)
                  (get-dimensions re
                                  (expect re maybe-tok |token.]|)
                                  (cons dim rev-dims)))))
              ((eq? tok op.=)
               (read-expression
                re
                #f
                max-precedence
                #f
                (lambda (re maybe-tok init)
                  (get-tail re
                            maybe-tok
                            rev-dims
                            init))))
              (else
               (get-tail re
                         tok
                         rev-dims
                         (##wrap re start-pos #f))))))

    (define (get-tail re maybe-tok rev-dims init)
      (cont re
            (if semi?
              (expect re maybe-tok |token.;|)
              maybe-tok)
            (##wrap-op4 re
                        start-pos
                        'infix.define-variable
                        identifier
                        type
                        (##wrap re start-pos (reverse rev-dims))
                        init)))

    (let ((tok (get-token re maybe-tok)))
      (if (eq? tok |token.(|)
        (read-procedure
         re
         #f
         start-pos
         type
         (lambda (re maybe-tok proc)
           (cont re
                 maybe-tok
                 (##wrap-op2 re
                             start-pos
                             'infix.define-procedure
                             identifier
                             proc))))
        (get-dimensions re
                        tok
                        '()))))

  (define (read-procedure re maybe-tok start-pos type cont)
    (let* ((tok
            (get-token re maybe-tok))
           (params-start-pos
            (macro-readenv-filepos re)))

      (define (get-body re maybe-tok rev-params)
        (read-compound-statement
         re
         (expect re maybe-tok |token.{|)
         start-pos
         (lambda (re maybe-tok stat)
           (cont re
                 maybe-tok
                 (##wrap-op3 re
                             start-pos
                             'infix.procedure
                             type
                             (##wrap re params-start-pos (reverse rev-params))
                             stat)))))

      (define (err re tok rev-params)
        (invalid-infix-syntax re)
        (get-body re
                  (if (eq? tok |token.)|)
                    #f
                    tok)
                  rev-params))

      (if (not (infix-type? re tok))
        (get-body re
                  (expect re tok |token.)|)
                  '())
        (let loop ((tok tok) (rev-params '()))
          (if (not (infix-type? re tok))
            (err re tok rev-params)
            (let* ((type
                    (macro-readenv-wrap re tok))
                   (tok
                    (get-token re #f)))
              (if (or (not (symbol? tok))
                      (infix-type? re tok))
                (err re tok rev-params)
                (let* ((param
                        (macro-readenv-wrap re tok))
                       (new-rev-params
                        (cons (macro-readenv-wrap re (list param type))
                              rev-params))
                       (tok
                        (get-token re #f)))
                  (if (eq? tok |op.,|)
                    (loop (get-token re #f)
                          new-rev-params)
                    (get-body re
                              (expect re tok |token.)|)
                              new-rev-params))))))))))

  (define (read-compound-statement re maybe-tok start-pos cont)
    (read-statements-tail
     re
     maybe-tok
     start-pos
     '()
     (lambda (re maybe-tok stats)
       (cont re
             maybe-tok
             (##wrap-op re
                        start-pos
                        'infix.compound
                        stats)))))

  (define (read-statements-tail re maybe-tok start-pos rev-stats cont)
    (let ((tok (get-token re maybe-tok)))
      (if (statement-starter? re tok)
        (read-statement
         re
         tok
         (lambda (re maybe-tok stat)
           (read-statements-tail
            re
            maybe-tok
            start-pos
            (cons stat rev-stats)
            cont)))
        (cont re
              (expect re tok |token.}|)
              (reverse rev-stats)))))

  (define (invalid-infix-syntax re)
    (##raise-datum-parsing-exception
     'invalid-infix-syntax
     re))

  (define (invalid-infix-syntax-character re)
    (##raise-datum-parsing-exception
     'invalid-infix-syntax-character
     re))

  (define (invalid-infix-syntax-number re)
    (##raise-datum-parsing-exception
     'invalid-infix-syntax-number
     re))

  (define (infix-type? re tok)
    ((##readtable-infix-type? (macro-readenv-readtable re)) tok))

  (macro-read-next-char-or-eof re) ; skip backslash

  (read-statement
   re
   #f
   (lambda (re maybe-tok expr)
     expr)))

(define (##infix-type? x)
  (##assq x ##infix-types))

(define ##infix-types '())
(set! ##infix-types
  '((int    . #f)
    (char   . #f)
    (bool   . #f)
    (void   . #f)
    (float  . #f)
    (double . #f)
    (object . #f)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Setup the standard readtable.

(define (##make-standard-readtable)
  (let ((rt
         (##make-readtable
          #f ; preserve case in symbols, character names, etc
          #t ; keywords ending with ":" are allowed
          ##standard-escaped-char-table
          ##standard-named-char-table
          ##standard-sharp-bang-table
          (##make-chartable #f) ; all chars are non-delimiters
          (##make-chartable ##read-number/keyword/symbol)
          (##make-chartable ##read-sharp-other)
          255          ; max-unescaped-char
          #t           ; escape-ctrl-chars?
          #f           ; sharing-allowed?
          #f           ; eval-allowed?
          ##max-fixnum ; max-write-level
          ##max-fixnum ; max-write-length
          ##standard-pretty-print-formats
          'syntax      ; sharp-quote-keyword
          ##infix-type?
          )))

    (if (##fixnum.< 0 (##get-standard-level)) ; comply to standard Scheme?
      (begin
        (##readtable-case-conversion?-set! rt #t)
        (##readtable-keywords-allowed?-set! rt #f)))

    ; setup control characters

    (let loop ((i 31))
      (if (not (< i 0))
        (begin
          (##readtable-char-class-set!
           rt
           (ucs4->character i)
           #t
           ##read-illegal)
          (loop (- i 1)))))

    ; setup whitespace characters

    (##readtable-char-class-set! rt #\space    #t ##read-whitespace)
    (##readtable-char-class-set! rt #\linefeed #t ##read-whitespace)
    (##readtable-char-class-set! rt #\return   #t ##read-whitespace)
    (##readtable-char-class-set! rt #\tab      #t ##read-whitespace)
    (##readtable-char-class-set! rt #\page     #t ##read-whitespace)

    ; setup handlers for non-whitespace delimiters

    (##readtable-char-class-set! rt #\; #t ##read-single-line-comment)

    (##readtable-char-class-set! rt #\" #t ##read-escaped-string)
    (##readtable-char-class-set! rt #\| #t ##read-number/keyword/symbol)

    (##readtable-char-class-set! rt #\' #t ##read-quotation)
    (##readtable-char-class-set! rt #\` #t ##read-quasiquotation)
    (##readtable-char-class-set! rt #\, #t ##read-unquotation)

    (##readtable-char-class-set! rt #\( #t ##read-list)
    (##readtable-char-class-set! rt #\) #t ##read-none)

    (##readtable-char-class-set! rt #\[ #t ##read-list)
    (##readtable-char-class-set! rt #\] #t ##read-none)

    (##readtable-char-class-set! rt #\{ #t ##read-illegal)
    (##readtable-char-class-set! rt #\} #t ##read-illegal)

    (##readtable-char-class-set! rt #\\ #t ##read-infix)

    ; setup handlers for "#" and "." (these are NOT delimiters)

    (##readtable-char-class-set! rt #\# #f ##read-sharp)
    (##readtable-char-class-set! rt #\. #f ##read-dot)

    ; setup handlers for sharp read-macros

    (let loop ((i 57))
      (if (not (< i 48))
        (begin
          (##readtable-char-sharp-handler-set!
           rt
           (ucs4->character i)
           ##read-sharp-digit)
          (loop (- i 1)))))

    (##readtable-char-sharp-handler-set! rt #\( ##read-sharp-vector)
    (##readtable-char-sharp-handler-set! rt #\\ ##read-sharp-char)
    (##readtable-char-sharp-handler-set! rt #\| ##read-sharp-comment)
    (##readtable-char-sharp-handler-set! rt #\! ##read-sharp-bang)
    (##readtable-char-sharp-handler-set! rt #\# ##read-sharp-sharp)
    (##readtable-char-sharp-handler-set! rt #\: ##read-sharp-colon)
    (##readtable-char-sharp-handler-set! rt #\' ##read-sharp-quotation)
    (##readtable-char-sharp-handler-set! rt #\& ##read-sharp-ampersand)
    (##readtable-char-sharp-handler-set! rt #\. ##read-sharp-dot)

    rt))

(if (not ##main-readtable)
  (set! ##main-readtable
    (##make-standard-readtable)))

;==============================================================================

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(macro-denv-input-port-set! (macro-thread-denv (macro-current-thread)) ##stdin)
(macro-denv-output-port-set! (macro-thread-denv (macro-current-thread)) ##stdout)
(define ##current-error-port
  (##make-parameter ##stderr))

(define ##current-readtable
  (##make-parameter ##main-readtable))
